/*
 * %NT_SOFTWARE_LICENSE%
 */

//
// Includes:
//
#include "ntlog.h"
#include "nthw_drv.h"
#include "nthw_register.h"

#include "flow_nthw_csu.h"

#include <stdlib.h>
#include <string.h>

void csuNthw_setDebugMode(struct CsuNthw *p, unsigned nDebugMode)
{
  Module_setDebugMode(p->mCsu, nDebugMode);
}

struct CsuNthw *CsuNthw_new(void)
{
  struct CsuNthw *p = malloc(sizeof(struct CsuNthw));
  if (p) {
    (void)memset(p, 0, sizeof(*p));
  }
  return p;
}

void CsuNthw_delete(struct CsuNthw *p)
{
  if (p) {
    (void)memset(p, 0, sizeof(*p));
    free(p);
  }
}

int CsuNthw_init(struct CsuNthw *p, Fpga_t *pFpga, int nInstance)
{
  const char *const pAdapterIdStr = pFpga->pFpgaInfo->mpAdapterIdStr;
  struct Module_s* pMod = Fpga_queryModule(pFpga, MOD_CSU, nInstance);

  assert(nInstance >= 0 && nInstance < 256);

  if (p == NULL) {
    return pMod == NULL ? -1 : 0;
  }

  if (pMod == NULL) {
    NT_LOG(ERR, NTHW, "%s: Csu %d: no such instance\n", pAdapterIdStr, nInstance);
    return -1;
  }

  p->mpFpga = pFpga;
  p->mPhysicalAdapterNo = (uint8_t)nInstance;
  p->mCsu = pMod;

  p->mpRcpCtrl = Module_getRegister(p->mCsu, CSU_RCP_CTRL);
  p->mpRcpCtrlAdr = Register_getField(p->mpRcpCtrl, CSU_RCP_CTRL_ADR);
  p->mpRcpCtrlCnt = Register_getField(p->mpRcpCtrl, CSU_RCP_CTRL_CNT);
  p->mpRcpData = Module_getRegister(p->mCsu, CSU_RCP_DATA);
  p->mpRcpDataOl3Cmd = Register_getField(p->mpRcpData, CSU_RCP_DATA_OL3_CMD);
  p->mpRcpDataOl4Cmd = Register_getField(p->mpRcpData, CSU_RCP_DATA_OL4_CMD);
  p->mpRcpDataIl3Cmd = Register_getField(p->mpRcpData, CSU_RCP_DATA_IL3_CMD);
  p->mpRcpDataIl4Cmd = Register_getField(p->mpRcpData, CSU_RCP_DATA_IL4_CMD);

  return 0;
}

void CsuNthw_RcpSelect(const struct CsuNthw *p, uint32_t val)
{
  Field_setVal32(p->mpRcpCtrlAdr, val);
}

void CsuNthw_RcpCnt(const struct CsuNthw *p, uint32_t val)
{
  Field_setVal32(p->mpRcpCtrlCnt, val);
}

void CsuNthw_RcpOuterL3Cmd(const struct CsuNthw *p, uint32_t val)
{
  // Select L3 calc method for outer layer3.
  // 0: Do not touch checksum field.
  // 1: Check, but do not touch checksum field.
  // 2: Insert checksum header value for BAD checksum.
  // 3: Insert checksum header value for GOOD checksum.
  Field_setVal32(p->mpRcpDataOl3Cmd, val);
}

void CsuNthw_RcpOuterL4Cmd(const struct CsuNthw *p, uint32_t val)
{
  // Select L4 calc method for outer layer4.
  // 0: Do not touch checksum field.
  // 1: Check, but do not touch checksum field.
  // 2: Insert checksum header value for BAD checksum.
  // 3: Insert checksum header value for GOOD checksum.
  // 4: Set UDP checksum value of ZERO for both IPv4/IPv6, set good checksum for TCP.
  // 5: Set UDP checksum value of ZERO for IPv4, set good checksum for TCP.
  // 6: Set UDP checksum value of ZERO for outer tunnel when tunnel is IPv4/IPv6 and UDP, otherwise GOOD checksum.
  // 7: Set UDP checksum value of ZERO for outer tunnel when tunnel is IPv4 and UDP, otherwise GOOD checksum.
  Field_setVal32(p->mpRcpDataOl4Cmd, val);
}

void CsuNthw_RcpInnerL3Cmd(const struct CsuNthw *p, uint32_t val)
{
  // Select L3 calc method for inner layer3 (tunneled).
  // 0: Do not touch checksum field.
  // 1: Check, but do not touch checksum field.
  // 2: Insert checksum header value for BAD checksum.
  // 3: Insert checksum header value for GOOD checksum.
  Field_setVal32(p->mpRcpDataIl3Cmd, val);
}

void CsuNthw_RcpInnerL4Cmd(const struct CsuNthw *p, uint32_t val)
{
  // Select L4 calc method for inner layer4 (tunneled).
  // 0: Do not touch checksum field.
  // 1: Check, but do not touch checksum field.
  // 2: Insert checksum header value for BAD checksum.
  // 3: Insert checksum header value for GOOD checksum.
  // 4: Set UDP checksum value of ZERO for both IPv4/IPv6, set good checksum for TCP.
  // 5: Set UDP checksum value of ZERO for IPv4, set good checksum for TCP.
  // 6: Set UDP checksum value of ZERO for outer tunnel when tunnel is IPv4/IPv6 and UDP, otherwise GOOD checksum.
  // 7: Set UDP checksum value of ZERO for outer tunnel when tunnel is IPv4 and UDP, otherwise GOOD checksum.
  Field_setVal32(p->mpRcpDataIl4Cmd, val);
}

void CsuNthw_RcpFlush(const struct CsuNthw *p)
{
  Register_flush(p->mpRcpCtrl, 1);
  Register_flush(p->mpRcpData, 1);
}
