/*
 * %NT_SOFTWARE_LICENSE%
 */

//
// Includes:
//
#include "nt_util.h"
#include "ntlog.h"
//
#include "nthw_drv.h"
#include "nthw_register.h"
#include "nthw_fpga_model.h"

//
//
//
#include "flow_nthw_info.h"

#include <stdlib.h> // malloc
#include <string.h> // memset

static inline unsigned clamp_one(unsigned val)
{
  return val > 1 ? 1 : val;
}

struct InfoNthw *InfoNthw_new(void)
{
  struct InfoNthw *p = malloc(sizeof(struct InfoNthw));
  if (p)
    (void)memset(p, 0, sizeof(*p));

  return p;
}

void InfoNthw_delete(struct InfoNthw *p)
{
  if (p) {
    (void)memset(p, 0, sizeof(*p));
    free(p);
  }
}

int InfoNthw_init(struct InfoNthw *p, Fpga_t *pFpga, int nInstance)
{
  assert(nInstance >= 0 && nInstance < 256);

  p->mpFpga = pFpga;
  p->mPhysicalAdapterNo = (uint8_t)nInstance;

  unsigned KmPresent  = clamp_one(Fpga_getProductParam(pFpga, NT_KM_PRESENT, 0));
  unsigned KccPresent = clamp_one(Fpga_getProductParam(pFpga, NT_CAT_KCC_PRESENT, 0));
  unsigned IoaPresent = clamp_one(Fpga_getProductParam(pFpga, NT_IOA_PRESENT, 0));
  unsigned RoaPresent = clamp_one(Fpga_getProductParam(pFpga, NT_ROA_PRESENT, 0));
  unsigned DbsPresent = clamp_one(Fpga_getProductParam(pFpga, NT_DBS_PRESENT, 0));
  unsigned FlmPresent = clamp_one(Fpga_getProductParam(pFpga, NT_FLM_PRESENT, 0));
  unsigned HstPresent = clamp_one(Fpga_getProductParam(pFpga, NT_HST_PRESENT, 0));

  // Modules for Tx Packet Edit function
  unsigned HfuPresent = clamp_one(Fpga_getProductParam(pFpga, NT_HFU_PRESENT, 0));
  unsigned TxCpyPresent = clamp_one(Fpga_getProductParam(pFpga, NT_TX_CPY_PRESENT, 0));
  unsigned TxInsPresent = clamp_one(Fpga_getProductParam(pFpga, NT_TX_INS_PRESENT, 0));
  unsigned TxRplPresent = clamp_one(Fpga_getProductParam(pFpga, NT_TX_RPL_PRESENT, 0));
  unsigned CsuPresent = clamp_one(Fpga_getProductParam(pFpga, NT_CSU_PRESENT, 0));
  unsigned TpePresent = (HfuPresent && TxCpyPresent && TxInsPresent && TxRplPresent && CsuPresent) ? 1 : 0;

  p->nPhyPorts = Fpga_getProductParam(pFpga, NT_PHY_PORTS, 0);
  p->nRxPorts = Fpga_getProductParam(pFpga, NT_RX_PORTS, 0);
  p->nLtxAvail = Fpga_getProductParam(pFpga, NT_LR_PRESENT, 0);
  p->nbCatFunc = Fpga_getProductParam(pFpga, NT_CAT_FUNCS, 0);
  p->nbCategories = Fpga_getProductParam(pFpga, NT_CATEGORIES, 0);
  p->nbQueues = Fpga_getProductParam(pFpga, NT_QUEUES, 0);
  p->nbFlowTypes = Fpga_getProductParam(pFpga, NT_KM_FLOW_TYPES, 0) * clamp_one(KmPresent + FlmPresent);
  p->nbPmExt = Fpga_getProductParam(pFpga, NT_CAT_N_EXT, 0);
  p->nbLen = Fpga_getProductParam(pFpga, NT_CAT_N_LEN, 0);
  p->nbKccSize = Fpga_getProductParam(pFpga, NT_CAT_KCC_SIZE, 0) * KccPresent;
  p->nbKccBanks = Fpga_getProductParam(pFpga, NT_CAT_KCC_BANKS, 0) * KccPresent;
  p->nbKmCategories = Fpga_getProductParam(pFpga, NT_KM_CATEGORIES, 0) * KmPresent;
  p->nbKmCamBanks = Fpga_getProductParam(pFpga, NT_KM_CAM_BANKS, 0) * KmPresent;
  p->nbKmCamRecordWords = Fpga_getProductParam(pFpga, NT_KM_CAM_REC_WORDS, 0) * KmPresent;
  p->nbKmCamRecords = Fpga_getProductParam(pFpga, NT_KM_CAM_RECORDS, 0) * KmPresent;
  p->nbKmTcamBanks = Fpga_getProductParam(pFpga, NT_KM_TCAM_BANKS, 0) * KmPresent;
  p->nbKmTcamBankWidth = Fpga_getProductParam(pFpga, NT_KM_TCAM_BANK_WIDTH, 0) * KmPresent;
  p->nbFlmCategories = Fpga_getProductParam(pFpga, NT_FLM_CATEGORIES, 0) * FlmPresent;
  p->nbFlmSizeMb = Fpga_getProductParam(pFpga, NT_FLM_SIZE_MB, 0);
  p->nbFlmEntrySize = Fpga_getProductParam(pFpga, NT_FLM_ENTRY_SIZE, 0);
  p->nbFlmVariant = Fpga_getProductParam(pFpga, NT_FLM_VARIANT, 0);
  p->nbFlmPrios = Fpga_getProductParam(pFpga, NT_FLM_PRIOS, 0) * FlmPresent;
  p->nbFlmPstProfiles = Fpga_getProductParam(pFpga, NT_FLM_PST_PROFILES, 0) * FlmPresent;
  p->nbHstCategories = Fpga_getProductParam(pFpga, NT_HST_CATEGORIES, 0) * HstPresent;
  p->nbQslCategories = Fpga_getProductParam(pFpga, NT_QSL_CATEGORIES, 0);
  p->nbQslQstEntries = Fpga_getProductParam(pFpga, NT_QSL_QST_SIZE, 0);
  p->nbPdbCategories = Fpga_getProductParam(pFpga, NT_PDB_CATEGORIES, 0);
  p->nbIoaCategories = Fpga_getProductParam(pFpga, NT_IOA_CATEGORIES, 0) * IoaPresent;
  p->nbRoaCategories = Fpga_getProductParam(pFpga, NT_ROA_CATEGORIES, 0) * RoaPresent;
  p->nbDbsCategories = MIN(Fpga_getProductParam(pFpga, NT_DBS_RX_QUEUES, 0), Fpga_getProductParam(pFpga, NT_DBS_TX_QUEUES, 0)) * DbsPresent;
  p->nbCatKmIfCnt = Fpga_getProductParam(pFpga, NT_CAT_KM_IF_CNT, KmPresent + FlmPresent);
  p->mCatKmIfM0 = Fpga_getProductParam(pFpga, NT_CAT_KM_IF_M0, -1);
  p->mCatKmIfM1 = Fpga_getProductParam(pFpga, NT_CAT_KM_IF_M1, -1);
  p->nbTpeCategories = Fpga_getProductParam(pFpga, NT_TPE_CATEGORIES, 0) * TpePresent;
  p->nbTxCpyWriters = Fpga_getProductParam(pFpga, NT_TX_CPY_WRITERS, 0) * TpePresent;
  p->nbTxCpyMaskMem = Fpga_getProductParam(pFpga, NT_CPY_MASK_MEM, 0) * TpePresent;
  p->nbTxRplDepth = Fpga_getProductParam(pFpga, NT_TX_RPL_DEPTH, 0) * TpePresent;
  p->nbTxRplExtCategories = Fpga_getProductParam(pFpga, NT_TX_RPL_EXT_CATEGORIES, 0) * TpePresent;
  return 0;
}

unsigned InfoNthw_getNbPhyPorts(const struct InfoNthw *p)
{
  return p->nPhyPorts;
}

unsigned InfoNthw_getNbRxPorts(const struct InfoNthw *p)
{
  return p->nRxPorts;
}

unsigned InfoNthw_getLtxAvail(const struct InfoNthw *p)
{
  return p->nLtxAvail;
}

unsigned InfoNthw_getNbCategories(const struct InfoNthw *p)
{
  return p->nbCategories;
}

unsigned InfoNthw_getKccSize(const struct InfoNthw *p)
{
  return p->nbKccSize;
}

unsigned InfoNthw_getKccBanks(const struct InfoNthw *p)
{
  return p->nbKccBanks;
}

unsigned InfoNthw_getNbQueues(const struct InfoNthw *p)
{
  return p->nbQueues;
}

unsigned InfoNthw_getNbCatFuncs(const struct InfoNthw *p)
{
  return p->nbCatFunc;
}

unsigned InfoNthw_getNbKMFlowTypes(const struct InfoNthw *p)
{
  return p->nbFlowTypes;
}

unsigned InfoNthw_getNbPmExt(const struct InfoNthw *p)
{
  return p->nbPmExt;
}

unsigned InfoNthw_getNbLen(const struct InfoNthw *p)
{
  return p->nbLen;
}

unsigned InfoNthw_getNbKmCategories(const struct InfoNthw *p)
{
  return p->nbKmCategories;
}

unsigned InfoNthw_getNbKmCamBanks(const struct InfoNthw *p)
{
  return p->nbKmCamBanks;
}

unsigned InfoNthw_getNbKmCamRecordWords(const struct InfoNthw *p)
{
  return p->nbKmCamRecordWords;
}

unsigned InfoNthw_getNbKmCamRecords(const struct InfoNthw *p)
{
  return p->nbKmCamRecords;
}

unsigned InfoNthw_getNbKmTcamBanks(const struct InfoNthw *p)
{
  return p->nbKmTcamBanks;
}

unsigned InfoNthw_getNbKmTcamBankWidth(const struct InfoNthw *p)
{
  return p->nbKmTcamBankWidth;
}

unsigned InfoNthw_getNbFlmCategories(const struct InfoNthw *p)
{
  return p->nbFlmCategories;
}

unsigned InfoNthw_getNbFlmSizeMb(const struct InfoNthw *p)
{
  return p->nbFlmSizeMb;
}

unsigned InfoNthw_getNbFlmEntrySize(const struct InfoNthw *p)
{
  return p->nbFlmEntrySize;
}

unsigned InfoNthw_getNbFlmVariant(const struct InfoNthw *p)
{
  return p->nbFlmVariant;
}

unsigned InfoNthw_getNbFlmPrios(const struct InfoNthw *p)
{
  return p->nbFlmPrios;
}

unsigned InfoNthw_getNbFlmPstProfiles(const struct InfoNthw *p)
{
  return p->nbFlmPstProfiles;
}

unsigned InfoNthw_getNbHstCategories(const struct InfoNthw *p)
{
  return p->nbHstCategories;
}

unsigned InfoNthw_getNbQslCategories(const struct InfoNthw *p)
{
  return p->nbQslCategories;
}

unsigned InfoNthw_getNbQslQstEntries(const struct InfoNthw *p)
{
  return p->nbQslQstEntries;
}

unsigned InfoNthw_getNbPdbCategories(const struct InfoNthw *p)
{
  return p->nbPdbCategories;
}

unsigned InfoNthw_getNbIoaCategories(const struct InfoNthw *p)
{
  return p->nbIoaCategories;
}

unsigned InfoNthw_getNbRoaCategories(const struct InfoNthw *p)
{
  return p->nbRoaCategories;
}

unsigned InfoNthw_getNbDbsCategories(const struct InfoNthw *p)
{
  return p->nbDbsCategories;
}

unsigned InfoNthw_getNbCatKmIfCnt(const struct InfoNthw *p)
{
  return p->nbCatKmIfCnt;
}

unsigned InfoNthw_getNbCatKmIfM0(const struct InfoNthw *p)
{
  return p->mCatKmIfM0;
}

unsigned InfoNthw_getNbCatKmIfM1(const struct InfoNthw *p)
{
  return p->mCatKmIfM1;
}

unsigned InfoNthw_getNbTpeCategories(const struct InfoNthw *p)
{
  return p->nbTpeCategories;
}

unsigned InfoNthw_getNbTxCpyWriters(const struct InfoNthw *p)
{
  return p->nbTxCpyWriters;
}

unsigned InfoNthw_getNbTxCpyMaskMem(const struct InfoNthw *p)
{
  return p->nbTxCpyMaskMem;
}

unsigned InfoNthw_getNbTxRplDepth(const struct InfoNthw *p)
{
  return p->nbTxRplDepth;
}

unsigned InfoNthw_getNbTxRplExtCategories(const struct InfoNthw *p)
{
  return p->nbTxRplExtCategories;
}
