/*
 * %NT_SOFTWARE_LICENSE%
 */

//
// Includes:
//
#include "ntlog.h"
//
#include "nthw_drv.h"
#include "nthw_register.h"

//
//
//
#include "flow_nthw_ioa.h"

#include <stdlib.h> // malloc
#include <string.h> // memset

void ioaNthw_setDebugMode(struct IoaNthw *p, unsigned nDebugMode)
{
  Module_setDebugMode(p->mIoa, nDebugMode);
}

struct IoaNthw *IoaNthw_new(void)
{
  struct IoaNthw *p = malloc(sizeof(struct IoaNthw));
  if (p)
    (void)memset(p, 0, sizeof(*p));

  return p;
}

void IoaNthw_delete(struct IoaNthw *p)
{
  if (p) {
    (void)memset(p, 0, sizeof(*p));
    free(p);
  }
}

int IoaNthw_init(struct IoaNthw *p, Fpga_t *pFpga, int nInstance)
{
  const char *const pAdapterIdStr = pFpga->pFpgaInfo->mpAdapterIdStr;
  struct Module_s* pMod = Fpga_queryModule(pFpga, MOD_IOA, nInstance);

  assert(nInstance >= 0 && nInstance < 256);

  if (p == NULL) {
    return pMod == NULL ? -1 : 0;
  }

  if (pMod == NULL) {
    NT_LOG(ERR, NTHW, "%s: Ioa %d: no such instance\n", pAdapterIdStr, nInstance);
    return -1;
  }

  p->mpFpga = pFpga;
  p->mPhysicalAdapterNo = (uint8_t)nInstance;
  p->mIoa = pMod;

  // RCP
  p->mpRcpCtrl  = Module_getRegister(p->mIoa, IOA_RECIPE_CTRL);
  p->mpRcpAddr  = Register_getField(p->mpRcpCtrl, IOA_RECIPE_CTRL_ADR);
  p->mpRcpCnt   = Register_getField(p->mpRcpCtrl, IOA_RECIPE_CTRL_CNT);
  p->mpRcpData  = Module_getRegister(p->mIoa, IOA_RECIPE_DATA);
  p->mpRcpDataTunnelPop = Register_getField(p->mpRcpData, IOA_RECIPE_DATA_TUNNEL_POP);
  p->mpRcpDataVlanPop = Register_getField(p->mpRcpData, IOA_RECIPE_DATA_VLAN_POP);
  p->mpRcpDataVlanPush = Register_getField(p->mpRcpData, IOA_RECIPE_DATA_VLAN_PUSH);
  p->mpRcpDataVlanVid = Register_getField(p->mpRcpData, IOA_RECIPE_DATA_VLAN_VID);
  p->mpRcpDataVlanDei = Register_getField(p->mpRcpData, IOA_RECIPE_DATA_VLAN_DEI);
  p->mpRcpDataVlanPcp = Register_getField(p->mpRcpData, IOA_RECIPE_DATA_VLAN_PCP);
  p->mpRcpDataVlanTpidSel = Register_getField(p->mpRcpData, IOA_RECIPE_DATA_VLAN_TPID_SEL);
  p->mpRcpDataQueueOverrideEn = Register_getField(p->mpRcpData, IOA_RECIPE_DATA_QUEUE_OVERRIDE_EN);
  p->mpRcpDataQueueId = Register_getField(p->mpRcpData, IOA_RECIPE_DATA_QUEUE_ID);

  // Special Vlan Tpid
  p->mpSpecial = Module_getRegister(p->mIoa, IOA_VLAN_TPID_SPECIAL);
  p->mpSpecialVlanTpidCustTpid0 = Register_getField(p->mpSpecial, IOA_VLAN_TPID_SPECIAL_CUSTTPID0);
  p->mpSpecialVlanTpidCustTpid1 = Register_getField(p->mpSpecial, IOA_VLAN_TPID_SPECIAL_CUSTTPID1);
  {

    /*
     * This extension in IOA is a messy way FPGA have chosen to
     * put control bits for EPP module in IOA. It is accepted as
     * we are going towards exchange IOA and ROA modules later
     * to get higher scalability in future.
     */
    p->mpRoaEppCtrl = Module_queryRegister(p->mIoa, IOA_ROA_EPP_CTRL);
    if (p->mpRoaEppCtrl) {
      p->mpRoaEppAddr = Register_getField(p->mpRoaEppCtrl, IOA_ROA_EPP_CTRL_ADR);
      p->mpRoaEppCnt  = Register_getField(p->mpRoaEppCtrl, IOA_ROA_EPP_CTRL_CNT);
    } else  {
      p->mpRoaEppAddr = NULL;
      p->mpRoaEppCnt  = NULL;
    }

    p->mpRoaEppData = Module_queryRegister(p->mIoa, IOA_ROA_EPP_DATA);
    if (p->mpRoaEppData) {
      p->mpRoaEppDataPushTunnel = Register_getField(p->mpRoaEppData, IOA_ROA_EPP_DATA_PUSH_TUNNEL);
      p->mpRoaEppDataTxPort = Register_getField(p->mpRoaEppData, IOA_ROA_EPP_DATA_TX_PORT);
    } else {
      p->mpRoaEppDataPushTunnel = NULL;
      p->mpRoaEppDataTxPort = NULL;
    }
  }
  return 0;
}

// RCP
void IoaNthw_RcpSelect(const struct IoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpRcpAddr, val);
};

void IoaNthw_RcpCnt(const struct IoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpRcpCnt, val);
};

void IoaNthw_RcpTunnelPop(const struct IoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpRcpDataTunnelPop, val);
};

void IoaNthw_RcpVlanPop(const struct IoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpRcpDataVlanPop, val);
};

void IoaNthw_RcpVlanPush(const struct IoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpRcpDataVlanPush, val);
};

void IoaNthw_RcpVlanVid(const struct IoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpRcpDataVlanVid, val);
};

void IoaNthw_RcpVlanDei(const struct IoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpRcpDataVlanDei, val);
};

void IoaNthw_RcpVlanPcp(const struct IoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpRcpDataVlanPcp, val);
};

void IoaNthw_RcpVlanTpidSel(const struct IoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpRcpDataVlanTpidSel, val);
};

void IoaNthw_RcpQueueOverrideEn(const struct IoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpRcpDataQueueOverrideEn, val);
};

void IoaNthw_RcpQueueId(const struct IoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpRcpDataQueueId, val);
};

void IoaNthw_RcpFlush(const struct IoaNthw *p)
{
  Register_flush(p->mpRcpCtrl, 1);
  Register_flush(p->mpRcpData, 1);
};

// Vlan Tpid Special
void IoaNthw_SpecialVlanTpidCustTpid0(const struct IoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpSpecialVlanTpidCustTpid0, val);
};

void IoaNthw_SpecialVlanTpidCustTpid1(const struct IoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpSpecialVlanTpidCustTpid1, val);
};

void IoaNthw_SpecialVlanTpidFlush(const struct IoaNthw *p)
{
  Register_flush(p->mpSpecial, 1);
};

void IoaNthw_RoaEppSelect(const struct IoaNthw *p, uint32_t val)
{
  if (p->mpRoaEppAddr) {
    Field_setVal32(p->mpRoaEppAddr, val);
  }
};

void IoaNthw_RoaEppCnt(const struct IoaNthw *p, uint32_t val)
{
  if (p->mpRoaEppCnt) {
    Field_setVal32(p->mpRoaEppCnt, val);
  }
};

void IoaNthw_RoaEppPushTunnel(const struct IoaNthw *p, uint32_t val)
{
  if (p->mpRoaEppDataPushTunnel) {
    Field_setVal32(p->mpRoaEppDataPushTunnel, val);
  }
};
void IoaNthw_RoaEppTxPort(const struct IoaNthw *p, uint32_t val)
{
  if (p->mpRoaEppDataTxPort) {
    Field_setVal32(p->mpRoaEppDataTxPort, val);
  }
};

void IoaNthw_RoaEppFlush(const struct IoaNthw *p)
{
  if (p->mpRoaEppCtrl) {
    Register_flush(p->mpRoaEppCtrl, 1);
  }
  if (p->mpRoaEppData) {
    Register_flush(p->mpRoaEppData, 1);
  }
};


