/*
 * %NT_SOFTWARE_LICENSE%
 */

//
// Includes:
//
#include "ntlog.h"
//
#include "nthw_drv.h"
#include "nthw_register.h"

//
//
//
#include "flow_nthw_qsl.h"

#include <stdlib.h> // malloc
#include <string.h> // memset

void qslNthw_setDebugMode(struct QslNthw *p, unsigned nDebugMode)
{
  Module_setDebugMode(p->mQsl, nDebugMode);
}

struct QslNthw *QslNthw_new(void)
{
  struct QslNthw *p = malloc(sizeof(struct QslNthw));
  if (p)
    (void)memset(p, 0, sizeof(*p));

  return p;
}

void QslNthw_delete(struct QslNthw *p)
{
  if (p) {
    (void)memset(p, 0, sizeof(*p));
    free(p);
  }
}

int QslNthw_init(struct QslNthw *p, Fpga_t *pFpga, int nInstance)
{
  const char *const pAdapterIdStr = pFpga->pFpgaInfo->mpAdapterIdStr;
  struct Module_s* pMod = Fpga_queryModule(pFpga, MOD_QSL, nInstance);

  assert(nInstance >= 0 && nInstance < 256);

  if (p == NULL) {
    return pMod == NULL ? -1 : 0;
  }

  if (pMod == NULL) {
    NT_LOG(ERR, NTHW, "%s: QSL %d: no such instance\n", pAdapterIdStr, nInstance);
    return -1;
  }

  p->mpFpga = pFpga;
  p->mPhysicalAdapterNo = (uint8_t)nInstance;
  p->mQsl = pMod;

#if defined(DEBUG) && (0)
  Module_setDebugMode(p->mQsl, 0xFF); // DEBUG:
#endif

  // RCP
  p->mpRcpCtrl = Module_getRegister(p->mQsl, QSL_RCP_CTRL);
  p->mpRcpAddr = Register_getField(p->mpRcpCtrl, QSL_RCP_CTRL_ADR);
  p->mpRcpCnt = Register_getField(p->mpRcpCtrl, QSL_RCP_CTRL_CNT);
  p->mpRcpData = Module_getRegister(p->mQsl, QSL_RCP_DATA);
  p->mpRcpDataDiscard = Register_getField(p->mpRcpData, QSL_RCP_DATA_DISCARD);
  p->mpRcpDataDrop = Register_getField(p->mpRcpData, QSL_RCP_DATA_DROP);
  p->mpRcpDataTblLo = Register_getField(p->mpRcpData, QSL_RCP_DATA_TBL_LO);
  p->mpRcpDataTblHi = Register_getField(p->mpRcpData, QSL_RCP_DATA_TBL_HI);
  p->mpRcpDataTblIdx = Register_getField(p->mpRcpData, QSL_RCP_DATA_TBL_IDX);
  p->mpRcpDataTblMsk = Register_getField(p->mpRcpData, QSL_RCP_DATA_TBL_MSK);
  p->mpRcpDataCao = Register_queryField(p->mpRcpData, QSL_RCP_DATA_CAO);
  p->mpRcpDataLr = Register_queryField(p->mpRcpData, QSL_RCP_DATA_LR);
  p->mpRcpDataTsa = Register_queryField(p->mpRcpData, QSL_RCP_DATA_TSA);
  p->mpRcpDataVli = Register_queryField(p->mpRcpData, QSL_RCP_DATA_VLI);

  // QST
  p->mpQstCtrl = Module_getRegister(p->mQsl, QSL_QST_CTRL);
  p->mpQstAddr = Register_getField(p->mpQstCtrl, QSL_QST_CTRL_ADR);
  p->mpQstCnt = Register_getField(p->mpQstCtrl, QSL_QST_CTRL_CNT);
  p->mpQstData = Module_getRegister(p->mQsl, QSL_QST_DATA);
  p->mpQstDataQueue = Register_getField(p->mpQstData, QSL_QST_DATA_QUEUE);
  p->mpQstDataEn = Register_queryField(p->mpQstData, QSL_QST_DATA_EN);
  p->mpQstDataTxPort = Register_queryField(p->mpQstData, QSL_QST_DATA_TX_PORT);
  p->mpQstDataLre = Register_queryField(p->mpQstData, QSL_QST_DATA_LRE);
  p->mpQstDataTci = Register_queryField(p->mpQstData, QSL_QST_DATA_TCI);
  p->mpQstDataVen = Register_queryField(p->mpQstData, QSL_QST_DATA_VEN);
  // QEN
  p->mpQenCtrl = Module_getRegister(p->mQsl, QSL_QEN_CTRL);
  p->mpQenAddr = Register_getField(p->mpQenCtrl, QSL_QEN_CTRL_ADR);
  p->mpQenCnt = Register_getField(p->mpQenCtrl, QSL_QEN_CTRL_CNT);
  p->mpQenData = Module_getRegister(p->mQsl, QSL_QEN_DATA);
  p->mpQenDataEn = Register_getField(p->mpQenData, QSL_QEN_DATA_EN);
  // UNMQ
  p->mpUnmqCtrl = Module_getRegister(p->mQsl, QSL_UNMQ_CTRL);
  p->mpUnmqAddr = Register_getField(p->mpUnmqCtrl, QSL_UNMQ_CTRL_ADR);
  p->mpUnmqCnt = Register_getField(p->mpUnmqCtrl, QSL_UNMQ_CTRL_CNT);
  p->mpUnmqData = Module_getRegister(p->mQsl, QSL_UNMQ_DATA);
  p->mpUnmqDataDestQueue = Register_getField(p->mpUnmqData, QSL_UNMQ_DATA_DEST_QUEUE);
  p->mpUnmqDataEn = Register_getField(p->mpUnmqData, QSL_UNMQ_DATA_EN);

  if (!p->mpQstDataEn) {
    // changed name from EN to QEN in v0.7
    p->mpQstDataEn = Register_getField(p->mpQstData, QSL_QST_DATA_QEN);
  }

  // LTX - not there anymore from v0.7+
  p->mpLtxCtrl = Module_queryRegister(p->mQsl, QSL_LTX_CTRL);
  if (p->mpLtxCtrl) {
    p->mpLtxAddr = Register_getField(p->mpLtxCtrl, QSL_LTX_CTRL_ADR);
    p->mpLtxCnt = Register_getField(p->mpLtxCtrl, QSL_LTX_CTRL_CNT);
  }
  else {
    p->mpLtxAddr = NULL;
    p->mpLtxCnt = NULL;
  }
  p->mpLtxData = Module_queryRegister(p->mQsl, QSL_LTX_DATA);
  if (p->mpLtxData) {
    p->mpLtxDataLr = Register_getField(p->mpLtxData, QSL_LTX_DATA_LR);
    p->mpLtxDataTxPort = Register_getField(p->mpLtxData, QSL_LTX_DATA_TX_PORT);
    p->mpLtxDataTsa = Register_getField(p->mpLtxData, QSL_LTX_DATA_TSA);
  }
  else {
    p->mpLtxDataLr = NULL;
    p->mpLtxDataTxPort = NULL;
    p->mpLtxDataTsa = NULL;
  }
  return 0;
}

int QslNthw_setup(struct QslNthw *p, int nIdx, int nIdxCnt)
{
  (void)p;
  (void)nIdx;
  (void)nIdxCnt;

  return 0;
}

// RCP
void QslNthw_RcpSelect(const struct QslNthw *p, uint32_t val)
{
  Field_setVal32(p->mpRcpAddr, val);
};

void QslNthw_RcpCnt(const struct QslNthw *p, uint32_t val)
{
  Field_setVal32(p->mpRcpCnt, val);
}

void QslNthw_RcpDiscard(const struct QslNthw *p, uint32_t val)
{
  Field_setVal32(p->mpRcpDataDiscard, val);
}

void QslNthw_RcpDrop(const struct QslNthw *p, uint32_t val)
{
  Field_setVal32(p->mpRcpDataDrop, val);
}

void QslNthw_RcpTblLo(const struct QslNthw *p, uint32_t val)
{
  Field_setVal32(p->mpRcpDataTblLo, val);
}
void QslNthw_RcpTblHi(const struct QslNthw *p, uint32_t val)

{
  Field_setVal32(p->mpRcpDataTblHi, val);
}
void QslNthw_RcpTblIdx(const struct QslNthw *p, uint32_t val)

{
  Field_setVal32(p->mpRcpDataTblIdx, val);
}

void QslNthw_RcpTblMsk(const struct QslNthw *p, uint32_t val)
{
  Field_setVal32(p->mpRcpDataTblMsk, val);
}

void QslNthw_RcpCao(const struct QslNthw *p, uint32_t val)
{
  if (p->mpRcpDataCao) {
    Field_setVal32(p->mpRcpDataCao, val);
  }
}

void QslNthw_RcpLr(const struct QslNthw *p, uint32_t val)
{
  if (p->mpRcpDataLr) {
    Field_setVal32(p->mpRcpDataLr, val);
  }
}

void QslNthw_RcpTsa(const struct QslNthw *p, uint32_t val)
{
  if (p->mpRcpDataTsa) {
    Field_setVal32(p->mpRcpDataTsa, val);
  }
}

void QslNthw_RcpVli(const struct QslNthw *p, uint32_t val)
{
  if (p->mpRcpDataVli) {
    Field_setVal32(p->mpRcpDataVli, val);
  }
}

void QslNthw_RcpFlush(const struct QslNthw *p)
{
  Register_flush(p->mpRcpCtrl, 1);
  Register_flush(p->mpRcpData, 1);
}

  // LTX
void QslNthw_LtxSelect(const struct QslNthw *p, uint32_t val)
{
  if (p->mpLtxAddr) {
    Field_setVal32(p->mpLtxAddr, val);
  }
};

void QslNthw_LtxCnt(const struct QslNthw *p, uint32_t val)
{
  if (p->mpLtxAddr) {
    Field_setVal32(p->mpLtxCnt, val);
  }
};

void QslNthw_LtxLr(const struct QslNthw *p, uint32_t val)
{
  if (p->mpLtxDataLr) {
    Field_setVal32(p->mpLtxDataLr, val);
  }
};

void QslNthw_LtxTxPort(const struct QslNthw *p, uint32_t val)
{
  if (p->mpLtxDataTxPort) {
    Field_setVal32(p->mpLtxDataTxPort, val);
  }
};

void QslNthw_LtxTsa(const struct QslNthw *p, uint32_t val)
{
  if (p->mpLtxDataTsa) {
    Field_setVal32(p->mpLtxDataTsa, val);
  }
};

void QslNthw_LtxFlush(const struct QslNthw *p)
{
  Register_flush(p->mpLtxCtrl, 1);
  Register_flush(p->mpLtxData, 1);
};

// QST
void QslNthw_QstSelect(const struct QslNthw *p, uint32_t val)
{
  Field_setVal32(p->mpQstAddr, val);
};

void QslNthw_QstCnt(const struct QslNthw *p, uint32_t val)
{
  Field_setVal32(p->mpQstCnt, val);
};

void QslNthw_QstQueue(const struct QslNthw *p, uint32_t val)
{
  Field_setVal32(p->mpQstDataQueue, val);
};

void QslNthw_QstEn(const struct QslNthw *p, uint32_t val)
{
  Field_setVal32(p->mpQstDataEn, val);
};

void QslNthw_QstTxPort(const struct QslNthw *p, uint32_t val)
{
  if (p->mpQstDataTxPort) {
    Field_setVal32(p->mpQstDataTxPort, val);
  }
};

void QslNthw_QstLre(const struct QslNthw *p, uint32_t val)
{
  if (p->mpQstDataLre) {
    Field_setVal32(p->mpQstDataLre, val);
  }
};

void QslNthw_QstTci(const struct QslNthw *p, uint32_t val)
{
  if (p->mpQstDataTci) {
    Field_setVal32(p->mpQstDataTci, val);
  }
};

void QslNthw_QstVen(const struct QslNthw *p, uint32_t val)
{
  if (p->mpQstDataVen) {
    Field_setVal32(p->mpQstDataVen, val);
  }
};

void QslNthw_QstFlush(const struct QslNthw *p)
{
  Register_flush(p->mpQstCtrl, 1);
  Register_flush(p->mpQstData, 1);
};

// QEN
void QslNthw_QenSelect(const struct QslNthw *p, uint32_t val)
{
  Field_setVal32(p->mpQenAddr, val);
};

void QslNthw_QenCnt(const struct QslNthw *p, uint32_t val)
{
  Field_setVal32(p->mpQenCnt, val);
};

void QslNthw_QenEn(const struct QslNthw *p, uint32_t val)
{
  Field_setVal32(p->mpQenDataEn, val);
};

void QslNthw_QenFlush(const struct QslNthw *p)
{
  Register_flush(p->mpQenCtrl, 1);
  Register_flush(p->mpQenData, 1);
};

// UNMQ
void QslNthw_UnmqSelect(const struct QslNthw *p, uint32_t val)
{
  Field_setVal32(p->mpUnmqAddr, val);
};

void QslNthw_UnmqCnt(const struct QslNthw *p, uint32_t val)
{
  Field_setVal32(p->mpUnmqCnt, val);
};

void QslNthw_UnmqDestQueue(const struct QslNthw *p, uint32_t val)
{
  Field_setVal32(p->mpUnmqDataDestQueue, val);
};

void QslNthw_UnmqEn(const struct QslNthw *p, uint32_t val)
{
  Field_setVal32(p->mpUnmqDataEn, val);
};

void QslNthw_UnmqFlush(const struct QslNthw *p)
{
  Register_flush(p->mpUnmqCtrl, 1);
  Register_flush(p->mpUnmqData, 1);
};
