/*
 * %NT_SOFTWARE_LICENSE%
 */

//
// Includes:
//
#include "ntlog.h"
//
#include "nthw_drv.h"
#include "nthw_register.h"

//
//
//
#include "flow_nthw_roa.h"

#include <stdlib.h> // malloc
#include <string.h> // memset

void roaNthw_setDebugMode(struct RoaNthw *p, unsigned nDebugMode)
{
  Module_setDebugMode(p->mRoa, nDebugMode);
}

struct RoaNthw *RoaNthw_new(void)
{
  struct RoaNthw *p = malloc(sizeof(struct RoaNthw));
  if (p)
    (void)memset(p, 0, sizeof(*p));

  return p;
}

void RoaNthw_delete(struct RoaNthw *p)
{
  if (p) {
    (void)memset(p, 0, sizeof(*p));
    free(p);
  }
}

int RoaNthw_init(struct RoaNthw *p, Fpga_t *pFpga, int nInstance)
{
  const char *const pAdapterIdStr = pFpga->pFpgaInfo->mpAdapterIdStr;
  struct Module_s* pMod = Fpga_queryModule(pFpga, MOD_ROA, nInstance);

  assert(nInstance >= 0 && nInstance < 256);

  if (p == NULL) {
    return pMod == NULL ? -1 : 0;
  }

  if (pMod == NULL) {
    NT_LOG(ERR, NTHW, "%s: ROA %d: no such instance\n", pAdapterIdStr, nInstance);
    return -1;
  }

  p->mpFpga = pFpga;
  p->mPhysicalAdapterNo = (uint8_t)nInstance;
  p->mRoa = pMod;

#if defined(DEBUG) && (0)
  Module_setDebugMode(p->mRoa, 0xFF); // DEBUG:
#endif

  // TUN HDR
  p->mpTunHdrCtrl = Module_getRegister(p->mRoa, ROA_TUNHDR_CTRL);
  p->mpTunHdrAddr = Register_getField(p->mpTunHdrCtrl, ROA_TUNHDR_CTRL_ADR);
  p->mpTunHdrCnt = Register_getField(p->mpTunHdrCtrl, ROA_TUNHDR_CTRL_CNT);
  p->mpTunHdrData = Module_getRegister(p->mRoa, ROA_TUNHDR_DATA);
  p->mpTunHdrDataTunnelHdr =     Register_getField(p->mpTunHdrData, ROA_TUNHDR_DATA_TUNNEL_HDR);
  // TUN CFG
  p->mpTunCfgCtrl = Module_getRegister(p->mRoa, ROA_TUNCFG_CTRL);
  p->mpTunCfgAddr = Register_getField(p->mpTunCfgCtrl, ROA_TUNCFG_CTRL_ADR);
  p->mpTunCfgCnt = Register_getField(p->mpTunCfgCtrl, ROA_TUNCFG_CTRL_CNT);
  p->mpTunCfgData = Module_getRegister(p->mRoa, ROA_TUNCFG_DATA);
  p->mpTunCfgDataTunLen = Register_getField(p->mpTunCfgData, ROA_TUNCFG_DATA_TUN_LEN);
  p->mpTunCfgDataTunType = Register_getField(p->mpTunCfgData, ROA_TUNCFG_DATA_TUN_TYPE);
  p->mpTunCfgDataTunVlan = Register_getField(p->mpTunCfgData, ROA_TUNCFG_DATA_TUN_VLAN);
  p->mpTunCfgDataIpType = Register_getField(p->mpTunCfgData, ROA_TUNCFG_DATA_TUN_IP_TYPE);
  p->mpTunCfgDataIpcsUpd = Register_getField(p->mpTunCfgData, ROA_TUNCFG_DATA_TUN_IPCS_UPD);
  p->mpTunCfgDataIpcsPrecalc = Register_getField(p->mpTunCfgData, ROA_TUNCFG_DATA_TUN_IPCS_PRECALC);
  p->mpTunCfgDataIptlUpd = Register_getField(p->mpTunCfgData, ROA_TUNCFG_DATA_TUN_IPTL_UPD);
  p->mpTunCfgDataIptlPrecalc = Register_getField(p->mpTunCfgData, ROA_TUNCFG_DATA_TUN_IPTL_PRECALC);
  p->mpTunCfgDataVxlanUdpLenUpd = Register_getField(p->mpTunCfgData, ROA_TUNCFG_DATA_TUN_VXLAN_UDP_LEN_UPD);
  p->mpTunCfgDataTxLagIx = Register_getField(p->mpTunCfgData, ROA_TUNCFG_DATA_TX_LAG_IX);
  p->mpTunCfgDataRecirculate = Register_getField(p->mpTunCfgData, ROA_TUNCFG_DATA_RECIRCULATE);
  p->mpTunCfgDataPushTunnel = Register_getField(p->mpTunCfgData, ROA_TUNCFG_DATA_PUSH_TUNNEL);
  p->mpTunCfgDataRecircPort = Register_getField(p->mpTunCfgData, ROA_TUNCFG_DATA_RECIRC_PORT);
  p->mpTunCfgDataRecircBypass = Register_getField(p->mpTunCfgData, ROA_TUNCFG_DATA_RECIRC_BYPASS);
  // CONFIG
  p->mpConfig = Module_getRegister(p->mRoa, ROA_CONFIG);
  p->mpConfigFwdRecirculate = Register_getField(p->mpConfig, ROA_CONFIG_FWD_RECIRCULATE);
  p->mpConfigFwdNormalPcks = Register_getField(p->mpConfig, ROA_CONFIG_FWD_NORMAL_PCKS);
  p->mpConfigFwdTxPort0 = Register_getField(p->mpConfig, ROA_CONFIG_FWD_TXPORT0);
  p->mpConfigFwdTxPort1 = Register_getField(p->mpConfig, ROA_CONFIG_FWD_TXPORT1);
  p->mpConfigFwdCellBuilderPcks = Register_getField(p->mpConfig, ROA_CONFIG_FWD_CELLBUILDER_PCKS);
  p->mpConfigFwdNonNormalPcks = Register_getField(p->mpConfig, ROA_CONFIG_FWD_NON_NORMAL_PCKS);
  // LAG
  p->mpLagCfgCtrl = Module_getRegister(p->mRoa, ROA_LAGCFG_CTRL);
  p->mpLagCfgAddr = Register_getField(p->mpLagCfgCtrl, ROA_LAGCFG_CTRL_ADR);
  p->mpLagCfgCnt = Register_getField(p->mpLagCfgCtrl, ROA_LAGCFG_CTRL_CNT);
  p->mpLagCfgData = Module_getRegister(p->mRoa, ROA_LAGCFG_DATA);
  p->mpLagCfgDataTxPhyPort = Register_getField(p->mpLagCfgData, ROA_LAGCFG_DATA_TXPHY_PORT);

  return 0;
}

// TUN HDR
void RoaNthw_TunHdrSelect(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpTunHdrAddr, val);
};

void RoaNthw_TunHdrCnt(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpTunHdrCnt, val);
};

void RoaNthw_TunHdrTunnelHdr(const struct RoaNthw *p, uint32_t *val)
{
  Field_setVal(p->mpTunHdrDataTunnelHdr, val, 4);
};

void RoaNthw_TunHdrFlush(const struct RoaNthw *p)
{
  Register_flush(p->mpTunHdrCtrl, 1);
  Register_flush(p->mpTunHdrData, 1);
};

// TUN CFG
void RoaNthw_TunCfgSelect(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpTunCfgAddr, val);
};

void RoaNthw_TunCfgCnt(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpTunCfgCnt, val);
};

void RoaNthw_TunCfgTunLen(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpTunCfgDataTunLen, val);
};

void RoaNthw_TunCfgTunType(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpTunCfgDataTunType, val);
};

void RoaNthw_TunCfgTunVlan(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpTunCfgDataTunVlan, val);
};

void RoaNthw_TunCfgIpType(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpTunCfgDataIpType, val);
};

void RoaNthw_TunCfgIpcsUpd(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpTunCfgDataIpcsUpd, val);
};

void RoaNthw_TunCfgIpcsPrecalc(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpTunCfgDataIpcsPrecalc, val);
};

void RoaNthw_TunCfgIptlUpd(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpTunCfgDataIptlUpd, val);
};

void RoaNthw_TunCfgIptlPrecalc(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpTunCfgDataIptlPrecalc, val);
};

void RoaNthw_TunCfgVxlanUdpLenUpd(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpTunCfgDataVxlanUdpLenUpd, val);
};

void RoaNthw_TunCfgTxLagIx(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpTunCfgDataTxLagIx, val);
};

void RoaNthw_TunCfgRecirculate(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpTunCfgDataRecirculate, val);
};

void RoaNthw_TunCfgPushTunnel(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpTunCfgDataPushTunnel, val);
};

void RoaNthw_TunCfgRecircPort(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpTunCfgDataRecircPort, val);
};

void RoaNthw_TunCfgRecircBypass(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpTunCfgDataRecircBypass, val);
};

void RoaNthw_TunCfgFlush(const struct RoaNthw *p)
{
  Register_flush(p->mpTunCfgCtrl, 1);
  Register_flush(p->mpTunCfgData, 1);
};

// ROA CONFIG
void RoaNthw_ConfigFwdRecirculate(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpConfigFwdRecirculate, val);
};

void RoaNthw_ConfigFwdNormalPcks(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpConfigFwdNormalPcks, val);
};

void RoaNthw_ConfigFwdTxPort0(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpConfigFwdTxPort0, val);
};

void RoaNthw_ConfigFwdTxPort1(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpConfigFwdTxPort1, val);
};

void RoaNthw_ConfigFwdCellBuilderPcks(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpConfigFwdCellBuilderPcks, val);
};

void RoaNthw_ConfigFwdNonNormalPcks(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpConfigFwdNonNormalPcks, val);
};

void RoaNthw_ConfigFlush(const struct RoaNthw *p)
{
  Register_flush(p->mpConfig, 1);
};

// LAG
void RoaNthw_LagCfgSelect(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpLagCfgAddr, val);
};

void RoaNthw_LagCfgCnt(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpLagCfgCnt, val);
};

void RoaNthw_LagCfgTxPhyPort(const struct RoaNthw *p, uint32_t val)
{
  Field_setVal32(p->mpLagCfgDataTxPhyPort, val);
};

void RoaNthw_LagCfgFlush(const struct RoaNthw *p)
{
  Register_flush(p->mpLagCfgCtrl, 1);
  Register_flush(p->mpLagCfgData, 1);
};
