//
// nthw_fpga_model.c
//

/*
 * %NT_SOFTWARE_LICENSE%
 */

#include <time.h> // ctime

#include "nthw_drv.h"  // FpgaInfo_s
#include "nthw_register.h"
#include "nthw_fpga_model.h"
#include "nthw_rac.h"
#include "nthw_debug.h"
#include "ntlog.h"


//
// generated code
//
#include "nthw_fpga_instances.c"

//
// generated code
//
#include "nthw_fpga_modules.c"

//
// NOTE: this needs to be (manually) synced with enum
//
static const char* const aBusType[] = {
  "ERR",  // BUS_TYPE_UNKNOWN,
  "BAR",  // BUS_TYPE_BAR,
  "PCI",  // BUS_TYPE_PCI,
  "CCIP", // BUS_TYPE_CCIP,
  "RAB0", // BUS_TYPE_RAB0,
  "RAB1", // BUS_TYPE_RAB1,
  "RAB2", // BUS_TYPE_RAB2,
  "NMB",  // BUS_TYPE_NMB,
  "NDM",  // BUS_TYPE_NDM,
};

//
//
//
static const char* getBusName(int nBusTypeId)
{
  if ((nBusTypeId >= 1) && (nBusTypeId <= (int)ARRAY_SIZE(aBusType))) {
    return aBusType[nBusTypeId];
  } else {
    return "ERR";
  }
}


//
// module name lookup by id from array
//
// Uses naive linear search as performance is not an issue here...
//
static const char* nthwFpgaModIdToStr(uint64_t nFpgaModId)
{
  int i;
  for (i = 0; i <= (int)ARRAY_SIZE(saNthwFpgaModMap); i++) {
    if ((uint64_t)saNthwFpgaModMap[i].a == nFpgaModId) {
      break;
    }
  }
  return (saNthwFpgaModMap[i].b ? saNthwFpgaModMap[i].b : "unknown");
}

//
// force C linkage for xxx_addr_bases and xxx_module_versions
//
#if (0)
#ifdef __cplusplus
extern "C" {
#endif

// nthw_gen1_legacy.h
#if !defined(_NTSD_MAX_NUM_PHYS_ADAPTERS_)
#define _NTSD_MAX_NUM_PHYS_ADAPTERS_ (16)
#endif // _NTSD_MAX_NUM_PHYS_ADAPTERS_

// nthw_gen1_legacy.h
#include "nthw_gen1_legacy.h"

//
uint32_t nthw_adapter_module_addr_bases[_NTSD_MAX_NUM_PHYS_ADAPTERS_][MOD_UNKNOWN_MAX][NTHW_FPGA_MODULE_INSTANCES_MAX];
//
nthw_gen1_module_version_t nthw_adapter_module_versions[_NTSD_MAX_NUM_PHYS_ADAPTERS_][(NTHW_FPGA_MODULE_BUS_COUNT * MOD_UNKNOWN_MAX)];

#ifdef __cplusplus
} // extern "C"
#endif
#endif

//
//
//
static int ReadData(struct FpgaInfo_s* pFpgaInfo, int nBusTypeId, uint32_t addr, uint32_t len, uint32_t* pData)
{
  int rc = -1;

  assert(pFpgaInfo);
  assert(pData);

  switch(nBusTypeId) {
  case BUS_TYPE_BAR:
  case BUS_TYPE_PCI:
    assert(len==1);
    NthwRac_regRead32(pFpgaInfo, addr, pData);
    rc = 0;
    break;
  case BUS_TYPE_RAB0:
    //NthwRac_rab_read32(NthwRac_t* p, uint32_t address, RabBusId_t busId, uint32_t wordCnt, uint32_t* pData, char* pErr);
    assert(pFpgaInfo->mpNthwRac);
    rc = NthwRac_rab_read32(pFpgaInfo->mpNthwRac, addr, 0, len, pData);
    break;
  case BUS_TYPE_RAB1:
    assert(pFpgaInfo->mpNthwRac);
    rc = NthwRac_rab_read32(pFpgaInfo->mpNthwRac, addr, 1, len, pData);
    break;
  case BUS_TYPE_RAB2:
    assert(pFpgaInfo->mpNthwRac);
    rc = NthwRac_rab_read32(pFpgaInfo->mpNthwRac, addr, 2, len, pData);
    break;
  default:
    assert(false);
    return -1;
  }

  return rc;
}

//
//
//
int DebugOnlyNthwReadData(struct FpgaInfo_s* pFpgaInfo, int nBusTypeId, uint32_t addr, uint32_t len, uint32_t* pData)
{
  return ReadData(pFpgaInfo, nBusTypeId, addr, len, pData);
}

//
//
//
static int ReadDataTsc(struct FpgaInfo_s* pFpgaInfo, int nBusTypeId, uint32_t addr, uint32_t len, uint32_t* pData, uint64_t* pTsc1, uint64_t* pTsc2)
{
  int rc = -1;

  (void) pTsc1;
  (void) pTsc2;

  rc = ReadData(pFpgaInfo, nBusTypeId, addr, len, pData);

  return rc;
}

//
//
//
static int WriteData(struct FpgaInfo_s* pFpgaInfo, int nBusTypeId, uint32_t addr, uint32_t len, const uint32_t* pData)
{
  int rc = -1;
  assert(pFpgaInfo);
  assert(pData);

  switch(nBusTypeId) {
  case BUS_TYPE_BAR:
  case BUS_TYPE_PCI:
    assert(len==1);
    NthwRac_regWrite32(pFpgaInfo, addr, *pData);
    rc = 0;
    break;
  case BUS_TYPE_RAB0:
    assert(pFpgaInfo->mpNthwRac);
    rc = NthwRac_rab_write32(pFpgaInfo->mpNthwRac, addr, 0, len, pData);
    break;
  case BUS_TYPE_RAB1:
    assert(pFpgaInfo->mpNthwRac);
    rc = NthwRac_rab_write32(pFpgaInfo->mpNthwRac, addr, 1, len, pData);
    break;
  case BUS_TYPE_RAB2:
    assert(pFpgaInfo->mpNthwRac);
    rc = NthwRac_rab_write32(pFpgaInfo->mpNthwRac, addr, 2, len, pData);
    break;
  default:
    assert(false);
    return -1;
  }

  return rc;
}

//
//
//
int DebugOnlyNthwWriteData(struct FpgaInfo_s* pFpgaInfo, int nBusTypeId, uint32_t addr, uint32_t len, const uint32_t* pData)
{
  return WriteData(pFpgaInfo, nBusTypeId, addr, len, pData);
}

//
// FpgaMgr
//
FpgaMgr_t* FpgaMgr_new(void)
{
  FpgaMgr_t* p = malloc(sizeof(FpgaMgr_t));
  return p;
}

//
//
//
void FpgaMgr_delete(FpgaMgr_t* p)
{
  memset(p, 0, sizeof(FpgaMgr_t));
  free(p);
}

//
//
//
void FpgaMgr_init(FpgaMgr_t* p)
{
  size_t i;

  // count fpga instance in array
  p->mpaFpgaProdInit = nthw_fpga_instances;
  for (i = 0; i < ARRAY_SIZE(nthw_fpga_instances); i++) {
    if (p->mpaFpgaProdInit[i] == NULL) {
      break;
    }
  }
  p->mnFpgas = (int) i;
#if defined(DEBUG) && (0)
  NT_LOG(DBG, NTHW, "%s: fpgas=%d\n", __func__, p->mnFpgas);
#endif
}

//
//
//
Fpga_t* FpgaMgr_queryFpga(FpgaMgr_t* p, uint64_t nFpgaId, struct FpgaInfo_s* pFpgaInfo)
{
  int i;

  const int nFpgaProd = FPGAID_TO_PRODUCTCODE(nFpgaId);
  const int nFpgaVer = FPGAID_TO_VERSIONCODE(nFpgaId);
  const int nFpgaRev = FPGAID_TO_REVISIONCODE(nFpgaId);

  for (i = 0; i < p->mnFpgas; i++) {
    FpgaProdInit_t* pInit = p->mpaFpgaProdInit[i];
    if (pInit->fpgaProductId == nFpgaProd && pInit->fpgaVersion == nFpgaVer && pInit->fpgaRevision == nFpgaRev ) {
      //NT_LOG(DBG, NTHW, "FPGA Id 0x%"PRIX64": %04d-%02d-%02d: match found\n", nFpgaId, nFpgaProd, nFpgaVer, nFpgaRev);
      {
        Fpga_t* pFpga = Fpga_new();
        Fpga_init(pFpga, pInit, pFpgaInfo);
        return pFpga;
      }
    }
  }

  NT_LOG(ERR, NTHW, "FPGA Id 0x%"PRIX64": %04d: %d.%d: no match found\n", nFpgaId, nFpgaProd, nFpgaVer, nFpgaRev);

  return NULL;
}

//
//
//
void FpgaMgr_show(FpgaMgr_t* p, FILE* fh_out, int detail_level)
{
  int i;
  fprintf(fh_out, "\n"); // start of records
  for (i = 0; i < p->mnFpgas; i++) {
    FpgaProdInit_t* pInit = p->mpaFpgaProdInit[i];
    if (detail_level == 0) {
      fprintf(fh_out, "%04d-%02d-%02d\n", pInit->fpgaProductId, pInit->fpgaVersion, pInit->fpgaRevision);
    } else {
      time_t fpgaBuildTime = pInit->fpgaBuildTime;
      fprintf(fh_out, "%04d-%02d-%02d: 0x%08lX: %s\n", pInit->fpgaProductId, pInit->fpgaVersion, pInit->fpgaRevision, fpgaBuildTime, (fpgaBuildTime ? ctime(&fpgaBuildTime) : "NA\n"));
    }
  }
  fprintf(fh_out, "\n"); // end of records
  fflush(fh_out);
}

//
//
//
void FpgaMgr_log_dump(FpgaMgr_t* p)
{
  int i;

  NT_LOG(DBG, NTHW, "%s: fpgas=%d\n", __func__, p->mnFpgas);
  for (i = 0; i < p->mnFpgas; i++) {
    FpgaProdInit_t* pInit _unused = p->mpaFpgaProdInit[i];
    NT_LOG(DBG, NTHW, "%s: fpga=%d/%d: %04d-%02d-%02d\n", __func__, i, p->mnFpgas, pInit->fpgaProductId, pInit->fpgaVersion, pInit->fpgaRevision);
  }
}


//
// Fpga
//
Fpga_t* Fpga_new(void)
{
  Fpga_t* p = malloc(sizeof(Fpga_t));
  if (p) {
    memset(p, 0, sizeof(Fpga_t));
  }
  return p;
}

//
//
//
void Fpga_delete(Fpga_t* p)
{
  memset(p, 0, sizeof(Fpga_t));
  free(p);
}

//
//
//
void Fpga_delete_all(Fpga_t* p)
{
  int i;

  for (i = 0; i < p->mnModules; i++) {
    Module_t* pMod = p->mpaModules[i];
    if (pMod) {
      Module_delete(pMod);
    }
  }

  Fpga_delete(p);
}

//
//
//
void Fpga_init(Fpga_t* p, FpgaProdInit_t* fpgaProdInit, struct FpgaInfo_s* pFpgaInfo)
{
  int i;
  //
  p->pFpgaInfo = pFpgaInfo;
  //
  p->mpInit = fpgaProdInit;
  //
  p->mItemId = fpgaProdInit->fpgaItemId;
  p->mProductId = fpgaProdInit->fpgaProductId;
  p->mFpgaVersion = fpgaProdInit->fpgaVersion;
  p->mFpgaRevision = fpgaProdInit->fpgaRevision;
  p->mFpgaPatchNo = fpgaProdInit->fpgaPatchNo;
  p->mFpgaBuildNo = fpgaProdInit->fpgaBuildNo;
  p->mFpgaBuildTime = fpgaProdInit->fpgaBuildTime;
  //
  p->mnParams = fpgaProdInit->nbProdParams;
  //
  if (p->mnParams) {
    p->mpaParams = malloc(p->mnParams*sizeof(Param_t*));
    if (p->mpaParams) {
      memset(p->mpaParams, 0, (p->mnParams*sizeof(Param_t*)));
      for (i = 0; i < p->mnParams; i++) {
        Param_t* pParam = Param_new();
        Param_init(pParam, p, &fpgaProdInit->productParams[i]);
        p->mpaParams[i] = pParam;
      }
    }
  }
  //
  p->mnModules = fpgaProdInit->nbModules;
  //
  if (p->mnModules) {
    p->mpaModules = malloc(fpgaProdInit->nbModules*sizeof(Fpga_t*));
    if (p->mpaModules) {
      memset(p->mpaModules, 0, (p->mnModules*sizeof(Fpga_t*)));
      for (i = 0; i < p->mnModules; i++) {
        Module_t* pMod = Module_new();
        Module_init(pMod, p, &fpgaProdInit->modules[i]);
        p->mpaModules[i] = pMod;
      }
    }
  }
}

//
//
//
void Fpga_setDebugMode(Fpga_t* p, int nDebugMode)
{
  int i;

  p->mDebugMode = nDebugMode;

  for (i = 0; i < p->mnModules; i++) {
    Module_t* pMod = p->mpaModules[i];
    if (pMod) {
      Module_setDebugMode(pMod, nDebugMode);
    }
  }
}

//
//
//
Module_t* Fpga_queryModule(const Fpga_t* p, int id, int instance)
{
  int i;

  for (i = 0; i < p->mnModules; i++) {
    Module_t* pMod = p->mpaModules[i];
    if (pMod->mModId == id && pMod->mInstance == instance) {
      return pMod;
    }
  }
  // NOTE:
  // No log messages from this function as it is used to probe if a module is present or not
  // If log messages are needed - do the log messages where the actual actual probing takes place
  return NULL;
}

//
//
//
bool Fpga_query(Fpga_t* p, int id, int instance)
{
  return (Fpga_queryModule(p, id, instance) != NULL);
}

//
//
//
FpgaModuleInit_t* Fpga_lookup_init(Fpga_t* p, int id, int instance)
{
  int i;

  for (i = 0; i < p->mpInit->nbModules; i++) {
    FpgaModuleInit_t* pModInit = &p->mpInit->modules[i];
    if (pModInit->id == id && pModInit->instance == instance) {
      return pModInit;
    }
  }
  //throw NtException(NT_ERROR_INTERNAL_ERROR, "Module '%s' (mod_id=%d mod_instance=%d) not found in: %s", nthwFpgaModIdToStr(id).c_str(), id, instance, toString().c_str());
  return NULL;

}

//
//
//
int Fpga_getProductParam(const Fpga_t* p, const int nParamId, const int nDefaultValue)
{
  int i;

  for (i = 0; i < p->mnParams; i++) {
    Param_t* pParam = p->mpaParams[i];
    if (pParam->paramId == nParamId) {
      return pParam->paramValue;
    }
  }

  return nDefaultValue;
}

//
//
//
int Fpga_getProductId(const Fpga_t* p)
{
  return p->mProductId;
}

//
//
//
int Fpga_getFpgaVersion(const Fpga_t* p)
{
  return p->mFpgaVersion;
}

//
//
//
int Fpga_getFpgaRevision(const Fpga_t* p)
{
  return p->mFpgaRevision;
}

//
//
//
void Fpga_logInfo(const Fpga_t* p _unused)
{
  NT_LOG(INF, NTHW, "FPGA: %d-%d-%d-%d-%d-%d (%08X)\n", p->mItemId, p->mProductId, p->mFpgaVersion, p->mFpgaRevision, p->mFpgaPatchNo, p->mFpgaBuildNo, p->mFpgaBuildTime);
}

//
//
//
void Fpga_dump(const Fpga_t* p)
{
  NT_LOG(DBG, NTHW, "%s: id=%d ver=%d.%d params=%d modules=%d\n", __func__, p->mProductId, p->mFpgaVersion, p->mFpgaRevision, p->mnParams, p->mnModules);
  Fpga_dumpParams(p);
  Fpga_dumpModules(p);
}

//
//
//
void Fpga_dumpParams(const Fpga_t* p)
{
  int i;
  NT_LOG(DBG, NTHW, "%s: params=%d\n", __func__, p->mnParams);

  for (i = 0; i < p->mnParams; i++) {
    Param_t* pPar = p->mpaParams[i];
    Param_dump(pPar);
  }
}

//
//
//
void Fpga_dumpModules(const Fpga_t* p)
{
  int i;
  NT_LOG(DBG, NTHW, "%s: modules=%d\n", __func__, p->mnModules);

  for (i = 0; i < p->mnModules; i++) {
    Module_t* pMod = p->mpaModules[i];
    Module_dump(pMod);
  }
}

//
// Param
//
Param_t* Param_new(void)
{
  Param_t* p = malloc(sizeof(Param_t));
  return p;
}

//
//
//
void Param_delete(Param_t* p)
{
  if (p) {
    memset(p, 0, sizeof(Param_t));
    free(p);
  }
}

//
//
//
void Param_init(Param_t* p, Fpga_t* pFpga, FpgaProdParam_t* pInit)
{
  p->mpOwner = pFpga;
  //
  p->mpInit = pInit;
  //
  p->paramId = pInit->paramId;
  p->paramValue = pInit->paramValue;
}

//
//
//
void Param_dump(const Param_t * p _unused)
{
  NT_LOG(DBG, NTHW, "%s: id=%d value=%d\n", __func__, p->paramId, p->paramValue);
}


//
// Module
//
Module_t* Module_new(void)
{
  Module_t* p = malloc(sizeof(Module_t));
  return p;
}

//
//
//
void Module_delete(Module_t* p)
{
  int i;
  for (i = 0; i < p->mnRegisters; i++) {
    Register_t* pReg = p->mpaRegisters[i];
    if (pReg) {
      Register_delete(pReg);
    }
  }
  memset(p, 0, sizeof(Module_t));
  free(p);
}

//
//
//
void Module_init(Module_t* p, Fpga_t* pFpga, FpgaModuleInit_t* pInit)
{
  int i;

  //
  p->mpOwner = pFpga;
  p->mpInit = pInit;
  //
  p->mModId = pInit->id;
  p->mInstance = pInit->instance;

  // Copy debug mode from owner
  if (p->mpOwner) {
    p->mDebugMode = p->mpOwner->mDebugMode;
  } else {
    p->mDebugMode = 0;
  }
  //
  p->mModDefId = pInit->defId;
  p->mMajorVersion = pInit->majorVersion;
  p->mMinorVersion = pInit->minorVersion;
  p->mBus = pInit->busId; // NthwBus_getInst(physicalAdapterNo, convBusType(pInit->busId))),
  p->mAddrBase = pInit->addrBase;
  //
  p->mnRegisters = pInit->nbRegisters;
  if (p->mnRegisters) {
    p->mpaRegisters = malloc(p->mnRegisters*sizeof(Register_t*));
    if (p->mpaRegisters) {
      memset(p->mpaRegisters, 0, (p->mnRegisters*sizeof(Register_t*)));
      for (i = 0; i < p->mnRegisters; i++) {
        Register_t* pReg = Register_new();
        Register_init(pReg, p, &pInit->registers[i]);
        p->mpaRegisters[i] = pReg;
      }
    }
  }
}

//
//
//
void Module_init2(Module_t* p, Fpga_t* pFpga, int modId, int instance, int debugMode)
{
  FpgaModuleInit_t* pInit = NULL;
  pInit = Fpga_lookup_init(pFpga, modId, instance);
  Module_init(p, pFpga, pInit);

  // set debug mode after regulat init...
  p->mDebugMode = debugMode;

}


//
//
//
void Module_dump(const Module_t* p)
{
  NT_LOG(DBG, NTHW, "%s: id=%d inst=%d def=%d ver=%d.%d busid=%d base=0x%X regs=%d\n", __func__, p->mModId, p->mInstance, p->mModDefId, p->mMajorVersion, p->mMinorVersion, p->mBus, p->mAddrBase, p->mnRegisters);
  Module_dumpRegisters(p);
}

//
//
//
void Module_dumpRegisters(const Module_t* p)
{
  int i;
  NT_LOG(DBG, NTHW, "%s: regs=%d\n", __func__, p->mnRegisters);

  for (i = 0; i < p->mnRegisters; i++) {
    Register_t* pReg = p->mpaRegisters[i];
    Register_dump(pReg);
  }
}

//
//
//
#if 0
std_string Module_toString(Module_t* p)
{

  std_string s;

  s = mrOwner.toString();

  s.append(" mod_bus_id=" + std_to_string(mpInit->busId));
  s.append(" mod_bus_name=");
  s.append(nthwFpgaBusTypeIdToStr(mpInit->busId));

  s.append(" mod_id=" + std_to_string(mpInit->id));
  s.append(" mod_name=");
  s.append(nthwFpgaModIdToStr(mpInit->id));
  s.append(" mod_instance=" + std_to_string(mpInit->instance));

  s.append(" mod_def_id=" + std_to_string(mpInit->defId));
  s.append(" mod_def_name=");
  s.append(nthwFpgaModIdToStr(mpInit->defId));
  s.append(" mod_version=" + std_to_string(mpInit->majorVersion) + "." + std_to_string(mpInit->minorVersion));

  return s;
}
#endif


//
//
//
int Module_getMajorVersion(const Module_t* p)
{
  return p->mMajorVersion;
}

//
//
//
int Module_getMinorVersion(const Module_t* p)
{
  return p->mMinorVersion;
}

//
//
//
uint64_t Module_getVersionPacked64(const Module_t* p)
{
  return (((uint64_t)p->mMajorVersion & 0xFFFFFFFF)<<32) | (p->mMinorVersion & 0xFFFFFFFF);
}

//
//
//
bool Module_isVersionNewer(const Module_t* p, int majorVersion, int minorVersion)
{
  if (majorVersion == p->mMajorVersion) {
    return p->mMinorVersion >= minorVersion;
  }
  return p->mMajorVersion >= majorVersion;
}

//
//
//
static Register_t* Module_lookupRegister(Module_t* p, uint32_t id)
{
  int i;
  Register_t* pRegister = NULL;

  for (i = 0; i < p->mnRegisters; i++) {
    if (p->mpaRegisters[i]->mId == id) {
      pRegister = p->mpaRegisters[i];
      break;
    }
  }
  return pRegister;
}
//
//
//
Register_t* Module_getRegister(Module_t* p, uint32_t id)
{
  Register_t* pRegister;

  if (p == NULL) {
    NT_LOG(ERR, NTHW, "Illegal module context for register %d\n", id);
    return NULL;
  }

  pRegister = Module_lookupRegister(p, id);
  if (!pRegister) {
    //throw NtException(NT_ERROR_INTERNAL_ERROR, "Register %d not found in module: %s", id, s.c_str());
    NT_LOG(ERR, NTHW, "Register %d not found in module: %s (%d)\n", id, nthwFpgaModIdToStr(p->mModId), p->mModId);
  }
  return pRegister;
}

//
//
//
Register_t* Module_queryRegister(Module_t* p, uint32_t id)
{
  return Module_lookupRegister(p, id);
}

//
//
//
int Module_getDebugMode(const Module_t* p)
{
  return p->mDebugMode;
}

//
//
//
void Module_setDebugMode(Module_t* p, unsigned nDebugMode)
{
  int i;
  Register_t* pRegister = NULL;

  p->mDebugMode = nDebugMode;

  for (i = 0; i < p->mnRegisters; i++) {
    pRegister = p->mpaRegisters[i];
    if (pRegister) {
      Register_setDebugMode(pRegister, nDebugMode);
    }
  }
}

//
//
//
int Module_getBus(const Module_t* p)
{
  return p->mBus;
}

//
//
//
uint32_t Module_getAddrBase(const Module_t* p)
{
  return p->mAddrBase;
}

//
//
//
void Module_unsuppported(const Module_t* p)
{
  NT_LOG(ERR, NTHW, "Module %d not supported", p->mpInit->id);
}


//
// Register
//
Register_t* Register_new(void)
{
  Register_t* p = malloc(sizeof(Register_t));
  return p;
}

//
//
//
void Register_delete(Register_t* p)
{
  int i;

  for (i = 0; i < p->mnFields; i++) {
    Field_t* pField = p->mpaFields[i];
    if (pField) {
      Field_delete(pField);
    }
  }

  if (p->mpShadow) {
    free(p->mpShadow);
  }

  if (p->mpDirty) {
    free(p->mpDirty);
  }

  memset(p, 0, sizeof(Register_t));
  free(p);
}

//
//
//
void Register_init(Register_t* p, Module_t* pModule, FpgaRegisterInit_t* pInit)
{
  int i;
  //
  p->mpOwner = pModule;
  //
  p->mId = pInit->id;
  p->mnBitWidth = pInit->bw;
  p->mnAddrRel = pInit->addrRel;
  p->mAddr = pModule->mAddrBase + pInit->addrRel;
  p->mType = pInit->type;
  p->mLen = ((pInit->bw != (uint16_t)-1) ? ((pInit->bw+31) >> 5) : 1); // Old P200 registers have no bw at register level - default to BW=-1
  p->mDebugMode = pModule->mDebugMode;
  //
  p->mnFields = pInit->nbFields;
  if (p->mnFields) {
    p->mpaFields = malloc(p->mnFields*sizeof(Field_t*));
    //
    if (p->mpaFields) {
      memset(p->mpaFields, 0, (p->mnFields*sizeof(Field_t*)));
      for (i = 0; i < p->mnFields; i++) {
        Field_t* pField = Field_new();
        Field_init(pField, p, &pInit->fields[i]);
        p->mpaFields[i] = pField;
      }
      //
      p->mpShadow = malloc(p->mLen*sizeof(uint32_t));
      if (p->mpShadow) {
        memset(p->mpShadow, 0x00, (p->mLen*sizeof(uint32_t)));
      }
      //
      p->mpDirty = malloc(p->mLen*sizeof(bool));
      if (p->mpDirty) {
        memset(p->mpDirty, 0x00, (p->mLen*sizeof(bool)));
      }
    }
  }
  //
}


//
//
//
void Register_dump(const Register_t* p)
{
  NT_LOG(DBG, NTHW, "Register_dump(id=%d type=%d addr=0x%08X addrrel=0x%08X len=%d bw=%d\n", p->mId, p->mType, p->mAddr, p->mnAddrRel, p->mLen, p->mnBitWidth);
  Register_dumpFields(p);
}

//
//
//
void Register_dumpFields(const Register_t* p)
{
  int i;
  //
  NT_LOG(DBG, NTHW, "Register_dumpFields(addr=0x%08X fields=%d\n", p->mAddr, p->mnFields);
  for (i = 0; i < p->mnFields; i++) {
    Field_dump(p->mpaFields[i]);
  }
  NT_LOG(DBG, NTHW, "\n");
}

//
//
//
uint32_t Register_getAddress(const Register_t* p)
{
  return p->mAddr;
}

//
//
//
void Register_reset(const Register_t* p)
{
  int i;
  Field_t* pField = NULL;

  for (i = 0; i < p->mnFields; i++) {
    pField = p->mpaFields[i];
    if (pField) {
      Field_reset(pField);
    }
  }
}

//
//
//
static Field_t* Register_lookupField(const Register_t* p, uint32_t id)
{
  int i;
  Field_t* pField = NULL;

  if (!p) {
    return NULL;
  }

  for (i = 0; i < p->mnFields; i++) {
    if (p->mpaFields[i]->mId == id) {
      pField = p->mpaFields[i];
      break;
    }
  }
  return pField;
}

//
//
//
Field_t* Register_getField(const Register_t* p, uint32_t id)
{
  Field_t* pField;

  if (p == NULL) {
    NT_LOG(ERR, NTHW, "Illegal register context for field %d\n", id);
    return NULL;
  }

  pField = Register_lookupField(p, id);
  if (!pField) {
    //throw NtException(NT_ERROR_INTERNAL_ERROR, "Field %d not found in module: %s", id, s.c_str());
    NT_LOG(ERR, NTHW, "Field %d not found in module: %s (%d)\n", id, nthwFpgaModIdToStr(p->mpOwner->mModId), p->mpOwner->mModId);
    //
    //assert(false);
  }
  return pField;
}

//
//
//
Field_t* Register_queryField(const Register_t* p, uint32_t id)
{
  return Register_lookupField(p, id);
}

//
//
//
int Register_getBitWidth(const Register_t* p)
{
  return p->mnBitWidth;
}

//
//
//
uint32_t Register_getAddrRel(const Register_t* p)
{
  return p->mnAddrRel;
}

//
//
//
int Register_getDebugMode(const Module_t* p)
{
  return p->mDebugMode;
}

//
// NOTE: do not set debug on fields - as register operation dumps typically are enough
//
void Register_setDebugMode(Register_t* p, unsigned nDebugMode)
{
  int i;
  Field_t* pField = NULL;

  p->mDebugMode = nDebugMode;

  for (i = 0; i < p->mnFields; i++) {
    pField = p->mpaFields[i];
    if (pField) {
      Field_setDebugMode(pField, nDebugMode);
    }
  }

}

//
//
//
static int Register_ReadData(const Register_t* p)
{
  int rc = -1;

  const int nBusTypeId = Module_getBus(p->mpOwner);
  const uint32_t addr = p->mAddr;
  const uint32_t len = p->mLen;
  uint32_t* const pData = p->mpShadow;

  struct FpgaInfo_s* pFpgaInfo = NULL;
  if (p && p->mpOwner && p->mpOwner->mpOwner) {
    pFpgaInfo = p->mpOwner->mpOwner->pFpgaInfo;
  }
  assert(pFpgaInfo);
  assert(pData);

  rc = ReadData(pFpgaInfo, nBusTypeId, addr, len, pData);
  return rc;
}

//
//
//
static int Register_ReadDataTsc(const Register_t* p, uint64_t* pTsc1, uint64_t* pTsc2)
{
  int rc = -1;

  const int nBusTypeId = Module_getBus(p->mpOwner);
  const uint32_t addr = p->mAddr;
  const uint32_t len = p->mLen;
  uint32_t* const pData = p->mpShadow;

  struct FpgaInfo_s* pFpgaInfo = NULL;
  if (p && p->mpOwner && p->mpOwner->mpOwner) {
    pFpgaInfo = p->mpOwner->mpOwner->pFpgaInfo;
  }

  rc = ReadDataTsc(pFpgaInfo, nBusTypeId, addr, len, pData, pTsc1, pTsc2);

  return rc;
}

//
//
//
static int Register_WriteData(const Register_t* p, uint32_t cnt)
{
  int rc = -1;

  const int nBusTypeId = Module_getBus(p->mpOwner);
  const uint32_t addr = p->mAddr;
  const uint32_t len = p->mLen;
  uint32_t* const pData = p->mpShadow;

  struct FpgaInfo_s* pFpgaInfo = NULL;
  if (p && p->mpOwner && p->mpOwner->mpOwner) {
    pFpgaInfo = p->mpOwner->mpOwner->pFpgaInfo;
  }
  assert(pFpgaInfo);
  assert(pData);

  rc = WriteData(pFpgaInfo, nBusTypeId, addr, (len * cnt), pData);

  return rc;
}

//
//
//
void Register_getVal(const Register_t* p, uint32_t* pData, uint32_t len)
{
  uint32_t i;

  if (len == (uint32_t)-1 || len > p->mLen) {
    len = p->mLen;
  }

  assert(len <= p->mLen);
  assert(pData);

  for (i = 0; i < len; i++) {
    pData[i] = p->mpShadow[i];
  }
}

//
//
//
uint32_t Register_getVal32(const Register_t* p)
{
  uint32_t val = 0;
  Register_getVal(p, &val, 1);
  return val;
}

//
//
//
void Register_update(const Register_t* p)
{
  if (p && p->mType != REGISTER_TYPE_WO) {
    const char* const pDevName _unused = "NA"; //p???->name;
    const int nBusTypeId = Module_getBus(p->mpOwner);
    const char* const pBusName _unused = getBusName(nBusTypeId);
    const uint32_t addr _unused = p->mAddr;
    const uint32_t len = p->mLen;
    uint32_t* const pData = p->mpShadow;

    Register_ReadData(p);
    if (p->mDebugMode & ON_READ) {
      uint32_t i = len;
      uint32_t* ptr _unused = pData;
      NT_LOG(DBG, NTHW, "Register::read(Dev: %s, Bus: %s, Addr: 0x%08X, Cnt: %d, Data:", pDevName, pBusName, addr, len);
      while (i--) {
        NT_LOG(DBG, NTHW, " 0x%08X ", *ptr++);
      }
      NT_LOG(DBG, NTHW, ")\n");
    }
  }
}

//
//
//
uint32_t Register_getValUpdated32(const Register_t* p)
{
  uint32_t val = 0;
  Register_update(p);
  Register_getVal(p, &val, 1);
  return val;
}

//
//
//
void Register_makeDirty(Register_t* p)
{
  uint32_t i;
  for (i = 0; i < p->mLen; i++) {
    p->mpDirty[i] = true;
  }
}


//
//
//
void Register_setVal(Register_t* p, const uint32_t* pData, uint32_t len)
{
  assert(len <= p->mLen);
  assert(pData);

  if (len == (uint32_t)-1 || len > p->mLen) {
    len = p->mLen;
  }

  if (p->mpShadow != pData) {
    memcpy(p->mpShadow, pData, (len*sizeof(uint32_t)));
  }
}

//
//
//
void Register_setValFlush(Register_t* p, const uint32_t* pData, uint32_t len)
{
  Register_setVal(p, pData, len);
  Register_flush(p, 1);
}

//
//
//
void Register_flush(const Register_t* p, uint32_t cnt)
{
  int rc;
  if (p->mType != REGISTER_TYPE_RO) {
    const char* const pDevName = "NA"; //p???->name;
    const int nBusTypeId = Module_getBus(p->mpOwner);
    const char* pBusName = getBusName(nBusTypeId);
    const uint32_t addr = p->mAddr;
    const uint32_t len = p->mLen;
    uint32_t* const pData = p->mpShadow;
    uint32_t i;
    assert(len*cnt <= 256);

    if (p->mDebugMode & ON_WRITE) {
      uint32_t i = len*cnt;
      uint32_t* ptr = pData;
      char *tmp_string = ntlog_helper_str_alloc("Register::write");
      ntlog_helper_str_add(tmp_string, "(Dev: %s, Bus: %s, Addr: 0x%08X, Cnt: %d, Data:", pDevName, pBusName, addr, i);
      while (i--) {
        ntlog_helper_str_add(tmp_string, " 0x%08X", *ptr++);
      }
      ntlog_helper_str_add(tmp_string, ")\n");
      NT_LOG(DBG, NTHW, "%s", tmp_string);
      ntlog_helper_str_free(tmp_string);
    }

    rc = Register_WriteData(p, cnt);

    if (rc) {
      NT_LOG(ERR, NTHW, "Register write error %d\n", rc);
    }

    for (i = 0; i < cnt; i++) {
      p->mpDirty[i] = false;
    }
  }
}


//
//
//
void Register_doReadTrigTs(const Register_t* p, uint64_t* tsc1, uint64_t* tsc2)
{
  Register_ReadDataTsc(p, tsc1, tsc2);
}

//
//
//
void Register_clr(Register_t* p)
{
  memset(p->mpShadow, 0, p->mLen * sizeof(uint32_t));
  Register_makeDirty(p);
}

//
//
//
void Register_set(Register_t* p)
{
  memset(p->mpShadow, 0xff, p->mLen * sizeof(uint32_t));
  Register_makeDirty(p);
}


//
// Field
//
Field_t* Field_new(void)
{
  Field_t* p = malloc(sizeof(Field_t));
  return p;
}

//
//
//
void Field_delete(Field_t* p)
{
  memset(p, 0, sizeof(Field_t));
  free(p);
};

//
//
//
void Field_init(Field_t* p, Register_t* pReg, const FpgaFieldInit_t* pInit)
{
  p->mpOwner = pReg;
  //
  p->mDebugMode = pReg->mDebugMode;
  //
  p->mId = pInit->id;
  p->mnBitWidth = pInit->bw;
  p->mnBitPosLow = pInit->low;
  p->mResetVal = (uint32_t)pInit->resetVal;
  p->mFirstWord =pInit->low/32;
  p->mFirstBit = pInit->low%32;
  p->mFrontMask = 0;
  p->mBodyLength = 0;
  p->mnWords = (pInit->bw + 0x1f) / 0x20;
  p->mTailMask = 0;

  {
    int bitsRemaining = pInit->bw;
    int frontMaskLength = 32 - p->mFirstBit;

    if (frontMaskLength > bitsRemaining) {
      frontMaskLength = bitsRemaining;
    }
    bitsRemaining -= frontMaskLength;

    p->mFrontMask = (uint32_t)(((1ULL<<frontMaskLength) - 1) << p->mFirstBit);

    p->mBodyLength = bitsRemaining / 32;
    bitsRemaining -= p->mBodyLength * 32;
    p->mTailMask = (1<<bitsRemaining) - 1;

    if (p->mDebugMode >= 0x100) {
      NT_LOG(DBG, NTHW, "%s: fldid=%08d: [%08d:%08d] %08d/%08d: (%08d,%08d) (0x%08X,%08d,0x%08X)\n", __func__, pInit->id, pInit->low, (pInit->low+pInit->bw), pInit->bw, ((pInit->bw+31)/32), p->mFirstWord, p->mFirstBit, p->mFrontMask, p->mBodyLength, p->mTailMask);
    }
  }

}

//
//
//
int Field_getDebugMode(const Module_t* p)
{
  return p->mDebugMode;
}

//
//
//
void Field_setDebugMode(Field_t* p, unsigned nDebugMode)
{
  p->mDebugMode = nDebugMode;
}

//
//
//
#if 0
std_string Field_toString(Field_t* p)
{
  std_string s;

  if (p->mpOwner) {
    s = p->mpOwner->toString();
  }

  s.append(" field=(" + std_to_string(p->mFirstWord)
           + "," + std_to_string(p->mFirstBit) + "," + std_to_string(p->mFrontMask)
           + "," + std_to_string(p->mBodyLength) + "," + std_to_string(p->mTailMask) + ")");

  return s;
}
#endif

//
//
//
int Field_getBitWidth(const Field_t* p)
{
  return p->mnBitWidth;
}

//
//
//
int Field_getBitPosLow(const Field_t* p)
{
  return p->mnBitPosLow;
}

//
//
//
int Field_getBitPosHigh(const Field_t* p)
{
  return p->mnBitPosLow + p->mnBitWidth - 1;
}

//
//
//
uint32_t Field_getMask(const Field_t* p)
{
  return p->mFrontMask;
}

//
//
//
void Field_reset(const Field_t* p)
{
  Field_setVal32(p, (uint32_t) p->mResetVal);
}

//
//
//
uint32_t Field_getValMask(const Field_t* p)
{
  return (p->mFrontMask >> p->mnBitPosLow);
}

//
//
//
//
uint32_t Field_getResetVal(const Field_t* p)
{
  return (uint32_t) p->mResetVal;
}

//
//
//
void Field_getVal(const Field_t* p, uint32_t* pData, uint32_t len)
{
  uint32_t i;
  uint32_t dataIndex = 0;
  uint32_t shadowIndex = p->mFirstWord;

  union {
    uint32_t w32[2];
    uint64_t w64;
  } buf;

  (void)len;
  assert(len == p->mnWords);

  // handle front
  buf.w32[0] = p->mpOwner->mpShadow[shadowIndex++] & p->mFrontMask;

  // handle body
  for (i = 0; i < p->mBodyLength; i++) {
    buf.w32[1] = p->mpOwner->mpShadow[shadowIndex++];
    buf.w64 = buf.w64 >> (p->mFirstBit);
    assert(dataIndex < len);
    pData[dataIndex++] = buf.w32[0];
    buf.w64 = buf.w64 >> (32 - p->mFirstBit);
  }

  // handle tail
  if (p->mTailMask) {
    buf.w32[1] = p->mpOwner->mpShadow[shadowIndex++] & p->mTailMask;
  } else {
    buf.w32[1] = 0;
  }
  buf.w64 = buf.w64 >> (p->mFirstBit);
  pData[dataIndex++] = buf.w32[0];
  if (dataIndex < p->mnWords) {
    pData[dataIndex++] = buf.w32[1];
  }
}

//
//
//
void Field_setVal(const Field_t* p, const uint32_t* pData, uint32_t len)
{
  uint32_t i;
  uint32_t dataIndex = 0;
  uint32_t shadowIndex = p->mFirstWord;

  union {
    uint32_t w32[2];
    uint64_t w64;
  } buf;

  (void)len;
  assert(len == p->mnWords);

  // handle front
  buf.w32[0] = 0;
  buf.w32[1] = pData[dataIndex++];
  buf.w64 = buf.w64 >> (32 - p->mFirstBit);
  p->mpOwner->mpShadow[shadowIndex] = (p->mpOwner->mpShadow[shadowIndex] & ~p->mFrontMask) | (buf.w32[0] & p->mFrontMask);
  shadowIndex++;

  // handle body
  for (i = 0; i < p->mBodyLength; i++) {
    buf.w64 = buf.w64 >> (p->mFirstBit);
    assert(dataIndex < len);
    buf.w32[1] = pData[dataIndex++];
    buf.w64 = buf.w64 >> (32 - p->mFirstBit);
    p->mpOwner->mpShadow[shadowIndex++] = buf.w32[0];
  }

  // handle tail
  if (p->mTailMask) {
    buf.w64 = buf.w64 >> (p->mFirstBit);
    if (dataIndex < len) {
      buf.w32[1] = pData[dataIndex];
    }
    buf.w64 = buf.w64 >> (32 - p->mFirstBit);
    p->mpOwner->mpShadow[shadowIndex] = (p->mpOwner->mpShadow[shadowIndex] & ~p->mTailMask) | (buf.w32[0] & p->mTailMask);
  }

  Register_makeDirty(p->mpOwner);
}

//
//
//
void Field_setValFlush(const Field_t* p, const uint32_t* pData, uint32_t len)
{
  Field_setVal(p, pData, len);
  Field_flushRegister(p);
}


//
//
//
uint32_t Field_getVal32(const Field_t* p)
{
  uint32_t val;
  Field_getVal(p, &val, 1);
  return val;
}

//
//
//
uint32_t Field_getUpdated(const Field_t* p)
{
  uint32_t val;

  Register_update(p->mpOwner);
  Field_getVal(p, &val, 1);

  return val;
}

//
//
//
void Field_readTrigWithTsc(const Field_t* p, uint64_t* tsc1, uint64_t* tsc2)
{
  Register_doReadTrigTs(p->mpOwner, tsc1, tsc2);
}

//
//
//
void Field_updateRegister(const Field_t* p)
{
  Register_update(p->mpOwner);
}

//
//
//
void Field_flushRegister(const Field_t* p)
{
  Register_flush(p->mpOwner, 1);
}

//
//
//
void Field_setVal32(const Field_t* p, uint32_t val)
{
  Field_setVal(p, &val, 1);
}

//
//
//
void Field_setValFlush32(const Field_t* p, uint32_t val)
{
  Field_setVal(p, &val, 1);
  Register_flush(p->mpOwner, 1);
}

//
//
//
void Field_clrAll(const Field_t* p)
{
  assert(p->mBodyLength==0);
  Field_setVal32(p, 0);
}

//
//
//
void Field_clrFlush(const Field_t* p)
{
  Field_clrAll(p);
  Register_flush(p->mpOwner, 1);
}

//
//
//
void Field_setAll(const Field_t* p)
{
  assert(p->mBodyLength==0);
  Field_setVal32(p, ~0);
}

//
//
//
void Field_setFlush(const Field_t* p)
{
  Field_setAll(p);
  Register_flush(p->mpOwner, 1);
}


//
//
//
enum Field_Match {
  FIELD_MATCH_CLR_ALL,
  FIELD_MATCH_SET_ALL,
  FIELD_MATCH_CLR_ANY,
  FIELD_MATCH_SET_ANY,
};

//
//
//
static int Field_waitCond32(const Field_t* p, enum Field_Match eMatch, int nPollIterations, int nPollInterval)
{
  const uint32_t nMask = (1<<p->mnBitWidth)-1;

  if (nPollIterations == -1) {
    nPollIterations = 10000;
  }
  if (nPollInterval == -1) {
    nPollInterval = 100; // usec
  }

  if (p->mDebugMode) {
    const char* const pCondName _unused = ((eMatch == FIELD_MATCH_SET_ALL) ? "SetAll" : ((eMatch == FIELD_MATCH_CLR_ALL) ? "ClrAll" : ((eMatch == FIELD_MATCH_CLR_ANY) ? "ClrAny" : "SetAny")));
    const char* const pDevName _unused = "NA";
    const char* const pBusName _unused = getBusName(Module_getBus(p->mpOwner->mpOwner));
    uint32_t nRegAddr _unused = Register_getAddress(p->mpOwner);
    uint32_t nRegMask _unused = (((1 << p->mnBitWidth)-1) << p->mnBitPosLow);

    NT_LOG(DBG, NTHW, "Register::Field::wait%s32(Dev: %s, Bus: %s, Addr: 0x%08X, Mask: 0x%08X, Iterations: %d, Interval: %d)\n", pCondName, pDevName, pBusName, nRegAddr, nRegMask, nPollIterations, nPollInterval);
  }

  while (true) {
    uint32_t val = Field_getUpdated(p);
    if ((eMatch == FIELD_MATCH_SET_ANY) && (val != 0)) {
      return 0;
    } else if ((eMatch == FIELD_MATCH_SET_ALL) && (val == nMask)) {
      return 0;
    } else if ((eMatch == FIELD_MATCH_CLR_ALL) && (val == 0)) {
      return 0;
    } else if (eMatch == FIELD_MATCH_CLR_ANY) {
      uint32_t mask = Field_getMask(p);
      if (val != mask) {
        return 0;
      }
    }

    nPollIterations--;
    if (nPollIterations <= 0) {
      return -1;
    }
    NtOs_WaituSec(nPollInterval);
  }
  return 0;
}

//
//
//
int Field_waitSetAll32(const Field_t* p, int nPollIterations, int nPollInterval)
{
  return Field_waitCond32(p, FIELD_MATCH_SET_ALL, nPollIterations, nPollInterval);
}

//
//
//
int Field_waitClrAll32(const Field_t* p, int nPollIterations, int nPollInterval)
{
  return Field_waitCond32(p, FIELD_MATCH_CLR_ALL, nPollIterations, nPollInterval);
}

//
//
//
int Field_waitSetAny32(const Field_t* p, int nPollIterations, int nPollInterval)
{
  return Field_waitCond32(p, FIELD_MATCH_SET_ANY, nPollIterations, nPollInterval);
}

//
//
//
int Field_waitClrAny32(const Field_t* p, int nPollIterations, int nPollInterval)
{
  return Field_waitCond32(p, FIELD_MATCH_CLR_ANY, nPollIterations, nPollInterval);
}

//
//
//
int Field_waitValMask32(const Field_t* p, uint32_t nWaitCondValue, uint32_t nWaitCondMask, int nPollIterations, int nPollInterval)
{
  if (nPollIterations == -1) {
    nPollIterations = 10000;
  }
  if (nPollInterval == -1) {
    nPollInterval = 100;
  }

  while (true) {
    uint32_t val = Field_getUpdated(p);
    if (val == (nWaitCondValue & nWaitCondMask)) {
      break;
    }
    nPollIterations--;
    if (nPollIterations <= 0) {
      //NT_LOG(ERR, NTHW, "Timeout waiting\n"); // no logging within nthw layer
      return -1;
    }
    NtOs_WaituSec(nPollInterval);
  }
  return 0;
}

//
//
//
void Field_dump(const Field_t* p _unused)
{
  NT_LOG(DBG, NTHW, "%s: %02d: %02d %02d %02d: %02d: %X\n", __func__, p->mId, p->mnBitPosLow, (p->mnBitPosLow+p->mnBitWidth), p->mnBitWidth, p->mnWords, p->mResetVal);
}


//
//
//
void Field_dumpVal(const Field_t* p)
{
  int i;
  uint32_t buf[32];
  //
  Field_getVal(p, buf, p->mnWords);
  NT_LOG(DBG, NTHW, " @%d:", p->mFirstBit + p->mFirstWord * 32);
  NT_LOG(DBG, NTHW, "%X", buf[p->mnWords-1]);
  for (i = p->mnWords-1; i > 0; i--) {
    NT_LOG(DBG, NTHW, "%08X", buf[i-1]);
  }
  NT_LOG(DBG, NTHW, "\n");
}

//
//
//
void Field_dump_init(const FpgaFieldInit_t* p _unused)
{
  NT_LOG(DBG, NTHW, "%s: %02d: %02d %02d %02d: %"PRIu64"\n", __func__, p->id, p->low, p->low+p->bw, p->bw, p->resetVal);
}

//
// nthw fpga model helpers
//

//
//
//
Fpga_t* Nthw_getFpga(struct FpgaInfo_s* pFpgaInfo, uint64_t nFpgaIdent)
{
  FpgaMgr_t* pFpgaMgr = NULL;
  Fpga_t* pFpga = NULL;
  int nFpgaTypeId, nFpgaProdId, nFpgaVerId, nFpgaRevId;
  char sFpgaProdVerRevStr[32];

  pFpgaMgr = FpgaMgr_new();
  FpgaMgr_init(pFpgaMgr);
  pFpga = FpgaMgr_queryFpga(pFpgaMgr, nFpgaIdent, pFpgaInfo);

  nFpgaTypeId = FPGAID_TO_PRODUCTTYPE(nFpgaIdent);
  nFpgaProdId = FPGAID_TO_PRODUCTCODE(nFpgaIdent);
  nFpgaVerId = FPGAID_TO_VERSIONCODE(nFpgaIdent);
  nFpgaRevId = FPGAID_TO_REVISIONCODE(nFpgaIdent);

  snprintf(sFpgaProdVerRevStr, sizeof(sFpgaProdVerRevStr), "%04d-%04d-%02d-%02d", nFpgaTypeId, nFpgaProdId, nFpgaVerId, nFpgaRevId);

  if (pFpga == NULL) {
    NT_LOG(ERR, NTHW, "%s: no match for FPGA: %s\n", __func__, sFpgaProdVerRevStr);
    // do not return here...
  }

  if (pFpgaMgr) {
    FpgaMgr_delete(pFpgaMgr);
    pFpgaMgr = NULL;
  }

  return pFpga;
}

//
//
//
Module_t* Nthw_getModule(Fpga_t* pFpga, int nMod, int nInstance)
{
  struct Module_s* pMod = Fpga_queryModule(pFpga, nMod, nInstance);
  return pMod;
}

//
//
//
Register_t* Nthw_getRegister(Module_t* pMod, int nReg)
{
  Register_t* pReg = Module_getRegister(pMod, nReg);
  return pReg;
}

//
//
//
Field_t* Nthw_getField(Register_t* pReg, int nFld)
{
  Field_t* pFld = Register_getField(pReg, nFld);
  return pFld;
}



//
// EOF
//
