//
// nthw_dbs.c
//

/*
 * %NT_SOFTWARE_LICENSE%
 */

//
// Includes:
//
#include <errno.h>
#include "ntlog.h"
//
#include "nthw_drv.h"
#include "nthw_register.h"

//
//
//
#include "nthw_dbs.h"

#if (0)
#define DBS_PRINT_REGS
#else
#undef DBS_PRINT_REGS
#endif

static void setShadow_TxQosData(NthwDbs_t* p,
                                uint32_t index,
                                uint32_t enable,
                                uint32_t ir,
                                uint32_t bs);
static void flush_TxQosData(NthwDbs_t* p, uint32_t index);
static void setShadow_TxQpData(NthwDbs_t* p, uint32_t index, uint32_t virtualPort);
static void flush_TxQpData(NthwDbs_t* p, uint32_t index);
static void setShadow_TxDrData(NthwDbs_t* p,
                           uint32_t index,
                           uint64_t guestPhysicalAddress,
                           uint32_t hostId,
                           uint32_t queueSize,
                           uint32_t port,
                           uint32_t header,
                           uint32_t packed);
static void flush_TxDrData(NthwDbs_t* p, uint32_t index);
static void setShadow_RxDrData(NthwDbs_t* p,
                           uint32_t index,
                           uint64_t guestPhysicalAddress,
                           uint32_t hostId,
                           uint32_t queueSize,
                           uint32_t header,
                           uint32_t packed);
static void flush_RxDrData(NthwDbs_t* p, uint32_t index);
static void setShadow_TxUwData(NthwDbs_t* p,
                           uint32_t index,
                           uint64_t guestPhysicalAddress,
                           uint32_t hostId,
                           uint32_t queueSize,
                           uint32_t packed,
                           uint32_t int_enable,
                           uint32_t vec,
                           uint32_t istk,
                           uint32_t inOrder);
static void flush_TxUwData(NthwDbs_t* p, uint32_t index);
static void setShadow_RxUwData(NthwDbs_t* p,
                           uint32_t index,
                           uint64_t guestPhysicalAddress,
                           uint32_t hostId,
                           uint32_t queueSize,
                           uint32_t packed,
                           uint32_t int_enable,
                           uint32_t vec,
                           uint32_t istk);
static void flush_RxUwData(NthwDbs_t* p, uint32_t index);
static void setShadow_RxAmData(NthwDbs_t* p,
                               uint32_t index,
                               uint64_t guestPhysicalAddress,
                               uint32_t enable,
                               uint32_t hostId,
                               uint32_t packed,
                               uint32_t int_enable);
static void flush_RxAmData(NthwDbs_t* p, uint32_t index);
static void setShadow_TxAmData(NthwDbs_t* p,
                               uint32_t index,
                               uint64_t guestPhysicalAddress,
                               uint32_t enable,
                               uint32_t hostId,
                               uint32_t packed,
                               uint32_t int_enable);
static void flush_TxAmData(NthwDbs_t* p, uint32_t index);

//
//
//
NthwDbs_t* NthwDbs_new(void)
{
  NthwDbs_t* p = malloc(sizeof(NthwDbs_t));
  if (p) {
    memset(p, 0, sizeof(NthwDbs_t));
  }
  return p;
}

//
//
//
void NthwDbs_delete(NthwDbs_t* p)
{
  if (p) {
    memset(p, 0, sizeof(NthwDbs_t));
    free(p);
  }
}

//
//
//
int dbs_init(NthwDbs_t *p, Fpga_t *pFpga, int nInstance)
{
  struct Module_s *mod = Fpga_queryModule(pFpga, MOD_DBS, nInstance);

  if (p == NULL)
    return mod == NULL ? -1 : 0;

  if (mod == NULL) {
    NT_LOG(ERR, NTHW, "%s: DBS %d: no such instance\n", pFpga->pFpgaInfo->mpAdapterIdStr,
        nInstance);
    return -1;
  }

  p->mpFpga = pFpga;
  p->mnInstance = nInstance;
  p->mpModDbs = mod;

#if defined(DEBUG) && (0)
  Module_setDebugMode(p->mpModDbs, 0x00); // DEBUG:
#endif

  p->mnParamDbsPresent = Fpga_getProductParam(pFpga, NT_DBS_PRESENT, 0);
  if (p->mnParamDbsPresent == 0) {
    NT_LOG(WRN, NTHW, "%s: DBS %d: logical error: module found but not flagged at present\n", p->mpFpga->pFpgaInfo->mpAdapterIdStr, p->mnInstance);
  }

  p->mpRegRxControl = Module_getRegister(p->mpModDbs, DBS_RX_CONTROL);
  p->mpFldRxControlLastQueue = Register_getField(p->mpRegRxControl, DBS_RX_CONTROL_LQ);
  p->mpFldRxControlAvailMonitorEnable = Register_getField(p->mpRegRxControl, DBS_RX_CONTROL_AME);
  p->mpFldRxControlAvailMonitorScanSpeed = Register_getField(p->mpRegRxControl, DBS_RX_CONTROL_AMS);
  p->mpFldRxControlUsedWriteEnable = Register_getField(p->mpRegRxControl, DBS_RX_CONTROL_UWE);
  p->mpFldRxControlUsedWriterUpdateSpeed = Register_getField(p->mpRegRxControl, DBS_RX_CONTROL_UWS);
  p->mpFldRxControlRxQueuesEnable = Register_getField(p->mpRegRxControl, DBS_RX_CONTROL_QE);

  p->mpRegTxControl = Module_getRegister(p->mpModDbs, DBS_TX_CONTROL);
  p->mpFldTxControlLastQueue = Register_getField(p->mpRegTxControl, DBS_TX_CONTROL_LQ);
  p->mpFldTxControlAvailMonitorEnable = Register_getField(p->mpRegTxControl, DBS_TX_CONTROL_AME);
  p->mpFldTxControlAvailMonitorScanSpeed = Register_getField(p->mpRegTxControl, DBS_TX_CONTROL_AMS);
  p->mpFldTxControlUsedWriteEnable = Register_getField(p->mpRegTxControl, DBS_TX_CONTROL_UWE);
  p->mpFldTxControlUsedWriterUpdateSpeed = Register_getField(p->mpRegTxControl, DBS_TX_CONTROL_UWS);
  p->mpFldTxControlTxQueuesEnable = Register_getField(p->mpRegTxControl, DBS_TX_CONTROL_QE);

  p->mpRegRxInit = Module_getRegister(p->mpModDbs, DBS_RX_INIT);
  p->mpFldRxInitInit = Register_getField(p->mpRegRxInit, DBS_RX_INIT_INIT);
  p->mpFldRxInitQueue = Register_getField(p->mpRegRxInit, DBS_RX_INIT_QUEUE);
  p->mpFldRxInitBusy = Register_getField(p->mpRegRxInit, DBS_RX_INIT_BUSY);

  p->mpRegRxInitVal = Module_queryRegister(p->mpModDbs, DBS_RX_INIT_VAL);
  if (p->mpRegRxInitVal) {
    p->mpFldRxInitValIdx = Register_queryField(p->mpRegRxInitVal, DBS_RX_INIT_VAL_IDX);
    p->mpFldRxInitValPtr = Register_queryField(p->mpRegRxInitVal, DBS_RX_INIT_VAL_PTR);
  }

  p->mpRegRxPtr = Module_queryRegister(p->mpModDbs, DBS_RX_PTR);
  if (p->mpRegRxPtr) {
    p->mpFldRxPtrPtr   = Register_queryField(p->mpRegRxPtr, DBS_RX_PTR_PTR);
    p->mpFldRxPtrQueue = Register_queryField(p->mpRegRxPtr, DBS_RX_PTR_QUEUE);
    p->mpFldRxPtrValid = Register_queryField(p->mpRegRxPtr, DBS_RX_PTR_VALID);
  }

  p->mpRegTxInit = Module_getRegister(p->mpModDbs, DBS_TX_INIT);
  p->mpFldTxInitInit = Register_getField(p->mpRegTxInit, DBS_TX_INIT_INIT);
  p->mpFldTxInitQueue = Register_getField(p->mpRegTxInit, DBS_TX_INIT_QUEUE);
  p->mpFldTxInitBusy = Register_getField(p->mpRegTxInit, DBS_TX_INIT_BUSY);

  p->mpRegTxInitVal = Module_queryRegister(p->mpModDbs, DBS_TX_INIT_VAL);
  if (p->mpRegTxInitVal) {
    p->mpFldTxInitValIdx = Register_queryField(p->mpRegTxInitVal, DBS_TX_INIT_VAL_IDX);
    p->mpFldTxInitValPtr = Register_queryField(p->mpRegTxInitVal, DBS_TX_INIT_VAL_PTR);
  }

  p->mpRegTxPtr = Module_queryRegister(p->mpModDbs, DBS_TX_PTR);
  if (p->mpRegTxPtr) {
    p->mpFldTxPtrPtr   = Register_queryField(p->mpRegTxPtr, DBS_TX_PTR_PTR);
    p->mpFldTxPtrQueue = Register_queryField(p->mpRegTxPtr, DBS_TX_PTR_QUEUE);
    p->mpFldTxPtrValid = Register_queryField(p->mpRegTxPtr, DBS_TX_PTR_VALID);
  }

  p->mpRegRxIdle = Module_queryRegister(p->mpModDbs, DBS_RX_IDLE);
  if (p->mpRegRxIdle) {
    p->mpFldRxIdleIdle = Register_queryField(p->mpRegRxIdle, DBS_RX_IDLE_IDLE);
    p->mpFldRxIdleQueue = Register_queryField(p->mpRegRxIdle, DBS_RX_IDLE_QUEUE);
    p->mpFldRxIdleBusy = Register_queryField(p->mpRegRxIdle, DBS_RX_IDLE_BUSY);
  }

  p->mpRegTxIdle = Module_queryRegister(p->mpModDbs, DBS_TX_IDLE);
  if (p->mpRegTxIdle) {
    p->mpFldTxIdleIdle = Register_queryField(p->mpRegTxIdle, DBS_TX_IDLE_IDLE);
    p->mpFldTxIdleQueue = Register_queryField(p->mpRegTxIdle, DBS_TX_IDLE_QUEUE);
    p->mpFldTxIdleBusy = Register_queryField(p->mpRegTxIdle, DBS_TX_IDLE_BUSY);
  }

  p->mpRegRxAvailMonitorControl = Module_getRegister(p->mpModDbs, DBS_RX_AM_CTRL);
  p->mpFldRxAvailMonitorControlAdr = Register_getField(p->mpRegRxAvailMonitorControl, DBS_RX_AM_CTRL_ADR);
  p->mpFldRxAvailMonitorControlCnt = Register_getField(p->mpRegRxAvailMonitorControl, DBS_RX_AM_CTRL_CNT);

  p->mpRegRxAvailMonitorData = Module_getRegister(p->mpModDbs, DBS_RX_AM_DATA);
  p->mpFldRxAvailMonitorDataGuestPhysicalAddress = Register_getField(p->mpRegRxAvailMonitorData, DBS_RX_AM_DATA_GPA);
  p->mpFldRxAvailMonitorDataEnable = Register_getField(p->mpRegRxAvailMonitorData, DBS_RX_AM_DATA_ENABLE);
  p->mpFldRxAvailMonitorDataHostID = Register_getField(p->mpRegRxAvailMonitorData, DBS_RX_AM_DATA_HID);
  p->mpFldRxAvailMonitorDataPacked = Register_queryField(p->mpRegRxAvailMonitorData, DBS_RX_AM_DATA_PCKED);
  p->mpFldRxAvailMonitorDataInt = Register_queryField(p->mpRegRxAvailMonitorData, DBS_RX_AM_DATA_INT);

  p->mpRegTxAvailMonitorControl = Module_getRegister(p->mpModDbs, DBS_TX_AM_CTRL);
  p->mpFldTxAvailMonitorControlAdr = Register_getField(p->mpRegTxAvailMonitorControl, DBS_TX_AM_CTRL_ADR);
  p->mpFldTxAvailMonitorControlCnt = Register_getField(p->mpRegTxAvailMonitorControl, DBS_TX_AM_CTRL_CNT);

  p->mpRegTxAvailMonitorData = Module_getRegister(p->mpModDbs, DBS_TX_AM_DATA);
  p->mpFldTxAvailMonitorDataGuestPhysicalAddress = Register_getField(p->mpRegTxAvailMonitorData, DBS_TX_AM_DATA_GPA);
  p->mpFldTxAvailMonitorDataEnable = Register_getField(p->mpRegTxAvailMonitorData, DBS_TX_AM_DATA_ENABLE);
  p->mpFldTxAvailMonitorDataHostID = Register_getField(p->mpRegTxAvailMonitorData, DBS_TX_AM_DATA_HID);
  p->mpFldTxAvailMonitorDataPacked = Register_queryField(p->mpRegTxAvailMonitorData, DBS_TX_AM_DATA_PCKED);
  p->mpFldTxAvailMonitorDataInt = Register_queryField(p->mpRegTxAvailMonitorData, DBS_TX_AM_DATA_INT);

  p->mpRegRxUsedWriterControl = Module_getRegister(p->mpModDbs, DBS_RX_UW_CTRL);
  p->mpFldRxUsedWriterControlAdr = Register_getField(p->mpRegRxUsedWriterControl, DBS_RX_UW_CTRL_ADR);
  p->mpFldRxUsedWriterControlCnt = Register_getField(p->mpRegRxUsedWriterControl, DBS_RX_UW_CTRL_CNT);

  p->mpRegRxUsedWriterData = Module_getRegister(p->mpModDbs, DBS_RX_UW_DATA);
  p->mpFldRxUsedWriterDataGuestPhysicalAddress = Register_getField(p->mpRegRxUsedWriterData, DBS_RX_UW_DATA_GPA);
  p->mpFldRxUsedWriterDataHostID = Register_getField(p->mpRegRxUsedWriterData, DBS_RX_UW_DATA_HID);
  p->mpFldRxUsedWriterDataQueueSize = Register_getField(p->mpRegRxUsedWriterData, DBS_RX_UW_DATA_QS);
  p->mpFldRxUsedWriterDataPacked = Register_queryField(p->mpRegRxUsedWriterData, DBS_RX_UW_DATA_PCKED);
  p->mpFldRxUsedWriterDataInt = Register_queryField(p->mpRegRxUsedWriterData, DBS_RX_UW_DATA_INT);
  p->mpFldRxUsedWriterDataVec = Register_queryField(p->mpRegRxUsedWriterData, DBS_RX_UW_DATA_VEC);
  p->mpFldRxUsedWriterDataIstk = Register_queryField(p->mpRegRxUsedWriterData, DBS_RX_UW_DATA_ISTK);

  p->mpRegTxUsedWriterControl = Module_getRegister(p->mpModDbs, DBS_TX_UW_CTRL);
  p->mpFldTxUsedWriterControlAdr = Register_getField(p->mpRegTxUsedWriterControl, DBS_TX_UW_CTRL_ADR);
  p->mpFldTxUsedWriterControlCnt = Register_getField(p->mpRegTxUsedWriterControl, DBS_TX_UW_CTRL_CNT);

  p->mpRegTxUsedWriterData = Module_getRegister(p->mpModDbs, DBS_TX_UW_DATA);
  p->mpFldTxUsedWriterDataGuestPhysicalAddress = Register_getField(p->mpRegTxUsedWriterData, DBS_TX_UW_DATA_GPA);
  p->mpFldTxUsedWriterDataHostID = Register_getField(p->mpRegTxUsedWriterData, DBS_TX_UW_DATA_HID);
  p->mpFldTxUsedWriterDataQueueSize = Register_getField(p->mpRegTxUsedWriterData, DBS_TX_UW_DATA_QS);
  p->mpFldTxUsedWriterDataPacked = Register_queryField(p->mpRegTxUsedWriterData, DBS_TX_UW_DATA_PCKED);
  p->mpFldTxUsedWriterDataInt = Register_queryField(p->mpRegTxUsedWriterData, DBS_TX_UW_DATA_INT);
  p->mpFldTxUsedWriterDataVec = Register_queryField(p->mpRegTxUsedWriterData, DBS_TX_UW_DATA_VEC);
  p->mpFldTxUsedWriterDataIstk = Register_queryField(p->mpRegTxUsedWriterData, DBS_TX_UW_DATA_ISTK);
  p->mpFldTxUsedWriterDataInOrder = Register_queryField(p->mpRegTxUsedWriterData, DBS_TX_UW_DATA_INO);

  p->mpRegRxDescriptorReaderControl = Module_getRegister(p->mpModDbs, DBS_RX_DR_CTRL);
  p->mpFldRxDescriptorReaderControlAdr = Register_getField(p->mpRegRxDescriptorReaderControl, DBS_RX_DR_CTRL_ADR);
  p->mpFldRxDescriptorReaderControlCnt = Register_getField(p->mpRegRxDescriptorReaderControl, DBS_RX_DR_CTRL_CNT);

  p->mpRegRxDescriptorReaderData = Module_getRegister(p->mpModDbs, DBS_RX_DR_DATA);
  p->mpFldRxDescriptorReaderDataGuestPhysicalAddress = Register_getField(p->mpRegRxDescriptorReaderData, DBS_RX_DR_DATA_GPA);
  p->mpFldRxDescriptorReaderDataHostID = Register_getField(p->mpRegRxDescriptorReaderData, DBS_RX_DR_DATA_HID);
  p->mpFldRxDescriptorReaderDataQueueSize = Register_getField(p->mpRegRxDescriptorReaderData, DBS_RX_DR_DATA_QS);
  p->mpFldRxDescriptorReaderDataHeader = Register_getField(p->mpRegRxDescriptorReaderData, DBS_RX_DR_DATA_HDR);
  p->mpFldRxDescriptorReaderDataPacked = Register_queryField(p->mpRegRxDescriptorReaderData, DBS_RX_DR_DATA_PCKED);

  p->mpRegTxDescriptorReaderControl = Module_getRegister(p->mpModDbs, DBS_TX_DR_CTRL);
  p->mpFldTxDescriptorReaderControlAdr = Register_getField(p->mpRegTxDescriptorReaderControl, DBS_TX_DR_CTRL_ADR);
  p->mpFldTxDescriptorReaderControlCnt = Register_getField(p->mpRegTxDescriptorReaderControl, DBS_TX_DR_CTRL_CNT);

  p->mpRegTxDescriptorReaderData = Module_getRegister(p->mpModDbs, DBS_TX_DR_DATA);
  p->mpFldTxDescriptorReaderDataGuestPhysicalAddress = Register_getField(p->mpRegTxDescriptorReaderData, DBS_TX_DR_DATA_GPA);
  p->mpFldTxDescriptorReaderDataHostID = Register_getField(p->mpRegTxDescriptorReaderData, DBS_TX_DR_DATA_HID);
  p->mpFldTxDescriptorReaderDataQueueSize = Register_getField(p->mpRegTxDescriptorReaderData, DBS_TX_DR_DATA_QS);
  p->mpFldTxDescriptorReaderDataHeader = Register_getField(p->mpRegTxDescriptorReaderData, DBS_TX_DR_DATA_HDR);
  p->mpFldTxDescriptorReaderDataPort = Register_getField(p->mpRegTxDescriptorReaderData, DBS_TX_DR_DATA_PORT);
  p->mpFldTxDescriptorReaderDataPacked = Register_queryField(p->mpRegTxDescriptorReaderData, DBS_TX_DR_DATA_PCKED);

  p->mpRegTxQueuePropertyControl = Module_getRegister(p->mpModDbs, DBS_TX_QP_CTRL);
  p->mpFldTxQueuePropertyControlAdr = Register_getField(p->mpRegTxQueuePropertyControl, DBS_TX_QP_CTRL_ADR);
  p->mpFldTxQueuePropertyControlCnt = Register_getField(p->mpRegTxQueuePropertyControl, DBS_TX_QP_CTRL_CNT);

  p->mpRegTxQueuePropertyData = Module_getRegister(p->mpModDbs, DBS_TX_QP_DATA);
  p->mpFldTxQueuePropertyDataVPort = Register_getField(p->mpRegTxQueuePropertyData, DBS_TX_QP_DATA_VPORT);

    /* HW QoS Tx rate limiting policing RFC2697/RFC4111 */
  p->mpRegTxQueueQosControl = Module_queryRegister(p->mpModDbs, DBS_TX_QOS_CTRL);
  p->mpRegTxQueueQosData = Module_queryRegister(p->mpModDbs, DBS_TX_QOS_DATA);
  if (p->mpRegTxQueueQosControl) {
    p->mpRegTxQueueQosControlAdr = Register_queryField(p->mpRegTxQueueQosControl, DBS_TX_QOS_CTRL_ADR);
    p->mpRegTxQueueQosControlCnt = Register_queryField(p->mpRegTxQueueQosControl, DBS_TX_QOS_CTRL_CNT);

    if (p->mpRegTxQueueQosData) {
      p->mpRegTxQueueQosDataEn = Register_queryField(p->mpRegTxQueueQosData, DBS_TX_QOS_DATA_EN);
      p->mpRegTxQueueQosDataIR = Register_queryField(p->mpRegTxQueueQosData, DBS_TX_QOS_DATA_IR);
      p->mpRegTxQueueQosDataBS = Register_queryField(p->mpRegTxQueueQosData, DBS_TX_QOS_DATA_BS);
    }
  }

  p->mpRegTxQueueQosRate = Module_queryRegister(p->mpModDbs, DBS_TX_QOS_RATE);
  if (p->mpRegTxQueueQosRate) {
    p->mpRegTxQueueQosRateMul = Register_queryField(p->mpRegTxQueueQosRate, DBS_TX_QOS_RATE_MUL);
    p->mpRegTxQueueQosRateDiv = Register_queryField(p->mpRegTxQueueQosRate, DBS_TX_QOS_RATE_DIV);
  }

  return 0;
}

//
//
//
int dbs_reset_rx_control(NthwDbs_t* p)
{
  Field_setVal32(p->mpFldRxControlLastQueue, 0);
  Field_setVal32(p->mpFldRxControlAvailMonitorEnable, 0);
  Field_setVal32(p->mpFldRxControlAvailMonitorScanSpeed, 8);
  Field_setVal32(p->mpFldRxControlUsedWriteEnable, 0);
  Field_setVal32(p->mpFldRxControlUsedWriterUpdateSpeed, 5);
  Field_setVal32(p->mpFldRxControlRxQueuesEnable, 0);
  Register_flush(p->mpRegRxControl, 1);
  return 0;
}

int dbs_reset_tx_control(NthwDbs_t* p)
{
  Field_setVal32(p->mpFldTxControlLastQueue, 0);
  Field_setVal32(p->mpFldTxControlAvailMonitorEnable, 0);
  Field_setVal32(p->mpFldTxControlAvailMonitorScanSpeed, 5);
  Field_setVal32(p->mpFldTxControlUsedWriteEnable, 0);
  Field_setVal32(p->mpFldTxControlUsedWriterUpdateSpeed, 8);
  Field_setVal32(p->mpFldTxControlTxQueuesEnable, 0);
  Register_flush(p->mpRegTxControl, 1);
  return 0;
}

void dbs_reset(NthwDbs_t* p)
{
  uint32_t i;

  NT_LOG(DBG, NTHW, "NthwDbs::%s: resetting DBS", __func__);

  dbs_reset_rx_control(p);
  dbs_reset_tx_control(p);

  // Reset RX memory banks and shado
  for (i = 0; i < NT_DBS_RX_QUEUES_MAX; ++i) {

    setShadow_RxAmData(p, i, 0, 0, 0, 0, 0);
    flush_RxAmData(p, i);

    setShadow_RxUwData(p, i, 0, 0, 0, 0, 0, 0, 0);
    flush_RxUwData(p, i);

    setShadow_RxDrData(p, i, 0, 0, 0, 0, 0);
    flush_RxDrData(p, i);
  }

  // Reset TX memory banks and shado
  for (i = 0; i < NT_DBS_TX_QUEUES_MAX; ++i) {
    setShadow_TxAmData(p, i, 0, 0, 0, 0, 0);
    flush_TxAmData(p, i);

    setShadow_TxUwData(p, i, 0, 0, 0, 0, 0, 0, 0, 0);
    flush_TxUwData(p, i);

    setShadow_TxDrData(p, i, 0, 0, 0, 0, 0, 0);
    flush_TxDrData(p, i);

    setShadow_TxQpData(p, i, 0);
    flush_TxQpData(p, i);

    setShadow_TxQosData(p, i, 0, 0, 0);
    flush_TxQosData(p, i);
  }
}

int set_rx_control(NthwDbs_t* p,
                         uint32_t lastQueue,
                         uint32_t availMonitorEnable,
                         uint32_t availMonitorSpeed,
                         uint32_t usedWriteEnable,
                         uint32_t usedWriteSpeed,
                         uint32_t rxQueueEnable)
{
#ifdef DBS_PRINT_REGS
  printf("lastQueue %u\n", lastQueue);
  printf("availMonitorEnable %u\n", availMonitorEnable);
  printf("availMonitorSpeed %u\n", availMonitorSpeed);
  printf("usedWriteEnable %u\n", usedWriteEnable);
  printf("usedWriteSpeed %u\n", usedWriteSpeed);
  printf("rxQueueEnable %u\n", rxQueueEnable);
#endif

  Field_setVal32(p->mpFldRxControlLastQueue, lastQueue);
  Field_setVal32(p->mpFldRxControlAvailMonitorEnable, availMonitorEnable);
  Field_setVal32(p->mpFldRxControlAvailMonitorScanSpeed, availMonitorSpeed);
  Field_setVal32(p->mpFldRxControlUsedWriteEnable, usedWriteEnable);
  Field_setVal32(p->mpFldRxControlUsedWriterUpdateSpeed, usedWriteSpeed);
  Field_setVal32(p->mpFldRxControlRxQueuesEnable, rxQueueEnable);
  Register_flush(p->mpRegRxControl, 1);
  return 0;
}

int NthwDbs_getRxControl(NthwDbs_t* p,
                         uint32_t *lastQueue,
                         uint32_t *availMonitorEnable,
                         uint32_t *availMonitorSpeed,
                         uint32_t *usedWriteEnable,
                         uint32_t *usedWriteSpeed,
                         uint32_t *rxQueueEnable)
{
  *lastQueue = Field_getVal32(p->mpFldRxControlLastQueue);
  *availMonitorEnable = Field_getVal32(p->mpFldRxControlAvailMonitorEnable);
  *availMonitorSpeed = Field_getVal32(p->mpFldRxControlAvailMonitorScanSpeed);
  *usedWriteEnable = Field_getVal32(p->mpFldRxControlUsedWriteEnable);
  *usedWriteSpeed = Field_getVal32(p->mpFldRxControlUsedWriterUpdateSpeed);
  *rxQueueEnable = Field_getVal32(p->mpFldRxControlRxQueuesEnable);
  return 0;
}

int set_tx_control(NthwDbs_t* p,
                         uint32_t lastQueue,
                         uint32_t availMonitorEnable,
                         uint32_t availMonitorSpeed,
                         uint32_t usedWriteEnable,
                         uint32_t usedWriteSpeed,
                         uint32_t txQueueEnable)
{
#ifdef DBS_PRINT_REGS
  printf("lastQueue %u\n", lastQueue);
  printf("availMonitorEnable %u\n", availMonitorEnable);
  printf("availMonitorSpeed %u\n", availMonitorSpeed);
  printf("usedWriteEnable %u\n", usedWriteEnable);
  printf("usedWriteSpeed %u\n", usedWriteSpeed);
#endif

  Field_setVal32(p->mpFldTxControlLastQueue, lastQueue);
  Field_setVal32(p->mpFldTxControlAvailMonitorEnable, availMonitorEnable);
  Field_setVal32(p->mpFldTxControlAvailMonitorScanSpeed, availMonitorSpeed);
  Field_setVal32(p->mpFldTxControlUsedWriteEnable, usedWriteEnable);
  Field_setVal32(p->mpFldTxControlUsedWriterUpdateSpeed, usedWriteSpeed);
  Field_setVal32(p->mpFldTxControlTxQueuesEnable, txQueueEnable);
  Register_flush(p->mpRegTxControl, 1);
  return 0;
}

int NthwDbs_getTxControl(NthwDbs_t* p,
                         uint32_t *lastQueue,
                         uint32_t *availMonitorEnable,
                         uint32_t *availMonitorSpeed,
                         uint32_t *usedWriteEnable,
                         uint32_t *usedWriteSpeed,
                         uint32_t *txQueueEnable)
{
  *lastQueue = Field_getVal32(p->mpFldTxControlLastQueue);
  *availMonitorEnable = Field_getVal32(p->mpFldTxControlAvailMonitorEnable);
  *availMonitorSpeed = Field_getVal32(p->mpFldTxControlAvailMonitorScanSpeed);
  *usedWriteEnable = Field_getVal32(p->mpFldTxControlUsedWriteEnable);
  *usedWriteSpeed = Field_getVal32(p->mpFldTxControlUsedWriterUpdateSpeed);
  *txQueueEnable = Field_getVal32(p->mpFldTxControlTxQueuesEnable);
  return 0;
}


int set_rx_init(NthwDbs_t* p,
                      uint32_t startIdx,
                      uint32_t startPtr,
                      uint32_t init,
                      uint32_t queue)
{
  if (p->mpRegRxInitVal) {
    Field_setVal32(p->mpFldRxInitValIdx, startIdx);
    Field_setVal32(p->mpFldRxInitValPtr, startPtr);
    //printf("Set RxVq init [%i]: idx,ptr %08x,%08x\n", queue, startIdx, startPtr);
    Register_flush(p->mpRegRxInitVal, 1);
  }
  Field_setVal32(p->mpFldRxInitInit, init);
  Field_setVal32(p->mpFldRxInitQueue, queue);
  Register_flush(p->mpRegRxInit, 1);
  return 0;
}

int get_rx_init(NthwDbs_t* p,
                      uint32_t *init,
                      uint32_t *queue,
                      uint32_t *busy)
{
  *init = Field_getVal32(p->mpFldRxInitInit);
  *queue = Field_getVal32(p->mpFldRxInitQueue);
  *busy = Field_getVal32(p->mpFldRxInitBusy);
  return 0;
}

int set_tx_init(NthwDbs_t* p,
                      uint32_t startIdx,
                      uint32_t startPtr,
                      uint32_t init,
                      uint32_t queue)
{
  if (p->mpRegTxInitVal) {
    Field_setVal32(p->mpFldTxInitValIdx, startIdx);
    Field_setVal32(p->mpFldTxInitValPtr, startPtr);
    //printf("Set TxVq init [%i]: idx,ptr %08x,%08x\n",queue, startIdx, startPtr);
    Register_flush(p->mpRegTxInitVal, 1);
  }
  Field_setVal32(p->mpFldTxInitInit, init);
  Field_setVal32(p->mpFldTxInitQueue, queue);
  Register_flush(p->mpRegTxInit, 1);
  return 0;
}

int get_tx_init(NthwDbs_t* p,
                      uint32_t *init,
                      uint32_t *queue,
                      uint32_t *busy)
{
  *init = Field_getVal32(p->mpFldTxInitInit);
  *queue = Field_getVal32(p->mpFldTxInitQueue);
  *busy = Field_getVal32(p->mpFldTxInitBusy);
  return 0;
}

int set_rx_idle(NthwDbs_t* p,
                      uint32_t idle,
                      uint32_t queue)

{
  if (!p->mpRegRxIdle)
    return -ENOTSUP;

  Field_setVal32(p->mpFldRxIdleIdle, idle);
  Field_setVal32(p->mpFldRxIdleQueue, queue);
  Register_flush(p->mpRegRxIdle, 1);
  return 0;
}

int get_rx_idle(NthwDbs_t* p,
                      uint32_t *idle,
                      uint32_t *queue,
                      uint32_t *busy)
{
  if (!p->mpRegRxIdle)
    return -ENOTSUP;

  *idle =  Field_getUpdated(p->mpFldRxIdleIdle);
  *queue =  0;
  *busy = Field_getUpdated(p->mpFldRxIdleBusy);
  return 0;
}

int set_tx_idle(NthwDbs_t* p,
                      uint32_t idle,
                      uint32_t queue)

{
  if (!p->mpRegTxIdle)
    return -ENOTSUP;

  //printf("set idle Tx, queue %i\n", queue);
  Field_setVal32(p->mpFldTxIdleIdle, idle);
  Field_setVal32(p->mpFldTxIdleQueue, queue);
  Register_flush(p->mpRegTxIdle, 1);
  return 0;
}

int get_tx_idle(NthwDbs_t* p,
                      uint32_t *idle,
                      uint32_t *queue,
                      uint32_t *busy)
{
  if (!p->mpRegTxIdle)
    return -ENOTSUP;

  *idle =  Field_getUpdated(p->mpFldTxIdleIdle);
  *queue =  0;
  *busy = Field_getUpdated(p->mpFldTxIdleBusy);
  //printf("get idle Tx queue %i\n", Field_getUpdated(p->mpFldTxIdleQueue));
  return 0;
}

int set_rx_ptr_queue(NthwDbs_t* p, uint32_t queue)
{
  if (!p->mpRegRxPtr)
    return -ENOTSUP;

  Field_setVal32(p->mpFldRxPtrQueue, queue);
  Register_flush(p->mpRegRxPtr, 1);
  return 0;
}

int get_rx_ptr(NthwDbs_t* p,
                     uint32_t *ptr,
                     uint32_t *queue,
                     uint32_t *valid)
{
  if (!p->mpRegRxPtr)
    return -ENOTSUP;

  *ptr = Field_getUpdated(p->mpFldRxPtrPtr);
  *queue =  0;
  *valid = Field_getUpdated(p->mpFldRxPtrValid);
  return 0;
}

int set_tx_ptr_queue(NthwDbs_t* p, uint32_t queue)
{
  if (!p->mpRegTxPtr)
    return -ENOTSUP;

  Field_setVal32(p->mpFldTxPtrQueue, queue);
  Register_flush(p->mpRegTxPtr, 1);
  return 0;
}

int get_tx_ptr(NthwDbs_t* p,
                     uint32_t *ptr,
                     uint32_t *queue,
                     uint32_t *valid)
{
  if (!p->mpRegTxPtr)
    return -ENOTSUP;

  *ptr =  Field_getUpdated(p->mpFldTxPtrPtr);
  *queue =  0;
  *valid = Field_getUpdated(p->mpFldTxPtrValid);
  return 0;
}
//
// ********************* set_rx_am_data *********************
//
static void setRxAmDataIndex(NthwDbs_t* p, uint32_t index)
{
  Field_setVal32(p->mpFldRxAvailMonitorControlAdr, index);
  Field_setVal32(p->mpFldRxAvailMonitorControlCnt, 1);
  Register_flush(p->mpRegRxAvailMonitorControl, 1);
}
//
//
//
static void setShadow_RxAmDataGuestPhysicalAddress(NthwDbs_t* p, uint32_t index, uint64_t guestPhysicalAddress)
{
  p->mRxAmShadow[index].guestPhysicalAddress = guestPhysicalAddress;
}
//
//
//
static void NthwDbs_setShadow_RxAmDataEnable(NthwDbs_t* p, uint32_t index, uint32_t enable)
{
  p->mRxAmShadow[index].enable = enable;
}
//
//
//
static void setShadow_RxAmDataHostId(NthwDbs_t* p, uint32_t index, uint32_t hostId)
{
  p->mRxAmShadow[index].hostId = hostId;
}
//
//
//
static void setShadow_RxAmDataPacked(NthwDbs_t* p, uint32_t index, uint32_t packed)
{
  p->mRxAmShadow[index].packed = packed;
}
//
//
//
static void setShadow_RxAmDataIntEnable(NthwDbs_t* p, uint32_t index, uint32_t int_enable)
{
  p->mRxAmShadow[index].int_enable = int_enable;
}
//
//
//
static void setShadow_RxAmData(NthwDbs_t* p,
                               uint32_t index,
                               uint64_t guestPhysicalAddress,
                               uint32_t enable,
                               uint32_t hostId,
                               uint32_t packed,
                               uint32_t int_enable)
{
  setShadow_RxAmDataGuestPhysicalAddress(p, index, guestPhysicalAddress);
  NthwDbs_setShadow_RxAmDataEnable(p, index, enable);
  setShadow_RxAmDataHostId(p, index, hostId);
  setShadow_RxAmDataPacked(p, index, packed);
  setShadow_RxAmDataIntEnable(p, index, int_enable);
}
//
//
//
static void flush_RxAmData(NthwDbs_t* p, uint32_t index)
{
  Field_setVal(p->mpFldRxAvailMonitorDataGuestPhysicalAddress, (uint32_t*)&p->mRxAmShadow[index].guestPhysicalAddress, 2);
  Field_setVal32(p->mpFldRxAvailMonitorDataEnable, p->mRxAmShadow[index].enable);
  Field_setVal32(p->mpFldRxAvailMonitorDataHostID, p->mRxAmShadow[index].hostId);
  if (p->mpFldRxAvailMonitorDataPacked) {
    Field_setVal32(p->mpFldRxAvailMonitorDataPacked, p->mRxAmShadow[index].packed);
  }
  if (p->mpFldRxAvailMonitorDataInt) {
    Field_setVal32(p->mpFldRxAvailMonitorDataInt, p->mRxAmShadow[index].int_enable);
  }

  setRxAmDataIndex(p, index);
  Register_flush(p->mpRegRxAvailMonitorData, 1);
}
//
//
//
int set_rx_am_data(NthwDbs_t* p,
                        uint32_t index,
                        uint64_t guestPhysicalAddress,
                        uint32_t enable,
                        uint32_t hostId,
                        uint32_t packed,
                        uint32_t int_enable)
{
  if (!p->mpRegRxAvailMonitorData)
    return -ENOTSUP;

  setShadow_RxAmData(p, index, guestPhysicalAddress, enable, hostId, packed, int_enable);
  flush_RxAmData(p, index);
  return 0;
}
//
// ********************* get_tx_init *********************
//
static void setTxAmDataIndex(NthwDbs_t* p, uint32_t index)
{
  Field_setVal32(p->mpFldTxAvailMonitorControlAdr, index);
  Field_setVal32(p->mpFldTxAvailMonitorControlCnt, 1);
  Register_flush(p->mpRegTxAvailMonitorControl, 1);
}
//
//
//
static void setShadow_TxAmData(NthwDbs_t* p,
                               uint32_t index,
                               uint64_t guestPhysicalAddress,
                               uint32_t enable,
                               uint32_t hostId,
                               uint32_t packed,
                               uint32_t int_enable)
{
  p->mTxAmShadow[index].guestPhysicalAddress = guestPhysicalAddress;
  p->mTxAmShadow[index].enable = enable;
  p->mTxAmShadow[index].hostId = hostId;
  p->mTxAmShadow[index].packed = packed;
  p->mTxAmShadow[index].int_enable = int_enable;
}
//
//
//
static void flush_TxAmData(NthwDbs_t* p, uint32_t index)
{
  Field_setVal(p->mpFldTxAvailMonitorDataGuestPhysicalAddress, (uint32_t*)&p->mTxAmShadow[index].guestPhysicalAddress, 2);
  Field_setVal32(p->mpFldTxAvailMonitorDataEnable, p->mTxAmShadow[index].enable);
  Field_setVal32(p->mpFldTxAvailMonitorDataHostID, p->mTxAmShadow[index].hostId);
  if (p->mpFldTxAvailMonitorDataPacked) {
    Field_setVal32(p->mpFldTxAvailMonitorDataPacked, p->mTxAmShadow[index].packed);
  }
  if (p->mpFldTxAvailMonitorDataInt) {
    Field_setVal32(p->mpFldTxAvailMonitorDataInt, p->mTxAmShadow[index].int_enable);
  }

  setTxAmDataIndex(p, index);
  Register_flush(p->mpRegTxAvailMonitorData, 1);
}
//
//
//
int set_tx_am_data(NthwDbs_t* p,
                        uint32_t index,
                        uint64_t guestPhysicalAddress,
                        uint32_t enable,
                        uint32_t hostId,
                        uint32_t packed,
                        uint32_t int_enable)
{
  if (!p->mpRegTxAvailMonitorData)
    return -ENOTSUP;

  setShadow_TxAmData(p, index, guestPhysicalAddress, enable, hostId, packed, int_enable);
  flush_TxAmData(p, index);
  return 0;
}
//
// ********************* set_rx_uw_data *********************
//
static void setRxUwDataIndex(NthwDbs_t* p, uint32_t index)
{
  Field_setVal32(p->mpFldRxUsedWriterControlAdr, index);
  Field_setVal32(p->mpFldRxUsedWriterControlCnt, 1);
  Register_flush(p->mpRegRxUsedWriterControl, 1);
}

//
//
//
static void setShadow_RxUwDataGuestPhysicalAddress(NthwDbs_t* p, uint32_t index, uint64_t guestPhysicalAddress)
{
  p->mRxUwShadow[index].guestPhysicalAddress = guestPhysicalAddress;
}

//
//
//
static void setShadow_RxUwDataHostId(NthwDbs_t* p, uint32_t index, uint32_t hostId)
{
  p->mRxUwShadow[index].hostId = hostId;
}

//
//
//
static void setShadow_RxUwDataQueueSize(NthwDbs_t* p, uint32_t index, uint32_t queueSize)
{
  p->mRxUwShadow[index].queueSize = queueSize;
}
//
//
//
static void setShadow_RxUwDataPacked(NthwDbs_t* p, uint32_t index, uint32_t packed)
{
  p->mRxUwShadow[index].packed = packed;
}
//
//
//
static void setShadow_RxUwDataIntEnable(NthwDbs_t* p, uint32_t index, uint32_t int_enable)
{
  p->mRxUwShadow[index].int_enable = int_enable;
}
//
//
//
static void setShadow_RxUwDataVec(NthwDbs_t* p, uint32_t index, uint32_t vec)
{
  p->mRxUwShadow[index].vec = vec;
}
//
//
//
static void setShadow_RxUwDataIstk(NthwDbs_t* p, uint32_t index, uint32_t istk)
{
  p->mRxUwShadow[index].istk = istk;
}
//
//
//
static void setShadow_RxUwData(NthwDbs_t* p,
                           uint32_t index,
                           uint64_t guestPhysicalAddress,
                           uint32_t hostId,
                           uint32_t queueSize,
                           uint32_t packed,
                           uint32_t int_enable,
                           uint32_t vec,
                           uint32_t istk)
{
  setShadow_RxUwDataGuestPhysicalAddress(p, index, guestPhysicalAddress);
  setShadow_RxUwDataHostId(p, index, hostId);
  setShadow_RxUwDataQueueSize(p, index, queueSize);
  setShadow_RxUwDataPacked(p, index, packed);
  setShadow_RxUwDataIntEnable(p, index, int_enable);
  setShadow_RxUwDataVec(p, index, vec);
  setShadow_RxUwDataIstk(p, index, istk);
}

//
//
//

static void flush_RxUwData(NthwDbs_t* p, uint32_t index)
{
  Field_setVal(p->mpFldRxUsedWriterDataGuestPhysicalAddress, (uint32_t*)&p->mRxUwShadow[index].guestPhysicalAddress, 2);
  Field_setVal32(p->mpFldRxUsedWriterDataHostID, p->mRxUwShadow[index].hostId);
  if (Module_isVersionNewer(p->mpModDbs, 0, 8)) {
    Field_setVal32(p->mpFldRxUsedWriterDataQueueSize, (1U << p->mRxUwShadow[index].queueSize) - 1U);
  } else {
    Field_setVal32(p->mpFldRxUsedWriterDataQueueSize, p->mRxUwShadow[index].queueSize);
  }
  if (p->mpFldRxUsedWriterDataPacked) {
    Field_setVal32(p->mpFldRxUsedWriterDataPacked, p->mRxUwShadow[index].packed);
  }
  if (p->mpFldRxUsedWriterDataInt) {
    Field_setVal32(p->mpFldRxUsedWriterDataInt, p->mRxUwShadow[index].int_enable);
    Field_setVal32(p->mpFldRxUsedWriterDataVec, p->mRxUwShadow[index].vec);
    Field_setVal32(p->mpFldRxUsedWriterDataIstk, p->mRxUwShadow[index].istk);
  }

  setRxUwDataIndex(p, index);
  Register_flush(p->mpRegRxUsedWriterData, 1);
}
//
//
//
int set_rx_uw_data(NthwDbs_t* p,
                        uint32_t index,
                        uint64_t guestPhysicalAddress,
                        uint32_t hostId,
                        uint32_t queueSize,
                        uint32_t packed,
                        uint32_t int_enable,
                        uint32_t vec,
                        uint32_t istk)
{
  if (!p->mpRegRxUsedWriterData)
    return -ENOTSUP;

  setShadow_RxUwData(p, index, guestPhysicalAddress, hostId, queueSize, packed, int_enable, vec, istk);
  flush_RxUwData(p, index);
  return 0;
}
//
// ********************* set_tx_uw_data *********************
//
static void setTxUwDataIndex(NthwDbs_t* p, uint32_t index)
{
  Field_setVal32(p->mpFldTxUsedWriterControlAdr, index);
  Field_setVal32(p->mpFldTxUsedWriterControlCnt, 1);
  Register_flush(p->mpRegTxUsedWriterControl, 1);
}
//
//
//
static void setShadow_TxUwDataGuestPhysicalAddress(NthwDbs_t* p, uint32_t index, uint64_t guestPhysicalAddress)
{
  p->mTxUwShadow[index].guestPhysicalAddress = guestPhysicalAddress;
}
//
//
//
static void setShadow_TxUwDataHostId(NthwDbs_t* p, uint32_t index, uint32_t hostId)
{
  p->mTxUwShadow[index].hostId = hostId;
}
//
//
//
static void setShadow_TxUwDataQueueSize(NthwDbs_t* p, uint32_t index, uint32_t queueSize)
{
  p->mTxUwShadow[index].queueSize = queueSize;
}
//
//
//
static void setShadow_TxUwDataPacked(NthwDbs_t* p, uint32_t index, uint32_t packed)
{
  p->mTxUwShadow[index].packed = packed;
}
//
//
//
static void setShadow_TxUwDataIntEnable(NthwDbs_t* p, uint32_t index, uint32_t int_enable)
{
  p->mTxUwShadow[index].int_enable = int_enable;
}
//
//
//
static void setShadow_TxUwDataVec(NthwDbs_t* p, uint32_t index, uint32_t vec)
{
  p->mTxUwShadow[index].vec = vec;
}
//
//
//
static void setShadow_TxUwDataIstk(NthwDbs_t* p, uint32_t index, uint32_t istk)
{
  p->mTxUwShadow[index].istk = istk;
}
//
//
//
static void setShadow_TxUwDataInOrder(NthwDbs_t* p, uint32_t index, uint32_t inOrder)
{
  p->mTxUwShadow[index].inOrder = inOrder;
}
//
//
//
static void setShadow_TxUwData(NthwDbs_t* p,
                           uint32_t index,
                           uint64_t guestPhysicalAddress,
                           uint32_t hostId,
                           uint32_t queueSize,
                           uint32_t packed,
                           uint32_t int_enable,
                           uint32_t vec,
                           uint32_t istk,
                           uint32_t inOrder)
{
  setShadow_TxUwDataGuestPhysicalAddress(p, index, guestPhysicalAddress);
  setShadow_TxUwDataHostId(p, index, hostId);
  setShadow_TxUwDataQueueSize(p, index, queueSize);
  setShadow_TxUwDataPacked(p, index, packed);
  setShadow_TxUwDataIntEnable(p, index, int_enable);
  setShadow_TxUwDataVec(p, index, vec);
  setShadow_TxUwDataIstk(p, index, istk);
  setShadow_TxUwDataInOrder(p, index, inOrder);
}
//
//
//
static void flush_TxUwData(NthwDbs_t* p, uint32_t index)
{
  Field_setVal(p->mpFldTxUsedWriterDataGuestPhysicalAddress, (uint32_t*)&p->mTxUwShadow[index].guestPhysicalAddress, 2);
  Field_setVal32(p->mpFldTxUsedWriterDataHostID, p->mTxUwShadow[index].hostId);
  if (Module_isVersionNewer(p->mpModDbs, 0,8)) {
    Field_setVal32(p->mpFldTxUsedWriterDataQueueSize, (1U << p->mTxUwShadow[index].queueSize) - 1U);
  } else {
    Field_setVal32(p->mpFldTxUsedWriterDataQueueSize, p->mTxUwShadow[index].queueSize);
  }
  if (p->mpFldTxUsedWriterDataPacked) {
    Field_setVal32(p->mpFldTxUsedWriterDataPacked, p->mTxUwShadow[index].packed);
  }
  if (p->mpFldTxUsedWriterDataInt) {
    Field_setVal32(p->mpFldTxUsedWriterDataInt, p->mTxUwShadow[index].int_enable);
    Field_setVal32(p->mpFldTxUsedWriterDataVec, p->mTxUwShadow[index].vec);
    Field_setVal32(p->mpFldTxUsedWriterDataIstk, p->mTxUwShadow[index].istk);
  }
  if (p->mpFldTxUsedWriterDataInOrder) {
    Field_setVal32(p->mpFldTxUsedWriterDataInOrder, p->mTxUwShadow[index].inOrder);
  }

  setTxUwDataIndex(p, index);
  Register_flush(p->mpRegTxUsedWriterData, 1);
}
//
//
//
int set_tx_uw_data(NthwDbs_t* p,
                        uint32_t index,
                        uint64_t guestPhysicalAddress,
                        uint32_t hostId,
                        uint32_t queueSize,
                        uint32_t packed,
                        uint32_t int_enable,
                        uint32_t vec,
                        uint32_t istk,
                        uint32_t inOrder)
{
  if (!p->mpRegTxUsedWriterData)
    return -ENOTSUP;

  setShadow_TxUwData(p, index, guestPhysicalAddress, hostId, queueSize, packed, int_enable, vec, istk, inOrder);
  flush_TxUwData(p, index);
  return 0;
}
//
// ********************* set_rx_dr_data *********************
//
static void setRxDrDataIndex(NthwDbs_t* p, uint32_t index)
{
  Field_setVal32(p->mpFldRxDescriptorReaderControlAdr, index);
  Field_setVal32(p->mpFldRxDescriptorReaderControlCnt, 1);
  Register_flush(p->mpRegRxDescriptorReaderControl, 1);
}
//
//
//
static void setShadow_RxDrDataGuestPhysicalAddress(NthwDbs_t* p, uint32_t index, uint64_t guestPhysicalAddress)
{
  p->mRxDrShadow[index].guestPhysicalAddress = guestPhysicalAddress;
}
//
//
//
static void setShadow_RxDrDataHostId(NthwDbs_t* p, uint32_t index, uint32_t hostId)
{
  p->mRxDrShadow[index].hostId = hostId;
}
//
//
//
static void setShadow_RxDrDataQueueSize(NthwDbs_t* p, uint32_t index, uint32_t queueSize)
{
  p->mRxDrShadow[index].queueSize = queueSize;
}
//
//
//
static void setShadow_RxDrDataHeader(NthwDbs_t* p, uint32_t index, uint32_t header)
{
  p->mRxDrShadow[index].header = header;
}
//
//
//
static void setShadow_RxDrDataPacked(NthwDbs_t* p, uint32_t index, uint32_t packed)
{
  p->mRxDrShadow[index].packed = packed;
}
//
//
//
static void setShadow_RxDrData(NthwDbs_t* p,
                           uint32_t index,
                           uint64_t guestPhysicalAddress,
                           uint32_t hostId,
                           uint32_t queueSize,
                           uint32_t header,
                           uint32_t packed)
{
  setShadow_RxDrDataGuestPhysicalAddress(p, index, guestPhysicalAddress);
  setShadow_RxDrDataHostId(p, index, hostId);
  setShadow_RxDrDataQueueSize(p, index, queueSize);
  setShadow_RxDrDataHeader(p, index, header);
  setShadow_RxDrDataPacked(p, index, packed);
}
//
//
//
static void flush_RxDrData(NthwDbs_t* p, uint32_t index)
{
  Field_setVal(p->mpFldRxDescriptorReaderDataGuestPhysicalAddress, (uint32_t*)&p->mRxDrShadow[index].guestPhysicalAddress, 2);
  Field_setVal32(p->mpFldRxDescriptorReaderDataHostID, p->mRxDrShadow[index].hostId);
  if (Module_isVersionNewer(p->mpModDbs, 0,8)) {
    Field_setVal32(p->mpFldRxDescriptorReaderDataQueueSize, (1U << p->mRxDrShadow[index].queueSize) - 1U);
  } else {
    Field_setVal32(p->mpFldRxDescriptorReaderDataQueueSize, p->mRxDrShadow[index].queueSize);
  }
  Field_setVal32(p->mpFldRxDescriptorReaderDataHeader, p->mRxDrShadow[index].header);
  if (p->mpFldRxDescriptorReaderDataPacked) {
    Field_setVal32(p->mpFldRxDescriptorReaderDataPacked, p->mRxDrShadow[index].packed);
  }

  setRxDrDataIndex(p, index);
  Register_flush(p->mpRegRxDescriptorReaderData, 1);
}
//
//
//
int set_rx_dr_data(NthwDbs_t* p,
                        uint32_t index,
                        uint64_t guestPhysicalAddress,
                        uint32_t hostId,
                        uint32_t queueSize,
                        uint32_t header,
                        uint32_t packed)
{
  if (!p->mpRegRxDescriptorReaderData)
    return -ENOTSUP;

  setShadow_RxDrData(p, index, guestPhysicalAddress, hostId, queueSize, header, packed);
  flush_RxDrData(p, index);
  return 0;
}
//
// ********************* set_tx_dr_data *********************
//
static void setTxDrDataIndex(NthwDbs_t* p, uint32_t index)
{
  Field_setVal32(p->mpFldTxDescriptorReaderControlAdr, index);
  Field_setVal32(p->mpFldTxDescriptorReaderControlCnt, 1);
  Register_flush(p->mpRegTxDescriptorReaderControl, 1);
}
//
//
//
static void setShadow_TxDrDataGuestPhysicalAddress(NthwDbs_t* p, uint32_t index, uint64_t guestPhysicalAddress)
{
  p->mTxDrShadow[index].guestPhysicalAddress = guestPhysicalAddress;
}
//
//
//
static void setShadow_TxDrDataHostId(NthwDbs_t* p, uint32_t index, uint32_t hostId)
{
  p->mTxDrShadow[index].hostId = hostId;
}
//
//
//
static void setShadow_TxDrDataQueueSize(NthwDbs_t* p, uint32_t index, uint32_t queueSize)
{
  p->mTxDrShadow[index].queueSize = queueSize;
}
//
//
//
static void setShadow_TxDrDataHeader(NthwDbs_t* p, uint32_t index, uint32_t header)
{
  p->mTxDrShadow[index].header = header;
}
//
//
//
static void setShadow_TxDrDataPort(NthwDbs_t* p, uint32_t index, uint32_t port)
{
  p->mTxDrShadow[index].port = port;
}
//
//
//
static void setShadow_TxDrDataPacked(NthwDbs_t* p, uint32_t index, uint32_t packed)
{
  p->mTxDrShadow[index].packed = packed;
}
//
//
//
static void setShadow_TxDrData(NthwDbs_t* p,
                           uint32_t index,
                           uint64_t guestPhysicalAddress,
                           uint32_t hostId,
                           uint32_t queueSize,
                           uint32_t port,
                           uint32_t header,
                           uint32_t packed)
{
  setShadow_TxDrDataGuestPhysicalAddress(p, index, guestPhysicalAddress);
  setShadow_TxDrDataHostId(p, index, hostId);
  setShadow_TxDrDataQueueSize(p, index, queueSize);
  setShadow_TxDrDataHeader(p, index, header);
  setShadow_TxDrDataPort(p, index, port);
  setShadow_TxDrDataPacked(p, index, packed);
}
//
//
//
static void flush_TxDrData(NthwDbs_t* p, uint32_t index)
{
  Field_setVal(p->mpFldTxDescriptorReaderDataGuestPhysicalAddress, (uint32_t*)&p->mTxDrShadow[index].guestPhysicalAddress, 2);
  Field_setVal32(p->mpFldTxDescriptorReaderDataHostID, p->mTxDrShadow[index].hostId);
  if (Module_isVersionNewer(p->mpModDbs, 0,8)) {
    Field_setVal32(p->mpFldTxDescriptorReaderDataQueueSize, (1U << p->mTxDrShadow[index].queueSize) - 1U);
  } else {
    Field_setVal32(p->mpFldTxDescriptorReaderDataQueueSize, p->mTxDrShadow[index].queueSize);
  }
  Field_setVal32(p->mpFldTxDescriptorReaderDataHeader, p->mTxDrShadow[index].header);
  Field_setVal32(p->mpFldTxDescriptorReaderDataPort, p->mTxDrShadow[index].port);
  if (p->mpFldTxDescriptorReaderDataPacked) {
    Field_setVal32(p->mpFldTxDescriptorReaderDataPacked, p->mTxDrShadow[index].packed);
  }

  setTxDrDataIndex(p, index);
  Register_flush(p->mpRegTxDescriptorReaderData, 1);
}
//
//
//
int set_tx_dr_data(NthwDbs_t* p,
                        uint32_t index,
                        uint64_t guestPhysicalAddress,
                        uint32_t hostId,
                        uint32_t queueSize,
                        uint32_t port,
                        uint32_t header,
                        uint32_t packed)
{
  if (!p->mpRegTxDescriptorReaderData)
    return -ENOTSUP;

  setShadow_TxDrData(p, index, guestPhysicalAddress, hostId, queueSize, port, header, packed);
  flush_TxDrData(p, index);
  return 0;
}
//
// ********************* NthwDbs_setTxQpData *********************
//
static void setTxQpDataIndex(NthwDbs_t* p, uint32_t index)
{
  Field_setVal32(p->mpFldTxQueuePropertyControlAdr, index);
  Field_setVal32(p->mpFldTxQueuePropertyControlCnt, 1);
  Register_flush(p->mpRegTxQueuePropertyControl, 1);
}

//
//
//
static void setShadow_TxQpDataVirtualPort(NthwDbs_t* p, uint32_t index, uint32_t virtualPort)
{
  p->mTxQpShadow[index].virtualPort = virtualPort;
}

//
//
//
static void setShadow_TxQpData(NthwDbs_t* p, uint32_t index, uint32_t virtualPort)
{
  setShadow_TxQpDataVirtualPort(p, index, virtualPort);
}

//1
//
//
static void flush_TxQpData(NthwDbs_t* p, uint32_t index)
{
  Field_setVal32(p->mpFldTxQueuePropertyDataVPort, p->mTxQpShadow[index].virtualPort);

  setTxQpDataIndex(p, index);
  Register_flush(p->mpRegTxQueuePropertyData, 1);
}

//
//
//
int NthwDbs_setTxQpData(NthwDbs_t* p,
                        uint32_t index,
                        uint32_t virtualPort)
{
  if (!p->mpRegTxQueuePropertyData)
    return -ENOTSUP;

  setShadow_TxQpData(p, index, virtualPort);
  flush_TxQpData(p, index);
  return 0;
}
//
// ********************* set_tx_qos_data *********************
//
static void setTxQosDataIndex(NthwDbs_t* p, uint32_t index)
{
  Field_setVal32(p->mpRegTxQueueQosControlAdr, index);
  Field_setVal32(p->mpRegTxQueueQosControlCnt, 1);
  Register_flush(p->mpRegTxQueueQosControl, 1);
}
//
//
//
static void setShadow_TxQosDataEnable(NthwDbs_t* p, uint32_t index, uint32_t enable)
{
  p->mTxQosShadow[index].enable = enable;
}
//
//
//
static void setShadow_TxQosDataIr(NthwDbs_t* p, uint32_t index, uint32_t ir)
{
  p->mTxQosShadow[index].ir = ir;
}
//
//
//
static void setShadow_TxQosDataBs(NthwDbs_t* p, uint32_t index, uint32_t bs)
{
  p->mTxQosShadow[index].bs = bs;
}
//
//
//
static void setShadow_TxQosData(NthwDbs_t* p,
                                uint32_t index,
                                uint32_t enable,
                                uint32_t ir,
                                uint32_t bs)
{
  setShadow_TxQosDataEnable(p, index, enable);
  setShadow_TxQosDataIr(p, index, ir);
  setShadow_TxQosDataBs(p, index, bs);
}
//
//
//
static void flush_TxQosData(NthwDbs_t* p, uint32_t index)
{
  Field_setVal32(p->mpRegTxQueueQosDataEn, p->mTxQosShadow[index].enable);
  Field_setVal32(p->mpRegTxQueueQosDataIR, p->mTxQosShadow[index].ir);
  Field_setVal32(p->mpRegTxQueueQosDataBS, p->mTxQosShadow[index].bs);

  setTxQosDataIndex(p, index);
  Register_flush(p->mpRegTxQueueQosData, 1);
}
//
//
//
int set_tx_qos_data(NthwDbs_t* p,
                         uint32_t index,
                         uint32_t enable,
                         uint32_t ir,
                         uint32_t bs)
{
  if (!p->mpRegTxQueueQosData)
    return -ENOTSUP;

  setShadow_TxQosData(p, index, enable, ir, bs);
  flush_TxQosData(p, index);
  return 0;
}

int set_tx_qos_rate(NthwDbs_t* p,
                         uint32_t mul,
                         uint32_t div)
{
  if (!p->mpRegTxQueueQosRate)
    return -ENOTSUP;

  Field_setVal32(p->mpRegTxQueueQosRateMul, mul);
  Field_setVal32(p->mpRegTxQueueQosRateDiv, div);
  Register_flush(p->mpRegTxQueueQosRate, 1);
  return 0;
}

//
// EOF
//
