//
// nthw_rab.c
//

/*
 * %NT_SOFTWARE_LICENSE%
 */

/**
 * @file
 * NWHW support for FPGA module PCI_REG_RAC
 */


#include "nt_util.h"
#include "ntlog.h"

//
#include "nthw_drv.h"
#include "nthw_register.h"
#include "nthw_rac.h"

//
#include <pthread.h>

//
// Prevent that RAB echo debug trace ever gets into a release build
//
#if defined(DEBUG)
#undef RAB_DEBUG_ECHO
//#define RAB_DEBUG_ECHO
#else
#undef RAB_DEBUG_ECHO
#endif // DEBUG

//
//
//
//
#define RAB_DMA_WAIT     (1000000)
#define RAB_DMA_BUF_CNT  (0x4000)
//
//
//
#define RAB_READ       (0x01)
#define RAB_WRITE      (0x02)
#define RAB_ECHO       (0x08)
#define RAB_COMPLETION (0x0F)
//
#define RAB_READ_ECHO  (RAB_READ | RAB_ECHO)
#define RAB_WRITE_ECHO (RAB_WRITE | RAB_ECHO)
//
//
//

#define RAB_OPR_LO     (28)
#define RAB_OPR_HI     (31)
#define RAB_OPR_BW      (4)
//
#define RAB_CNT_LO     (20)
#define RAB_CNT_HI     (27)
#define RAB_CNT_BW      (8)
//
#define RAB_BUSID_LO   (16)
#define RAB_BUSID_HI   (19)
#define RAB_BUSID_BW    (4)
//
#define RAB_ADDR_LO     (0)
#define RAB_ADDR_HI    (15)
#define RAB_ADDR_BW    (16)


//
//
//
NthwRac_t* NthwRac_new(void)
{
  NthwRac_t* p = malloc(sizeof(NthwRac_t));
  memset(p, 0, sizeof(NthwRac_t));
  return p;
}

//
//
//
void NthwRac_delete(NthwRac_t* p)
{
  if (p) {
    memset(p, 0, sizeof(NthwRac_t));
    free(p);
  }
}

//
//
//
int NthwRac_init(NthwRac_t *p, Fpga_t *pFpga, struct FpgaInfo_s* pFpgaInfo)
{
  assert(pFpgaInfo);

  const char* const pAdapterIdStr = pFpgaInfo->mpAdapterIdStr;
  struct Module_s *mod = Fpga_queryModule(pFpga, MOD_RAC, 0);

  if (p == NULL)
    return mod == NULL ? -1 : 0;

  if (mod == NULL) {
    NT_LOG(ERR, NTHW, "%s: RAC %d: no such instance\n", pAdapterIdStr, 0);
    return -1;
  }

  p->mpFpga = pFpga;
  p->mpModRac = mod;

#if defined(DEBUG) && (0)
  Module_setDebugMode(p->mpModRac, 0xff); // DEBUG
#endif

  {
    //
    // RAC is a primary communication channel
    // turn off debug by default
    // except for rac_rab_init
    //
    const int nDebugMode = Module_getDebugMode(p->mpModRac);
    if (nDebugMode && nDebugMode <= 0xff) {
      Module_setDebugMode(p->mpModRac, 0);
      Register_setDebugMode(p->mpRegRabInit, nDebugMode);
    }
  }

  // Params
  p->mnParamRacRabInterfaces = Fpga_getProductParam(p->mpFpga, NT_RAC_RAB_INTERFACES, 3);
  NT_LOG(DBG, NTHW, "%s: NT_RAC_RAB_INTERFACES=%d\n", pAdapterIdStr, p->mnParamRacRabInterfaces);

  p->mnParamRacRabObUpdate = Fpga_getProductParam(p->mpFpga, NT_RAC_RAB_OB_UPDATE, 0);
  NT_LOG(DBG, NTHW, "%s: NT_RAC_RAB_OB_UPDATE=%d\n", pAdapterIdStr, p->mnParamRacRabObUpdate);

  // Optional dummy test registers
  p->mpRegDummy0 = Module_queryRegister(p->mpModRac, RAC_DUMMY0);
  p->mpRegDummy1 = Module_queryRegister(p->mpModRac, RAC_DUMMY1);
  p->mpRegDummy2 = Module_queryRegister(p->mpModRac, RAC_DUMMY2);
  //
  p->mpRegRabInit = Module_getRegister(p->mpModRac, RAC_RAB_INIT);
  p->mpFldRabInit = Register_getField(p->mpRegRabInit, RAC_RAB_INIT_RAB);
  p->mnFldRabInitBw = Field_getBitWidth(p->mpFldRabInit);
  p->mnFldRabInitMask = Field_getMask(p->mpFldRabInit);
  //
  // RAC_RAB_INIT_RAB reg/field sanity checks:
  //
  assert(p->mnFldRabInitMask == ((1UL << p->mnFldRabInitBw)-1));
  assert(p->mnFldRabInitBw == p->mnParamRacRabInterfaces);
  //
  p->mpRegDbgCtrl = Module_queryRegister(p->mpModRac, RAC_DBG_CTRL);
  if (p->mpRegDbgCtrl) {
    p->mpFldDbgCtrl = Register_queryField(p->mpRegDbgCtrl, RAC_DBG_CTRL_C);
  } else {
    p->mpFldDbgCtrl = NULL;
  }
  //
  p->mpRegDbgData = Module_queryRegister(p->mpModRac, RAC_DBG_DATA);
  if (p->mpRegDbgData) {
    p->mpFldDbgData = Register_queryField(p->mpRegDbgData, RAC_DBG_DATA_D);
  } else {
    p->mpRegDbgData = NULL;
  }
  //
  p->mpRegRabIBData = Module_getRegister(p->mpModRac, RAC_RAB_IB_DATA);
  p->mpFldRabIBData = Register_getField(p->mpRegRabIBData, RAC_RAB_IB_DATA_D);
  //
  p->mpRegRabOBData = Module_getRegister(p->mpModRac, RAC_RAB_OB_DATA);
  p->mpFldRabOBData = Register_getField(p->mpRegRabOBData, RAC_RAB_OB_DATA_D);
  //
  p->mpRegRabBufFree = Module_getRegister(p->mpModRac, RAC_RAB_BUF_FREE);
  p->mpFldRabBufFreeIBFree = Register_getField(p->mpRegRabBufFree, RAC_RAB_BUF_FREE_IB_FREE);
  p->mpFldRabBufFreeIBOvf = Register_getField(p->mpRegRabBufFree, RAC_RAB_BUF_FREE_IB_OVF);
  p->mpFldRabBufFreeOBFree = Register_getField(p->mpRegRabBufFree, RAC_RAB_BUF_FREE_OB_FREE);
  p->mpFldRabBufFreeOBOvf = Register_getField(p->mpRegRabBufFree, RAC_RAB_BUF_FREE_OB_OVF);
  p->mpFldRabBufFreeTimeout = Register_getField(p->mpRegRabBufFree, RAC_RAB_BUF_FREE_TIMEOUT);
  //
  p->mpRegRabBufUsed = Module_getRegister(p->mpModRac, RAC_RAB_BUF_USED);
  p->mpFldRabBufUsedIBUsed = Register_getField(p->mpRegRabBufUsed, RAC_RAB_BUF_USED_IB_USED);
  p->mpFldRabBufUsedOBUsed = Register_getField(p->mpRegRabBufUsed, RAC_RAB_BUF_USED_OB_USED);
  p->mpFldRabBufUsedFlush = Register_getField(p->mpRegRabBufUsed, RAC_RAB_BUF_USED_FLUSH);
  //
  // RAC_RAB_DMA regs are optional - only found in real NT4GA - not found in 9231/9232 and earlier
  p->mpRegRabDmaIbLo = Module_getRegister(p->mpModRac, RAC_RAB_DMA_IB_LO);
  p->mpFldRabDmaIbLoPhyAddr = Register_getField(p->mpRegRabDmaIbLo, RAC_RAB_DMA_IB_LO_PHYADDR);
  //
  p->mpRegRabDmaIbHi = Module_getRegister(p->mpModRac, RAC_RAB_DMA_IB_HI);
  p->mpFldRabDmaIbHiPhyAddr = Register_getField(p->mpRegRabDmaIbHi, RAC_RAB_DMA_IB_HI_PHYADDR);
  //
  p->mpRegRabDmaObLo = Module_getRegister(p->mpModRac, RAC_RAB_DMA_OB_LO);
  p->mpFldRabDmaObLoPhyAddr = Register_getField(p->mpRegRabDmaObLo, RAC_RAB_DMA_OB_LO_PHYADDR);
  //
  p->mpRegRabDmaObHi = Module_getRegister(p->mpModRac, RAC_RAB_DMA_OB_HI);
  p->mpFldRabDmaObHiPhyAddr = Register_getField(p->mpRegRabDmaObHi, RAC_RAB_DMA_OB_HI_PHYADDR);
  //
  p->mpRegRabDmaIbWr = Module_getRegister(p->mpModRac, RAC_RAB_DMA_IB_WR);
  p->mpFldRabDmaIbWrPtr = Register_getField(p->mpRegRabDmaIbWr, RAC_RAB_DMA_IB_WR_PTR);
  //
  p->mpRegRabDmaIbRd = Module_getRegister(p->mpModRac, RAC_RAB_DMA_IB_RD);
  p->mpFldRabDmaIbRdPtr = Register_getField(p->mpRegRabDmaIbRd, RAC_RAB_DMA_IB_RD_PTR);
  //
  p->mpRegRabDmaObWr = Module_getRegister(p->mpModRac, RAC_RAB_DMA_OB_WR);
  p->mpFldRabDmaObWrPtr = Register_getField(p->mpRegRabDmaObWr, RAC_RAB_DMA_OB_WR_PTR);

  //
  //
  //
  p->RAC_RAB_INIT_ADDR = Register_getAddress(p->mpRegRabInit);
  p->RAC_RAB_IB_DATA_ADDR = Register_getAddress(p->mpRegRabIBData);
  p->RAC_RAB_OB_DATA_ADDR = Register_getAddress(p->mpRegRabOBData);
  p->RAC_RAB_BUF_FREE_ADDR = Register_getAddress(p->mpRegRabBufFree);
  p->RAC_RAB_BUF_USED_ADDR = Register_getAddress(p->mpRegRabBufUsed);
  //
  // RAC_RAB_DMA regs are optional - only found in real NT4GA - not found in 9231/9232 and earlier
  //

  p->RAC_RAB_DMA_IB_LO_ADDR = Register_getAddress(p->mpRegRabDmaIbLo);
  p->RAC_RAB_DMA_IB_HI_ADDR = Register_getAddress(p->mpRegRabDmaIbHi);
  p->RAC_RAB_DMA_OB_LO_ADDR = Register_getAddress(p->mpRegRabDmaObLo);
  p->RAC_RAB_DMA_OB_HI_ADDR = Register_getAddress(p->mpRegRabDmaObHi);
  p->RAC_RAB_DMA_IB_RD_ADDR = Register_getAddress(p->mpRegRabDmaIbRd);
  p->RAC_RAB_DMA_OB_WR_ADDR = Register_getAddress(p->mpRegRabDmaObWr);
  p->RAC_RAB_DMA_IB_WR_ADDR = Register_getAddress(p->mpRegRabDmaIbWr);

  //
  p->RAC_RAB_BUF_FREE_IB_FREE_MASK = Field_getMask(p->mpFldRabBufFreeIBFree);
  p->RAC_RAB_BUF_FREE_OB_FREE_MASK = Field_getMask(p->mpFldRabBufFreeOBFree);
  p->RAC_RAB_BUF_USED_IB_USED_MASK = Field_getMask(p->mpFldRabBufUsedIBUsed);
  p->RAC_RAB_BUF_USED_OB_USED_MASK = Field_getMask(p->mpFldRabBufUsedOBUsed);
  //
  p->RAC_RAB_BUF_USED_FLUSH_MASK = Field_getMask(p->mpFldRabBufUsedFlush);

  p->RAC_RAB_BUF_USED_OB_USED_LOW = Field_getBitPosLow(p->mpFldRabBufUsedOBUsed);

  //
  // NMB
  //
  p->mpRegRabNmbRd = Module_queryRegister(p->mpModRac, RAC_NMB_RD_ADR);
  if (p->mpRegRabNmbRd) {
    p->RAC_NMB_RD_ADR_ADDR = Register_getAddress(p->mpRegRabNmbRd);
  }

  p->mpRegRabNmbData = Module_queryRegister(p->mpModRac, RAC_NMB_DATA);
  if (p->mpRegRabNmbData) {
    p->RAC_NMB_DATA_ADDR = Register_getAddress(p->mpRegRabNmbData);
  }

  p->mpRegRabNmbWr = Module_queryRegister(p->mpModRac, RAC_NMB_WR_ADR);
  if (p->mpRegRabNmbWr) {
    p->RAC_NMB_WR_ADR_ADDR = Register_getAddress(p->mpRegRabNmbWr);
  }

  p->mpRegRabNmbStatus = Module_queryRegister(p->mpModRac, RAC_NMB_STATUS);
  if (p->mpRegRabNmbStatus) {
    p->RAC_NMB_STATUS_ADDR = Register_getAddress(p->mpRegRabNmbStatus);
  }

  //
  //
  //
  p->mDma = NULL;

  pthread_mutex_init(&p->mMutex, NULL);

  return 0;
}

//
//
//
int NthwRac_get_rab_interface_count(const NthwRac_t *p)
{
  return p->mnParamRacRabInterfaces;
}

//
//
//
static inline int
_NthwRac_waitForRabDone(const NthwRac_t *p, uint32_t address, uint32_t wordCnt)
{
  const struct FpgaInfo_s* const pFpgaInfo = p->mpFpga->pFpgaInfo;
  const char* const pAdapterIdStr = pFpgaInfo->mpAdapterIdStr;
  uint32_t used = 0;
  uint32_t retry;

  for (retry = 0; retry < 100000; retry++) {
    NthwRac_regRead32(pFpgaInfo, p->RAC_RAB_BUF_USED_ADDR, &used);
    used = (used & p->RAC_RAB_BUF_USED_OB_USED_MASK) >> p->RAC_RAB_BUF_USED_OB_USED_LOW;
    if (used >= wordCnt) {
      break;
    }
  }

  if (used < wordCnt) {
    NT_LOG(ERR, NTHW, "%s: Fail rab bus r/w addr=0x%08X used=%x wordcount=%d\n", pAdapterIdStr, address, used, wordCnt);
    return -1;
  }
  return 0;
}

//
// NT_PCI_REG_P9xyz_RAC_RAB_INIT
//
// Initializes (resets) the programmable registers on the Register Access Busses (RAB).
// This initialization must be performed by software as part of the driver load procedure.
//
// Bit n of this field initializes the programmable registers on RAB interface n.
// Software must write one to the bit and then clear the bit again.
//
// All RAB module registers will be reset to their defaults.
// This includes the product specific RESET module (eg RST9xyz)
// As a consequence of this behavior the official reset sequence
// must be excersised - as all RAB modules will be held in reset.
//
int NthwRac_rab_init(NthwRac_t* p, uint32_t nRabIntfMask)
{
  // Write rac_rab_init
  // Perform operation twice - first to get trace of operation - second to get things done...
  const struct FpgaInfo_s* const pFpgaInfo = p->mpFpga->pFpgaInfo;
  Field_setValFlush32(p->mpFldRabInit, nRabIntfMask);
  NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_INIT_ADDR, nRabIntfMask);
  return 0;
}

//
//
//
int NthwRac_rab_reset(NthwRac_t* p)
{
  const struct FpgaInfo_s* const pFpgaInfo = p->mpFpga->pFpgaInfo;
  const char* const pAdapterIdStr _unused = pFpgaInfo->mpAdapterIdStr;

  // RAC RAB bus "flip/flip" reset
  const int nRacRabBusCount = NthwRac_get_rab_interface_count(p);
  const int nRacRabBusMask = (1 << nRacRabBusCount) - 1;

  NT_LOG(DBG, NTHW, "%s: NT_RAC_RAB_INTERFACES=%d (0x%02X)\n", pAdapterIdStr, nRacRabBusCount, nRacRabBusMask);
  assert(nRacRabBusCount);
  assert(nRacRabBusMask);

  // RAC RAB bus "flip/flip" reset first stage - new impl (ref RMT#37020)
  NthwRac_rab_init(p, 0);
  NthwRac_rab_init(p, nRacRabBusMask);
  NthwRac_rab_init(p, nRacRabBusMask & ~0x01);

  return 0;
}

//
//
//
int NthwRac_rab_setup(NthwRac_t* p)
{
  int rc = 0;

  const struct FpgaInfo_s* const pFpgaInfo = p->mpFpga->pFpgaInfo;
  uint32_t nDmaBufSize = 2L * RAB_DMA_BUF_CNT * sizeof(uint32_t);
  const size_t align_size = ALIGN_SIZE(nDmaBufSize);
  int numaNode = pFpgaInfo->numa_node;
  uint64_t dmaAddr;
  uint32_t buf;

  if (!p->mDma) {
    struct nt_dma_s *vfio_dma;
    // FPGA needs Page alignment (4K)
    vfio_dma = nt_dma_alloc(align_size, 0x1000, numaNode);

    if (vfio_dma==NULL) {
      NT_LOG(ERR, ETHDEV, "%s: nt_dma_alloc failed\n", __func__);
      return -1;
    }
    p->mDmaInBuf = (uint32_t*)vfio_dma->addr;
    p->mDmaOutBuf = p->mDmaInBuf + RAB_DMA_BUF_CNT;
    p->mDma = vfio_dma;
  }

  // Setup DMA on the adapter
  dmaAddr = p->mDma->iova;
  NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_DMA_IB_LO_ADDR, dmaAddr&0xffffffff);
  NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_DMA_IB_HI_ADDR, (uint32_t)(dmaAddr>>32)&0xffffffff);
  dmaAddr += RAB_DMA_BUF_CNT * sizeof(uint32_t);
  NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_DMA_OB_LO_ADDR, dmaAddr&0xffffffff);
  NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_DMA_OB_HI_ADDR, (uint32_t)(dmaAddr>>32)&0xffffffff);

  // Set initial value of internal pointers
  NthwRac_regRead32(pFpgaInfo, p->RAC_RAB_DMA_IB_RD_ADDR, &buf);
  p->mDmaInPtrWr = (uint16_t)(buf/sizeof(uint32_t));
  NthwRac_regRead32(pFpgaInfo, p->RAC_RAB_DMA_OB_WR_ADDR, &buf);
  p->mDmaOutPtrRd = (uint16_t)(buf/sizeof(uint32_t));
  p->mInFree = RAB_DMA_BUF_CNT;

  return rc;
}

//
//
//
int NthwRac_rab_dma_begin(NthwRac_t* p)
{
  const struct FpgaInfo_s* const pFpgaInfo = p->mpFpga->pFpgaInfo;
  const char* const pAdapterIdStr = pFpgaInfo->mpAdapterIdStr;

  pthread_mutex_lock(&p->mMutex);

  if (p->mDmaActive) {
    pthread_mutex_unlock(&p->mMutex);
    NT_LOG(ERR, NTHW, "%s: DMA begin requested, but a DMA transaction is already active\n", pAdapterIdStr);
    return -1;
  }

  p->mDmaActive = true;

  return 0;
}

//
//
//
static void _NthwRac_rab_dma_activate(NthwRac_t* p)
{
  const struct FpgaInfo_s* const pFpgaInfo = p->mpFpga->pFpgaInfo;
  const uint32_t completion = RAB_COMPLETION << RAB_OPR_LO;

  // Write completion word
  p->mDmaInBuf[p->mDmaInPtrWr] = completion;
  p->mDmaInPtrWr = (uint16_t)((p->mDmaInPtrWr + 1) & (RAB_DMA_BUF_CNT - 1));

  // Clear output completion word
  p->mDmaOutBuf[p->mDmaOutPtrRd] = 0;

  // Update DMA pointer and start transfer
  NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_DMA_IB_WR_ADDR, (uint32_t)(p->mDmaInPtrWr*sizeof(uint32_t)));
}

//
//
//
static int _NthwRac_rab_dma_wait(NthwRac_t* p)
{
  const struct FpgaInfo_s* const pFpgaInfo = p->mpFpga->pFpgaInfo;
  const uint32_t completion = RAB_COMPLETION << RAB_OPR_LO;
  uint32_t i;

  for (i = 0; i < RAB_DMA_WAIT; i++) {
    NtOs_WaituSecPoll(1);
    if ((p->mDmaOutBuf[p->mDmaOutPtrRd]&completion) == completion) {
      break;
    }
  }

  if (i == RAB_DMA_WAIT) {
    NT_LOG(ERR, NTHW, "%s: RAB: Unexpected value of completion (0x%08X)\n", pFpgaInfo->mpAdapterIdStr, p->mDmaOutBuf[p->mDmaOutPtrRd]);
    return -1;
  }

  p->mDmaOutPtrRd = (uint16_t)((p->mDmaOutPtrRd + 1) & (RAB_DMA_BUF_CNT - 1));
  p->mInFree = RAB_DMA_BUF_CNT;

  return 0;
}

//
//
//
int NthwRac_rab_dma_commit(NthwRac_t* p)
{
  int ret;

  if (!p->mDmaActive) {
    // Expecting mutex not to be locked!
    assert(0); // alert developer that something is wrong
    return -1;
  }

  _NthwRac_rab_dma_activate(p);
  ret = _NthwRac_rab_dma_wait(p);

  p->mDmaActive = false;

  pthread_mutex_unlock(&p->mMutex);

  return ret;
}

//
//
//
#if defined(DEBUG) && (0)
static void NthwRac_checkRabTimeout(const struct FpgaInfo_s *pFpgaInfo)
{
  if (pFpgaInfo->mpNthwRac == NULL) {
    /* register addrs not known yet */
    return;
  }

  uint32_t value =(*(volatile uint32_t*)((uint8_t*)(pFpgaInfo->bar0_addr)+(pFpgaInfo->mpNthwRac->RAC_RAB_BUF_FREE_ADDR)));
  if (value & 0x80000000) {
    NT_LOG(ERR,NTHW,"RAB timeout bit set 0x%X\n",value);
  }
}
#endif

//
//
//
void NthwRac_regRead32(const struct FpgaInfo_s *pFpgaInfo, uint32_t regAddr, uint32_t *pData)
{
  *pData = *(volatile uint32_t*)((uint8_t*)pFpgaInfo->bar0_addr+regAddr);
#if defined(DEBUG) && (0)
  NthwRac_checkRabTimeout(pFpgaInfo);
  //NT_LOG(DB2,NTHW,"%s:%X=%X\n",__func__, regAddr, *pData);
#endif
}

//
//
//
void NthwRac_regWrite32(const struct FpgaInfo_s *pFpgaInfo, uint32_t regAddr, uint32_t pData)
{
  *(volatile uint32_t*)((uint8_t*)pFpgaInfo->bar0_addr+regAddr) = pData;
#if defined(DEBUG) && (0)
  NthwRac_checkRabTimeout(pFpgaInfo);
  //NT_LOG(DB2,NTHW,"%s:%X=%X\n", __func__, regAddr, pData);
#endif
}

//
//
//
int NthwRac_rab_write32_dma(NthwRac_t* p, uint32_t address, RabBusId_t busId, uint32_t wordCnt, const uint32_t *pData)
{
  const struct FpgaInfo_s* const pFpgaInfo = p->mpFpga->pFpgaInfo;

  if (wordCnt == 0 || wordCnt > 256) {
    NT_LOG(ERR, NTHW, "%s: Failed rab dma write length check - bus: %d addr: 0x%08X wordcount: %d - inBufFree: 0x%08X\n",
      pFpgaInfo->mpAdapterIdStr, busId, address, wordCnt, p->mInFree);
    assert(0); // alert developer that something is wrong
    return -1;
  }

  if (p->mInFree < (wordCnt + 3)) {
    // No more memory available.
    // NthwRac_rab_dma_commit() needs to be called to start and finish pending transfers.
    return -1;
  }

  p->mInFree -= (wordCnt + 1);

  /* Write the command word */
#if defined(RAB_DEBUG_ECHO)
  p->mDmaInBuf[p->mDmaInPtrWr] = (RAB_WRITE_ECHO<<RAB_OPR_LO) | ((wordCnt&((1<<RAB_CNT_BW)-1)) << RAB_CNT_LO) | (busId << RAB_BUSID_LO) | address;
  p->mDmaOutPtrRd = (uint16_t)((p->mDmaOutPtrRd+wordCnt+1)&(RAB_DMA_BUF_CNT-1));
#else
  p->mDmaInBuf[p->mDmaInPtrWr] = (RAB_WRITE<<RAB_OPR_LO) | ((wordCnt&((1<<RAB_CNT_BW)-1)) << RAB_CNT_LO) | (busId << RAB_BUSID_LO) | address;
#endif
  p->mDmaInPtrWr = (uint16_t)((p->mDmaInPtrWr+1)&(RAB_DMA_BUF_CNT-1));

  for (uint32_t i = 0; i < wordCnt; i++) {
    p->mDmaInBuf[p->mDmaInPtrWr] = pData[i];
    p->mDmaInPtrWr = (uint16_t)((p->mDmaInPtrWr+1)&(RAB_DMA_BUF_CNT-1));
  }

  return 0;
}

//
//
//
int NthwRac_rab_read32_dma(NthwRac_t* p, uint32_t address, RabBusId_t busId, uint32_t wordCnt, struct DmaBufPtr* bufPtr)
{
  const struct FpgaInfo_s* const pFpgaInfo = p->mpFpga->pFpgaInfo;

  if (wordCnt == 0 || wordCnt > 256) {
    NT_LOG(ERR, NTHW, "%s: Failed rab dma read length check - bus: %d addr: 0x%08X wordcount: %d - inBufFree: 0x%08X\n",
      pFpgaInfo->mpAdapterIdStr, busId, address, wordCnt, p->mInFree);
    assert(0); // alert developer that something is wrong
    return -1;
  }

  if ((wordCnt + 3) > RAB_DMA_BUF_CNT) {
    NT_LOG(ERR, NTHW, "%s: Failed rab dma read length check - bus: %d addr: 0x%08X wordcount: %d: 0x%08X",
           pFpgaInfo->mpAdapterIdStr, busId, address, wordCnt);
    return -1;
  }

  if (p->mInFree < 3) {
    // No more memory available.
    // NthwRac_rab_dma_commit() needs to be called to start and finish pending transfers.
    return -1;
  }

  p->mInFree -= 1;

  /* Write the command word */
#if defined(RAB_DEBUG_ECHO)
  p->mDmaInBuf[p->mDmaInPtrWr] = (RAB_READ_ECHO<<RAB_OPR_LO) | ((wordCnt&((1<<RAB_CNT_BW)-1)) << RAB_CNT_LO) | (busId << RAB_BUSID_LO) | address;
  p->mDmaOutPtrRd = (uint16_t)((p->mDmaOutPtrRd+1)&(RAB_DMA_BUF_CNT-1));
#else
  p->mDmaInBuf[p->mDmaInPtrWr] = (RAB_READ<<RAB_OPR_LO) | ((wordCnt&((1<<RAB_CNT_BW)-1)) << RAB_CNT_LO) | (busId << RAB_BUSID_LO) | address;
#endif
  p->mDmaInPtrWr = (uint16_t)((p->mDmaInPtrWr+1)&(RAB_DMA_BUF_CNT-1));

  bufPtr->index = p->mDmaOutPtrRd;
  bufPtr->size = RAB_DMA_BUF_CNT;
  bufPtr->base = p->mDmaOutBuf;
  p->mDmaOutPtrRd = (uint16_t)((p->mDmaOutPtrRd + wordCnt) & (RAB_DMA_BUF_CNT - 1U));

  return 0;
}

//
//
//
int NthwRac_rab_write32(NthwRac_t* p, uint32_t address, RabBusId_t busId, uint32_t wordCnt, const uint32_t *pData)
{
  const struct FpgaInfo_s* const pFpgaInfo = p->mpFpga->pFpgaInfo;
  const char* const pAdapterIdStr = pFpgaInfo->mpAdapterIdStr;
  int res = 0;
  uint32_t rabOperWr;
  uint32_t rabOperCmpl;
  uint32_t rabEchoOperCmpl;
  uint32_t wordCntExpected;
  uint32_t bufUsed;
  uint32_t bufFree;
  uint32_t inBufFree;
  uint32_t outBufFree;

  if (address > (1<<RAB_ADDR_BW)) {
    NT_LOG(ERR, NTHW, "%s: RAB: Illegal address: value too large %d - max %d\n", pAdapterIdStr, address, (1<<RAB_ADDR_BW));
    return -1;
  }

  if (busId > (1<<RAB_BUSID_BW)) {
    NT_LOG(ERR, NTHW, "%s: RAB: Illegal bus id: value too large %d - max %d\n", pAdapterIdStr, busId, (1<<RAB_BUSID_BW));
    return -1;
  }

  if (wordCnt == 0) {
    NT_LOG(ERR, NTHW, "%s: RAB: Illegal word count: value is zero (%d) \n", pAdapterIdStr, wordCnt);
    return -1;
  }

  if (wordCnt > (1<<RAB_CNT_BW)) {
    NT_LOG(ERR, NTHW, "%s: RAB: Illegal word count: value too large %d - max %d\n", pAdapterIdStr, wordCnt, (1<<RAB_CNT_BW));
    return -1;
  }

  pthread_mutex_lock(&p->mMutex);

  if (p->mDmaActive) {
    NT_LOG(ERR, NTHW, "%s: RAB: Illegal operation: DMA enabled\n", pAdapterIdStr);
    res = -1;
    goto exit_unlock_res;
  }

  // Read buffer free register
  NthwRac_regRead32(pFpgaInfo, p->RAC_RAB_BUF_FREE_ADDR, &bufFree);

  inBufFree = bufFree & p->RAC_RAB_BUF_FREE_IB_FREE_MASK;
  outBufFree = (bufFree & p->RAC_RAB_BUF_FREE_OB_FREE_MASK) >> 16;

  // Read buffer used register
  NthwRac_regRead32(pFpgaInfo, p->RAC_RAB_BUF_USED_ADDR, &bufUsed);

  bufUsed = bufUsed & (p->RAC_RAB_BUF_USED_IB_USED_MASK | p->RAC_RAB_BUF_USED_OB_USED_MASK);

  // Verify that output buffer can hold one completion word,
  // input buffer can hold the number of words to be written +
  // one write and one completion command
  // and that the input and and output "used" buffer is 0
  if (((outBufFree >= 1) && (inBufFree >= wordCnt+2)) && bufUsed == 0) {
    uint32_t i;

    wordCntExpected = 0;

    // Compose write command
#if defined(RAB_DEBUG_ECHO)
    rabOperWr = (RAB_WRITE_ECHO<<RAB_OPR_LO) | ((wordCnt&((1<<RAB_CNT_BW)-1)) << RAB_CNT_LO) | (busId << RAB_BUSID_LO) | address;
    wordCntExpected += wordCnt+1;
#else
    rabOperWr = (RAB_WRITE<<RAB_OPR_LO) | ((wordCnt&((1<<RAB_CNT_BW)-1)) << RAB_CNT_LO) | (busId << RAB_BUSID_LO) | address;
#endif // RAB_DEBUG_ECHO

    // Write command
    NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_IB_DATA_ADDR, rabOperWr);

    // Write da to input buffer
    for (i = 0; i < wordCnt; i++) {
      NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_IB_DATA_ADDR, *pData);
      pData++;
    }

    // Compose completion command
    rabOperCmpl = (RAB_COMPLETION<<RAB_OPR_LO);
    wordCntExpected++;

    // Write command
    NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_IB_DATA_ADDR, rabOperCmpl);

    // Wait until done
    if (_NthwRac_waitForRabDone(p, address, wordCntExpected)) {
      res = -1;
      goto exit_unlock_res;
    }

#if defined(RAB_DEBUG_ECHO)
    {
      uint32_t rabEchoOperWr;
      NthwRac_regRead32(pFpgaInfo, p->RAC_RAB_OB_DATA_ADDR, &rabEchoOperWr);
      if (p->mnParamRacRabObUpdate) {
        NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_OB_DATA_ADDR, 0);
      }
      if (rabOperWr != rabEchoOperWr) {
        NT_LOG(ERR, NTHW, "%s: expected rab read echo oper (0x%08X) - read (0x%08X)\n", pAdapterIdStr, rabOperWr, rabEchoOperWr);
      }
    }

    {
      // Read data from output buffer
      uint32_t data;
      for (i = 0; i < wordCnt; i++) {
        NthwRac_regRead32(pFpgaInfo, p->RAC_RAB_OB_DATA_ADDR, &data);
        if (p->mnParamRacRabObUpdate) {
          NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_OB_DATA_ADDR, 0);
        }
      }
    }
#endif // RAB_DEBUG_ECHO

    // Read completion from out buffer
    NthwRac_regRead32(pFpgaInfo, p->RAC_RAB_OB_DATA_ADDR, &rabEchoOperCmpl);
    if (p->mnParamRacRabObUpdate) {
      NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_OB_DATA_ADDR, 0);
    }
    if (rabEchoOperCmpl != rabOperCmpl) {
      NT_LOG(ERR, NTHW, "%s: RAB: Unexpected value of completion (0x%08X)- inBufFree: 0x%08X, outBufFree: 0x%08X, bufUsed: 0x%08X\n",
        pAdapterIdStr, rabEchoOperCmpl, inBufFree, outBufFree, bufUsed);
      res = -1;
      goto exit_unlock_res;
    }

#if defined(RAB_DEBUG_ECHO) && (0)
    loghex(DBG, NULL, 0, "%s: RAB: oper=0x%08x", __func__, physicalAdapterNo, rabEchoOperWr);
    NT_LOG(DBG, NTHW, "%s: RAB: oper=0x%08x", __func__, physicalAdapterNo, rabEchoOperWr);
#endif // RAB_DEBUG_ECHO

    // Read buffer free register
    NthwRac_regRead32(pFpgaInfo, p->RAC_RAB_BUF_FREE_ADDR, &bufFree);
    if (bufFree & 0x80000000) {
      // Clear Timeout and overflow bits
      NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_BUF_FREE_ADDR, 0x0);
      NT_LOG(ERR, NTHW, "%s: RAB: timeout - Access outside register - bus: %d addr: 0x%08X - inBufFree: 0x%08X, outBufFree: 0x%08X, bufUsed: 0x%08X\n",
        pAdapterIdStr, busId, address, inBufFree, outBufFree, bufUsed);
      res = -1;
      goto exit_unlock_res;
    }

    res = 0;
    goto exit_unlock_res;
  }
  else {
    NT_LOG(ERR, NTHW, "%s: RAB: Fail rab bus buffer check - bus: %d addr: 0x%08X wordcount: %d - inBufFree: 0x%08X, outBufFree: 0x%08X, bufUsed: 0x%08X\n",
      pAdapterIdStr, busId, address, wordCnt, inBufFree, outBufFree, bufUsed);
    res = -1;
    goto exit_unlock_res;
  }

exit_unlock_res:
  pthread_mutex_unlock(&p->mMutex);
  return res;
}


//
//
//
int NthwRac_rab_read32(NthwRac_t* p, uint32_t address, RabBusId_t busId, uint32_t wordCnt, uint32_t* pData)
{
  const struct FpgaInfo_s* const pFpgaInfo = p->mpFpga->pFpgaInfo;
  const char* const pAdapterIdStr = pFpgaInfo->mpAdapterIdStr;
  int res = 0;
  uint32_t rabOperRd;
  uint32_t wordCntExpected;
  uint32_t bufUsed;
  uint32_t bufFree;
  uint32_t inBufFree;
  uint32_t outBufFree;

  pthread_mutex_lock(&p->mMutex);

  if (address > (1<<RAB_ADDR_BW)) {
    NT_LOG(ERR, NTHW, "%s: RAB: Illegal address: value too large %d - max %d\n", pAdapterIdStr, address, (1<<RAB_ADDR_BW));
    res = -1;
    goto exit_unlock_res;
  }

  if (busId > (1<<RAB_BUSID_BW)) {
    NT_LOG(ERR, NTHW, "%s: RAB: Illegal bus id: value too large %d - max %d\n", pAdapterIdStr, busId, (1<<RAB_BUSID_BW));
    res = -1;
    goto exit_unlock_res;
  }

  if (wordCnt == 0) {
    NT_LOG(ERR, NTHW, "%s: RAB: Illegal word count: value is zero (%d) \n", pAdapterIdStr, wordCnt);
    res = -1;
    goto exit_unlock_res;
  }

  if (wordCnt > (1<<RAB_CNT_BW)) {
    NT_LOG(ERR, NTHW, "%s: RAB: Illegal word count: value too large %d - max %d\n", pAdapterIdStr, wordCnt, (1<<RAB_CNT_BW));
    res = -1;
    goto exit_unlock_res;
  }

  // Read buffer free register
  NthwRac_regRead32(pFpgaInfo, p->RAC_RAB_BUF_FREE_ADDR, &bufFree);

  inBufFree = bufFree & p->RAC_RAB_BUF_FREE_IB_FREE_MASK;
  outBufFree = (bufFree & p->RAC_RAB_BUF_FREE_OB_FREE_MASK) >> 16;

  // Read buffer used register
  NthwRac_regRead32(pFpgaInfo, p->RAC_RAB_BUF_USED_ADDR, &bufUsed);

  bufUsed = bufUsed & (p->RAC_RAB_BUF_USED_IB_USED_MASK | p->RAC_RAB_BUF_USED_OB_USED_MASK);

  // Verify that output buffer can hold the number of words to be read,
  // input buffer can hold one read command
  // and that the input and and output "used" buffer is 0
  if (((outBufFree >= wordCnt) && (inBufFree >= 1)) && bufUsed == 0) {
    wordCntExpected = wordCnt;

#if defined(RAB_DEBUG_ECHO)
    rabOperRd = (RAB_READ_ECHO<<RAB_OPR_LO) | ((wordCnt&((1<<RAB_CNT_BW)-1))<<RAB_CNT_LO) | (busId << RAB_BUSID_LO) | address;
    wordCntExpected++;
#else
    rabOperRd = (RAB_READ<<RAB_OPR_LO)|(wordCnt<<RAB_CNT_LO)|(busId << RAB_BUSID_LO) | address;
#endif // RAB_DEBUG_ECHO

    NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_IB_DATA_ADDR, rabOperRd);

    // Wait until done
    if (_NthwRac_waitForRabDone(p, address, wordCntExpected)) {
      res = -1;
      goto exit_unlock_res;
    }

#if defined(RAB_DEBUG_ECHO)
    uint32_t rabEchoOperRd;
    NthwRac_regRead32(pFpgaInfo, p->RAC_RAB_OB_DATA_ADDR, &rabEchoOperRd);
    if (p->mnParamRacRabObUpdate) {
      NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_OB_DATA_ADDR, 0);
    }
    if (rabOperRd != rabEchoOperRd) {
      NT_LOG(ERR, NTHW, "%s: RAB: expected rab read echo oper (0x%08X) - read (0x%08X)\n", pAdapterIdStr, rabOperRd, rabEchoOperRd);
    }
#endif // RAB_DEBUG_ECHO

    // Read data from output buffer
    {
      uint32_t i;
      for (i = 0; i < wordCnt; i++) {
        NthwRac_regRead32(pFpgaInfo, p->RAC_RAB_OB_DATA_ADDR, pData);
        if (p->mnParamRacRabObUpdate) {
          NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_OB_DATA_ADDR, 0);
        }
        pData++;
      }
    }

#if defined(RAB_DEBUG_ECHO) && (0)
    loghex(DBG, pData, wordCnt, "%s: RAB: oper=0x%08x", pAdapterIdStr, rabEchoOperRd);
    NT_LOG(DBG, NTHW, "%s: RAB: oper=0x%08x", pAdapterIdStr, rabEchoOperRd);
#endif // RAB_DEBUG_ECHO

#if defined(DEBUG) && (0)
    NT_LOG(DBG, NTHW, "%s: RAB: inBufFree: 0x%08X, outBufFree: 0x%08X, bufUsed: 0x%08X\n", pAdapterIdStr, inBufFree, outBufFree, bufUsed);
#endif

    // Read buffer free register
    NthwRac_regRead32(pFpgaInfo, p->RAC_RAB_BUF_FREE_ADDR, &bufFree);
    if (bufFree & 0x80000000) {
      // Clear Timeout and overflow bits
      NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_BUF_FREE_ADDR, 0x0);
      NT_LOG(ERR, NTHW, "%s: RAB: timeout - Access outside register - bus: %d addr: 0x%08X - inBufFree: 0x%08X, outBufFree: 0x%08X, bufUsed: 0x%08X\n",
        pAdapterIdStr, busId, address, inBufFree, outBufFree, bufUsed);
      res = -1;
      goto exit_unlock_res;
    }

    res = 0;
    goto exit_unlock_res;
  }
  else {
    NT_LOG(ERR, NTHW, "%s: RAB: Fail rab bus buffer check - bus: %d addr: 0x%08X wordcount: %d - inBufFree: 0x%08X, outBufFree: 0x%08X, bufUsed: 0x%08X\n",
      pAdapterIdStr, busId, address, wordCnt, inBufFree, outBufFree, bufUsed);
    res = -1;
    goto exit_unlock_res;
  }

exit_unlock_res:
  pthread_mutex_unlock(&p->mMutex);
  return res;
}

//
//
//
int NthwRac_rab_flush(NthwRac_t* p)
{
  const struct FpgaInfo_s* const pFpgaInfo = p->mpFpga->pFpgaInfo;
  const char* const pAdapterIdStr = pFpgaInfo->mpAdapterIdStr;
  uint32_t data = 0;
  uint32_t retry;
  int res = 0;

  pthread_mutex_lock(&p->mMutex);

  // Set the flush bit
  NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_BUF_USED_ADDR, p->RAC_RAB_BUF_USED_FLUSH_MASK);

  // Reset BUF FREE register
  NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_BUF_FREE_ADDR, 0x0);

  // Wait until OB_USED and IB_USED are 0
  for (retry = 0; retry < 100000; retry++) {
    NthwRac_regRead32(pFpgaInfo, p->RAC_RAB_BUF_USED_ADDR, &data);

    if ((data & 0xFFFFFFFF) == p->RAC_RAB_BUF_USED_FLUSH_MASK) {
      break;
    }
  }

  if (data != p->RAC_RAB_BUF_USED_FLUSH_MASK) {
    NT_LOG(ERR, NTHW, "%s: RAB: Rab bus flush error. \n", pAdapterIdStr);
    res = -1;
  }

  // Clear flush bit when done
  NthwRac_regWrite32(pFpgaInfo, p->RAC_RAB_BUF_USED_ADDR, 0x0);

  pthread_mutex_unlock(&p->mMutex);
  return res;
}


//
// EOF
//
