//
// nthw_stat.c
//

//
// Includes:
//

#include "nt_util.h"
#include "ntlog.h"
//
#include "nthw_drv.h"
#include "nthw_register.h"

//
//
//
#include "nthw_stat.h"

//
#include <malloc.h>

//
//
//
NthwStat_t* NthwStat_new(void)
{
  NthwStat_t* p = malloc(sizeof(NthwStat_t));
  if (p) {
    memset(p, 0, sizeof(NthwStat_t));
  }
  return p;
}

//
//
//
void NthwStat_delete(NthwStat_t* p)
{
  if (p) {
    free(p);
  }
}

//
//
//
int NthwStat_init(NthwStat_t *p, Fpga_t *pFpga, int nInstance)
{
  const char* const pAdapterIdStr = pFpga->pFpgaInfo->mpAdapterIdStr;
  uint64_t nModuleVersionPacked64 = -1;
  struct Module_s *mod = Fpga_queryModule(pFpga, MOD_STA, nInstance);

  if (p == NULL)
    return mod == NULL ? -1 : 0;

  if (mod == NULL) {
    NT_LOG(ERR, NTHW, "%s: STAT %d: no such instance\n", pAdapterIdStr, nInstance);
    return -1;
  }

  p->mpFpga = pFpga;
  p->mnInstance = nInstance;
  p->mpModStat = mod;

  nModuleVersionPacked64 = Module_getVersionPacked64(p->mpModStat);
  NT_LOG(DBG, NTHW, "%s: STAT %d: version=0x%08lX\n", pAdapterIdStr, p->mnInstance, nModuleVersionPacked64 );

#if defined(DEBUG) && (0)
  Module_setDebugMode(p->mpModStat, 0x00); // DEBUG:
#endif

  //
  {
    Register* pReg;
    // STA_CFG register
    pReg = Module_getRegister(p->mpModStat, STA_CFG);
    p->mpFldDmaEna = Register_getField(pReg, STA_CFG_DMA_ENA);
    p->mpFldCntClear = Register_getField(pReg, STA_CFG_CNT_CLEAR);

    // CFG: fields NOT available from v. 3
    p->mpFldTxDisable = Register_queryField(pReg, STA_CFG_TX_DISABLE);
    p->mpFldCntFreeze = Register_queryField(pReg, STA_CFG_CNT_FRZ);

    // STA_STATUS register
    pReg = Module_getRegister(p->mpModStat, STA_STATUS);
    p->mpFldStatToggleMissed = Register_getField(pReg, STA_STATUS_STAT_TOGGLE_MISSED);

    // HOST_ADR registers
    pReg = Module_getRegister(p->mpModStat, STA_HOST_ADR_LSB);
    p->mpFldDmaLsb = Register_getField(pReg, STA_HOST_ADR_LSB_LSB);

    pReg = Module_getRegister(p->mpModStat, STA_HOST_ADR_MSB);
    p->mpFldDmaMsb = Register_getField(pReg, STA_HOST_ADR_MSB_MSB);
  }

  //
  // Params
  //
  p->mbIsVswitch = pFpga->pFpgaInfo->profile == FPGA_INFO_PROFILE_VSWITCH;

  p->mNbNimPorts = Fpga_getProductParam(pFpga, NT_NIMS, 0);
  p->mNbPhyPorts = Fpga_getProductParam(pFpga, NT_PHY_PORTS, 0);

  p->mNbRxPorts = Fpga_getProductParam(pFpga, NT_STA_RX_PORTS, -1); // VSWITCH
  if (p->mNbRxPorts == -1) {
    p->mNbRxPorts = Fpga_getProductParam(pFpga, NT_RX_PORTS, -1); // non-VSWITCH
    if (p->mNbRxPorts == -1) {
      p->mNbRxPorts = Fpga_getProductParam(pFpga, NT_PORTS, 0); // non-VSWITCH
    }
  }


  p->mNbTxPorts = Fpga_getProductParam(pFpga, NT_TX_PORTS, 0);
  p->mRxPortReplicate = Fpga_getProductParam(pFpga, NT_RX_PORT_REPLICATE, 0);

  p->mNbColorCounters =  Fpga_getProductParam(pFpga, NT_STA_COLORS, 64)*2; // VSWITCH
  if (p->mNbColorCounters == 0) {
    p->mNbColorCounters = Fpga_getProductParam(pFpga, NT_CAT_FUNCS, 0)*2; // non-VSWITCH
  }

  p->mNbRxHostBuffers = Fpga_getProductParam(pFpga, NT_QUEUES, 0);
  p->mNbTxHostBuffers = p->mNbRxHostBuffers;

  p->mDbsPresent = Fpga_getProductParam(pFpga, NT_DBS_PRESENT, 0);

  p->mNbRxHbCounters = (p->mNbRxHostBuffers * (6 + 2 * (nModuleVersionPacked64 >= VERSION_PACKED64(0, 6) ? p->mDbsPresent : 0)));

  p->mNbTxHbCounters = 0;

  p->mNbRxPortCounters = 42 + 2 * (nModuleVersionPacked64 >= VERSION_PACKED64(0, 6) ? p->mDbsPresent : 0);
  p->mNbTxPortCounters = 0;

  p->mNbCounters = p->mNbColorCounters + p->mNbRxHbCounters + p->mNbTxHbCounters;

  p->mnStatLayoutVersion = 0;
  if (nModuleVersionPacked64 >= VERSION_PACKED64(0, 8)) {
    p->mnStatLayoutVersion = 6;
  } else if (nModuleVersionPacked64 >= VERSION_PACKED64(0, 6)) {
    p->mnStatLayoutVersion = 5;
  } else if (nModuleVersionPacked64 >= VERSION_PACKED64(0, 4)) {
    p->mnStatLayoutVersion = 4;
  } else if (nModuleVersionPacked64 >= VERSION_PACKED64(0, 3)) {
    p->mnStatLayoutVersion = 3;
  } else if (nModuleVersionPacked64 >= VERSION_PACKED64(0, 2)) {
    p->mnStatLayoutVersion = 2;
  } else if (nModuleVersionPacked64 > VERSION_PACKED64(0, 0)) {
    p->mnStatLayoutVersion = 1;
  } else {
    p->mnStatLayoutVersion = 0;
    NT_LOG(ERR, NTHW, "%s: unknown module_version 0x%08lX layout=%d\n", pAdapterIdStr, nModuleVersionPacked64, p->mnStatLayoutVersion);
  }
  assert(p->mnStatLayoutVersion);

  // STA module 0.2+ adds IPF counters per port (Rx feature)
  if (nModuleVersionPacked64 >= VERSION_PACKED64(0, 2)) {
    p->mNbRxPortCounters += 6;
  }

  // STA module 0.3+ adds TX stats
  if ((nModuleVersionPacked64 >= VERSION_PACKED64(0, 3)) || (p->mNbTxPorts >= 1))  {
    p->mbHasTxStats = true;
  }

  // STA module 0.3+ adds TX stat counters
  if (nModuleVersionPacked64 >= VERSION_PACKED64(0, 3))  {
    p->mNbTxPortCounters += 22;
  }

  // STA module 0.4+ adds TX drop event counter
  if (nModuleVersionPacked64 >= VERSION_PACKED64(0, 4))  {
    p->mNbTxPortCounters += 1; // TX drop event counter
  }

  // STA module 0.6+ adds pkt filter drop octets+pkts, retransmit and duplicate counters
  if (nModuleVersionPacked64 >= VERSION_PACKED64(0, 6))  {
    p->mNbRxPortCounters += 4;
    p->mNbTxPortCounters += 1;
  }

  //
  if (p->mbIsVswitch) {
    p->mNbRxPortCounters = 5;
    p->mNbTxPortCounters = 5;
  }

  p->mNbCounters += (p->mNbRxPorts * p->mNbRxPortCounters);

  if (p->mbHasTxStats) {
    p->mNbCounters += (p->mNbTxPorts * p->mNbTxPortCounters);
  }

  //
  // Output params (debug)
  //
  NT_LOG(DBG, NTHW, "%s: nims=%d rxports=%d txports=%d rxrepl=%d colors=%d queues=%d\n", pAdapterIdStr, p->mNbNimPorts, p->mNbRxPorts, p->mNbTxPorts, p->mRxPortReplicate, p->mNbColorCounters, p->mNbRxHostBuffers);
  NT_LOG(DBG, NTHW, "%s: hbs=%d hbcounters=%d rxcounters=%d txcounters=%d\n", pAdapterIdStr, p->mNbRxHostBuffers, p->mNbRxHbCounters, p->mNbRxPortCounters, p->mNbTxPortCounters);
  NT_LOG(DBG, NTHW, "%s: layout=%d \n", pAdapterIdStr, p->mnStatLayoutVersion);
  NT_LOG(DBG, NTHW, "%s: counters=%d (0x%X)\n", pAdapterIdStr, p->mNbCounters, p->mNbCounters);
  NT_LOG(DBG, NTHW, "%s: vswitch=%d\n", pAdapterIdStr, p->mbIsVswitch);

  //
  // Init
  //
  if (p->mpFldTxDisable) {
    Field_setFlush(p->mpFldTxDisable);
  }

  Field_updateRegister(p->mpFldCntClear);
  Field_setFlush(p->mpFldCntClear);
  Field_clrFlush(p->mpFldCntClear);

  Field_updateRegister(p->mpFldStatToggleMissed);
  Field_setFlush(p->mpFldStatToggleMissed);

  Field_updateRegister(p->mpFldDmaEna);
  Field_clrFlush(p->mpFldDmaEna);
  Field_updateRegister(p->mpFldDmaEna);

  return 0;
}

//
//
//
int NthwStat_setDmaAddress(NthwStat_t* p, uint64_t statDmaPhysical, uint32_t* pStatDmaVirtual)
{
  p->mpTimestamp = NULL;

  p->mStatDmaPhysical = statDmaPhysical;
  p->mpStatDmaVirtual = pStatDmaVirtual;

  if (p->mpStatDmaVirtual) {
    memset(p->mpStatDmaVirtual, 0, (p->mNbCounters * sizeof(uint32_t)));
  }

  Field_setValFlush32(p->mpFldDmaMsb, (uint32_t)((p->mStatDmaPhysical>>32) & 0xffffffff));
  Field_setValFlush32(p->mpFldDmaLsb, (uint32_t)(p->mStatDmaPhysical & 0xffffffff));

  p->mpTimestamp = (uint64_t *)(p->mpStatDmaVirtual + p->mNbCounters);
  NT_LOG(DBG, NTHW, "NthwStat_setDmaAddress: statDmaPhysical=%"PRIX64" pStatDmaVirtual=%"PRIX64" mpTimestamp=%"PRIX64"\n", p->mStatDmaPhysical, p->mpStatDmaVirtual, p->mpTimestamp);
  if (p->mbIsVswitch) {
    *p->mpTimestamp = NtOs_GetTimeNs();
  } else {
    *p->mpTimestamp = (uint64_t)(int64_t)-1;
  }
  return 0;
}

//
//
//
int NthwStat_trigger(NthwStat_t* p)
{
  int nToggleMiss = Field_getUpdated(p->mpFldStatToggleMissed);
  if (nToggleMiss) {
    Field_setFlush(p->mpFldStatToggleMissed);
  }

  if (p->mpTimestamp) {
    *p->mpTimestamp = -1; // Clear old ts
  }

  Field_updateRegister(p->mpFldDmaEna);
  Field_setFlush(p->mpFldDmaEna);

  return 0;
}

//
// EOF
//
