#include <unistd.h>
#include <time.h>
#include <rte_mbuf.h>
#include <rte_pci.h>

#include "ntos_drv.h"
#include "nt_util.h"
#include "ntnic_dbsconfig.h"
#include "ntnic_ethdev.h"
#include "ntlog.h"

#define STRUCT_ALIGNMENT (4*1024LU)
#define MAX_VIRT_QUEUES 128

#define LAST_QUEUE 127
#define DISABLE 0
#define ENABLE 1
#define RX_AM_DISABLE DISABLE
#define RX_AM_ENABLE ENABLE
#define RX_UW_DISABLE DISABLE
#define RX_UW_ENABLE ENABLE
#define RX_Q_DISABLE DISABLE
#define RX_Q_ENABLE ENABLE
#define RX_AM_POLL_SPEED 5
#define RX_UW_POLL_SPEED 9
#define HOST_ID 0
#define INIT_QUEUE 1


#define TX_AM_DISABLE DISABLE
#define TX_AM_ENABLE ENABLE
#define TX_UW_DISABLE DISABLE
#define TX_UW_ENABLE ENABLE
#define TX_Q_DISABLE DISABLE
#define TX_Q_ENABLE ENABLE
#define TX_AM_POLL_SPEED 5
#define TX_UW_POLL_SPEED 8


/**************************************************************************/

#define le64 uint64_t
#define le32 uint32_t
#define le16 uint16_t

/**************************************************************************/

#define VIRTQ_AVAIL_F_NO_INTERRUPT 1

struct virtq_avail {
  le16 flags;
  le16 idx;
  le16 ring[];  /* Queue Size */
  //le16 used_event; /* Only if VIRTIO_F_EVENT_IDX */
} __attribute__((__packed__));

/**************************************************************************/

/* le32 is used here for ids for padding reasons. */
struct virtq_used_elem {
  /* Index of start of used descriptor chain. */
  le32 id;
  /* Total length of the descriptor chain which was used (written to) */
  le32 len;
} __attribute__((__packed__));

#define VIRTQ_USED_F_NO_NOTIFY 1

struct virtq_used {
  le16 flags;
  le16 idx;
  struct virtq_used_elem ring[];  /* Queue Size */
  //le16 avail_event; /* Only if VIRTIO_F_EVENT_IDX */
} __attribute__((__packed__));



struct virtq_struct_layout_s
{
  size_t used_offset;
  size_t desc_offset;
//  size_t packet_size;
//  size_t aligned_packet_size;
//  size_t packets_per_block;
};

enum nthw_virt_queue_usage {UNUSED = 0, UNMANAGED, MANAGED};

#define PACKED(vq_type) (vq_type == PACKED_RING ? 1 : 0)

struct nthw_virt_queue
{
  // Pointers to virt-queue structs
  union {
    struct {
      /* SPLIT virtqueue */
      struct virtq_avail* pAvail;
      struct virtq_used* pUsed;
      struct virtq_desc* pDesc;
      // Control variables for virt-queue structs
      le16 amIdx;
      le16 usedIdx;
      le16 cachedIdx;
      le16 txDescrAvailIdx;
    };
    struct {
      /* PACKED virtqueue */
      struct pvirtq_event_suppress *driver_event;
      struct pvirtq_event_suppress *device_event;
      struct pvirtq_desc *desc;
//      int *sgs;                         // same size as descriptor elements. Index is a bufferID active in descriptor ring. value holds number of segments
      struct {
        le16 next;
        le16 num;
      } outs;                           // when in-order release used Tx packets from FPGA it may collapse into a batch. When getting new Tx buffers we may only need partial
      le16 next_avail;
      le16 next_used;
      le16 avail_wrap_count;
      le16 used_wrap_count;
    };
  };

  // Array with packet buffers
  struct nthw_memory_descriptor* pVirtualAddr;

  // Queue configuration info
  enum nthw_virt_queue_usage usage;
  uint16_t vq_type;
  uint16_t in_order;
  int irq_vector;

  NthwDbs_t *mpNthwDbs;
  uint32_t index;
  le16 queueSize;
  uint32_t am_enable;
  uint32_t hostId;
  uint32_t port;        // Only used by TX queues
  uint32_t virtualPort; // Only used by TX queues
  uint32_t header;      // Only used by TX queues  0: VirtIO-Net header (12 bytes). 1: Napatech DVIO0 descriptor (12 bytes).
  void* availStructPhysAddr;
  void* usedStructPhysAddr;
  void* descStructPhysAddr;

};

struct pvirtq_struct_layout_s {
  size_t driver_event_offset;
  size_t device_event_offset;
};


static struct nthw_virt_queue rxvq[MAX_VIRT_QUEUES];
static struct nthw_virt_queue txvq[MAX_VIRT_QUEUES];

static void dbs_init_rx_queue(NthwDbs_t *pNthwDbs, uint32_t queue, uint32_t start_idx, uint32_t start_ptr)
{
  uint32_t busy;
  uint32_t init;
  uint32_t dummy;

  do
  {
    get_rx_init(pNthwDbs, &init, &dummy, &busy);
  } while (busy != 0);

  set_rx_init(pNthwDbs, start_idx, start_ptr, INIT_QUEUE, queue);

  do
  {
    get_rx_init(pNthwDbs, &init, &dummy, &busy);
  } while (busy != 0);
}

static void dbs_init_tx_queue(NthwDbs_t *pNthwDbs, uint32_t queue, uint32_t start_idx, uint32_t start_ptr)
{
  uint32_t busy;
  uint32_t init;
  uint32_t dummy;

  do
  {
    get_tx_init(pNthwDbs, &init, &dummy, &busy);
  } while (busy != 0);

  set_tx_init(pNthwDbs, start_idx, start_ptr, INIT_QUEUE, queue);

  do
  {
    get_tx_init(pNthwDbs, &init, &dummy, &busy);
  } while (busy != 0);
}

//
//
//
int nthw_virt_queue_init(struct FpgaInfo_s* pFpgaInfo)
{
  assert(pFpgaInfo);

  Fpga_t* const pFpga = pFpgaInfo->mpFpga;
  NthwDbs_t* pNthwDbs;
  int res = 0;
  uint32_t i;

  pFpgaInfo->mpNthwDbs = NULL;

  pNthwDbs = NthwDbs_new();
  if (pNthwDbs == NULL) {
    return -1;
  }

  res = dbs_init(NULL, pFpga, 0);  // Check that DBS exists in FPGA
  if (res) {
    free(pNthwDbs);
    return res;
  }

  res = dbs_init(pNthwDbs, pFpga, 0);  // Create DBS module
  if (res) {
    free(pNthwDbs);
    return res;
  }

  pFpgaInfo->mpNthwDbs = pNthwDbs;

  for(i = 0; i < MAX_VIRT_QUEUES; ++i)
  {
    rxvq[i].usage = UNUSED;
    txvq[i].usage = UNUSED;
  }

  dbs_reset(pNthwDbs);

  for(i = 0; i < NT_DBS_RX_QUEUES_MAX; ++i) {
    dbs_init_rx_queue(pNthwDbs, i, 0, 0);
  }

  for(i = 0; i < NT_DBS_TX_QUEUES_MAX; ++i) {
    dbs_init_tx_queue(pNthwDbs, i, 0, 0);
  }

  set_rx_control(pNthwDbs, LAST_QUEUE, RX_AM_DISABLE, RX_AM_POLL_SPEED, RX_UW_DISABLE, RX_UW_POLL_SPEED, RX_Q_DISABLE);
  set_rx_control(pNthwDbs, LAST_QUEUE, RX_AM_ENABLE, RX_AM_POLL_SPEED, RX_UW_ENABLE, RX_UW_POLL_SPEED, RX_Q_DISABLE);
  set_rx_control(pNthwDbs, LAST_QUEUE, RX_AM_ENABLE, RX_AM_POLL_SPEED, RX_UW_ENABLE, RX_UW_POLL_SPEED, RX_Q_ENABLE);

  set_tx_control(pNthwDbs, LAST_QUEUE, TX_AM_DISABLE, TX_AM_POLL_SPEED, TX_UW_DISABLE, TX_UW_POLL_SPEED, TX_Q_DISABLE);
  set_tx_control(pNthwDbs, LAST_QUEUE, TX_AM_ENABLE, TX_AM_POLL_SPEED, TX_UW_ENABLE, TX_UW_POLL_SPEED, TX_Q_DISABLE);
  set_tx_control(pNthwDbs, LAST_QUEUE, TX_AM_ENABLE, TX_AM_POLL_SPEED, TX_UW_ENABLE, TX_UW_POLL_SPEED, TX_Q_ENABLE);

  return 0;
}

static struct virtq_struct_layout_s dbs_calc_struct_layout(uint32_t queueSize)
{
  size_t availMem = sizeof(struct virtq_avail) + queueSize * sizeof(le16); // + sizeof(le16); ("avail->used_event" is not used)
  size_t availMemAligned = ((availMem % STRUCT_ALIGNMENT) == 0) ? availMem : STRUCT_ALIGNMENT * (availMem/STRUCT_ALIGNMENT + 1);

  size_t usedMem = sizeof(struct virtq_used) + queueSize * sizeof(struct virtq_used_elem); // + sizeof(le16); ("used->avail_event" is not used)
  size_t usedMemAligned = ((usedMem % STRUCT_ALIGNMENT) == 0) ? usedMem : STRUCT_ALIGNMENT * (usedMem/STRUCT_ALIGNMENT + 1);

  struct virtq_struct_layout_s virtq_layout;
  virtq_layout.used_offset = availMemAligned;
  virtq_layout.desc_offset = availMemAligned + usedMemAligned;

  return virtq_layout;
}

static void dbs_initialize_avail_struct(void* addr, uint16_t queueSize, uint16_t initialAvailIdx)
{
  uint16_t i;
  struct virtq_avail* pAvail = (struct virtq_avail*)addr;

  pAvail->flags = VIRTQ_AVAIL_F_NO_INTERRUPT;
  pAvail->idx = initialAvailIdx;
  for (i=0; i < queueSize; ++i)
  {
    pAvail->ring[i]= i;
  }
}

static void dbs_initialize_used_struct(void* addr, uint16_t queueSize)
{
  int i;
  struct virtq_used* pUsed = (struct virtq_used*)addr;

  pUsed->flags = 1;
  pUsed->idx = 0;
  for (i=0; i < queueSize; ++i)
  {
    pUsed->ring[i].id = 0;
    pUsed->ring[i].len = 0;
  }
}

static void dbs_initialize_descriptor_struct(void* addr,
                                             struct nthw_memory_descriptor* packetBufferDescriptors,
                                             uint16_t queueSize, ule16 flgs)
{
  if (packetBufferDescriptors)
  {
    int i;
    struct virtq_desc* pDesc = (struct virtq_desc*)addr;

    for (i=0; i < queueSize; ++i)
    {
      pDesc[i].addr = (uint64_t)packetBufferDescriptors[i].physAddr;
      pDesc[i].len = packetBufferDescriptors[i].len;
      pDesc[i].flags = flgs;
      pDesc[i].next = 0;
    }
  }
}

static void dbs_initialize_virt_queue_structs(void* availStructAddr,
                                              void* usedStructAddr,
                                              void* descStructAddr,
                                              struct nthw_memory_descriptor* packetBufferDescriptors,
                                              uint16_t queueSize,
                                              uint16_t initialAvailIdx, ule16 flgs)
{
  dbs_initialize_avail_struct(availStructAddr, queueSize, initialAvailIdx);
  dbs_initialize_used_struct(usedStructAddr, queueSize);
  dbs_initialize_descriptor_struct(descStructAddr, packetBufferDescriptors, queueSize, flgs);
}


static le16 dbs_qsize_log2(le16 qsize)
{
  uint32_t qs = 0;
  while (qsize)
  {
    qsize = qsize >> 1;
    ++qs;
  }
  --qs;
  return qs;
}

struct nthw_virt_queue* nthw_setup_rx_virt_queue(NthwDbs_t *pNthwDbs,
                                                 uint32_t index,
                                                 uint16_t start_idx,
                                                 uint16_t start_ptr,
                                                 void* availStructPhysAddr,
                                                 void* usedStructPhysAddr,
                                                 void* descStructPhysAddr,
                                                 uint16_t queueSize,
                                                 uint32_t hostId,
                                                 uint32_t header,
                                                 uint32_t vq_type,
                                                 int irq_vector)
{
  uint32_t qs = dbs_qsize_log2(queueSize);
  uint32_t int_enable;
  uint32_t vec;
  uint32_t istk;

  // Setup DBS module - DSF00094
  // 3. Configure the DBS.RX_DR_DATA memory; good idea to initialize all DBS_RX_QUEUES entries.
  if (set_rx_dr_data(pNthwDbs, index, (uint64_t)descStructPhysAddr, hostId, qs, header, PACKED(vq_type)) != 0) {
      return NULL;
  }

  // 4. Configure the DBS.RX_UW_DATA memory; good idea to initialize all DBS_RX_QUEUES entries.
  //    Notice: We always start out with interrupts disabled (by setting the "irq_vector" argument to -1).
  //            Queues that require interrupts will have it enabled at a later time (after we have enabled
  //            vfio interrupts in the kernel).
  int_enable = 0;
  vec = 0;
  istk = 0;
  NT_LOG(DBG, ETHDEV,"%s: set_rx_uw_data int=0 irq_vector=%u\n", __func__, irq_vector);
  if (set_rx_uw_data(pNthwDbs, index, (vq_type == PACKED_RING) ? (uint64_t)descStructPhysAddr : (uint64_t)usedStructPhysAddr, hostId, qs, PACKED(vq_type), int_enable, vec, istk) != 0) {
    return NULL;
  }

  // 2. Configure the DBS.RX_AM_DATA memory and enable the queues you plan to use;
  //    good idea to initialize all DBS_RX_QUEUES entries.
  //    Notice: We do this only for queues that don't require interrups (i.e. if irq_vector < 0).
  //            Queues that require interrupts will have RX_AM_DATA enabled at a later time (after
  //            we have enabled vfio interrupts in the kernel).
  if (irq_vector < 0) {
    if (set_rx_am_data(pNthwDbs, index,  (uint64_t)availStructPhysAddr, RX_AM_DISABLE, hostId, PACKED(vq_type), (irq_vector >= 0) ? 1 : 0) != 0) {
      return NULL;
    }
  }

  // 5. Initialize all RX queues (all DBS_RX_QUEUES of them) using the DBS.RX_INIT register.
  dbs_init_rx_queue(pNthwDbs, index, start_idx, start_ptr);

  // 2. Configure the DBS.RX_AM_DATA memory and enable the queues you plan to use;
  //    good idea to initialize all DBS_RX_QUEUES entries.
  if (set_rx_am_data(pNthwDbs, index, (uint64_t)availStructPhysAddr, RX_AM_ENABLE, hostId, PACKED(vq_type), (irq_vector >= 0) ? 1 : 0) != 0) {
    return NULL;
  }

  // Save queue state
  rxvq[index].usage = UNMANAGED;
  rxvq[index].mpNthwDbs = pNthwDbs;
  rxvq[index].index = index;
  rxvq[index].queueSize = queueSize;
  rxvq[index].am_enable = (irq_vector < 0) ? RX_AM_ENABLE : RX_AM_DISABLE;
  rxvq[index].hostId = hostId;
  rxvq[index].availStructPhysAddr = availStructPhysAddr;
  rxvq[index].usedStructPhysAddr = usedStructPhysAddr;
  rxvq[index].descStructPhysAddr = descStructPhysAddr;
  rxvq[index].vq_type = vq_type;
  rxvq[index].in_order = 0; // not used
  rxvq[index].irq_vector = irq_vector;

  // Return queue handle
  return &rxvq[index];

}

static int dbs_wait_hw_queue_shutdown(struct nthw_virt_queue* vq, int rx);

int nthw_disable_rx_virt_queue(struct nthw_virt_queue* rx_vq)
{
  if (!rx_vq) {
    NT_LOG(ERR, ETHDEV,"%s: Invalid queue\n", __func__);
    return -1;
  }

  NthwDbs_t* pNthwDbs = rx_vq->mpNthwDbs;

  if (rx_vq->index >= MAX_VIRT_QUEUES)
  {
    return -1;
  }

  if (rx_vq->usage != UNMANAGED)
  {
    return -1;
  }

  uint32_t qs = dbs_qsize_log2(rx_vq->queueSize);

  // if ISTK is set, make sure to unset it
  if (set_rx_uw_data(pNthwDbs,
                          rx_vq->index,
                          (rx_vq->vq_type == PACKED_RING) ? (uint64_t)rx_vq->descStructPhysAddr : (uint64_t)rx_vq->usedStructPhysAddr,
                          rx_vq->hostId,
                          qs,
                          PACKED(rx_vq->vq_type),
                          0,
                          0,
                          0) != 0) {
    return -1;
  }

  // Disable AM
  rx_vq->am_enable = RX_AM_DISABLE;
  if (set_rx_am_data(pNthwDbs,
                          rx_vq->index,
                          (uint64_t)rx_vq->availStructPhysAddr,
                          rx_vq->am_enable,
                          rx_vq->hostId,
                          PACKED(rx_vq->vq_type),
                          0) != 0) {
    return -1;
  }

  // let the FPGA finish packet processing
  if (dbs_wait_hw_queue_shutdown(rx_vq, 1) != 0)
  {
    return -1;
  }

  return 0;
}



int nthw_enable_rx_virt_queue(struct nthw_virt_queue* rx_vq)
{
  uint32_t int_enable;
  uint32_t vec;
  uint32_t istk;

  if (!rx_vq) {
    NT_LOG(ERR, ETHDEV,"%s: Invalid queue\n", __func__);
    return -1;
  }

  NthwDbs_t* pNthwDbs = rx_vq->mpNthwDbs;

  if (rx_vq->index >= MAX_VIRT_QUEUES)
  {
    return -1;
  }

  if (rx_vq->usage != UNMANAGED)
  {
    return -1;
  }

  uint32_t qs = dbs_qsize_log2(rx_vq->queueSize);

  // set ISTK if
  if (rx_vq->irq_vector >= 0 && rx_vq->irq_vector < MAX_MSIX_VECTORS_PR_VF) {
    int_enable = 1;
    vec = rx_vq->irq_vector;
    istk = 1;
  } else {
    int_enable = 0;
    vec = 0;
    istk = 0;
  }
  NT_LOG(DBG, ETHDEV,"%s: set_rx_uw_data irq_vector=%u\n", __func__, rx_vq->irq_vector);
  if (set_rx_uw_data(pNthwDbs,
                     rx_vq->index,
                     (rx_vq->vq_type == PACKED_RING) ? (uint64_t)rx_vq->descStructPhysAddr : (uint64_t)rx_vq->usedStructPhysAddr,
                     rx_vq->hostId,
                     qs,
                     PACKED(rx_vq->vq_type),
                     int_enable, vec, istk) != 0) return -1;

  // Enable AM
  rx_vq->am_enable = RX_AM_ENABLE;
  if (set_rx_am_data(pNthwDbs,
                     rx_vq->index,
                     (uint64_t)rx_vq->availStructPhysAddr,
                     rx_vq->am_enable,
                     rx_vq->hostId,
                     PACKED(rx_vq->vq_type),
                     (rx_vq->irq_vector >= 0) ? 1 : 0) != 0) return -1;

  return 0;
}



int nthw_disable_tx_virt_queue(struct nthw_virt_queue* tx_vq)
{
  if (!tx_vq) {
    NT_LOG(ERR, ETHDEV,"%s: Invalid queue\n", __func__);
    return -1;
  }

  NthwDbs_t* pNthwDbs = tx_vq->mpNthwDbs;

  if (tx_vq->index >= MAX_VIRT_QUEUES)
  {
    return -1;
  }

  if (tx_vq->usage != UNMANAGED)
  {
    return -1;
  }

  uint32_t qs = dbs_qsize_log2(tx_vq->queueSize);

  // if ISTK is set, make sure to unset it
  if (set_tx_uw_data(pNthwDbs,
                     tx_vq->index,
                     (tx_vq->vq_type == PACKED_RING) ? (uint64_t)tx_vq->descStructPhysAddr : (uint64_t)tx_vq->usedStructPhysAddr,
                     tx_vq->hostId,
                     qs,
                     PACKED(tx_vq->vq_type),
                     0, 0, 0,
                     tx_vq->in_order) != 0) return -1;

  // Disable AM
  tx_vq->am_enable = TX_AM_DISABLE;
  if (set_tx_am_data(pNthwDbs,
                     tx_vq->index,
                     (uint64_t)tx_vq->availStructPhysAddr,
                     tx_vq->am_enable,
                     tx_vq->hostId,
                     PACKED(tx_vq->vq_type),
                     0) != 0) return -1;

  // let the FPGA finish packet processing
  if (dbs_wait_hw_queue_shutdown(tx_vq, 0) != 0)
  {
    return -1;
  }

  return 0;
}


int nthw_enable_tx_virt_queue(struct nthw_virt_queue* tx_vq)
{
  uint32_t int_enable;
  uint32_t vec;
  uint32_t istk;

  if (!tx_vq) {
    NT_LOG(ERR, ETHDEV,"%s: Invalid queue\n", __func__);
    return -1;
  }

  NthwDbs_t* pNthwDbs = tx_vq->mpNthwDbs;

  if (tx_vq->index >= MAX_VIRT_QUEUES)
  {
    return -1;
  }

  if (tx_vq->usage != UNMANAGED)
  {
    return -1;
  }

  uint32_t qs = dbs_qsize_log2(tx_vq->queueSize);

  // set ISTK if irq_vector is used
  if (tx_vq->irq_vector >= 0 && tx_vq->irq_vector < MAX_MSIX_VECTORS_PR_VF) {
    int_enable = 1;
    vec = tx_vq->irq_vector;
    istk = 1; // use sticky interrupt
  } else {
    int_enable = 0;
    vec = 0;
    istk = 0;
  }
  if (set_tx_uw_data(pNthwDbs,
                     tx_vq->index,
                     (tx_vq->vq_type == PACKED_RING) ? (uint64_t)tx_vq->descStructPhysAddr : (uint64_t)tx_vq->usedStructPhysAddr,
                     tx_vq->hostId,
                     qs,
                     PACKED(tx_vq->vq_type),
                     int_enable,
                     vec,
                     istk,
                     tx_vq->in_order) != 0) return -1;

  // Enable AM
  tx_vq->am_enable = TX_AM_ENABLE;
  if (set_tx_am_data(pNthwDbs,
                     tx_vq->index,
                     (uint64_t)tx_vq->availStructPhysAddr,
                     tx_vq->am_enable,
                     tx_vq->hostId,
                     PACKED(tx_vq->vq_type),
                     (tx_vq->irq_vector >= 0) ? 1 : 0) != 0) return -1;

  return 0;
}

int nthw_enable_and_change_port_tx_virt_queue(struct nthw_virt_queue* tx_vq, uint32_t outport)
{
  NthwDbs_t* pNthwDbs = tx_vq->mpNthwDbs;
  uint32_t qs = dbs_qsize_log2(tx_vq->queueSize);

  if (set_tx_dr_data(pNthwDbs,
                     tx_vq->index,
                     (uint64_t)tx_vq->descStructPhysAddr,
                     tx_vq->hostId,
                     qs,
                     outport,
                     0,
                     PACKED(tx_vq->vq_type)) != 0) {
    return -1;
  }
  return nthw_enable_tx_virt_queue(tx_vq);
}


int nthw_set_tx_qos_config(NthwDbs_t *pNthwDbs,
                           uint32_t port,
                           uint32_t enable,
                           uint32_t ir,
                           uint32_t bs)
{
  return set_tx_qos_data(pNthwDbs, port, enable, ir, bs);

}


int nthw_set_tx_qos_rate_global(NthwDbs_t *pNthwDbs,
                                uint32_t multiplier,
                                uint32_t divider)
{
  return set_tx_qos_rate(pNthwDbs, multiplier, divider);
}

#define INDEX_PTR_NOT_VALID  0x80000000
static int dbs_get_rx_ptr(NthwDbs_t *pNthwDbs, uint32_t *pIndex)
{
  uint32_t ptr;
  uint32_t queue;
  uint32_t valid;

  const int status = get_rx_ptr(pNthwDbs, &ptr, &queue, &valid);

  if (status == 0)
  {
    if (valid)
      *pIndex = ptr;
    else
      *pIndex = INDEX_PTR_NOT_VALID;
  }
  return status;
}

static int dbs_get_tx_ptr(NthwDbs_t *pNthwDbs, uint32_t *pIndex)
{
  uint32_t ptr;
  uint32_t queue;
  uint32_t valid;

  const int status = get_tx_ptr(pNthwDbs, &ptr, &queue, &valid);
  if (status == 0)
  {
    if (valid)
      *pIndex = ptr;
    else
      *pIndex = INDEX_PTR_NOT_VALID;
  }
  return status;
}


static int dbs_initialize_get_rx_ptr(NthwDbs_t* pNthwDbs, uint32_t queue)
{
  return set_rx_ptr_queue(pNthwDbs, queue);
}

static int dbs_initialize_get_tx_ptr(NthwDbs_t* pNthwDbs, uint32_t queue)
{
  return set_tx_ptr_queue(pNthwDbs, queue);
}

static int dbs_wait_on_busy(struct nthw_virt_queue* vq, uint32_t *idle, int rx)
{
  uint32_t busy;
  uint32_t queue;
  int err = 0;
  NthwDbs_t *pNthwDbs = vq->mpNthwDbs;

  do {
    if (rx)
      err = get_rx_idle(pNthwDbs, idle, &queue, &busy);
    else
      err = get_tx_idle(pNthwDbs, idle, &queue, &busy);
  } while (!err && busy);

  return err;
}

static int dbs_wait_hw_queue_shutdown(struct nthw_virt_queue* vq, int rx)
{
  int err = 0;
  uint32_t idle = 0;
  NthwDbs_t* pNthwDbs = vq->mpNthwDbs;

  err = dbs_wait_on_busy(vq, &idle, rx);
  if (err) {
    if (err == -ENOTSUP) {
      NtOs_WaituSec(200000);
      return 0;
    }
    return -1;
  }

  do {
    if (rx)
      err = set_rx_idle(pNthwDbs, 1, vq->index);
    else
      err = set_tx_idle(pNthwDbs, 1, vq->index);

    if (err)
      return -1;

    if (dbs_wait_on_busy(vq, &idle, rx) != 0)
      return -1;

  } while (idle == 0);

  return 0;
}

static int dbs_internal_release_rx_virt_queue(struct nthw_virt_queue* rxvq)
{
  NthwDbs_t* pNthwDbs = rxvq->mpNthwDbs;

  if (rxvq == NULL) return -1;

  // Clear UW
  rxvq->usedStructPhysAddr = NULL;
  if (set_rx_uw_data(pNthwDbs,
                          rxvq->index,
                          (uint64_t)rxvq->usedStructPhysAddr,
                          rxvq->hostId,
                          0,
                          PACKED(rxvq->vq_type),
                          0, 0, 0) != 0)  {
    return -1;
  }

  // Disable AM
  rxvq->am_enable = RX_AM_DISABLE;
  if (set_rx_am_data(pNthwDbs,
                          rxvq->index,
                          (uint64_t)rxvq->availStructPhysAddr,
                          rxvq->am_enable,
                          rxvq->hostId,
                          PACKED(rxvq->vq_type),
                          0) != 0) {
    return -1;
  }

  // let the FPGA finish packet processing
  if (dbs_wait_hw_queue_shutdown(rxvq, 1) != 0)
    return -1;

  // Clear rest of AM
  rxvq->availStructPhysAddr = NULL;
  rxvq->hostId = 0;
  if (set_rx_am_data(pNthwDbs,
                     rxvq->index,
                     (uint64_t)rxvq->availStructPhysAddr,
                     rxvq->am_enable,
                     rxvq->hostId,
                     PACKED(rxvq->vq_type),
                     0) != 0) return -1;

  // Clear DR
  rxvq->descStructPhysAddr = NULL;
  if (set_rx_dr_data(pNthwDbs,
                     rxvq->index,
                     (uint64_t)rxvq->descStructPhysAddr,
                     rxvq->hostId,
                     0,
                     rxvq->header,
                     PACKED(rxvq->vq_type)) != 0) return -1;

  // Initialize queue
  dbs_init_rx_queue(pNthwDbs, rxvq->index, 0, 0);

  // Reset queue state
  rxvq->usage = UNUSED;
  rxvq->mpNthwDbs = pNthwDbs;
  rxvq->index = 0;
  rxvq->queueSize = 0;

  return 0;
}

int nthw_release_rx_virt_queue(struct nthw_virt_queue* rxvq)
{
  if (rxvq == NULL || rxvq->usage != UNMANAGED) return -1;

  return dbs_internal_release_rx_virt_queue(rxvq);
}

int nthw_release_managed_rx_virt_queue(struct nthw_virt_queue* rxvq)
{
  if (rxvq == NULL || rxvq->usage != MANAGED) return -1;

  if (rxvq->pVirtualAddr)
  {
    free(rxvq->pVirtualAddr);
    rxvq->pVirtualAddr = NULL;
  }

  return dbs_internal_release_rx_virt_queue(rxvq);
}

static int dbs_internal_release_tx_virt_queue(struct nthw_virt_queue* txvq)
{
  NthwDbs_t* pNthwDbs = txvq->mpNthwDbs;

  if (txvq == NULL) return -1;

  // Clear UW
  txvq->usedStructPhysAddr = NULL;
  if (set_tx_uw_data(pNthwDbs,
                          txvq->index,
                          (uint64_t)txvq->usedStructPhysAddr,
                          txvq->hostId,
                          0,
                          PACKED(txvq->vq_type),
                          0, 0, 0,
                          txvq->in_order) != 0) {
    return -1;
  }

  // Disable AM
  txvq->am_enable = TX_AM_DISABLE;
  if (set_tx_am_data(pNthwDbs,
                          txvq->index,
                          (uint64_t)txvq->availStructPhysAddr,
                          txvq->am_enable,
                          txvq->hostId,
                          PACKED(txvq->vq_type),
                          0) != 0) {
    return -1;
  }

  // let the FPGA finish packet processing
  if (dbs_wait_hw_queue_shutdown(txvq, 0) != 0)
    return -1;

  // Clear rest of AM
  txvq->availStructPhysAddr = NULL;
  txvq->hostId = 0;
  if (set_tx_am_data(pNthwDbs,
                          txvq->index,
                          (uint64_t)txvq->availStructPhysAddr,
                          txvq->am_enable,
                          txvq->hostId,
                          PACKED(txvq->vq_type),
                          0) != 0) {
    return -1;
  }

  // Clear DR
  txvq->descStructPhysAddr = NULL;
  txvq->port = 0;
  txvq->header = 0;
  if (set_tx_dr_data(pNthwDbs,
                          txvq->index,
                          (uint64_t)txvq->descStructPhysAddr,
                          txvq->hostId,
                          0,
                          txvq->port,
                          txvq->header,
                          PACKED(txvq->vq_type)) != 0) {
    return -1;
  }

  // Clear QP
  txvq->virtualPort = 0;
  if (NthwDbs_setTxQpData(pNthwDbs,
                          txvq->index,
                          txvq->virtualPort) != 0) {
    return -1;
  }

  // Initialize queue
  dbs_init_tx_queue(pNthwDbs, txvq->index, 0, 0);

  // Reset queue state
  txvq->usage = UNUSED;
  txvq->mpNthwDbs = pNthwDbs;
  txvq->index = 0;
  txvq->queueSize = 0;

  return 0;
}

int nthw_release_tx_virt_queue(struct nthw_virt_queue* txvq)
{
  if (txvq == NULL || txvq->usage != UNMANAGED) return -1;

  return dbs_internal_release_tx_virt_queue(txvq);
}

int nthw_release_managed_tx_virt_queue(struct nthw_virt_queue* txvq)
{
  if (txvq == NULL || txvq->usage != MANAGED) return -1;

  if (txvq->pVirtualAddr)
  {
    free(txvq->pVirtualAddr);
    txvq->pVirtualAddr = NULL;
  }

  return dbs_internal_release_tx_virt_queue(txvq);
}


struct nthw_virt_queue* nthw_setup_tx_virt_queue(NthwDbs_t *pNthwDbs,
                                                 uint32_t index,
                                                 uint16_t start_idx,
                                                 uint16_t start_ptr,
                                                 void* availStructPhysAddr,
                                                 void* usedStructPhysAddr,
                                                 void* descStructPhysAddr,
                                                 uint16_t queueSize,
                                                 uint32_t hostId,
                                                 uint32_t port,
                                                 uint32_t virtualPort,
                                                 uint32_t header,
                                                 uint32_t vq_type,
                                                 int irq_vector,
                                                 uint32_t in_order)
{
  uint32_t int_enable;
  uint32_t vec;
  uint32_t istk;
  uint32_t qs = dbs_qsize_log2(queueSize);

  // Setup DBS module - DSF00094
  // 3. Configure the DBS.TX_DR_DATA memory; good idea to initialize all DBS_TX_QUEUES entries.
  if (set_tx_dr_data(pNthwDbs, index, (uint64_t)descStructPhysAddr, hostId, qs, port, header, PACKED(vq_type)) != 0) {
    return NULL;
  }

  // 4. Configure the DBS.TX_UW_DATA memory; good idea to initialize all DBS_TX_QUEUES entries.
  //    Notice: We always start out with interrupts disabled (by setting the "irq_vector" argument to -1).
  //            Queues that require interrupts will have it enabled at a later time (after we have enabled
  //            vfio interrupts in the kernel).
  int_enable = 0;
  vec = 0;
  istk = 0;

  if (set_tx_uw_data(pNthwDbs, index, (vq_type == PACKED_RING) ? (uint64_t)descStructPhysAddr : (uint64_t)usedStructPhysAddr, hostId, qs, PACKED(vq_type), int_enable, vec, istk, in_order) != 0) {
    return NULL;
  }

  // 2. Configure the DBS.TX_AM_DATA memory and enable the queues you plan to use;
  //    good idea to initialize all DBS_TX_QUEUES entries.
  if (set_tx_am_data(pNthwDbs, index, (uint64_t)availStructPhysAddr, TX_AM_DISABLE, hostId, PACKED(vq_type), (irq_vector >= 0) ? 1 : 0) != 0) {
    return NULL;
  }

  // 5. Initialize all TX queues (all DBS_TX_QUEUES of them) using the DBS.TX_INIT register.
  dbs_init_tx_queue(pNthwDbs, index, start_idx, start_ptr);

  if (NthwDbs_setTxQpData(pNthwDbs, index, virtualPort) != 0) {
    return NULL;
  }

  // 2. Configure the DBS.TX_AM_DATA memory and enable the queues you plan to use;
  //    good idea to initialize all DBS_TX_QUEUES entries.
  //    Notice: We do this only for queues that don't require interrups (i.e. if irq_vector < 0).
  //            Queues that require interrupts will have TX_AM_DATA enabled at a later time (after
  //            we have enabled vfio interrupts in the kernel).
  if (irq_vector < 0) {
    if (set_tx_am_data(pNthwDbs, index, (uint64_t)availStructPhysAddr, TX_AM_ENABLE, hostId, PACKED(vq_type), (irq_vector >= 0) ? 1 : 0) != 0) {
      return NULL;
    }
  }

  if (vq_type == PACKED_RING) {
    // whenever FPGA will get the need to be kicked
    // (struct pvirtq_event_suppress *)usedStructPhysAdd   - device event suppression mem area
  }

  // Save queue state
  txvq[index].usage = UNMANAGED;
  txvq[index].mpNthwDbs = pNthwDbs;
  txvq[index].index = index;
  txvq[index].queueSize = queueSize;
  txvq[index].am_enable = (irq_vector < 0) ? TX_AM_ENABLE : TX_AM_DISABLE;
  txvq[index].hostId = hostId;
  txvq[index].port = port;
  txvq[index].virtualPort = virtualPort;
  txvq[index].header = header;
  txvq[index].availStructPhysAddr = availStructPhysAddr;
  txvq[index].usedStructPhysAddr = usedStructPhysAddr;
  txvq[index].descStructPhysAddr = descStructPhysAddr;
  txvq[index].vq_type = vq_type;
  txvq[index].in_order = in_order;
  txvq[index].irq_vector = irq_vector;

  // Return queue handle
  return &txvq[index];
}

static struct nthw_virt_queue* nthw_setup_managed_rx_virt_queue_split(NthwDbs_t *pNthwDbs,
                                                                     uint32_t index,
                                                                     uint32_t queueSize,
                                                                     uint32_t hostId,
                                                                     uint32_t header,
                                                                     struct nthw_memory_descriptor* pVirtStructArea,
                                                                     struct nthw_memory_descriptor* pPacketBuffers,
                                                                     int irq_vector)
{
  struct virtq_struct_layout_s virtq_struct_layout = dbs_calc_struct_layout(queueSize);

  dbs_initialize_virt_queue_structs(pVirtStructArea->virtAddr,
                             (char*)pVirtStructArea->virtAddr + virtq_struct_layout.used_offset,
                             (char*)pVirtStructArea->virtAddr + virtq_struct_layout.desc_offset,
                             pPacketBuffers,
                             (uint16_t)queueSize,
                             pPacketBuffers ? (uint16_t)queueSize : 0,
                             VIRTQ_DESC_F_WRITE /* Rx */);

  rxvq[index].pAvail = pVirtStructArea->virtAddr;
  rxvq[index].pUsed = (void*)((char*)pVirtStructArea->virtAddr + virtq_struct_layout.used_offset);
  rxvq[index].pDesc = (void*)((char*)pVirtStructArea->virtAddr + virtq_struct_layout.desc_offset);

  rxvq[index].amIdx = pPacketBuffers ? (uint16_t)queueSize : 0;
  rxvq[index].usedIdx = 0;
  rxvq[index].cachedIdx = 0;
  rxvq[index].pVirtualAddr = NULL;

  if (pPacketBuffers)
  {
    rxvq[index].pVirtualAddr = malloc(queueSize * sizeof *pPacketBuffers);
    memcpy(rxvq[index].pVirtualAddr, pPacketBuffers, queueSize * sizeof *pPacketBuffers);
  }

  nthw_setup_rx_virt_queue(pNthwDbs,
                           index,
                           0, 0,
                           (void*)pVirtStructArea->physAddr,
                           (char*)pVirtStructArea->physAddr + virtq_struct_layout.used_offset,
                           (char*)pVirtStructArea->physAddr + virtq_struct_layout.desc_offset,
                           (uint16_t)queueSize,
                           hostId,
                           header,
                           SPLIT_RING,
                           irq_vector);

  rxvq[index].usage = MANAGED;

  return &rxvq[index];
}

static struct nthw_virt_queue* nthw_setup_managed_tx_virt_queue_split(NthwDbs_t *pNthwDbs,
                                                                     uint32_t index,
                                                                     uint32_t queueSize,
                                                                     uint32_t hostId,
                                                                     uint32_t port,
                                                                     uint32_t virtualPort,
                                                                     uint32_t header,
                                                                     int irq_vector,
                                                                     uint32_t in_order,
                                                                     struct nthw_memory_descriptor* pVirtStructArea,
                                                                     struct nthw_memory_descriptor* pPacketBuffers)
{
  struct virtq_struct_layout_s virtq_struct_layout = dbs_calc_struct_layout(queueSize);

  dbs_initialize_virt_queue_structs(pVirtStructArea->virtAddr,
                             (char*)pVirtStructArea->virtAddr + virtq_struct_layout.used_offset,
                             (char*)pVirtStructArea->virtAddr + virtq_struct_layout.desc_offset,
                             pPacketBuffers,
                             (uint16_t)queueSize,
                             0,
                             0 /* Tx */);

  txvq[index].pAvail = pVirtStructArea->virtAddr;
  txvq[index].pUsed = (void*)((char*)pVirtStructArea->virtAddr + virtq_struct_layout.used_offset);
  txvq[index].pDesc = (void*)((char*)pVirtStructArea->virtAddr + virtq_struct_layout.desc_offset);
  txvq[index].queueSize = (le16)queueSize;
  txvq[index].amIdx = 0;
  txvq[index].usedIdx = 0;
  txvq[index].cachedIdx = 0;
  txvq[index].pVirtualAddr = NULL;

  txvq[index].txDescrAvailIdx = 0;

  if (pPacketBuffers)
  {
    txvq[index].pVirtualAddr = malloc(queueSize * sizeof *pPacketBuffers);
    memcpy(txvq[index].pVirtualAddr, pPacketBuffers, queueSize * sizeof *pPacketBuffers);
  }

  nthw_setup_tx_virt_queue(pNthwDbs,
                           index,
                           0, 0,
                           (void*)pVirtStructArea->physAddr,
                           (char*)pVirtStructArea->physAddr + virtq_struct_layout.used_offset,
                           (char*)pVirtStructArea->physAddr + virtq_struct_layout.desc_offset,
                           (uint16_t)queueSize,
                           hostId,
                           port,
                           virtualPort,
                           header,
                           SPLIT_RING,
                           irq_vector,
                           in_order);

  txvq[index].usage = MANAGED;

  return &txvq[index];
}



/*
 * Packed Ring
 */

static int nthw_setup_managed_virt_queue_packed(struct nthw_virt_queue *vq,
                                                struct pvirtq_struct_layout_s *pvirtq_layout,
                                                struct nthw_memory_descriptor* pVirtStructArea,
                                                struct nthw_memory_descriptor* pPacketBuffers,
                                                ule16 flags,
                                                int rx)
{

  // page aligned
  assert(((uintptr_t)pVirtStructArea->physAddr & 0xfff) == 0);
  assert(pPacketBuffers);

  /* clean canvas */
  memset(pVirtStructArea->virtAddr, 0, sizeof(struct pvirtq_desc) * vq->queueSize + sizeof(struct pvirtq_event_suppress) * 2 + sizeof(int) * vq->queueSize);

  pvirtq_layout->device_event_offset = sizeof(struct pvirtq_desc) * vq->queueSize;
  pvirtq_layout->driver_event_offset = pvirtq_layout->device_event_offset + sizeof(struct pvirtq_event_suppress);

  vq->desc = pVirtStructArea->virtAddr;
  vq->device_event = (void *)((uintptr_t)vq->desc + pvirtq_layout->device_event_offset);
  vq->driver_event = (void *)((uintptr_t)vq->desc + pvirtq_layout->driver_event_offset);

//  vq->sgs          = (int  *)((uintptr_t)vq->driver_event + sizeof(struct pvirtq_event_suppress));

  vq->next_avail = 0;
  vq->next_used = 0;
  vq->avail_wrap_count = 1;
  vq->used_wrap_count = 1;


  /*
   * Only possible if FPGA always delivers in-order
   * Buffer ID used is the index in the pPacketBuffers array
   */

    unsigned i;
    struct pvirtq_desc* pDesc = vq->desc;
    for (i = 0; i < vq->queueSize; i++)
    {
      if (rx) {
        pDesc[i].addr  = (uint64_t)pPacketBuffers[i].physAddr;
        pDesc[i].len   = pPacketBuffers[i].len;
      }
      pDesc[i].id    = i;
      pDesc[i].flags = flags;

//      vq->sgs[i] = 1;
    }

  if (rx)
    vq->avail_wrap_count ^= 1;  // filled up available buffers for Rx
  else
    vq->used_wrap_count ^= 1;   // pre-fill free buffer IDs

  if(vq->queueSize == 0) return -1; // don't allocate memory with size of 0 bytes
  vq->pVirtualAddr = malloc(vq->queueSize * sizeof *pPacketBuffers);
  if (vq->pVirtualAddr == NULL) return -1;

  memcpy(vq->pVirtualAddr, pPacketBuffers, vq->queueSize * sizeof *pPacketBuffers);


  /* not used yet by FPGA - make sure we disable */
  vq->device_event->flags = RING_EVENT_FLAGS_DISABLE;

  return 0;
}

static struct nthw_virt_queue* nthw_setup_managed_rx_virt_queue_packed(NthwDbs_t *pNthwDbs,
                                                                       uint32_t index,
                                                                       uint32_t queueSize,
                                                                       uint32_t hostId,
                                                                       uint32_t header,
                                                                       struct nthw_memory_descriptor* pVirtStructArea,
                                                                       struct nthw_memory_descriptor* pPacketBuffers,
                                                                       int irq_vector)
{
  struct pvirtq_struct_layout_s pvirtq_layout;
  struct nthw_virt_queue *vq = &rxvq[index];
  // set size and setup packed vq ring
  vq->queueSize = queueSize;
  // Use Avail flag bit == 1 because wrap bit is intially set to 1 - and Used is inverse
  if (nthw_setup_managed_virt_queue_packed(vq, &pvirtq_layout, pVirtStructArea, pPacketBuffers, VIRTQ_DESC_F_WRITE | VIRTQ_DESC_F_AVAIL, 1) != 0) return NULL;

  nthw_setup_rx_virt_queue(pNthwDbs,
                           index,
                           0x8000, 0,                  // start wrap ring counter as 1
                           (void *)((uintptr_t)pVirtStructArea->physAddr + pvirtq_layout.driver_event_offset),
                           (void *)((uintptr_t)pVirtStructArea->physAddr + pvirtq_layout.device_event_offset),
                           pVirtStructArea->physAddr,
                           (uint16_t)queueSize,
                           hostId,
                           header,
                           PACKED_RING,
                           irq_vector);

  vq->usage = MANAGED;
  return vq;
}

static struct nthw_virt_queue* nthw_setup_managed_tx_virt_queue_packed(NthwDbs_t *pNthwDbs,
                                                                      uint32_t index,
                                                                      uint32_t queueSize,
                                                                      uint32_t hostId,
                                                                      uint32_t port,
                                                                      uint32_t virtualPort,
                                                                      uint32_t header,
                                                                      int irq_vector,
                                                                      uint32_t in_order,
                                                                      struct nthw_memory_descriptor* pVirtStructArea,
                                                                      struct nthw_memory_descriptor* pPacketBuffers)
{
  struct pvirtq_struct_layout_s pvirtq_layout;
  struct nthw_virt_queue *vq = &txvq[index];
  // set size and setup packed vq ring
  vq->queueSize = queueSize;
  if (nthw_setup_managed_virt_queue_packed(vq, &pvirtq_layout, pVirtStructArea, pPacketBuffers, 0, 0) != 0) return NULL;

  nthw_setup_tx_virt_queue(pNthwDbs,
                           index,
                           0x8000, 0,                  // start wrap ring counter as 1
                           (void *)((uintptr_t)pVirtStructArea->physAddr + pvirtq_layout.driver_event_offset),
                           (void *)((uintptr_t)pVirtStructArea->physAddr + pvirtq_layout.device_event_offset),
                           pVirtStructArea->physAddr,
                           (uint16_t)queueSize,
                           hostId,
                           port,
                           virtualPort,
                           header,
                           PACKED_RING,
                           irq_vector,
                           in_order);

  vq->usage = MANAGED;
  return vq;
}

/*
 * Create a Managed Rx Virt Queue
 *
 * Notice: The queue will be created with interrupts disabled.
 *         If interrupts are required, make sure to call nthw_enable_rx_virt_queue() afterwards.
 */
struct nthw_virt_queue* nthw_setup_managed_rx_virt_queue(NthwDbs_t *pNthwDbs,
                                                         uint32_t index,
                                                         uint32_t queueSize,
                                                         uint32_t hostId,
                                                         uint32_t header,
                                                         struct nthw_memory_descriptor* pVirtStructArea,
                                                         struct nthw_memory_descriptor* pPacketBuffers,
                                                         uint32_t vq_type,
                                                         int irq_vector)
{
  switch (vq_type) {
  case SPLIT_RING:
    return nthw_setup_managed_rx_virt_queue_split (pNthwDbs, index, queueSize, hostId, header, pVirtStructArea, pPacketBuffers, irq_vector);
  case PACKED_RING:
    return nthw_setup_managed_rx_virt_queue_packed(pNthwDbs, index, queueSize, hostId, header, pVirtStructArea, pPacketBuffers, irq_vector);
  default:
    break;
  }
  return NULL;
}

/*
 * Create a Managed Tx Virt Queue
 *
 * Notice: The queue will be created with interrupts disabled.
 *         If interrupts are required, make sure to call nthw_enable_tx_virt_queue() afterwards.
 */
struct nthw_virt_queue* nthw_setup_managed_tx_virt_queue(NthwDbs_t *pNthwDbs,
                                                         uint32_t index,
                                                         uint32_t queueSize,
                                                         uint32_t hostId,
                                                         uint32_t port,
                                                         uint32_t virtualPort,
                                                         uint32_t header,
                                                         struct nthw_memory_descriptor* pVirtStructArea,
                                                         struct nthw_memory_descriptor* pPacketBuffers,
                                                         uint32_t vq_type,
                                                         int irq_vector,
                                                         uint32_t in_order)
{
  switch (vq_type) {
  case SPLIT_RING:
    return nthw_setup_managed_tx_virt_queue_split (pNthwDbs, index, queueSize, hostId, port, virtualPort, header, irq_vector, in_order, pVirtStructArea, pPacketBuffers);
  case PACKED_RING:
    return nthw_setup_managed_tx_virt_queue_packed(pNthwDbs, index, queueSize, hostId, port, virtualPort, header, irq_vector, in_order, pVirtStructArea, pPacketBuffers);
  default:
    break;
  }
  return NULL;
}



/*
 * Packed Ring helper macros
 */
#define avail_flag(vq)    (vq->avail_wrap_count ? VIRTQ_DESC_F_AVAIL : 0)
#define used_flag_inv(vq) (vq->avail_wrap_count ? 0 : VIRTQ_DESC_F_USED)

#define inc_avail(vq, num) do { \
  vq->next_avail += num; \
  if (vq->next_avail >= vq->queueSize) { \
    vq->next_avail -= vq->queueSize; \
    vq->avail_wrap_count ^= 1;} \
  } while (0)

#define inc_used(vq, num) do { \
  vq->next_used += num; \
  if (vq->next_used >= vq->queueSize) { \
    vq->next_used -= vq->queueSize; \
    vq->used_wrap_count ^= 1;} \
  } while (0)

uint16_t nthw_get_rx_packets(struct nthw_virt_queue* rxvq,
    uint16_t n,
    struct nthw_received_packets* rp,
    uint16_t *nbPkts)
{
  le16 segs = 0;
  uint16_t pkts = 0;

  if (rxvq->vq_type == SPLIT_RING) {
    le16 i;
    le16 entriesReady = (le16)(rxvq->cachedIdx - rxvq->usedIdx);

    if (entriesReady < n)
    {
      // Look for more packets
      rxvq->cachedIdx = rxvq->pUsed->idx;
      entriesReady = (le16)(rxvq->cachedIdx - rxvq->usedIdx);
      if (entriesReady == 0) {
        *nbPkts = 0;
        return 0;
      }

      if (n > entriesReady) n = entriesReady;
    }

    // Give packets - make sure all packets are whole packets
    const le16 queueMask = (le16)(rxvq->queueSize - 1); // Valid because queueSize is always 2^n
    const ule32 buf_len = rxvq->pDesc[0].len;

    le16 used = rxvq->usedIdx;
    for (i = 0; i < n; ++i) {
      le32 id = rxvq->pUsed->ring[used & queueMask].id;
      rp[i].addr = rxvq->pVirtualAddr[id].virtAddr;
      rp[i].len = rxvq->pUsed->ring[used & queueMask].len;

      uint32_t pkt_len = ((struct _pkt_hdr_rx *)rp[i].addr)->cap_len;

      if (pkt_len > buf_len) {
        /* segmented */
        int nbsegs = (pkt_len + buf_len - 1) / buf_len;

        if (((int)i + nbsegs) > n) {
          /* don't have enough segments - break out */
          break;
        }

        int ii;
        for (ii = 1; ii < nbsegs; ii++) {
          ++i;
          id = rxvq->pUsed->ring[(used + ii) & queueMask].id;
          rp[i].addr = rxvq->pVirtualAddr[id].virtAddr;
          rp[i].len = rxvq->pUsed->ring[(used + ii) & queueMask].len;
        }
        used += nbsegs;
      } else
        ++used;

      pkts++;
      segs = i + 1;
    }

    rxvq->usedIdx = used;
  } else
  if (rxvq->vq_type == PACKED_RING) {

    // not handling jumbo yet - 

    // this requires in-order behavior from FPGA
    int i;
    for (i = 0; i < n; i++) {
      struct pvirtq_desc* desc = &rxvq->desc[rxvq->next_used];

      ule16 flags = desc->flags;
      uint8_t avail = !!(flags & VIRTQ_DESC_F_AVAIL);
      uint8_t used  = !!(flags & VIRTQ_DESC_F_USED);

      if (avail != rxvq->used_wrap_count || used != rxvq->used_wrap_count) {
        break;
      }

      rp[pkts].addr = rxvq->pVirtualAddr[desc->id].virtAddr;
      rp[pkts].len  = desc->len;
      pkts++;

      inc_used(rxvq, 1);
    }

    // no jumbo for now!
    segs = pkts;
  }

  *nbPkts = pkts;
  return segs;
}


/*
 * Put buffers back into Avail Ring
 */
void nthw_release_rx_packets(struct nthw_virt_queue* rxvq, le16 n)
{
  if (rxvq->vq_type == SPLIT_RING) {
    rxvq->amIdx = (le16)(rxvq->amIdx + n);
    rxvq->pAvail->idx = rxvq->amIdx;
  } else
  if (rxvq->vq_type == PACKED_RING) {
    int i;
    // defer flags update on first segment - due to serialization towards HW and when jumbo segments are added

    ule16 first_flags =  VIRTQ_DESC_F_WRITE | avail_flag(rxvq) | used_flag_inv(rxvq);
    struct pvirtq_desc* first_desc = &rxvq->desc[rxvq->next_avail];

    uint32_t len = rxvq->pVirtualAddr[0].len; // all same size

// Optimization point: use in-order release

    for (i = 0; i < n; i++) {
      struct pvirtq_desc* desc = &rxvq->desc[rxvq->next_avail];

      desc->id   = rxvq->next_avail;
      desc->addr = (ule64)rxvq->pVirtualAddr[desc->id].physAddr;
      desc->len  = len;
      if (i)
        desc->flags = VIRTQ_DESC_F_WRITE | avail_flag(rxvq) | used_flag_inv(rxvq);

      inc_avail(rxvq, 1);
    }
    rte_rmb();
    first_desc->flags = first_flags;
  }
}

#if 0
#define vq_log_arg(vq, format, ...) printf("[%s]: VqTx %s: vPort %i (%s) hostId %i index %i: "format, __func__, (vq->vq_type==SPLIT_RING)?"SPLIT":"PACKED", \
  vq->virtualPort, (vq->virtualPort>127)?"excep":"vp", vq->hostId, vq->index, __VA_ARGS__)
#else
#define vq_log_arg(vq, format, ...)
#endif

uint16_t nthw_get_tx_buffers(struct nthw_virt_queue* txvq,
                             uint16_t n,
                             uint16_t *firstIdx,
                             struct nthw_cvirtq_desc *cvq,
                             struct nthw_memory_descriptor **pVirtAddr)
{
  int m = 0;
  le16 queueMask = (le16)(txvq->queueSize - 1); // Valid because queueSize is always 2^n
  *pVirtAddr = txvq->pVirtualAddr;

  if (txvq->vq_type == SPLIT_RING) {

    cvq->s            = txvq->pDesc;
    cvq->vq_type      = SPLIT_RING;

    *firstIdx = txvq->txDescrAvailIdx;

    le16 entriesUsed = (le16)((txvq->txDescrAvailIdx - txvq->cachedIdx) & queueMask);
    le16 entriesReady = (le16)(txvq->queueSize - 1 - entriesUsed);

    vq_log_arg(txvq, "ask %i: descrAvail %i, cachedidx %i, used: %i, ready %i used->idx %i\n",n, txvq->txDescrAvailIdx, txvq->cachedIdx,
        entriesUsed, entriesReady, txvq->pUsed->idx);

    if (entriesReady < n)
    {
      // Look for more packets
      // Using the used_idx in the avail ring since they are held synchronous because of in-order
      txvq->cachedIdx = txvq->pAvail->ring[(txvq->pUsed->idx - 1) & queueMask];

      vq_log_arg(txvq, "Update: get cachedidx %i (used_idx-1 %i)\n", txvq->cachedIdx, (txvq->pUsed->idx-1) & queueMask);
      entriesUsed = (le16)((txvq->txDescrAvailIdx - txvq->cachedIdx) & queueMask);
      entriesReady = (le16)(txvq->queueSize - 1 - entriesUsed);
      vq_log_arg(txvq, "new used: %i, ready %i\n", entriesUsed, entriesReady);
      if (n > entriesReady) n = entriesReady;
    }
  } else
  if (txvq->vq_type == PACKED_RING) {
    int i;

    cvq->p       = txvq->desc;
    cvq->vq_type = PACKED_RING;

    if (txvq->outs.num) {
      *firstIdx = txvq->outs.next;
      uint16_t num = MIN(n, txvq->outs.num);
      txvq->outs.next = (txvq->outs.next + num) & queueMask;
      txvq->outs.num -= num;

      if (n == num) {
        return n;
      }

      m = num;
      n -= num;
    } else {
      *firstIdx = txvq->next_used;
    }

    // iterate the ring - this requires in-order behavior from FPGA
    for (i = 0; i < n; i++) {
      struct pvirtq_desc* desc = &txvq->desc[txvq->next_used];

      ule16 flags = desc->flags;
      uint8_t avail = !!(flags & VIRTQ_DESC_F_AVAIL);
      uint8_t used  = !!(flags & VIRTQ_DESC_F_USED);

      if (avail != txvq->used_wrap_count || used != txvq->used_wrap_count) {
        n = i;
        break;
      }

      le16 incr = (desc->id - txvq->next_used) & queueMask;
      i += incr;
      inc_used(txvq, incr + 1);
    }

    if (i > n) {
      int outs_num = i - n;
      txvq->outs.next = (txvq->next_used - outs_num) & queueMask;
      txvq->outs.num  = outs_num;
    }

  } else
    return 0;

  return m+n;
}

void nthw_release_tx_buffers(struct nthw_virt_queue* txvq,
                      uint16_t n, uint16_t n_segs[])
{
  int i;
  if (txvq->vq_type == SPLIT_RING) {
    le16 queueMask = (le16)(txvq->queueSize - 1); // Valid because queueSize is always 2^n

    vq_log_arg(txvq, "pkts %i, avail idx %i, start at %i\n", n, txvq->amIdx, txvq->txDescrAvailIdx);
    for (i = 0; i < n; i++) {
      int idx = txvq->amIdx & queueMask;
      txvq->pAvail->ring[idx] = txvq->txDescrAvailIdx;
      txvq->txDescrAvailIdx = (txvq->txDescrAvailIdx + n_segs[i]) & queueMask;
      txvq->amIdx++;
    }
    /* make sure the ring has been updated before HW reads index update */
    rte_mb();
    txvq->pAvail->idx = txvq->amIdx;
    vq_log_arg(txvq, "new avail idx %i, descr_idx %i\n", txvq->pAvail->idx, txvq->txDescrAvailIdx);

  } else
  if (txvq->vq_type == PACKED_RING) {


    // defer flags update on first segment - due to serialization towards HW and when jumbo segments are added

    ule16 first_flags =  avail_flag(txvq) | used_flag_inv(txvq);
    struct pvirtq_desc* first_desc = &txvq->desc[txvq->next_avail];

    for (i = 0; i < n; i++) {
      struct pvirtq_desc* desc = &txvq->desc[txvq->next_avail];

      desc->id   = txvq->next_avail;
      desc->addr = (ule64)txvq->pVirtualAddr[desc->id].physAddr;

      if (i)
        desc->flags |= avail_flag(txvq) | used_flag_inv(txvq); // bitwise-or here because next flags may already have been setup

      inc_avail(txvq, 1);
    }
    // proper read barrier before FPGA may see first flags
    rte_rmb();
    first_desc->flags = first_flags;
  }

}

int nthw_get_rx_queue_ptr(struct nthw_virt_queue* rxvq, uint16_t *index)
{
  uint32_t rx_ptr;
  uint32_t loops = 100000;

  dbs_initialize_get_rx_ptr(rxvq->mpNthwDbs, rxvq->index);
  do {
    if (dbs_get_rx_ptr(rxvq->mpNthwDbs, &rx_ptr) != 0) return -1;
    if (--loops == 0) return -1;
    usleep(10);
  } while (rx_ptr == INDEX_PTR_NOT_VALID);

  *index = (uint16_t)(rx_ptr & 0xffff);
  return 0;
}


int nthw_get_tx_queue_ptr(struct nthw_virt_queue* txvq, uint16_t *index)
{
  uint32_t tx_ptr;
  uint32_t loops = 100000;

  dbs_initialize_get_tx_ptr(txvq->mpNthwDbs, txvq->index);
  do {
    if (dbs_get_tx_ptr(txvq->mpNthwDbs, &tx_ptr) != 0) return -1;
    if (--loops == 0) return -1;
    usleep(10);
  } while (tx_ptr == INDEX_PTR_NOT_VALID);

  *index = (uint16_t)(tx_ptr & 0xffff);
  return 0;
}
