/*-
 *   BSD LICENSE
 *
 *   Copyright(c) 2020 Napatech A/S. All rights reserved.
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of Napatech nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

//
// Includes:
//
#include <stdio.h>
#include <stdint.h>
#include <malloc.h>
//
#include "ntos_drv.h"
#include <rte_flow_driver.h>
#include <rte_pci.h>
#include "ntnic_ethdev.h"

#include "ntlog.h"
#include "nt_util.h"
#include "create_elements.h"
#include "ntnic_filter.h"

#define MAX_RTE_FLOWS  8192
#define MAX_PORTIDS  64

#if (MAX_COLOR_FLOW_STATS != NT_MAX_COLOR_FLOW_STATS)
#error Difference in COLOR_FLOW_STATS. Please synchronize the defines.
#endif

struct rte_flow nt_flows[MAX_RTE_FLOWS];

#define MAC_ADDRESS(a)  a[0] & 0xFF, a[1] & 0xFF, a[2] & 0xFF, a[3] & 0xFF, a[4] & 0xFF, a[5] & 0xFF
#define IPV4_ADDRESS(a) ((const char *)&a)[0] & 0xFF, ((const char *)&a)[1] & 0xFF, ((const char *)&a)[2] & 0xFF, ((const char *)&a)[3] & 0xFF

static int is_flow_handle_typecast(struct rte_flow* flow)
{
  const void* first_element = &nt_flows[0];
  const void* last_element = &nt_flows[MAX_RTE_FLOWS - 1];
  return (void*)flow < first_element || (void*)flow > last_element;
}

static int convert_flow(struct rte_eth_dev *eth_dev,
   const struct rte_flow_attr *attr,
   const struct rte_flow_item items[],
   const struct rte_flow_action actions[],
   struct cnv_attr_s *attribute,
   struct cnv_match_s *match,
   struct cnv_action_s *action,
   struct rte_flow_error *error)
{
  struct pmd_internals *dev = eth_dev->data->dev_private;
  struct FpgaInfo_s *fpga_info = &dev->pDrv->ntdrv.adapter_info.fpga_info;

  static struct flow_error flow_error = {.type = FLOW_ERROR_NONE, .message = "none"};
  uint32_t flow_stat_id = 0;
  uint32_t queue_offset = 0;

#ifdef RTE_FLOW_DEBUG
  NT_LOG(DBG, FILTER, "ntnic_flow_create port_id %u - %s\n", eth_dev->data->port_id, eth_dev->data->name);
#endif

  if (dev->type == PORT_TYPE_OVERRIDE && dev->vpq_nb_vq > 0) {
    // The queues comming from the main PMD will always start from 0
    // When the port is a the VF/vDPA port the queues must be changed
    // to match the queues allocated for VF/vDPA.
    queue_offset = dev->vpq[0].id;
  }

  /* set initial error */
  convert_error(error, &flow_error);

  if (!dev) {
    rte_flow_error_set(error, EINVAL, RTE_FLOW_ERROR_TYPE_UNSPECIFIED, NULL, "Missing eth_dev");
    return -1;
  }

  if (create_attr(attribute, attr) != 0) {
    rte_flow_error_set(error, EINVAL, RTE_FLOW_ERROR_TYPE_ATTR, NULL, "Error in attr");
    return -1;
  }
  if (create_match_elements(match, items, MAX_ELEMENTS) != 0) {
    rte_flow_error_set(error, EINVAL, RTE_FLOW_ERROR_TYPE_ITEM, NULL, "Error in items");
    return -1;
  }
  if (fpga_info->profile == FPGA_INFO_PROFILE_INLINE) {
    if (create_action_elements_inline(action, actions, MAX_ACTIONS, queue_offset) != 0) {
      rte_flow_error_set(error, EINVAL, RTE_FLOW_ERROR_TYPE_ACTION, NULL, "Error in actions");
      return -1;
    }
    if (attribute->attr.group > 0) {
      return 0;
    }
  }
  else if (fpga_info->profile == FPGA_INFO_PROFILE_VSWITCH) {

    if (create_action_elements_vswitch(action, actions, MAX_ACTIONS, &flow_stat_id) != 0) {
      rte_flow_error_set(error, EINVAL, RTE_FLOW_ERROR_TYPE_ACTION, NULL, "Error in actions");
      return -1;
    }
  }
  else {
    rte_flow_error_set(error, EPERM, RTE_FLOW_ERROR_TYPE_UNSPECIFIED, NULL, "Unsupported adapter profile");
    return -1;
  }
  return 0;
}


static int eth_flow_destroy(struct rte_eth_dev *eth_dev,
    struct rte_flow *flow,
    struct rte_flow_error *error)
{
  struct pmd_internals *dev = eth_dev->data->dev_private;
  static struct flow_error flow_error = {.type = FLOW_ERROR_NONE, .message = "none"};

  int res = 0;

  /* set initial error */
  convert_error(error, &flow_error);

  if (!flow)
    return 0;

  if (is_flow_handle_typecast(flow)) {
    res = flow_destroy(dev->flw_dev, (void*)flow, &flow_error);
    convert_error(error, &flow_error);
  }
  else {
    res = flow_destroy(dev->flw_dev, flow->flw_hdl, &flow_error);
    convert_error(error, &flow_error);

    rte_spinlock_lock(&flow_lock);
    delete_flow_stat_id_locked(flow->flow_stat_id);
    flow->used = 0;
    rte_spinlock_unlock(&flow_lock);
  }

  // Clear the flow statistics if succesfully destroyed
  if (res == 0) {
    flow->stat_pkts = 0UL;
    flow->stat_bytes = 0UL;
    flow->stat_tcp_flags = 0;
  }

  return res;
}

static int eth_flow_validate(struct rte_eth_dev *eth_dev,
   const struct rte_flow_attr *attr,
   const struct rte_flow_item items[],
   const struct rte_flow_action actions[],
   struct rte_flow_error *error)
{
  static struct flow_error flow_error = {.type = FLOW_ERROR_NONE, .message = "none"};
  struct pmd_internals *dev = eth_dev->data->dev_private;
  struct cnv_attr_s   attribute;
  struct cnv_match_s  match;
  struct cnv_action_s action;
  int res;

  if (convert_flow(eth_dev, attr, items, actions, &attribute, &match, &action, error) < 0) {
    return -EINVAL;
  }

  res = flow_validate(dev->flw_dev, match.flow_elem, action.flow_actions, &flow_error);

  if (res < 0) {
    convert_error(error, &flow_error);
  }

  return res;
}

static struct rte_flow *eth_flow_create(struct rte_eth_dev *eth_dev,
           const struct rte_flow_attr *attr,
           const struct rte_flow_item items[],
           const struct rte_flow_action actions[],
           struct rte_flow_error *error)
{
  struct pmd_internals *dev = eth_dev->data->dev_private;
  struct FpgaInfo_s *fpga_info = &dev->pDrv->ntdrv.adapter_info.fpga_info;

  struct cnv_attr_s   attribute;
  struct cnv_match_s  match;
  struct cnv_action_s action;

  static struct flow_error flow_error = {.type = FLOW_ERROR_NONE, .message = "none"};
  uint32_t flow_stat_id = 0;

#ifdef RTE_FLOW_DEBUG
  NT_LOG(DBG, FILTER, "ntnic_flow_create port_id %u - %s\n", eth_dev->data->port_id, eth_dev->data->name);
#endif

  if (convert_flow(eth_dev, attr, items, actions, &attribute, &match, &action, error) < 0) {
    return NULL;
  }

  if ((fpga_info->profile == FPGA_INFO_PROFILE_INLINE) && (attribute.attr.group > 0)) {
    void *flw_hdl = flow_create(dev->flw_dev, &attribute.attr, match.flow_elem, action.flow_actions, &flow_error);
    convert_error(error, &flow_error);
    return (struct rte_flow*)flw_hdl;
  }

  struct rte_flow *flow = NULL;
  rte_spinlock_lock(&flow_lock);
  int i;
  for (i = 0; i < MAX_RTE_FLOWS; i++) {
    if (!nt_flows[i].used) {
      nt_flows[i].flow_stat_id = flow_stat_id;
      if (nt_flows[i].flow_stat_id < NT_MAX_COLOR_FLOW_STATS) {
        nt_flows[i].used = 1;
        flow = &nt_flows[i];
      }
      break;
    }
  }
  rte_spinlock_unlock(&flow_lock);
  if (flow) {
    flow->flw_hdl = flow_create(dev->flw_dev, &attribute.attr, match.flow_elem, action.flow_actions, &flow_error);
    convert_error(error, &flow_error);
    if (!flow->flw_hdl) {
      rte_spinlock_lock(&flow_lock);
      delete_flow_stat_id_locked(flow->flow_stat_id);
      flow->used = 0;
      flow = NULL;
      rte_spinlock_unlock(&flow_lock);
    } else {
#ifdef RTE_FLOW_DEBUG
      NT_LOG(INF, FILTER, "Create Flow %p using stat_id %i\n", flow, flow->flow_stat_id);
#endif
    }
  }
  return flow;
}


uint64_t last_stat_rtc = 0;

int poll_statistics(struct pmd_internals *internals)
{
  int flow;
  //int hb_idx;
  struct Drv_s* pDrv = internals->pDrv;
  struct Ntdrv_4ga_s* pNtDrv = &pDrv->ntdrv;
  Nt4gaStat_t* pNt4gaStat = &pNtDrv->adapter_info.nt4ga_stat;
  const int if_index = internals->if_index;

  if (!pNt4gaStat || if_index < 0 || if_index > NUM_ADAPTER_PORTS_MAX) {
    return -1;
  }

  assert(rte_tsc_freq > 0);

  rte_spinlock_lock(&hwlock);

  uint64_t now_rtc = rte_get_tsc_cycles();

  // Check per port max once a second
  /* if more than a second since last stat read, do a new one */
  if ((now_rtc - internals->last_stat_rtc) < rte_tsc_freq) {
    rte_spinlock_unlock(&hwlock);
    return 0;
  }

  internals->last_stat_rtc = now_rtc;

  pthread_mutex_lock(&pNtDrv->stat_lck);

  // Add the RX statistics increments since last time we polled. (No difference if physical or virtual port)
  internals->rxq_scg[0].rx_pkts += pNt4gaStat->aPortRxPacketsTotal[if_index] - pNt4gaStat->aPortRxPacketsBase[if_index];
  internals->rxq_scg[0].rx_bytes += pNt4gaStat->aPortRxOctetsTotal[if_index] - pNt4gaStat->aPortRxOctetsBase[if_index];
  internals->rxq_scg[0].err_pkts += 0; // What to user here ??
  internals->rx_missed += pNt4gaStat->aPortRxDropsTotal[if_index] - pNt4gaStat->aPortRxDropsBase[if_index];

  // Update the increment bases
  pNt4gaStat->aPortRxPacketsBase[if_index] = pNt4gaStat->aPortRxPacketsTotal[if_index];
  pNt4gaStat->aPortRxOctetsBase[if_index] = pNt4gaStat->aPortRxOctetsTotal[if_index];
  pNt4gaStat->aPortRxDropsBase[if_index] = pNt4gaStat->aPortRxDropsTotal[if_index];

  //NT_LOG(INF, FILTER, "[%i] Get port stat pkts %lu\n", internals->port, nt_flows[flow].stat_pkts);

  // Tx (here we must distinguish between physical and virtual ports)
  if (internals->type == PORT_TYPE_PHYSICAL) {
    // LAG management of Tx stats.
    if (lag_active && if_index == 0) {
      unsigned int i;
      /*
       * Collect all LAG ports Tx stat into this one. Simplified to only collect from port 0 and 1
       * TODO: must sum over all participating ports in each LAG - for future extension
       */
      for (i = 0; i < 2; i++) {
        // Add the statistics increments since last time we polled
        internals->txq_scg[0].tx_pkts += pNt4gaStat->aPortTxPacketsTotal[i] - pNt4gaStat->aPortTxPacketsBase[i];
        internals->txq_scg[0].tx_bytes += pNt4gaStat->aPortTxOctetsTotal[i] - pNt4gaStat->aPortTxOctetsBase[i];
        internals->txq_scg[0].err_pkts += 0; // What to user here ??

        // Update the increment bases
        pNt4gaStat->aPortTxPacketsBase[i] = pNt4gaStat->aPortTxPacketsTotal[i];
        pNt4gaStat->aPortTxOctetsBase[i] = pNt4gaStat->aPortTxOctetsTotal[i];
      }
    } else {
      // Add the statistics increments since last time we polled
      internals->txq_scg[0].tx_pkts += pNt4gaStat->aPortTxPacketsTotal[if_index] - pNt4gaStat->aPortTxPacketsBase[if_index];
      internals->txq_scg[0].tx_bytes += pNt4gaStat->aPortTxOctetsTotal[if_index] - pNt4gaStat->aPortTxOctetsBase[if_index];
      internals->txq_scg[0].err_pkts += 0; // What to user here ??

      // Update the increment bases
      pNt4gaStat->aPortTxPacketsBase[if_index] = pNt4gaStat->aPortTxPacketsTotal[if_index];
      pNt4gaStat->aPortTxOctetsBase[if_index] = pNt4gaStat->aPortTxOctetsTotal[if_index];
    }
  }
  if (internals->type == PORT_TYPE_VIRTUAL) {
    // Update TX counters from HB queue counter
    unsigned int i;
    struct HostBufferCounters* const pHbCounters = pNt4gaStat->mpStatStructsHb;
    uint64_t vPortPacketsTotal = 0, vPortOctetsTotal = 0;

    /*
     * This is a bit odd. But typically nb_tx_queues must be only 1 since it denotes the number of exception queues
     * which must be 1 - for now. The code is kept if we want it in future, but it will not be likely.
     * Therefore adding all vPorts queue tx counters into Tx[0] is ok for now.
     *
     * Only use the vPort Tx counter to update OVS, since these are the real ones. The rep port into OVS that represents this port
     * will always replicate the traffic here, also when no offload occurs
     */
    for(i = 0; i < internals->vpq_nb_vq; ++i) {
      vPortPacketsTotal += pHbCounters[internals->vpq[i].id].fwdPackets;
      vPortOctetsTotal += pHbCounters[internals->vpq[i].id].fwdBytes;
    }
    // Add the statistics increments since last time we polled
    internals->txq_scg[0].tx_pkts += vPortPacketsTotal - pNt4gaStat->aPortTxPacketsBase[if_index];
    internals->txq_scg[0].tx_bytes += vPortOctetsTotal - pNt4gaStat->aPortTxOctetsBase[if_index];
    internals->txq_scg[0].err_pkts += 0; // What to user here ??

    // Update the increment bases
    pNt4gaStat->aPortTxPacketsBase[if_index] = vPortPacketsTotal;
    pNt4gaStat->aPortTxOctetsBase[if_index] = vPortOctetsTotal;
  }

  // Globally only once a second
  if ((now_rtc - last_stat_rtc) < rte_tsc_freq) {
    rte_spinlock_unlock(&hwlock);
    pthread_mutex_unlock(&pNtDrv->stat_lck);
    return 0;
  }

  last_stat_rtc = now_rtc;

  //NT_LOG(INF, FILTER, "Get flow stats  called by port %i\n", internals->port);
  // All color counter are global, therefore only 1 pmd must update them
  const struct ColorCounters* pColorCounters = pNt4gaStat->mpStatStructsColor;
  struct ColorCounters* pColorCountersBase = pNt4gaStat->aStatStructsColorBase;
  uint64_t colorPacketsAccumulated, colorBytesAccumulated;

  for (flow = 0; flow < MAX_RTE_FLOWS; flow++) {
    if (nt_flows[flow].used) {
      unsigned color = nt_flows[flow].flow_stat_id;

      //NT_LOG(INF, FILTER, "Get stat flow %i, color %u\n", flow, color);

      if (color < NT_MAX_COLOR_FLOW_STATS) {

        colorPacketsAccumulated = pColorCounters[color].colorPackets;
        nt_flows[flow].stat_pkts      += (colorPacketsAccumulated - pColorCountersBase[color].colorPackets);

        nt_flows[flow].stat_tcp_flags |= pColorCounters[color].tcp_flags;

        colorBytesAccumulated = pColorCounters[color].colorBytes;
        nt_flows[flow].stat_bytes     += (colorBytesAccumulated - pColorCountersBase[color].colorBytes);

        // Update the counter bases
        pColorCountersBase[color].colorPackets = colorPacketsAccumulated;
        pColorCountersBase[color].colorBytes = colorBytesAccumulated;

        //NT_LOG(INF, FILTER, "Get stat pkts %lu\n", nt_flows[flow].stat_pkts);
      }
    }
  }

  rte_spinlock_unlock(&hwlock);
  pthread_mutex_unlock(&pNtDrv->stat_lck);

  return 0;
}




static int eth_flow_query(struct rte_eth_dev *dev,
         struct rte_flow *flow,
         const struct rte_flow_action *action,
         void *data,
         struct rte_flow_error *err)
{
  struct pmd_internals *internals = dev->data->dev_private;
  err->cause = NULL;
  err->message = NULL;

  if (is_flow_handle_typecast(flow)) {
    rte_flow_error_set(err, EFAULT, RTE_FLOW_ERROR_TYPE_UNSPECIFIED, NULL, "Error in flow handle");
    return -1;
  }

  poll_statistics(internals);

  //NT_LOG(INF, FILTER, "Flow %p query stat_id %i\n", flow, flow->flow_stat_id);
  if (action->type == RTE_FLOW_ACTION_TYPE_COUNT) {
    struct rte_flow_query_count *qcnt = (struct rte_flow_query_count *)data;
    if (qcnt) {
      if (flow) {
        //NT_LOG(INF, FILTER, "Update flow cnt %lu\n", flow->stat_pkts);
        qcnt->hits = flow->stat_pkts;
        /* tcp_flags is not DPDK std, but added functionality */
        qcnt->tcp_flags = 0 /*flow->stat_tcp_flags*/; // 

        qcnt->hits_set = 1;
        qcnt->bytes = flow->stat_bytes;
        qcnt->bytes_set = 1;

        if (qcnt->reset) {
          flow->stat_pkts = 0UL;
          flow->stat_bytes = 0UL;
          flow->stat_tcp_flags = 0;
        }
      } else {
        qcnt->hits_set = 0;
        qcnt->bytes_set = 0;
      }
    }
  } else {
    rte_flow_error_set(err, EINVAL, RTE_FLOW_ERROR_TYPE_ACTION, NULL, "Unsupported query");
    return -1;
  }
  rte_flow_error_set(err, 0, RTE_FLOW_ERROR_TYPE_NONE, NULL, "Success");
  return 0;
}




#ifdef DEBUGGING

static void _print_tunnel(struct rte_flow_tunnel *tunnel)
{
  struct in_addr addr;

  printf("    tun type: %i\n", tunnel->type);
  printf("    tun ID: %016lx\n", tunnel->tun_id);
  addr.s_addr = tunnel->ipv4.src_addr;
  printf("    tun src IP: %s\n", inet_ntoa(addr));
  addr.s_addr = tunnel->ipv4.dst_addr;
  printf("    tun dst IP: %s\n",   inet_ntoa(addr));
  printf("    tun tp_src: %i\n",   htons(tunnel->tp_src));
  printf("    tun tp_dst: %i\n",   htons(tunnel->tp_dst));
  printf("    tun flags:  %i\n",   tunnel->tun_flags);
  printf("    tun ipv6:  %i\n",   tunnel->is_ipv6);

  printf("    tun tos:   %i\n",   tunnel->tos);
  printf("    tun ttl:   %i\n",   tunnel->ttl);

}
#endif


static struct rte_flow_action _pmd_actions[] = {{.type = (enum rte_flow_action_type)NT_RTE_FLOW_ACTION_TYPE_TUNNEL_SET, .conf = NULL},
    {.type = 0, .conf = NULL}
};

static int ntnic_tunnel_decap_set(struct rte_eth_dev *dev _unused,
   struct rte_flow_tunnel *tunnel,
   struct rte_flow_action **pmd_actions,
   uint32_t *num_of_actions,
   struct rte_flow_error *err _unused)
{

#ifdef DEBUGGING
  printf("called ntnic_tunnel_decap_set\n");
//  _print_tunnel(tunnel);
#endif

  if (tunnel->type == RTE_FLOW_ITEM_TYPE_VXLAN)
    _pmd_actions[1].type = RTE_FLOW_ACTION_TYPE_VXLAN_DECAP;
  else
    return -ENOTSUP;

  *pmd_actions = _pmd_actions;
  *num_of_actions = 2;

  return 0;
}


static struct rte_flow_item _pmd_items = {.type = (enum rte_flow_item_type)NT_RTE_FLOW_ITEM_TYPE_TUNNEL, .spec = NULL, .last = NULL, .mask = NULL};

static int ntnic_tunnel_match(struct rte_eth_dev *dev _unused,
   struct rte_flow_tunnel *tunnel _unused,
   struct rte_flow_item **pmd_items,
   uint32_t *num_of_items,
   struct rte_flow_error *err _unused)
{

#ifdef DEBUGGING
  printf("called ntnic_tunnel_match\n");
#endif

  *pmd_items = &_pmd_items;
  *num_of_items = 1;
  return 0;
}



/*
 * Restoration API support
 */
static int ntnic_get_restore_info(struct rte_eth_dev *dev _unused,
   struct rte_mbuf *m,
   struct rte_flow_restore_info *info,
   struct rte_flow_error *err _unused)
{
#ifdef DEBUGGING
  printf("called ntnic_get_restore_info\n");
  printf("dev name: %s - port_id %i\n", dev->data->name, dev->data->port_id);
  printf("dpdk tunnel mark %08x\n", m->hash.fdir.hi);
#endif


  if ((m->ol_flags & RTE_MBUF_F_RX_FDIR_ID) && m->hash.fdir.hi) {
    uint8_t port_id = (m->hash.fdir.hi >> 24) & 0xff;
    uint32_t stat_id = m->hash.fdir.lo & 0xffffff;

    //printf("dpdk tunnel mark %08x, flow id %i\n", m->hash.fdir.hi, m->hash.fdir.lo);

    struct tunnel_cfg_s tuncfg;
    int ret = flow_get_tunnel_definition(&tuncfg, stat_id, port_id);

    if (ret)
      return -EINVAL;

    if (tuncfg.ipversion == 4) {
      info->tunnel.ipv4.dst_addr = tuncfg.v4.dst_ip;
      info->tunnel.ipv4.src_addr = tuncfg.v4.src_ip;
      info->tunnel.is_ipv6 = 0;
    } else {
      // IPv6
      for (int i = 0; i < 16; i++) {
        info->tunnel.ipv6.src_addr[i] = tuncfg.v6.src_ip[i];
        info->tunnel.ipv6.dst_addr[i] = tuncfg.v6.dst_ip[i];
      }
      info->tunnel.is_ipv6 = 1;
    }

    info->tunnel.tp_dst = tuncfg.d_port;
    info->tunnel.tp_src = tuncfg.s_port;

    info->tunnel.ttl = 64;
    info->tunnel.tos = 0;

    info->tunnel.tun_flags = (1 << 3) | (1 << 1); // FLOW_TNL_F_KEY | FLOW_TNL_F_DONT_FRAGMENT

    info->tunnel.type = RTE_FLOW_ITEM_TYPE_VXLAN;
    info->tunnel.tun_id = m->hash.fdir.hi & 0xffffff;


    info->flags = RTE_FLOW_RESTORE_INFO_TUNNEL;  //         | RTE_FLOW_RESTORE_INFO_ENCAPSULATED; if restored packet is sent back
    info->group_id = 0;

#ifdef DEBUGGING
    _print_tunnel(&info->tunnel);
#endif

    return 0;
  }
  return -EINVAL;  // supported, but no hit found
}

static int ntnic_tunnel_action_decap_release(struct rte_eth_dev *dev _unused,
   struct rte_flow_action *pmd_actions _unused,
   uint32_t num_of_actions _unused,
   struct rte_flow_error *err _unused)
{
#ifdef DEBUGGING
  printf("called ntnic_tunnel_action_decap_release\n");
#endif
  return 0;
}

static int ntnic_tunnel_item_release(struct rte_eth_dev *dev _unused,
   struct rte_flow_item *pmd_items _unused,
   uint32_t num_of_items _unused,
   struct rte_flow_error *err _unused)
{
#ifdef DEBUGGING
  printf("called ntnic_tunnel_item_release\n");
#endif
  return 0;
}

// void cleanup_flows(struct pmd_internals *dev)
// {
//   if (dev->flw_dev)
//     flow_destroy(dev->flw_dev, NULL, NULL);
//   /*
//    * cleanup all stat_id
//    */
//   memset(flow_stat_id_map, 0, sizeof(flow_stat_id_map));
//   /*
//    * cleanup recorded flows
//    */
//   memset(nt_flows, 0, sizeof(nt_flows));
// }

const struct rte_flow_ops _dev_flow_ops = {
  .validate = eth_flow_validate,
  .create = eth_flow_create,
  .destroy = eth_flow_destroy,
  .flush = NULL,
  .query = eth_flow_query,
  .tunnel_decap_set = ntnic_tunnel_decap_set,
  .tunnel_match = ntnic_tunnel_match,
  .get_restore_info = ntnic_get_restore_info,
  .tunnel_action_decap_release = ntnic_tunnel_action_decap_release,
  .tunnel_item_release = ntnic_tunnel_item_release

};
