#include "sensors.h"
#include "ntlog.h"

void sensor_deinit(struct nt_sensor_group* sg) {
  if(sg) {
    if(sg->sensor){free(sg->sensor);}
    if(sg->monitor){free(sg->monitor);}
    free(sg);
  }
}

struct nt_adapter_sensor* allocate_sensor(
  uint8_t adapterOrPortIndex,
  const char* pName,
  enum NtSensorSource_e ssrc,
  enum NtSensorType_e type,
  unsigned index,
  enum NtSensorEventAlarm_e eventAlarm,
  enum sensor_mon_sign si
) {
  struct nt_adapter_sensor* sensor = (struct nt_adapter_sensor*)malloc(sizeof(struct nt_adapter_sensor));
  if(sensor == NULL) {
    NT_LOG(ERR, ETHDEV, "%s: sensor is NULL", __func__);
    return NULL;
  }

  sensor->alarm = eventAlarm;
  sensor->mEnableAlarm = true;
  sensor->mIntfNo = 0xFF;
  sensor->mAdapterNo = 0xFF;
  sensor->si = si;

  sensor->info.source = ssrc;
  sensor->info.sourceIndex = adapterOrPortIndex;
  sensor->info.sensorIndex = index;
  sensor->info.type = type;
  sensor->info.subType = NT_SENSOR_SUBTYPE_NA;
  sensor->info.state = NT_SENSOR_STATE_INITIALIZING;
  sensor->info.value = NT_SENSOR_NAN;
  sensor->info.valueLowest = NT_SENSOR_NAN;
  sensor->info.valueHighest = NT_SENSOR_NAN;
  memset(sensor->info.name, 0, NT_INFO_SENSOR_NAME);
  memcpy(sensor->info.name, pName, (strlen(pName) > NT_INFO_SENSOR_NAME) ? NT_INFO_SENSOR_NAME : strlen(pName));
  sensor->info.name[NT_INFO_SENSOR_NAME] = '\0';

  return sensor;
}

void update_sensor_value(struct nt_adapter_sensor* sensor, int32_t value)
{
  if(sensor == NULL){return;}
  sensor->info.value = value;
  if (sensor->info.valueHighest < value || sensor->info.valueHighest == NT_SENSOR_NAN) {
    sensor->info.valueHighest = value;
  }
  if (sensor->info.valueLowest > value || sensor->info.valueLowest == NT_SENSOR_NAN) {
    sensor->info.valueLowest = value;
  }
}

struct nt_adapter_sensor* allocate_sensor_by_description(uint8_t adapterOrPortIndex, enum NtSensorSource_e ssrc, struct nt_adapter_sensor_description* descr)
{
  struct nt_adapter_sensor* sensor = (struct nt_adapter_sensor*)malloc(sizeof(struct nt_adapter_sensor));
  if(sensor == NULL) {
    NT_LOG(ERR, ETHDEV, "%s: sensor is NULL", __func__);
    return NULL;
  }

  sensor->alarm = descr->eventAlarm;
  sensor->mEnableAlarm = true;
  sensor->mIntfNo = 0xFF;
  sensor->mAdapterNo = 0xFF;
  sensor->si = SENSOR_MON_UNSIGNED;

  sensor->info.sourceIndex = adapterOrPortIndex;
  sensor->info.source = ssrc;
  sensor->info.type = descr->type;
  sensor->info.sensorIndex = descr->index;
  memset(sensor->info.name, 0, NT_INFO_SENSOR_NAME);
  memcpy(sensor->info.name, descr->name, (strlen(descr->name) > NT_INFO_SENSOR_NAME) ? NT_INFO_SENSOR_NAME : strlen(descr->name));
  sensor->info.name[NT_INFO_SENSOR_NAME] = '\0';

  return sensor;
}

void init_sensor_group(struct nt_sensor_group* sg)
{
  // set all pointers to NULL
  sg->sensor = NULL;
  sg->monitor = NULL;
  sg->next = NULL;
  sg->read = NULL;
  sg->conv_func = NULL;
}

// getters
int32_t get_value(struct nt_sensor_group* sg) {return sg->sensor->info.value;};
int32_t get_lowest(struct nt_sensor_group* sg) {return sg->sensor->info.valueLowest;};
int32_t get_highest(struct nt_sensor_group* sg) {return sg->sensor->info.valueHighest;};
char* get_name(struct nt_sensor_group* sg) {return sg->sensor->info.name;};

// conversion functions
int null_signed(uint32_t pSensorResult) {
  return (int16_t)pSensorResult;
}

int null_unsigned(uint32_t pSensorResult) {
  return (uint16_t)pSensorResult;
}


//******************************************************************************
//For EXAR7724: Convert a read Vch value to Napatech internal representation
//Doc: Vout = ReadVal * 0.015 (PRESCALE is accounted for)
//******************************************************************************
int Exar7724_Vch(uint32_t pSensorResult)
{
  return pSensorResult * 15; //NT unit: 1mV
}

//******************************************************************************
//For EXAR7724: Convert a read Vin value to Napatech internal representation
//Doc: Vout = ReadVal * 0.0125
//******************************************************************************
int Exar7724_Vin(uint32_t pSensorResult)
{
  return (pSensorResult * 25) / 2; //NT unit: 1mV
}

//******************************************************************************
//For EXAR7724: Convert a read Tj value to Napatech internal representation
//Doc: Temp (in Kelvin) = (((ReadVal * 10mV) - 600mV) / (2mV/K)) + 300K =
//                      = ReadVal * 5K
//******************************************************************************
int Exar7724_Tj(uint32_t pSensorResult)
{
  //A value of 2730 is used instead of 2732 which is more correct but since
  //the temperature step is 5 degrees it is more natural to show these steps
  return pSensorResult * 50 - 2730; //NT unit: 0.1C
}

//******************************************************************************
//Convertion function for Linear Tecnology Linear_5s_11s format.
//The functions returns Y * 2**N, where N = b[15:11] is a 5-bit two's complement
//integer and Y = b[10:0] is an 11-bit two's complement integer.
//The multiplier value is used for scaling to Napatech units.
//******************************************************************************
static int conv5s_11s(uint16_t value, int multiplier)
{
  int  N, Y;

  Y = value & 0x07FF;

  if (value & 0x0400) {
    Y -= 0x0800; //The MSBit is a sign bit
  }

  N = (value >> 11) & 0x1F;

  if (N & 0x10) {
    N -= 0x20;   //The MSBit is a sign bit
  }

  Y *= multiplier;

  if (N > 0) {
    Y *= (1 << N);
  }
  else if (N < 0) {
    Y /= (1 << (-N));
  }

  return Y;
}

//******************************************************************************
//Temperature conversion from Linear_5s_11s format.
//******************************************************************************
int Ltm4676_Tj(uint32_t pSensorResult)
{
  return (uint16_t)conv5s_11s(pSensorResult, 10);  //NT unit: 0.1C
}

//******************************************************************************
//For MP2886a: Convert a read Tj value to Napatech internal representation
//******************************************************************************
int Mp2886a_Tj(uint32_t pSensorResult)
{
  /* MPS-2886p: READ_TEMPERATURE (register 0x8Dh)
     READ_TEMPERATURE is a 2-byte, unsigned integer. */
  return (uint16_t)pSensorResult; //NT unit: 0.1C
}

//******************************************************************************
//For MAX6642: Convert a read temperature value to Napatech internal representation
//******************************************************************************
int Max6642_T(uint32_t pSensorResult)
{
  if ((pSensorResult >> 8) == 0xFF) {
    return NT_SENSOR_NAN;
  }

  //The six lower bits are not used
  return (int)(((pSensorResult >> 6) * 5) / 2); //NT unit: 0.25 deg, Native unit: 0.1C
}

//******************************************************************************
//For DS1775: Convert a read temperature value to Napatech internal representation
//******************************************************************************
int Ds1775_T(uint32_t pSensorResult)
{
  return (pSensorResult * 10) / 256; //NT unit: 0.1 deg, Native unit: 1/256 C
}

//******************************************************************************
//For FAN: Convert a tick count to RPM
//NT unit: RPM, Native unit: 2 ticks/revolution
//******************************************************************************
int Fan(uint32_t pSensorResult)
{
  return (pSensorResult * 60U / 4);
}
