#ifndef _SENSORS_H
#define _SENSORS_H

#include "sensor_types.h"
#include "stream_info.h"
#include "nthw_platform_drv.h"
#include "nthw_drv.h"
#include "nthw_spi_v3.h"
#include "nthw_fpga_model.h"

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdlib.h>
#include <limits.h>
#include "avr_intf.h"


enum NtSensorEventAlarm_e {
  NT_SENSOR_ENABLE_ALARM,
  NT_SENSOR_LOG_ALARM,
  NT_SENSOR_DISABLE_ALARM,
};

/**
 * Sensor Class types
 */
enum NtSensorClass_e {
  NT_SENSOR_CLASS_FPGA        = 0, //!< Class for FPGA based sensors e.g FPGA temperature
  NT_SENSOR_CLASS_MCU         = 1, //!< Class for MCU based sensors e.g MCU temperature
  NT_SENSOR_CLASS_PSU         = 2, //!< Class for PSU based sensors e.g PSU temperature
  NT_SENSOR_CLASS_PCB         = 3, //!< Class for PCB based sensors e.g PCB temperature
  NT_SENSOR_CLASS_NIM         = 4, //!< Class for NIM based sensors e.g NIM temperature
  NT_SENSOR_CLASS_ANY         = 5, //!< Class for ANY sensors e.g any sensors
};

typedef enum NtSensorClass_e NtSensorClass_t;

/*
* Port of the sensor class
*/
struct nt_adapter_sensor
{
  uint8_t mAdapterNo;
  uint8_t mIntfNo;
  uint8_t fpgaIdx;  // for AVR sensors
  enum sensor_mon_sign si;
  struct NtInfoSensor_s info;
  enum NtSensorEventAlarm_e alarm;
  bool mEnableAlarm;
};

struct nt_fpga_sensor_monitor
{
  Fpga_t* FPGA;
  Module_t* mod;

  Register_t* reg;
  Field** fields;
  uint8_t fields_num;
};


/**
 * Sensor description.
 * Describe the static behavior of the sensor.
 */
struct nt_adapter_sensor_description
{
  enum NtSensorType_e    type;          //!< Sensor type.
  enum NtSensorSubType_e subType;       //!< Sensor subtype (if any applicable)
  unsigned  index;                      //!< Sensor group index.
  enum NtSensorEventAlarm_e eventAlarm; //!< Enable/Disable event alarm
  char name[20];                        //!< Sensor name.
};

struct nt_sensor_group
{
  struct nt_adapter_sensor* sensor;
  struct nt_fpga_sensor_monitor* monitor;
  void (*read)(struct nt_sensor_group* sg, NthwSpis_t* t_spi);

  // conv params are needed to call current conversion functions
  int (*conv_func)(uint32_t pSensorResult);
  // i2c interface for NIM sensors

  struct nt_sensor_group* next;
};

void init_sensor_group(struct nt_sensor_group* sg);

void update_sensor_value(struct nt_adapter_sensor* sensor, int32_t value);

void sensor_deinit(struct nt_sensor_group* sg);

// getters
int32_t get_value(struct nt_sensor_group* sg);
int32_t get_lowest(struct nt_sensor_group* sg);
int32_t get_highest(struct nt_sensor_group* sg);
char* get_name(struct nt_sensor_group* sg);

struct nt_adapter_sensor* allocate_sensor(
  uint8_t adapterOrPortIndex,
  const char* pName,
  enum NtSensorSource_e ssrc,
  enum NtSensorType_e type,
  unsigned index,
  enum NtSensorEventAlarm_e eventAlarm,
  enum sensor_mon_sign si
);
struct nt_adapter_sensor* allocate_sensor_by_description(uint8_t adapterOrPortIndex, enum NtSensorSource_e ssrc, struct nt_adapter_sensor_description* descr);


// conversion functions
int null_signed(uint32_t pSensorResult);
int null_unsigned(uint32_t pSensorResult);
int Exar7724_Tj(uint32_t pSensorResult);
int Max6642_T(uint32_t pSensorResult);
int Ds1775_T(uint32_t pSensorResult);
int Ltm4676_Tj(uint32_t pSensorResult);
int Exar7724_Vch(uint32_t pSensorResult);
int Exar7724_Vin(uint32_t pSensorResult);
int Mp2886a_Tj(uint32_t pSensorResult);
int Fan(uint32_t pSensorResult);
#endif
