/*-
 *   BSD LICENSE
 *
 *   Copyright(c) 2010-2015 Intel Corporation. All rights reserved.
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <inttypes.h>
#include <limits.h>
#include <signal.h>
#include <stdatomic.h>
#include <stdint.h>
#include <time.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>

#include <rte_eal.h>
#include <rte_ethdev.h>
#include <rte_cycles.h>
#include <rte_lcore.h>
#include <rte_mbuf.h>
#include <rte_mbuf_dyn.h>
#include <rte_flow.h>
#include <rte_mtr.h>
#include <rte_bitops.h>
#include <rte_hash.h>
#include <rte_jhash.h>

#include "filter.h"
#include "port.h"
#include "meter_setup.h"
#include "pre_learn.h"

#define BURST_SIZE 32
#define MAX_OPT_LEN 256
#define FLOW_DB_SIZE 100000000 // 100M flows to be handled
#define MAX_FLOW_DB_SIZE (RTE_MIN(FLOW_DB_SIZE, RTE_HASH_ENTRIES_MAX))

extern const uint32_t N3_IF_IPV4;
extern const uint32_t N3_IPV4_PEER;

extern const uint32_t N6_IF_IPV4_NET;
extern const uint32_t N6_IF_IPV4_SUBNET_MASK;

extern const uint32_t IF_IPV4_MASK;

extern const uint8_t N3_IF_IPV6[16];
extern const uint8_t N3_IPV6_PEER[16];

extern const uint8_t N6_IF_IPV6[16];
extern const uint8_t IF_IPV6_MASK[16];

extern const uint8_t IPv6_addr_any[16];

extern const uint16_t NAT_ports[2];

struct flow_key {

  union
  {
    uint32_t ipv4_src;
    uint8_t ipv6_src[16];
  };

  union
  {
    uint32_t ipv4_dst;
    uint8_t ipv6_dst[16];
  };

  uint32_t encap_gtp_teid;
  uint16_t ether_type; // IPv4 or IPv6
  uint16_t port_src;
  uint16_t port_dst;
  // uint32_t proto; // protocol is not needed for this example as only UDP is supported
} __rte_packed;

// ------------------------- arguments
#define MAX_QUEUE_NUMBER (128 / MAX_PORT_NUMBER)

static bool opt_verbose = false;
static bool NAT_EN = false;
static uint32_t start_mtr_id = 0;
static uint32_t end_mtr_id = 0;

static char pre_learn_file[MAX_OPT_LEN] = {0};
static int rss_queues_count[MAX_PORT_NUMBER] = {1}; // One queue always allowed
static uint16_t rss_queues_ul[MAX_QUEUE_NUMBER] = {0};
static int rx_rings[MAX_PORT_NUMBER] = {0};
static int tx_rings[MAX_PORT_NUMBER] = {0};
uint64_t *green_pkt = NULL;
uint64_t stats_mask = 0;

static const char *usageExamples =
  "\nImplemented commands:\n"
  "\n"
  "* dpdk-inline_upf_demo\n"
  "\n"
  "* dpdk-inline_upf_demo -- -v\n"
  "\n"
  "* dpdk-inline_upf_demo -- -p 0,pir=10G\n"
  "\n"
  "* dpdk-inline_upf_demo -- -p 0,pir=10M,pbs=100000\n"
  "\n"
  "* dpdk-inline_upf_demo -- -f ./test.pcap\n"
  "\n"
  "* dpdk-inline_upf_demo -- -f ./test.pcap -m 1\n"
  "\n"
  "* dpdk-inline_upf_demo -- -f ./test.pcap -m 1 -u\n"
  "\n"
  "* dpdk-inline_upf_demo -- -m 1\n"
  "\n"
  "* dpdk-inline_upf_demo -- -m 1,pir=10G\n"
  "\n"
  "* dpdk-inline_upf_demo -- -m 2,pir=10G,pir=5G\n"
  "\n"
  "* dpdk-inline_upf_demo -- -r 2,3\n"
  "\n"
  "* dpdk-inline_upf_demo -a 0000:10:00.0,rxqs=4 -- -r all\n"
  "\n"
  "* dpdk-inline_upf_demo -- -s 3-12"
  "\n"
  "commands:\n"

  "\t-p 0[,pir={number}{M|G}[,pbs={number}{M|G}]] [-m <count of policy per flow>[,pir={number}{M|G}[,pbs={number}{M|G}]] \
  [-u] [-s <start_mete_id>-<end_meter_id]] \
  -r {<comma separated ids>|all} ]\n"

  "\t-v\t verbose logging\n"
  "\t-h\t show this message\n"
  "\t-f\t set pcap file for pre learning\n"
  "\t-m\t set number of policy per flow\n"
  "\t-u\t enable unique policy per flow, used in combination with -m param\n"
  "\t-s\t displays statistics for selected flows range\n"
  "\t-n\t enable NAT\n"
  "\t-r\t set queues IDs for RSS\n"
  ;

static void set_PIR(char *str, uint64_t *PIR)
{
  if(str && PIR && strncmp(str, "pir=", 4) == 0) {
    int rate = atoi(str + 4);
    if (rate == 0) {
      *PIR = DEFAULT_PIR;
      return;
    }

    *PIR = rate;

    // convert rate to Bytes per second
    if(str[strlen(str) - 1] == 'G') {
      *PIR *= (1000 * 1000 * 1000);
    } else if (str[strlen(str) - 1] == 'M'){
      *PIR *= (1000 * 1000);
    }
  }
}

static void set_PBS(char *str, uint64_t *PBS)
{
  if(str && PBS && strncmp(str, "pbs=", 4) == 0) {
    int bytes = atoi(str + 4);
    if (bytes == 0) {
      *PBS = DEFAULT_PBS;
      return;
    }

    *PBS = bytes;

    // convert to bytes
    if(str[strlen(str) - 1] == 'G') {
      *PBS *= (1000 * 1000 * 1000);
    } else if (str[strlen(str) - 1] == 'M'){
      *PBS *= (1000 * 1000);
    }
  }
}

static void parse_opt_port(char *str)
{
  char *token = strtok(str, ",");
  uint16_t port_id = 0;

  if(token != NULL) {
    port_id = atoi(token);
    if (port_id != PORT_UL) {
      printf("ERROR: MBR configuration for port %d is ignored\n", port_id);
      return;
    }
    token = strtok(NULL, ",");

    set_PIR(token, &port_pir_value[PORT_UL]); // for UL traffic only
    token = strtok(NULL, ",");
    if (token && strncmp(token, "pbs=", 4) == 0) {
      set_PBS(token, &port_pbs_value[PORT_UL]);
    }
  }
}

static void parse_opt_MTR(char *str)
{
  char *token = strtok(str, ",");
  uint16_t MTR_nb = 0;
  uint16_t i = 0;

  if(token != NULL) {
    MTR_nb = atoi(token);
    if (MTR_nb > MAX_POLICY_PER_FLOW) {
      printf("ERROR: while applying meter policy per flow, must be 1 or 2, not %i\n", flow_meter_policy);
      flow_meter_policy = 0;
      return;
    }
  }
  flow_meter_policy = MTR_nb;
  policy_per_flow = MTR_nb;

  token = strtok(NULL, ",");
  while (token != NULL && i < MTR_nb)
  {
    set_PIR(token, &flow_meter_policy_PIR[i]);
    token = strtok(NULL, ",");
    if (token && strncmp(token, "pbs=", 4) == 0) {
      set_PBS(token, &flow_meter_policy_PBS[i]);
      token = strtok(NULL, ",");
    }
    ++i;
  }
}

static void parse_opt_rss(char *str) {
  char *token = strtok(str, ",");
  int queues_nb = 0;

  while (token != NULL) {

    if (strncmp(token, "all", 3) == 0){
      for (; queues_nb < rx_rings[PORT_UL]; queues_nb++)
      {
        rss_queues_ul[queues_nb] = queues_nb;
      }
      break;
    }

    uint16_t queue_id = atoi(token);
    if (queue_id >= rx_rings[PORT_UL]) {
      printf("ERROR: requested queue %d is out of range [0-%d]\n", queue_id, rx_rings[PORT_UL]-1);
      token = strtok(NULL, ",");
      continue;
    }

    rss_queues_ul[queues_nb++] = atoi(token);

    token = strtok(NULL, ",");
  }

  rss_queues_count[PORT_UL] = queues_nb > 1 ? queues_nb : 1; // Always allow 1 queue
}

static void launch_args_parse(int argc, char** argv) {
  int opt;
  uint32_t start = 0;
  uint32_t end = 0;

  while ((opt = getopt(argc, argv, "hvp:f:m:uns:r:")) != -1) {
    switch (opt) {
      case 'v':
        opt_verbose = true;
        break;
      case 'p':
        if (optarg != NULL) {
          parse_opt_port(optarg);
        }
        break;
      case 'm':
        if (optarg != NULL) {
          parse_opt_MTR(optarg);
        }
        break;
      case 'u':
        unique_policy_per_flow_en = true;
        break;
      case 'f':
        strlcpy(pre_learn_file, optarg, MAX_OPT_LEN);
        break;
      case 'n':
        NAT_EN = true;
        break;
      case 's':
        start = atoi(strtok(optarg, "-"));
        end = atoi(strtok(NULL, "-"));
        start_mtr_id = RTE_MIN(start, end);
        end_mtr_id = RTE_MAX(start, end);
        stats_mask = RTE_MTR_STATS_N_PKTS_GREEN | RTE_MTR_STATS_N_BYTES_GREEN;
        break;
      case 'r':
        if (optarg != NULL) {
          parse_opt_rss(optarg);
        }
        break;
      case 'h':
      default:
        printf("%s\n", usageExamples);
        exit(0);
    }
  }
}
// -----------------------------------

_Atomic volatile uint8_t quit_signal;

static void
int_handler(int sig_num)
{
  printf("Exiting on signal %d\n", sig_num);
  atomic_store(&quit_signal, 1);
}

////////////////////////////////////////////////////////////////////////////////
//  Threads setup
////////////////////////////////////////////////////////////////////////////////

struct worker_thread_context_s {
  int number;
  int port;
  int queue;
};

struct rte_flow_list {
  struct rte_flow* flow_downlink;
  struct rte_flow* flow_uplink;
  struct rte_flow_list* next;
};

static void print_meter_stat(void)
{
  struct rte_mtr_stats stats;
  uint64_t stats_mask;
  struct rte_mtr_error error;
  for (uint32_t mtr_id = start_mtr_id; mtr_id < end_mtr_id; ++mtr_id) {
    if (rte_mtr_stats_read(0, mtr_id, &stats, &stats_mask, 0, &error) == 0) {
      if (green_pkt[mtr_id-start_mtr_id] != stats.n_pkts[RTE_COLOR_GREEN]) {
        green_pkt[mtr_id-start_mtr_id] = stats.n_pkts[RTE_COLOR_GREEN];
        printf("Meter ID %-9d : Green packets %-16" PRIu64 " Green bytes %-16" PRIu64 "\n",
          mtr_id, stats.n_pkts[RTE_COLOR_GREEN], stats.n_bytes[RTE_COLOR_GREEN]);
      }
    } else {
      printf("ERROR: rte_mtr_stats_read failed");
    }
  }
}

static int lcore_worker(void* p)
{

  struct worker_thread_context_s *context = (struct worker_thread_context_s*)p;

  struct rte_flow_error error;
  memset(&error, 0x0, sizeof(struct rte_flow_error));

  uint32_t N6_IPV4_SUBNET = {0};
  uint8_t N6_IPV6_SUBNET[16];
  memcpy(N6_IPV6_SUBNET, N6_IF_IPV6, 16);

  uint16_t NAT_udp_src = NAT_ports[0];
  uint8_t dscp_tag = DEFAULT_DSCP;
  uint8_t qfi = DEFAULT_QFI;
  uint8_t action_out_port_id = context->port == 0 ? 1 : 0;
  uint8_t ipv6_offset = 15;

  struct rte_flow_list *root = calloc(1, sizeof(struct rte_flow_list));
  if (!root) {
    printf("ERROR: cannot allocate memory\n");
    return -1;
  }
  struct rte_flow_list *tail = root;

  sleep(1);
  printf("lcore %u processing packets from port %d queue %d\n",
         rte_lcore_id(), context->port, context->queue);

  /*
    SUBNET = 20.20.0.0
    MASK = 255.240.0.0 = 0xFFF00000
                              ^-------- Set queue id here to to achieve subnets distribution between queues with RSS
    queue = 0 ----> subnet = 20.20.0.0
    queue = 1 ----> subnet = 20.21.0.0
    queue = 2 ----> subnet = 20.22.0.0
    queue = 3 ----> subnet = 20.23.0.0
  */
  N6_IPV4_SUBNET = N6_IF_IPV4_NET | RTE_BE32(( context->queue & 0xff) << 16);
  N6_IPV6_SUBNET[2] = (uint8_t)context->queue;

  if ( NAT_EN ) {
    printf("\t\tN3 IF ADDRESS IPv4: %d.%d.%d.%d\n",
                              ((N3_IF_IPV4 >> 0) & 0xff),
                              ((N3_IF_IPV4 >> 8) & 0xff),
                              ((N3_IF_IPV4 >> 16) & 0xff),
                              ((N3_IF_IPV4 >> 24) & 0xff));

    printf("\t\tN6 IF SUBNET  IPv4: %d.%d.%d.%d/16\n",
                              ((N6_IPV4_SUBNET >> 0) & 0xff),
                              ((N6_IPV4_SUBNET >> 8) & 0xff),
                              ((N6_IPV4_SUBNET >> 16) & 0xff),
                              ((N6_IPV4_SUBNET >> 24) & 0xff));

    printf("\t\tNAT UDP src port range: %d - %d\n", NAT_ports[0], NAT_ports[1]);
  }

  char name[RTE_HASH_NAMESIZE];
  snprintf(name, sizeof(name), "flows_DB_c%d_q%d",  rte_lcore_id(), context->queue);

  struct rte_hash_parameters ht_params = {
    .name = name,
    .entries = MAX_FLOW_DB_SIZE / rss_queues_count[PORT_UL], //each worker has separate hash table
    .key_len = sizeof(struct flow_key),
    .hash_func = rte_jhash,
    .hash_func_init_val = 0,
    .socket_id = rte_socket_id(),
    .extra_flag = 0
  };

  struct rte_hash *ht = rte_hash_create(&ht_params);
  if (!ht){
    printf("ERROR: cannot create hash table. The worker %u(port %d queue %d) will be dropped\n",
                                                  rte_lcore_id(), context->port, context->queue);
    free(root);
    return -1;
  }

  while (!atomic_load(&quit_signal)) {

    /* Get burst of RX packets */
    struct rte_mbuf *bufs[BURST_SIZE];
    const uint16_t nb_rx = rte_eth_rx_burst(context->port, context->queue, bufs, BURST_SIZE);

    for (uint16_t buf = 0; buf < nb_rx; buf++) {
      struct rte_mbuf *mbuf = bufs[buf];
      const uint8_t *data = (const uint8_t*)mbuf->buf_addr + mbuf->data_off;

      struct flow_key flw_key;
      memset(&flw_key, 0, sizeof(struct flow_key));

      static int debug_packet_counter = 1;
      if(opt_verbose) {
        printf("Reading packet %d on port - %d, queue - %d\n", debug_packet_counter++, context->port, context->queue);
      }

      struct packet_fields fields;
      if (packet_decoder(data, &fields) != 0) {
        goto done_with_packet;
      }

      if (fields.inner_proto == 4) { // encapsulated IPv4 packet
        flw_key.ipv4_src = fields.match_ipv4_src_inner;
        flw_key.ipv4_dst = fields.match_ipv4_dst_inner;
      } else if (fields.inner_proto == 6) {// encapsulated IPv6 packet
        memcpy(&flw_key.ipv6_src, fields.match_ipv6_src_inner, 16);
        memcpy(&flw_key.ipv6_dst, fields.match_ipv6_dst_inner, 16);
      }

      flw_key.encap_gtp_teid = fields.encap_gtp_teid;
      flw_key.ether_type = fields.ether_type;
      flw_key.port_src = fields.match_udp_src_inner;
      flw_key.port_dst = fields.match_udp_dst_inner;

      if (rte_hash_lookup(ht, &flw_key) >= 0) {
        goto done_with_packet;
      }

      rte_hash_add_key(ht, &flw_key);

      if (opt_verbose) {
        dump_decoded_packet(&fields);
      }

      uint32_t meter_id_arr[MAX_POLICY_PER_FLOW + 1] = {0,0,0}; // + 1 for mbr policy
      int policy_count = get_meter_ids(USE_MTR_COUNT_LOCK, meter_id_arr);

      if (fields.ether_type == RTE_BE16(RTE_ETHER_TYPE_IPV4)) {
        /* Uplink IPv4 - decap */
        tail->flow_uplink = setup_tunnel_IPv4_uplink_flow(context->port,
          fields.match_ipv4_src, fields.match_ipv4_dst,
          fields.match_udp_src, fields.match_udp_dst,
          fields.encap_gtp_teid, dscp_tag,
          NAT_EN,
          N6_IPV4_SUBNET,
          RTE_BE16(NAT_udp_src),
          action_out_port_id,
          FLOW_MATCH_IPV4_GROUP_UPLINK,
          meter_id_arr, policy_count);

        /* Downlink - encap */
        uint32_t ipv4_src_match = NAT_EN ? fields.match_ipv4_dst_inner : fields.match_ipv4_src_inner;
        uint32_t ipv4_dst_match = NAT_EN ? N6_IPV4_SUBNET : fields.match_ipv4_dst_inner;

        tail->flow_downlink = setup_tunnel_IPv4_downlink_flow(action_out_port_id,
          ipv4_src_match, ipv4_dst_match,
          fields.match_udp_src_inner, fields.match_udp_dst_inner,
          fields.match_ipv4_src_inner,
          RTE_BE16(NAT_udp_src),
          fields.encap_gtp_teid, qfi,
          NAT_EN,
          context->port,
          FLOW_MATCH_IPV4_GROUP_DOWNLINK,
          NULL, 0);

      }
      else if (fields.ether_type == RTE_BE16(RTE_ETHER_TYPE_IPV6)) {
        /* Uplink IPv6 - decap */
        tail->flow_uplink = setup_tunnel_IPv6_uplink_flow(context->port,
          fields.match_ipv6_src, fields.match_ipv6_dst,
          fields.match_udp_src, fields.match_udp_dst,
          fields.encap_gtp_teid, dscp_tag,
          NAT_EN,
          RTE_BE16(NAT_udp_src),
          action_out_port_id,
          FLOW_MATCH_IPV6_GROUP_UPLINK,
          meter_id_arr, policy_count);

        /* Downlink - encap */
        uint8_t *ipv6_src_match = NAT_EN ? fields.match_ipv6_dst_inner : fields.match_ipv6_src_inner;
        uint8_t *ipv6_dst_match = NAT_EN ? N6_IPV6_SUBNET : fields.match_ipv6_dst_inner;

        tail->flow_downlink = setup_tunnel_IPv6_downlink_flow(action_out_port_id,
          ipv6_src_match, ipv6_dst_match,
          fields.match_udp_src_inner, fields.match_udp_dst_inner,
          fields.match_ipv6_src_inner,
          RTE_BE16(NAT_udp_src),
          fields.encap_gtp_teid, qfi,
          NAT_EN,
          context->port,
          FLOW_MATCH_IPV6_GROUP_DOWNLINK,
          NULL, 0);
      } else {
        goto done_with_packet;
      }

      tail->next = calloc(1, sizeof(struct rte_flow_list));
      tail = tail->next;

      if (NAT_EN) {
        if ( NAT_udp_src++ == NAT_ports[1] ) {
          NAT_udp_src = NAT_ports[0]; // Reset port counter

          N6_IPV4_SUBNET += RTE_BE32(1); // Change subnet ipv4

          if(++N6_IPV6_SUBNET[ipv6_offset] == 255) // Change subnet ipv6
            ipv6_offset = ipv6_offset == 0 ? 15 : ipv6_offset-1;
        }
    }

done_with_packet:
      rte_pktmbuf_free(bufs[buf]);
    }
  }

  while (root != NULL) {
    if ( root->flow_downlink && rte_flow_destroy(action_out_port_id, root->flow_downlink, &error)) {
    }

    if (root->flow_uplink && rte_flow_destroy(context->port, root->flow_uplink, &error)) {
    }

    struct rte_flow_list *temp = root;
    root = root->next;
    free(temp);
  }

  rte_hash_free(ht);

  return 0;
}

static void lcore_main(void)
{
  printf("Main core %u running. [Ctrl+C to quit]\n", rte_lcore_id());

  while (!atomic_load(&quit_signal)) {
    sleep(1);
    if(stats_mask != 0) {
      print_meter_stat();
    }
  }
}

////////////////////////////////////////////////////////////////////////////////
// Main function
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char *argv[])
{
  int ret;
  struct worker_thread_context_s *worker_contexts = NULL;

  uint16_t nb_ports;
  struct rte_ether_addr port_addr[MAX_PORT_NUMBER];

  /* catch ctrl-c so we can print on exit */
  atomic_store(&quit_signal, 0);
  signal(SIGINT, int_handler);

  /* Initialize the Environment Abstraction Layer (EAL). */
  ret = rte_eal_init(argc, argv);
  if (ret < 0) {
    rte_exit(EXIT_FAILURE, "Error with EAL initialization\n");
  }

  argc -= ret;
  argv += ret;

  for(int p = 0; p < MAX_PORT_NUMBER; p++) {
    struct rte_eth_dev_info dev_info;
    rte_eth_dev_info_get(p, &dev_info);
    rx_rings[p] = dev_info.max_rx_queues;
    tx_rings[p] = dev_info.max_tx_queues;
  }

  if (argc > 1) {
    launch_args_parse(argc, argv);
  }

  if (port_pir_value[PORT_UL] != 0) {
    printf("1 port meter for port %d will be added:\n", PORT_UL);
    printf("\tPIR=%" PRIu64 " bytes/s PBS=%"PRIu64" bytes\n",
           port_pir_value[PORT_UL], port_pbs_value[PORT_UL]);
  }
  if (policy_per_flow) {
    printf("%d flow meter%s for each flow will be added:\n",
           policy_per_flow, (policy_per_flow > 1) ? "s" : "");
    for (int i = 0; i < policy_per_flow; i++) {
      printf("\tPIR=%" PRIu64 " bytes/s PBS=%" PRIu64 " bytes\n",
             flow_meter_policy_PIR[i], flow_meter_policy_PBS[i]);
    }
  }

  /* Retrieve the firmware version of a device */
  {
    char version[256];
    if (rte_eth_dev_fw_version_get(0, version, 256) > 0) {
      printf("Firmware version: %s\n", version);
    }
  }

  /* Get the number of ports which are usable for the application */
  nb_ports = rte_eth_dev_count_avail();

  if (nb_ports == 0) {
    rte_exit(EXIT_FAILURE, "No ports found\n");
  }
  else if (nb_ports > MAX_PORT_NUMBER) {
    rte_exit(EXIT_FAILURE, "Unsupported number of ports found\n");
  }

  /* Initialize all ports. */
  for (uint16_t port_id = 0; port_id < nb_ports; port_id++) {
    if (port_init(port_id, &port_addr[port_id], rx_rings[port_id], tx_rings[port_id]) != 0) {
      rte_exit(EXIT_FAILURE, "Cannot init port %"PRIu8 "\n", port_id);
    }
  }

  /* Check required resources */
  if (rte_lcore_count() < nb_ports) {
    rte_exit(EXIT_FAILURE, "Need atleast %d cores specified", nb_ports);
  }

  /* Setup main filters and templates */
  struct rte_flow_error error;
  memset(&error, 0x0, sizeof(struct rte_flow_error));

  init_GTP_tun_template(N3_IF_IPV4, N3_IPV4_PEER, N3_IF_IPV6, N3_IPV6_PEER);

  setup_meters();

  uint32_t ipv4_src_filt = 0; // Any src IP
  uint32_t ipv4_src_mask = 0; // Any src IP
  uint32_t ipv4_dst_filt = NAT_EN ? N3_IF_IPV4 : 0;
  uint32_t ipv4_dst_mask = NAT_EN ? IF_IPV4_MASK : 0;

  const uint8_t *ipv6_src_filt = IPv6_addr_any;
  const uint8_t *ipv6_src_mask = IPv6_addr_any;
  const uint8_t *ipv6_dst_filt = NAT_EN ? N3_IF_IPV6 : IPv6_addr_any;
  const uint8_t *ipv6_dst_mask = NAT_EN ? IF_IPV6_MASK : IPv6_addr_any;

  /* Uplink IPv4 - decap */
  struct rte_flow *uplink_ipv4_flow = setup_main_uplink_IPv4_filter(PORT_UL,
    ipv4_src_filt, ipv4_src_mask,
    ipv4_dst_filt, ipv4_dst_mask,
    rss_queues_ul, rss_queues_count[PORT_UL], FLOW_MATCH_IPV4_GROUP_UPLINK);

  /* Downlink IPv4 - encap */
  ipv4_dst_filt = NAT_EN ? N6_IF_IPV4_NET : 0;
  ipv4_dst_mask = NAT_EN ? N6_IF_IPV4_SUBNET_MASK : 0;

  struct rte_flow *downlink_flow_ipv4 = setup_main_downlink_IPv4_filter(PORT_DL,
    ipv4_src_filt, ipv4_src_mask,
    ipv4_dst_filt, ipv4_dst_mask,
    FLOW_MATCH_IPV4_GROUP_DOWNLINK);

  /* Uplink IPv6 - decap */
  struct rte_flow *uplink_ipv6_flow = setup_main_uplink_IPv6_filter(PORT_UL,
    ipv6_src_filt, ipv6_src_mask,
    ipv6_dst_filt, ipv6_dst_mask,
    rss_queues_ul, rss_queues_count[PORT_UL], FLOW_MATCH_IPV6_GROUP_UPLINK);

  /* Downlink IPv6 - encap */
  ipv6_dst_filt = NAT_EN ? N6_IF_IPV6 : IPv6_addr_any;
  ipv6_dst_mask = NAT_EN ? IF_IPV6_MASK : IPv6_addr_any;

  struct rte_flow *downlink_flow_ipv6 = setup_main_downlink_IPv6_filter(PORT_DL,
    ipv6_src_filt, ipv6_src_mask,
    ipv6_dst_filt, ipv6_dst_mask,
    FLOW_MATCH_IPV6_GROUP_DOWNLINK);

  green_pkt = calloc(end_mtr_id - start_mtr_id, sizeof(uint64_t));

  if ((*pre_learn_file) != 0) {
    printf("Started flow programming based on %s - please wait\n", pre_learn_file);
    int flow_number = 0;
    uint64_t delta_us = 0;
    struct timespec start, end;

    clock_gettime(CLOCK_MONOTONIC_RAW, &start);
    flow_number = parse_pcap(pre_learn_file, opt_verbose, NAT_EN);
    clock_gettime(CLOCK_MONOTONIC_RAW, &end);
    delta_us = (end.tv_sec - start.tv_sec) * 1000000 + (end.tv_nsec - start.tv_nsec) / 1000;
    printf("Flow programming took %f seconds to execute\n", (double)delta_us/1000000);
    printf("Created %i flows\n", flow_number);
    if (delta_us/1000000 >= 1) {
      printf("Flow program rate: %li flow/second\n", flow_number/(delta_us/1000000));
    }
    else {
      printf("Flow program rate: %i flow/second\n", flow_number/1);
    }
  }

  // Create workers only for RX traffic (separate worker for each queue)
  const int workers_nb = rss_queues_count[PORT_UL];

  /* Initialize worker contexts */
  worker_contexts = calloc(workers_nb, sizeof(struct worker_thread_context_s));

  for (int i = 0; i < workers_nb; ++i) {
    int port  = PORT_UL;
    int queue = rss_queues_ul[i];

    worker_contexts[i].number = i;
    worker_contexts[i].port = port;
    worker_contexts[i].queue = queue;
  }

  /* Launch worker threads. */
  unsigned lcore_id = rte_get_next_lcore(-1, 1, 0);

  for (int i = 0; i < workers_nb; ++i) {
    rte_eal_remote_launch((lcore_function_t *)lcore_worker, &worker_contexts[i], lcore_id);
    lcore_id = rte_get_next_lcore(lcore_id, 1, 0);
  }

  /* Call lcore_main on the master core only. */
  lcore_main();

  /* Wait for worker threads to stop. */
  lcore_id = rte_get_next_lcore(-1, 1, 0);
  for (int i = 0; i < workers_nb; ++i) {
    if (rte_eal_wait_lcore(lcore_id) < 0) {
      return -1;
    }
    lcore_id = rte_get_next_lcore(lcore_id, 1, 0);
  }

  /* Flow tear down */
  if ((*pre_learn_file) != 0) {
    printf("Removing flows programmed based on %s - please wait\n", pre_learn_file);
    remove_pre_learn_flows();
    printf("Flows removed\n");
  }

  /* Filter tear down */
  if (uplink_ipv4_flow && rte_flow_destroy(PORT_UL, uplink_ipv4_flow, &error)) {
    printf("Error: Cannot destroy flow: %s\n", error.message);
  }

  if (uplink_ipv6_flow && rte_flow_destroy(PORT_UL, uplink_ipv6_flow, &error)) {
    printf("Error: Cannot destroy flow: %s\n", error.message);
  }

  if (downlink_flow_ipv4 && rte_flow_destroy(PORT_DL, downlink_flow_ipv4, &error)) {
    printf("Error: Cannot destroy flow: %s\n", error.message);
  }

  if (downlink_flow_ipv6 && rte_flow_destroy(PORT_DL, downlink_flow_ipv6, &error)) {
    printf("Error: Cannot destroy flow: %s\n", error.message);
  }

  destroy_meters();

  free(worker_contexts);
  free(green_pkt);

  sleep(3);

  /* Stop and close down ports. */
  for (uint16_t port_id = 0; port_id < nb_ports; port_id++) {
    port_close(port_id);
  }

  return ret;
}
