/*-
 *   BSD LICENSE
 *
 *   Copyright(c) 2010-2015 Intel Corporation. All rights reserved.
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided with the
 *       distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "port.h"

#include <rte_ethdev.h>
#include <rte_mbuf.h>
#include <rte_mbuf_dyn.h>

#define RX_RING_SIZE 128
#define TX_RING_SIZE 512

#define NUM_MBUFS 8192
#define MBUF_CACHE_SIZE 250

static struct rte_mempool* mbuf_pool = NULL;

int port_init(uint8_t port_id, struct rte_ether_addr *addr, int rx_rings, int tx_rings)
{
  struct rte_eth_conf port_conf;
  int retval;

  uint16_t nb_ports = rte_eth_dev_count_avail();

  memset(&port_conf, 0x0, sizeof(struct rte_eth_conf));

  if (mbuf_pool == NULL) {

    /* Creates a new mempool in memory to hold the mbufs. */
    mbuf_pool = rte_pktmbuf_pool_create(
      "MBUF_POOL",               /* The name of the mbuf pool  */
      NUM_MBUFS * nb_ports,      /* The number of elements in the mbuf pool. The optimum size (in terms of memory usage) for a mempool is when n is a power of two minus one: n = (2^q - 1) */
      MBUF_CACHE_SIZE,           /* Size of the per-core object cache. See rte_mempool_create() for details */
      0,                         /* Size of application private are between the rte_mbuf structure and the data buffer. This value must be aligned to RTE_MBUF_PRIV_ALIGN */
      RTE_MBUF_DEFAULT_BUF_SIZE, /* Size of data buffer in each mbuf, including RTE_PKTMBUF_HEADROOM */
      rte_socket_id());          /* The socket identifier where the memory should be allocated. The value can be SOCKET_ID_ANY if there is no NUMA constraint for the reserved zone */
                                 /* rte_socket_id: Return the ID of the physical socket of the logical core we are running on */

    if (mbuf_pool == NULL) {
      rte_exit(EXIT_FAILURE, "Cannot create mbuf pool\n");
    }
  }

  /* Get the number of ports which are usable for the application */
  if (port_id >= rte_eth_dev_count_avail()) {
    return -1;
  }

  /* Configure the Ethernet device. This function must be invoked first before any other function in the Ethernet API. This function can also be re-invoked when a device is in the stopped state */
  retval = rte_eth_dev_configure(
    port_id,     /* The port identifier of the Ethernet device to configure */
    rx_rings,    /* The number of receive queues to set up for the Ethernet device */
    tx_rings,    /* The number of transmit queues to set up for the Ethernet device */
    &port_conf); /* The pointer to the configuration data to be used for the Ethernet device */

  if (retval != 0) {
    return retval;
  }

  /* Allocate and set up "RX_RINGS" RX queue per Ethernet port. */
  for (int q = 0; q < rx_rings; q++) {
    retval = rte_eth_rx_queue_setup(
      port_id,                        /* The port identifier of the Ethernet device */
      q,                              /* The index of the receive queue to set up. The value must be in the range [0, nb_rx_queue - 1] previously supplied to rte_eth_dev_configure() */
      RX_RING_SIZE,                   /* The number of receive descriptors to allocate for the receive ring */
      rte_eth_dev_socket_id(port_id), /* The socket_id argument is the socket identifier in case of NUMA. The value can be SOCKET_ID_ANY if there is no NUMA constraint for the DMA memory allocated for the receive descriptors of the ring */
      NULL,                           /* The pointer to the configuration data to be used for the receive queue. NULL value is allowed, in which case default Rx configuration will be used */
      mbuf_pool);                     /* The pointer to the memory pool from which to allocate rte_mbuf network memory buffers to populate each descriptor of the receive ring */

    if (retval < 0) {
      return retval;
    }
  }

  /* Allocate and set up "TX_RINGS" TX queue per Ethernet port. */
  for (int q = 0; q < tx_rings; q++) {
    retval = rte_eth_tx_queue_setup(
      port_id,
      q,
      TX_RING_SIZE,
      rte_eth_dev_socket_id(port_id),
      NULL);

    if (retval < 0) {
      return retval;
    }
  }

  /* Start the Ethernet port. */
  retval = rte_eth_dev_start(port_id);
  if (retval < 0) {
    return retval;
  }

  /* Get the port MAC address. */
  if (addr) {
    rte_eth_macaddr_get(port_id, addr);
  }

  rte_eth_promiscuous_enable(port_id);

  return 0;
}

int port_close(uint8_t port_id)
{
  rte_eth_dev_stop(port_id);
  rte_eth_dev_close(port_id);

  return 0;
}
