#include <pcap.h>
#include <rte_flow.h>
#include <unistd.h>
#include <rte_ethdev.h>

#include "pre_learn.h"
#include "filter.h"
#include "port.h"
#include "meter_setup.h"

#define FLM_MTR_PROFILE_SIZE  0x100000

extern const uint32_t N3_IF_IPV4;
extern const uint32_t N3_IPV4_PEER;

extern const uint32_t N6_IF_IPV4_NET;
extern const uint32_t N6_IF_IPV4_SUBNET_MASK;

extern const uint32_t IF_IPV4_MASK;

extern const uint8_t N3_IF_IPV6[16];
extern const uint8_t N3_IPV6_PEER[16];

extern const uint8_t N6_IF_IPV6[16];
extern const uint8_t IF_IPV6_MASK[16];

extern const uint8_t IPv6_addr_any[16];

extern const uint16_t NAT_ports[2];

extern uint64_t stats_mask;

uint32_t N6_IPV4_SUBNET;
uint8_t N6_IPV6_SUBNET[16];
uint16_t NAT_udp_src;

struct rte_flow_pre_learn_list {
  struct rte_flow* flow_downlink;
  struct rte_flow* flow_uplink;
  struct rte_flow_pre_learn_list* next;
};

static int meter_count = START_UNIQUE_METER_ID; // 1 and 2 used for mbr policy
static struct rte_flow_pre_learn_list *root = NULL;
static struct meter_context_s common_policy[MAX_POLICY_PER_FLOW] = {0};
static struct meter_context_s mtr_ctx_unique[MAX_POLICY_PER_FLOW];
static struct meter_context_s mtr_ctx_ul;
static int flow_number = 0;

// arguments
static bool NAT_EN = false;
static uint64_t pir_meter_id[MAX_PORT_NUMBER] = {0};
uint64_t port_pir_value[MAX_PORT_NUMBER] = {0};
uint64_t port_pbs_value[MAX_PORT_NUMBER] = {0};
static pthread_mutex_t meter_guard = PTHREAD_MUTEX_INITIALIZER;
static bool opt_verbose = false;
int policy_per_flow = 0;
bool unique_policy_per_flow_en = false;
uint64_t flow_meter_policy_PIR[MAX_POLICY_PER_FLOW] = {DEFAULT_PIR, DEFAULT_PIR};
uint64_t flow_meter_policy_PBS[MAX_POLICY_PER_FLOW] = {DEFAULT_PBS, DEFAULT_PBS};
int flow_meter_policy = 0;

static int icmpCount = 0;
static int tcpCount = 0;
static int udpCount = 0;

static struct rte_flow_error error = {0};
static uint8_t action_out_port_id = 1;
static uint8_t dscp_tag = DEFAULT_DSCP;
static uint8_t qfi = DEFAULT_QFI;

void dump_decoded_packet(struct packet_fields *fields) {
  if (fields->ether_type == RTE_BE16(RTE_ETHER_TYPE_IPV4)) {
    printf("Learn flow. IPv4 %d.%d.%d.%d\n", ((fields->match_ipv4_src >> 0) & 0xff), ((fields->match_ipv4_src >> 8) & 0xff), ((fields->match_ipv4_src >> 16) & 0xff), ((fields->match_ipv4_src >> 24) & 0xff));
    printf("            IPv4 %d.%d.%d.%d\n", ((fields->match_ipv4_dst >> 0) & 0xff), ((fields->match_ipv4_dst >> 8) & 0xff), ((fields->match_ipv4_dst >> 16) & 0xff), ((fields->match_ipv4_dst >> 24) & 0xff));
    printf("            UDP  %d %d\n", (int)RTE_BE16(fields->match_udp_src), (int)RTE_BE16(fields->match_udp_dst));
    printf("            NAT IPv4 %d.%d.%d.%d\n", ((N6_IPV4_SUBNET >> 0) & 0xff), ((N6_IPV4_SUBNET >> 8) & 0xff), ((N6_IPV4_SUBNET>> 16) & 0xff), ((N6_IPV4_SUBNET >> 24) & 0xff));
    printf("            NAT UDP port  %d\n", (int)RTE_BE16(NAT_udp_src));
  } else if (fields->ether_type == RTE_BE16(RTE_ETHER_TYPE_IPV6)) {
    printf("Learn flow. IPv6 %02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x\n",
           fields->match_ipv6_src[0], fields->match_ipv6_src[1], fields->match_ipv6_src[2], fields->match_ipv6_src[3],
           fields->match_ipv6_src[4], fields->match_ipv6_src[5], fields->match_ipv6_src[6], fields->match_ipv6_src[7],
           fields->match_ipv6_src[8], fields->match_ipv6_src[9], fields->match_ipv6_src[10], fields->match_ipv6_src[11],
           fields->match_ipv6_src[12], fields->match_ipv6_src[13], fields->match_ipv6_src[14], fields->match_ipv6_src[15]);
    printf("IPv6 %02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x\n",
           fields->match_ipv6_dst[0], fields->match_ipv6_dst[1], fields->match_ipv6_dst[2], fields->match_ipv6_dst[3],
           fields->match_ipv6_dst[4], fields->match_ipv6_dst[5], fields->match_ipv6_dst[6], fields->match_ipv6_dst[7],
           fields->match_ipv6_dst[8], fields->match_ipv6_dst[9], fields->match_ipv6_dst[10], fields->match_ipv6_dst[11],
           fields->match_ipv6_dst[12], fields->match_ipv6_dst[13], fields->match_ipv6_dst[14], fields->match_ipv6_dst[15]);
  }
  printf("Inner UDP  %d %d\n", (int)RTE_BE16(fields->match_udp_src_inner), (int)RTE_BE16(fields->match_udp_dst_inner));
  printf("TEID 0x%08x\n", RTE_BE32(fields->encap_gtp_teid));
  if (fields->inner_proto == 4) {
    printf("Inner IPv4 %d.%d.%d.%d\n", ((fields->match_ipv4_src_inner >> 0) & 0xff), ((fields->match_ipv4_src_inner >> 8) & 0xff), ((fields->match_ipv4_src_inner >> 16) & 0xff), ((fields->match_ipv4_src_inner >> 24) & 0xff));
    printf("Inner IPv4 %d.%d.%d.%d\n", ((fields->match_ipv4_dst_inner >> 0) & 0xff), ((fields->match_ipv4_dst_inner >> 8) & 0xff), ((fields->match_ipv4_dst_inner >> 16) & 0xff), ((fields->match_ipv4_dst_inner >> 24) & 0xff));
    printf("Inner UDP  %d %d\n", (int)RTE_BE16(fields->match_udp_src_inner), (int)RTE_BE16(fields->match_udp_dst_inner));
  } else if (fields->inner_proto == 6) {
    printf("Inner IPv6 %02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x\n",
           fields->match_ipv6_src_inner[0], fields->match_ipv6_src_inner[1], fields->match_ipv6_src_inner[2], fields->match_ipv6_src_inner[3],
           fields->match_ipv6_src_inner[4], fields->match_ipv6_src_inner[5], fields->match_ipv6_src_inner[6], fields->match_ipv6_src_inner[7],
           fields->match_ipv6_src_inner[8], fields->match_ipv6_src_inner[9], fields->match_ipv6_src_inner[10], fields->match_ipv6_src_inner[11],
           fields->match_ipv6_src_inner[12], fields->match_ipv6_src_inner[13], fields->match_ipv6_src_inner[14], fields->match_ipv6_src_inner[15]);
    printf("Inner IPv6 %02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x\n",
           fields->match_ipv6_dst_inner[0], fields->match_ipv6_dst_inner[1], fields->match_ipv6_dst_inner[2], fields->match_ipv6_dst_inner[3],
           fields->match_ipv6_dst_inner[4], fields->match_ipv6_dst_inner[5], fields->match_ipv6_dst_inner[6], fields->match_ipv6_dst_inner[7],
           fields->match_ipv6_dst_inner[8], fields->match_ipv6_dst_inner[9], fields->match_ipv6_dst_inner[10], fields->match_ipv6_dst_inner[11],
           fields->match_ipv6_dst_inner[12], fields->match_ipv6_dst_inner[13], fields->match_ipv6_dst_inner[14], fields->match_ipv6_dst_inner[15]);
  }
  printf("Inner UDP  %d %d\n", (int)RTE_BE16(fields->match_udp_src_inner), (int)RTE_BE16(fields->match_udp_dst_inner));
}

int packet_decoder(const uint8_t *data, struct packet_fields *fields) {
  memset(fields, 0, sizeof *fields);  // 

  // L2 - Ethernet frame
  data += 12;
  fields->ether_type = ((const uint16_t *)data)[0];
  while (fields->ether_type == RTE_BE16(RTE_ETHER_TYPE_VLAN) ||
         fields->ether_type == RTE_BE16(RTE_ETHER_TYPE_QINQ)) {
    data += 4;
    fields->ether_type = ((const uint16_t *)data)[0];
  }
  data += 2;

  // L3
  if (fields->ether_type == RTE_BE16(RTE_ETHER_TYPE_IPV4)) {
    fields->protocol = data[9]; // offset to protocol field
    fields->match_ipv4_src = ((const uint32_t *)data)[3];
    fields->match_ipv4_dst = ((const uint32_t *)data)[4];
    data += 20;
  } else if (fields->ether_type == RTE_BE16(RTE_ETHER_TYPE_IPV6)) {
    fields->protocol = data[6];
    memcpy(fields->match_ipv6_src, data+8, 16);
    memcpy(fields->match_ipv6_dst, data+24, 16);
    data += 40;
  } else {
    return -1;  // Unsupported packet
  }

  // L4
  if (fields->protocol != 0x11) {  // UDP
    return -1;  // Unsupported packet
  }
  udpCount++;

  fields->match_udp_src = ((const uint16_t *)data)[0];
  fields->match_udp_dst = ((const uint16_t *)data)[1];
  if (RTE_BE16(fields->match_udp_dst) != 2152) { // UDP PORT 2152 is GTP TRAFFIC
    return -1;
  }
  data += 8;

  fields->encap_gtp_teid = ((const uint32_t *)data)[1];
  data += 16;

  fields->inner_proto = (*data & 0xf0) >> 4;

  if (fields->inner_proto == 4) { // encapsulated IPv4 packet
    // L3 inner
    fields->protocol = data[9];
    fields->match_ipv4_src_inner = ((const uint32_t *)data)[3];
    fields->match_ipv4_dst_inner = ((const uint32_t *)data)[4];
    data += 20;
  } else if (fields->inner_proto == 6) {// encapsulated IPv6 packet
    // L3 inner
    fields->protocol = data[6];
    memcpy(fields->match_ipv6_src_inner, data+8, 16);
    memcpy(fields->match_ipv6_dst_inner, data+24, 16);
    data += 40;
  }

  // L4 inner
  if (fields->protocol != 0x11) {  // UDP
    return -1;  // Unsupported packet
  }

  fields->match_udp_src_inner = ((const uint16_t *)data)[0];
  fields->match_udp_dst_inner = ((const uint16_t *)data)[1];

  return 0;
}

int get_meter_ids(int use_lock, uint32_t *meter_id_arr) {
  int policy_count = 0;

  // Check for port-meter
  if (port_pir_value[PORT_UL] != 0) {
    meter_id_arr[policy_count] = pir_meter_id[PORT_UL];
    policy_count++;
  }

  if (policy_per_flow && unique_policy_per_flow_en) {
    if (use_lock) pthread_mutex_lock(&meter_guard);
    for (int i = 0; i < policy_per_flow; i++) {
      meter_learn(&mtr_ctx_unique[i], meter_count, stats_mask);
      meter_id_arr[policy_count] = meter_count;
      policy_count++;
      meter_count++;
    }
    if (use_lock) pthread_mutex_unlock(&meter_guard);
  }
  else if (policy_per_flow && !unique_policy_per_flow_en) { // apply common policy for flow
    for (int i = 0; i < policy_per_flow; i++) {
        meter_id_arr[policy_count] = common_policy[i].id;
        policy_count++;
    }
  }
  return policy_count;
}

////////////////////////////////////////////////////////////////////////////////
// Pcap file parse handler
////////////////////////////////////////////////////////////////////////////////

static void packetHandler(u_char *userData _unused, const struct pcap_pkthdr* pkthdr _unused, const u_char* packet) {

  const uint8_t *data = (const uint8_t*)packet;
  struct rte_flow_pre_learn_list *rule = NULL;
  static int debug_packet_counter = 1;
  static uint8_t ipv6_offset = 15;

  if (opt_verbose) {
    printf("Reading packet %d\n", debug_packet_counter++);
  }

  struct packet_fields fields;
  if (packet_decoder(data, &fields) != 0) {
    return;
  }

  if (opt_verbose) {
    dump_decoded_packet(&fields);
  }

  rule = calloc(1, sizeof(struct rte_flow_pre_learn_list));
  if (!rule) {
    return;
  }

  uint32_t meter_id_arr[MAX_POLICY_PER_FLOW + 1] = {0,0,0}; // + 1 for mbr policy
  int policy_count = get_meter_ids(DONT_USE_MTR_COUNT_LOCK, meter_id_arr);

  if (fields.ether_type == RTE_BE16(RTE_ETHER_TYPE_IPV4)) {
    /* Uplink IPv4 - decap */
    rule->flow_uplink = setup_tunnel_IPv4_uplink_flow(PORT_UL,
          fields.match_ipv4_src, fields.match_ipv4_dst,
          fields.match_udp_src, fields.match_udp_dst,
          fields.encap_gtp_teid, dscp_tag,
          NAT_EN,
          N6_IPV4_SUBNET,
          RTE_BE16(NAT_udp_src),
          action_out_port_id,
          FLOW_MATCH_IPV4_GROUP_UPLINK,
          meter_id_arr, policy_count);


    /* Downlink - encap */
    uint32_t ipv4_src_match = NAT_EN ? fields.match_ipv4_dst_inner : fields.match_ipv4_src_inner;
    uint32_t ipv4_dst_match = NAT_EN ? N6_IPV4_SUBNET : fields.match_ipv4_dst_inner;

    rule->flow_downlink = setup_tunnel_IPv4_downlink_flow(action_out_port_id,
          ipv4_src_match, ipv4_dst_match,
          fields.match_udp_src_inner, fields.match_udp_dst_inner,
          fields.match_ipv4_src_inner,
          RTE_BE16(NAT_udp_src),
          fields.encap_gtp_teid, qfi, NAT_EN,
          PORT_UL,
          FLOW_MATCH_IPV4_GROUP_DOWNLINK,
          NULL, 0);

    if (root == NULL) {
      root = rule;
    }
    else {
      rule->next = root;
      root = rule;
    }

    if (rule->flow_downlink) {
      flow_number++;
    }
    if (rule->flow_uplink) {
      flow_number++;
    }

  }
  else if (fields.ether_type == RTE_BE16(RTE_ETHER_TYPE_IPV6)) {
    /* Uplink IPv6 - decap */
    rule->flow_uplink = setup_tunnel_IPv6_uplink_flow(PORT_UL,
      fields.match_ipv6_src, fields.match_ipv6_dst,
      fields.match_udp_src, fields.match_udp_dst,
      fields.encap_gtp_teid, dscp_tag,
      NAT_EN,
      RTE_BE16(NAT_udp_src),
      action_out_port_id,
      FLOW_MATCH_IPV6_GROUP_UPLINK,
      meter_id_arr, policy_count);

    /* Downlink - encap */
    uint8_t *ipv6_src_match = NAT_EN ? fields.match_ipv6_dst_inner : fields.match_ipv6_src_inner;
    uint8_t *ipv6_dst_match = NAT_EN ? N6_IPV6_SUBNET : fields.match_ipv6_dst_inner;

    rule->flow_downlink = setup_tunnel_IPv6_downlink_flow(action_out_port_id,
      ipv6_src_match, ipv6_dst_match,
      fields.match_udp_src_inner, fields.match_udp_dst_inner,
      fields.match_ipv6_src_inner,
      RTE_BE16(NAT_udp_src),
      fields.encap_gtp_teid, qfi,
      NAT_EN,
      PORT_UL,
      FLOW_MATCH_IPV6_GROUP_DOWNLINK,
      NULL, 0);

    if (root == NULL) {
      root = rule;
    }
    else {
      rule->next = root;
      root = rule;
    }

    if (rule->flow_downlink) {
      flow_number++;
    }
    if (rule->flow_uplink) {
      flow_number++;
    }
  } else {
    free(rule);
  }

  if (NAT_EN) {
    if ( NAT_udp_src++ == NAT_ports[1] ) {
      NAT_udp_src =  NAT_ports[0]; // Reset port counter

      N6_IPV4_SUBNET += RTE_BE32(1); // Change subnet ipv4

      if(++N6_IPV6_SUBNET[ipv6_offset] == 255) // Change subnet ipv6
        ipv6_offset = ipv6_offset == 0 ? 15 : ipv6_offset-1;
    }
  }
}

////////////////////////////////////////////////////////////////////////////////
// Open pcap and parse pcap file
////////////////////////////////////////////////////////////////////////////////

int parse_pcap(char *file_name, bool verbose, bool do_nat) {
  pcap_t *fp;
  char errbuf[PCAP_ERRBUF_SIZE];

  opt_verbose = verbose;
  NAT_EN = do_nat;

  N6_IPV4_SUBNET = N6_IF_IPV4_NET;
  memcpy(N6_IPV6_SUBNET, N6_IF_IPV6, 16);
  NAT_udp_src = NAT_ports[0];

  //open file and create pcap handler
  fp = pcap_open_offline(file_name, errbuf);

  if (fp == NULL) {
    fprintf(stderr, "pcap_open_offline() failed: %s\n", errbuf);
    return 0;
  }
  else {
    printf("Start Pre Learning, file: %s\n", file_name);
  }

  if (pcap_loop(fp, 0, packetHandler, NULL) < 0) {
    fprintf(stderr, "pcap_loop() failed: %s\n", pcap_geterr(fp));
    return 0;
  }

  if (opt_verbose) {
    printf("Protocol Summary: %d ICMP packets, %d TCP packets, %d UDP packets\n", icmpCount, tcpCount, udpCount);
  }

  return flow_number;

}

int remove_pre_learn_flows(void) {
  struct rte_flow_pre_learn_list *temp = NULL;

  // Remove flow rules
  while (root != NULL) {

    if (root->flow_downlink) {
      rte_flow_destroy(action_out_port_id, root->flow_downlink, &error);
    }

    if (root->flow_uplink) {
      rte_flow_destroy(PORT_UL, root->flow_uplink, &error);
    }

    temp = root;
    root = root->next;
    free(temp);
  }

  return 0;
}

void setup_meters(void) {
  if (port_pir_value[PORT_UL]) {
    meter_handle_create(&mtr_ctx_ul,
                        PORT_UL,
                        MTR_ID_UL,
                        port_pir_value[PORT_UL],
                        port_pbs_value[PORT_UL]);
    meter_learn(&mtr_ctx_ul, MTR_ID_UL, stats_mask);
    pir_meter_id[0] = MTR_ID_UL;
  }

  // create common policy for all flows
  if (policy_per_flow && !unique_policy_per_flow_en) {
    for (int i = 0; i < policy_per_flow; i++) {
      meter_handle_create(&common_policy[i], PORT_UL, MTR_ID_FIRST_COMMON + i,
                          flow_meter_policy_PIR[i],
                          flow_meter_policy_PBS[i]);
      meter_learn(&common_policy[i], MTR_ID_FIRST_COMMON + i, stats_mask);
      common_policy[i].id = MTR_ID_FIRST_COMMON + i;
    }
  }

  // create profile and policy for uniques MTRs
  for (int i = 0; i < policy_per_flow; i++) {
    meter_handle_create(&mtr_ctx_unique[i],
                PORT_UL,                    /* port*/
                START_UNIQUE_METER_ID,      /* profile_id */
                flow_meter_policy_PIR[i],
                flow_meter_policy_PBS[i]);
  }

}

void destroy_meters(void) {
  if (policy_per_flow && unique_policy_per_flow_en) {
    for (int i = (START_UNIQUE_METER_ID); i < meter_count; i++) {
      meter_unlearn(&mtr_ctx_unique[i%policy_per_flow], i);
    }

    for (int i = 0; i < flow_meter_policy; i++)
    {
      meter_handle_destroy(&mtr_ctx_unique[i]);
    }
  }

  if (port_pir_value[PORT_UL]) {
    meter_unlearn(&mtr_ctx_ul, MTR_ID_UL);
    meter_handle_destroy(&mtr_ctx_ul);
  }

  if (policy_per_flow && !unique_policy_per_flow_en) {
    for (int i = 0; i < policy_per_flow; i++) {
      meter_unlearn(&common_policy[i], common_policy[i].id);
      meter_handle_destroy(&common_policy[i]);
    }
  }
}

