//
// monitoring.c
//

/*
 * %NT_SOFTWARE_LICENSE%
 */
#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <sys/time.h>
#include "utils.h"
#include "monitoring.h"

static struct mon_data_s mon;

static uint8_t quit_signal;

static void int_handler(int sig_num _unused)
{
  quit_signal = 1;
}



static int __update_data(void)
{
  uint8_t *data;
  int len;
  static struct timeval last;
  struct timeval now;
  struct reply_err_s err;

  gettimeofday(&now, NULL);
  if (now.tv_sec* 1000000 + now.tv_usec - (last.tv_sec * 1000000 + last.tv_usec) < 500000)
    return 0;
  last = now;

  int fd = ntconn_client_connect();

  if (fd < 0) {
    return -1;
  }

  do {
    if ((len = ntconn_client_request(fd, mon.nics->nic_pci_id[mon.sel], "adapter", "get,interfaces", &data, &err)) < 0) goto err_out;
    if (len > 0) {
      if (mon.ifs) {
        free(mon.ifs);
        mon.ifs = NULL;
      }
      mon.ifs = (struct ntc_interfaces_s *)data;

      if (mon.ifs->final_list == 0) {
        // not all interfaces are finally configured
        sleep(1);
      } else
        break;

    } else
      break;

  } while (1);

  if ((len = ntconn_client_request(fd, mon.nics->nic_pci_id[mon.sel], "stat", "get,layout_version", &data, NULL)) < 0) goto err_out;
  if (len) {
    if (len != 4) {
      free(data);
      goto err_out;
    }
    mon.stat_layout_version = *(int *)data;
    free(data);
  } else
    goto err_out;


  if ((len = ntconn_client_request(fd, mon.nics->nic_pci_id[mon.sel], "stat", "snapshot", &data, NULL)) < 0) goto err_out;
  if (len) {
    free(data);
    goto err_out;
  }

  if ((len = ntconn_client_request(fd, mon.nics->nic_pci_id[mon.sel], "stat", "get,flm", &data, NULL)) < 0) goto err_out;
  if (len > 0) {
    if (mon.flm) {
      free(mon.flm);
      mon.flm = NULL;
    }
    mon.flm = (struct ntc_stat_get_data_s *)data;
  }

  if ((len = ntconn_client_request(fd, mon.nics->nic_pci_id[mon.sel], "stat", "get,snapshot,colors", &data, NULL)) < 0) goto err_out;
  if (len > 0) {
    if (mon.clr) {
      free(mon.clr);
      mon.clr = NULL;
    }
    mon.clr = (struct ntc_stat_get_data_s *)data;
  }

  if ((len = ntconn_client_request(fd, mon.nics->nic_pci_id[mon.sel], "stat", "get,snapshot,queues", &data, NULL)) < 0) goto err_out;
  if (len > 0) {
    if (!mon.que) {
      mon.que = (struct ntc_stat_get_data_s *)data;
    } else {
      /*
       * Need to keep last timestamp set of queue data to calculate Tx throughput on virt port
       */
      if (mon.que->timestamp != ((struct ntc_stat_get_data_s *)data)->timestamp) {
       // update last
        if (mon.last_que) {
          free(mon.last_que);
        }
        mon.last_que = mon.que;
      } else {
        free(mon.que);
      }
      mon.que = (struct ntc_stat_get_data_s *)data;
    }
  }

  if ((len = ntconn_client_request(fd, mon.nics->nic_pci_id[mon.sel], "stat", "get,snapshot,rx_counters", &data, NULL)) < 0) goto err_out;
  if (len > 0) {

    if (!mon.rxc) {
      mon.rxc = (struct ntc_stat_get_data_s *)data;
    } else {
      /*
       * Need to keep last timestamp set of queue data to calculate Rx throughput on phys/virt port
       */
      if (mon.rxc->timestamp != ((struct ntc_stat_get_data_s *)data)->timestamp) {
       // update last
        if (mon.last_rxc)
          free(mon.last_rxc);
        mon.last_rxc = mon.rxc;
      } else {
        free(mon.rxc);
      }
      mon.rxc = (struct ntc_stat_get_data_s *)data;
    }
  }

  if ((len = ntconn_client_request(fd, mon.nics->nic_pci_id[mon.sel], "stat", "get,snapshot,tx_counters", &data, NULL)) < 0) goto err_out;
  if (len > 0) {
    if (!mon.txc) {
      mon.txc = (struct ntc_stat_get_data_s *)data;
    } else {
      /*
       * Need to keep last timestamp set of queue data to calculate Tx throughput on phys port
       */
      if (mon.txc->timestamp != ((struct ntc_stat_get_data_s *)data)->timestamp) {
       // update last
        if (mon.last_txc)
          free(mon.last_txc);
        mon.last_txc = mon.txc;
      } else {
        free(mon.txc);
      }
      mon.txc = (struct ntc_stat_get_data_s *)data;
    }
  }

  close(fd);
  return 0;

err_out:
  close(fd);
  return -1;
}


static int run_main(void)
{
  while (quit_signal == 0) {
    int failed = __update_data();
    display(failed);
  }
  deinit_display();
  return 0;
}



int main(void)
{
  int len;
  uint8_t *data;
  int fd = ntconn_client_connect();

  if (fd < 0) {
    fprintf(stderr, "ERROR while connecting to %s\n", NTCONNECT_SOCKET);
    perror(NULL);
    return 0;
  }

  mon.server_version = ntconn_client_get_version(fd, "0000:00:00.0", "server");
  if (mon.server_version != 1ULL) {
    fprintf(stderr, "Server version %u.%u not supported\n", (unsigned)(mon.server_version >> 32), (unsigned)(mon.server_version & 0xffffffff));
    close(fd);
    return 0;
  }

  if ((len = ntconn_client_request(fd, "0000:00:00.0", "server", "get,nic_pci_ids", &data, NULL)) < 0) return 0;
  if (len) {
    mon.nics = (struct ntc_nic_pci_ids_s *)data;
  } else {
    fprintf(stderr, "No modules registered on server\n");
    return 0;
  }

  if (mon.nics->num_nics <= 0) {
    fprintf(stderr, "No SmartNICs found\n");
    return 0;
  }

  mon.sel = 0; // initially select first NIC found

  mon.mod_stat_version = ntconn_client_get_version(fd, mon.nics->nic_pci_id[mon.sel], "stat");
  if (mon.mod_stat_version != 1ULL) {
    fprintf(stderr, "Server version %u.%u not supported\n", (unsigned)(mon.mod_stat_version >> 32), (unsigned)(mon.mod_stat_version & 0xffffffff));
    close(fd);
    return 0;
  }

  mon.mod_adap_version = ntconn_client_get_version(fd, mon.nics->nic_pci_id[mon.sel], "adapter");
  if (mon.mod_adap_version != 1ULL) {
    fprintf(stderr, "Server version %u.%u not supported\n", (unsigned)(mon.mod_adap_version >> 32), (unsigned)(mon.mod_adap_version & 0xffffffff));
    close(fd);
    return 0;
  }

  if ((len = ntconn_client_request(fd, mon.nics->nic_pci_id[mon.sel], "adapter", "get,info", &data, NULL)) < 0) return 0;
  if (len > 0) {
    mon.info = (struct ntc_adap_get_info_s *)data;
  }
  close(fd);

  if (len <= 0) {
    fprintf(stderr, "Getting adapter info failed\n");
    return 0;
  }

  signal(SIGINT, int_handler);

  __update_data(); // first time update and read data
  init_display(&mon);
  run_main();

  if (mon.nics)
    free(mon.nics);
  if (mon.info)
    free(mon.info);
  if (mon.ifs)
    free(mon.ifs);
  if (mon.clr)
    free(mon.clr);
  if (mon.que)
    free(mon.que);
  if (mon.last_que)
    free(mon.last_que);
  if (mon.rxc)
    free(mon.rxc);
  if (mon.last_rxc)
    free(mon.last_rxc);
  if (mon.txc)
    free(mon.txc);
  if (mon.last_txc)
    free(mon.last_txc);

  return 0;
}
