#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>

#include "ntconn_client.h"
#include "ntconnect_api_test.h"
#include "utils.h"

#define event_count (1)

static struct test_data_s {
  uint64_t server_version;
  uint64_t mod_test_version;
  struct ntc_nic_pci_ids_s *nics;
  int sel;
} test_data;

int test_init_data(void)
{
  int len;
  uint8_t *data;

  int fd = ntconn_client_connect();
  if (fd < 0) {
    printf("ERROR while connecting to %s\n", NTCONNECT_SOCKET);
    return -1;
  }


  test_data.server_version = ntconn_client_get_version(fd, "0000:00:00.0", "server");
  if (test_data.server_version != 1ULL) {
    printf("Server version %u.%u not supported\n", (unsigned)(test_data.server_version >> 32), (unsigned)(test_data.server_version & 0xffffffff));
    close(fd);
    return -1;
  }
  printf("Server version %u.%u\n", (unsigned)(test_data.server_version >> 32), (unsigned)(test_data.server_version & 0xffffffff));

  if ((len = ntconn_client_request(fd, "0000:00:00.0", "server", "get,nic_pci_ids", &data, NULL)) < 0) return 0;
  if (len) {
    test_data.nics = (struct ntc_nic_pci_ids_s *)data;
  } else {
    printf("No modules registered on server\n");
    close(fd);
    return -1;
  }

  if (test_data.nics->num_nics <= 0) {
    printf("No SmartNICs found\n");
    close(fd);
    return -1;
  }

  {
    int i;
    for (i = 0; i <   test_data.nics->num_nics; i++) {
      printf("PCI ID: %s\n", test_data.nics->nic_pci_id[i]);
    }
  }

  test_data.sel = 0; // initially select first NIC found

  test_data.mod_test_version = ntconn_client_get_version(fd, test_data.nics->nic_pci_id[test_data.sel], "filter");
  if (test_data.mod_test_version != 1ULL) {
    printf("Test version %u.%u not supported\n", (unsigned)(test_data.mod_test_version >> 32), (unsigned)(test_data.mod_test_version & 0xffffffff));
    close(fd);
    return -1;
  }
  printf("Test version %u.%u\n", (unsigned)(test_data.mod_test_version >> 32), (unsigned)(test_data.mod_test_version & 0xffffffff));

  close(fd);
  return 0;
}

void test_free_data(void)
{
  if (test_data.nics) {
    free(test_data.nics);
    test_data.nics = NULL;
  }
}

static int ntconnect_fd = -1;

int ntconnect_test(int number, int test_size, int fd_close)
{
  static struct reply_err_s reply_error;
  int len;
  uint8_t *data = NULL;

  if (ntconnect_fd == -1) {
    ntconnect_fd = ntconn_client_connect();
    if (ntconnect_fd < 0) {
      fprintf(stderr, "\nntconnection failed\n\n");
      return -1;
    }
  }

  struct test_s *test = malloc(test_size);
  if (!test) {
    fprintf(stderr, "\nFailed to allocate test structure\n\n");
    return -1;
  }

  test->number = number;
  int count = number;
  for (int i = 0; i < number; i++) {
    test->test[i] = count++;
  }

  if ((len = ntconn_client_request_with_data(ntconnect_fd, test_data.nics->nic_pci_id[test_data.sel], "ntconnect_test", "test", &data, &reply_error, (void *)test, test_size)) < 0) {
    fprintf(stderr, "\nError in communication: %s\n\n", reply_error.msg);
    free(test);
    return -1;
  }

  if (len == test_size) {
    struct test_s *return_data = (struct test_s *)data;
    for (int i = 0; i < number; i++) {
      if (return_data->test[i] != test->test[i]) {
        fprintf(stderr, "\nError in returned data pos %u. Was %llu, expected %llu\n\n", i, (long long unsigned int )return_data->test[i], (long long unsigned int )test->test[i]);
      }
    }
    free(test);
    free(data);
    if (fd_close) {
      close(ntconnect_fd);
      ntconnect_fd = -1;
    }
    return 0;
  } else {
    free(test);
    free(data);
    fprintf(stderr, "\nIllegal size returned\n\n");
    if (fd_close) {
      close(ntconnect_fd);
      ntconnect_fd = -1;
    }
    return -1;
  }
}

static __inline__ unsigned long long rdtsc(void)
{
    unsigned hi, lo;
    __asm__ __volatile__ ("rdtsc" : "=a"(lo), "=d"(hi));
    return ( (unsigned long long)lo)|( ((unsigned long long)hi)<<32 );
}

int main(int argc, char const *argv[])
{
  long long unsigned int tsc1 = 0;
  long long unsigned int tsc2 = 0;
  long long unsigned int counter = 0;
  long long unsigned int diff = 0;
  int from;
  int to;
  int times;

  if (test_init_data() < 0) {
    return -1;
  }

  for (int t = 0; t < 2; t++) {
    if (t == 0) {
      times = 100;
    }
    else {
      times = 10;
    }

    from = 8000;
    to = 1;
    printf("Test from %u to %u - %u times with %s socket IF\r", from, to, times, t == 0 ? "non closed" : "closed"); fflush(stdout);
    counter = 0;
    diff = 0;
    for (int j = 0; j < times; j++) {
      for (int i = from; i >= to; i--) {
        int test_size = sizeof(struct test_s) + sizeof(uint64_t) * i;
        tsc1 = rdtsc();
        if (ntconnect_test(i, test_size, t) < 0) {
          return -1;
        }
        tsc2 = rdtsc();
        counter++;
        diff += tsc2-tsc1;
      }
      printf("Test from %u to %u - %u times with %s socket IF - No %u\r", from, to, times, t == 0 ? "non closed" : "closed", j); fflush(stdout);
    }
    printf("Test from %u to %u - %u times with %s socket IF - Time %16llu                                   \n", from, to, times, t == 0 ? "non closed" : "closed", diff/counter);

    from = 1;
    to = 8000;
    printf("Test from %u to %u - %u times with %s socket IF\r", from, to, times, t == 0 ? "non closed" : "closed"); fflush(stdout);
    counter = 0;
    diff = 0;

    for (int j = 0; j < times; j++) {
      for (int i = from; i <= to; i++) {
        int test_size = sizeof(struct test_s) + sizeof(uint64_t) * i;

        tsc1 = rdtsc();
        if (ntconnect_test(i, test_size, t) < 0) {
          return -1;
        }
        tsc2 = rdtsc();
        counter++;
        diff += tsc2-tsc1;
      }
      printf("Test from %u to %u - %u times with %s socket IF - No %u\r", from, to, times, t == 0 ? "non closed" : "closed", j); fflush(stdout);
    }

    printf("Test from %u to %u - %u times with %s socket IF - Time %16llu                                   \n", from, to, times, t == 0 ? "non closed" : "closed", diff/counter);

    from = 8000;
    to = 1;
    printf("Test random from %u to %u - %u times with %s socket IF\r", from, to, times, t == 0 ? "non closed" : "closed"); fflush(stdout);
    counter = 0;
    diff = 0;
    for (int j = 0; j < times; j++) {
      for (int i = from; i >= to; i--) {
        int r = (i * 100) % 8000 + 1;
        int test_size = sizeof(struct test_s) + sizeof(uint64_t) * r;
        tsc1 = rdtsc();
        if (ntconnect_test(r, test_size, t) < 0) {
          return -1;
        }
        tsc2 = rdtsc();
        counter++;
        diff += tsc2-tsc1;
      }
      printf("Test random from %u to %u - %u times with %s socket IF - No %u\r", from, to, times, t == 0 ? "non closed" : "closed", j); fflush(stdout);
    }
    printf("Test random from %u to %u - %u times with %s socket IF - Time %16llu                                   \n", from, to, times, t == 0 ? "non closed" : "closed", diff/counter);
  }
  printf("\n");
  return 0;
}
