#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "sensors.h"
#include "argparse.h"

/**
 * Sensor types
 */
enum NtSensorType_e {
  NT_SENSOR_TYPE_UNKNOWN      = 0,
  NT_SENSOR_TYPE_TEMPERATURE  = 1, //!< Unit: 0.1 degree Celsius
  NT_SENSOR_TYPE_VOLTAGE      = 2, //!< Unit: 1 mV
  NT_SENSOR_TYPE_CURRENT      = 3, //!< Unit: 1 uA
  NT_SENSOR_TYPE_POWER        = 4, //!< Unit: 0.1 uW
  NT_SENSOR_TYPE_FAN          = 5, //!< Unit: 1 RPM (Revolutions Per Minute)
  NT_SENSOR_TYPE_HIGH_POWER   = 6, //!< Unit: 1 mW
  NT_SENSOR_TYPE_NUMBER       = 7,
};

/**
 * Sensor value
 */
#define NT_SENSOR_NAN (int)0x80000000 //!< Indicates that sensor value or sensor limit is not valid (Not a Number)

typedef enum NtSensorType_e NtSensorType_t;

static int adapter = 0;
static int port = -1;
static const char SensorFmtStr[] = "%-13s %-7s %-6s %7s %7s %7s\n";

struct argparse_option options[] = {
  OPT_HELP(),
  OPT_INTEGER('a', "adapter", &adapter, "Specifies the adapter that the tool is to operate on", NULL, 0, 0),
  OPT_INTEGER('p', "port", &port, "Specifies the port that the tool is to operate on", NULL, 0, 0),
  OPT_END(),
};

static const char *usage[] = {
  "The Sensors tool commands used to get sensors values\n"
  "\n"
  "Syntax:\n"
  "---------------------\n"
  "\n"
  "sensors [-a adapter] [-p port] [-h]\n"
  "\n",
  NULL,
};

static const char *usageExamples =
  "\nImplemented commands:\n"
  "* sensors -p 0\n"
  "\n"
  "* sensors -p 1\n"
  "\n"
  "* sensors -a 0 (there is no more than 1 adapter yet)\n"
  "\n";

void DisplaySeparator1(void)
{
  // Print separator line
  for (int i = 0; i < 78; ++i) {
    putchar('-');
  }
  printf("\n");
}

const char* NT_GetSensorUnit(NtSensorType_t Type)
{
  switch (Type)
  {
  case NT_SENSOR_TYPE_UNKNOWN      : return "?";
  case NT_SENSOR_TYPE_TEMPERATURE  : return "C";
  case NT_SENSOR_TYPE_VOLTAGE      : return "V";
  case NT_SENSOR_TYPE_CURRENT      : return "A";
  case NT_SENSOR_TYPE_POWER        : return "W";
  case NT_SENSOR_TYPE_FAN          : return "RPM";
  case NT_SENSOR_TYPE_HIGH_POWER   : return "W";
  default:
    return "Unhandled";
  }
}

const char* NT_TranslateSensorType(NtSensorType_t Type)
{
  switch (Type)
  {
  case NT_SENSOR_TYPE_UNKNOWN      : return "Unknown";
  case NT_SENSOR_TYPE_TEMPERATURE  : return "Temp.";
  case NT_SENSOR_TYPE_VOLTAGE      : return "Voltage";
  case NT_SENSOR_TYPE_CURRENT      : return "Current";
  case NT_SENSOR_TYPE_POWER        : return "Power";
  case NT_SENSOR_TYPE_FAN          : return "Speed";
  case NT_SENSOR_TYPE_HIGH_POWER   : return "Power";
  default:
    return "Unhandled";
  }
}

float NT_GetSensorMultiplier(NtSensorType_t Type)
{
  switch (Type)
  {
  case NT_SENSOR_TYPE_TEMPERATURE  : return (float)0.1;   //!< Unit: 0.1 degree Celsius
  case NT_SENSOR_TYPE_VOLTAGE      : return (float)1e-3;  //!< Unit: 1 mV
  case NT_SENSOR_TYPE_CURRENT      : return (float)1e-6;  //!< Unit: 1 uA
  case NT_SENSOR_TYPE_POWER        : return (float)1e-7;  //!< Unit: 0.1 uW
  case NT_SENSOR_TYPE_FAN          : return (float)1;     //!< Unit: 1 RPM (Revolutions Per Minute)
  case NT_SENSOR_TYPE_HIGH_POWER   : return (float)1e-3;  //!< Unit: 1 mW
  default:
    return 0.0;
  }
}

char* NT_TranslateSensorValue(char*          Str,
                                     size_t         Size,
                                     int            Value,
                                     NtSensorType_t Type,
                                     const char*    FmtStr,
                                     char*          Prefix)
{
  char  UnitStr[32];
  float FloatVal;
  float PrefixMultiplier;

  if (Value == NT_SENSOR_NAN)
  {
    snprintf(Str, Size, "%s", "");
    return Str;
  }

  //Multiply with sensor multiplier in order to get a value that is in
  //correspondence with the sensor unit
  FloatVal = (float)Value * NT_GetSensorMultiplier(Type);

  if (strcmp(Prefix, "m") == 0)
  {
    PrefixMultiplier = 1e3;
    strcpy(UnitStr, "m");
  }
  else if (strcmp(Prefix, "u") == 0)
  {
    PrefixMultiplier = 1e6;
    strcpy(UnitStr, "u");
  }
  else
  {
    PrefixMultiplier = 1;
    strcpy(UnitStr, "");
  }

  strncat(UnitStr, NT_GetSensorUnit(Type), sizeof(UnitStr) - strlen(UnitStr) - 1);
  FloatVal *= PrefixMultiplier;

  //Test if value and/or unit should be printet
  if (strstr(FmtStr, "s") != NULL)
  {
    if (strstr(FmtStr, "f") != NULL)
      snprintf(Str, Size, FmtStr, FloatVal, UnitStr);
    else
      snprintf(Str, Size, FmtStr, UnitStr);
  }
  else
    snprintf(Str, Size, FmtStr, FloatVal);

  return Str;
}

void DisplaySensorsData(struct sensor* sensor)
{
  char  NameStr[50];
  char  UnitStr[16];
  char  ValStr[16]    = "";
  char  ValLowStr[16] = "";
  char  ValHghStr[16] = "";
  char  Prefix[4];

  //Define the format for printing various sensor type values.
  //                       Unk     Temp    Volt    Curr    OptPwr  FAN     ElecPwr
  const char* ValFmt[7] = {"%.1f", "%.1f", "%.2f", "%.2f", "%.0f", "%.0f", "%.1f"};

  snprintf(NameStr, sizeof(NameStr), "%s", sensor->name);

  //Specify the prefix. Since we dont print any adapter currents it is ok to
  //set the general current prefix to "m"
  if (sensor->type == NT_SENSOR_TYPE_CURRENT) {   //Used for Tx bias current only in mA
    (void)strncpy(Prefix, "m", sizeof(Prefix));
  } else if (sensor->type == NT_SENSOR_TYPE_POWER) { //Optical power in uW
    (void)strncpy(Prefix, "u", sizeof(Prefix));
  } else {
    (void)strncpy(Prefix, "", sizeof(Prefix));
  }

  //Get the unit string alone by not specifying any float translation in the
  //output format using the arbitrary value -10
  NT_TranslateSensorValue(UnitStr, sizeof(UnitStr), 0, sensor->type, "[%s]", Prefix);
  // (void)snprintf(StateStr, sizeof(StateStr), "%s", NT_TranslateSensorState(pSensor->state));

  const char* pValFmt;
  pValFmt = ValFmt[sensor->type];
  //Dump sensor data
  NT_TranslateSensorValue(ValStr, sizeof(ValStr)-1, (uint16_t)sensor->current_value, sensor->type, pValFmt, Prefix);
  NT_TranslateSensorValue(ValLowStr, sizeof(ValLowStr)-1, (uint16_t)sensor->min_value, sensor->type, pValFmt, Prefix);
  NT_TranslateSensorValue(ValHghStr, sizeof(ValHghStr)-1, (uint16_t)sensor->max_value, sensor->type, pValFmt, Prefix);

  printf(SensorFmtStr, NameStr, NT_TranslateSensorType(sensor->type),
         UnitStr, ValStr, ValLowStr, ValHghStr);
}

int main(int argc, char const *argv[])
{
  uint64_t version;
  int len;
  uint8_t *data;
  int fd;
  struct ntc_sensors_s *sensors_info = NULL;
  struct sensor sensor;
  struct ntc_nic_pci_ids_s *nics = NULL;
  struct argparse argparse;

  fd = ntconn_client_connect();

  if (fd < 0) {
    fprintf(stderr, "ERROR while connecting to %s\n", NTCONNECT_SOCKET);
    perror(NULL);
    return 0;
  }

  /*
   * Get version of NtConnect server
   */
  version = ntconn_client_get_version(fd, "0000:00:00.0", "server");
  if (version != 1ULL) {
    fprintf(stderr, "Server version %u.%u not supported\n", VERSION_HI(version), VERSION_LO(version));
    goto exit;
  }

  /*
   * Get all physical adapters and their PCI IDs
   */
  len = ntconn_client_request(fd, "0000:00:00.0", "server", "get,nic_pci_ids", &data, NULL);
  if (len == 0) {
    fprintf(stderr, "No modules registered on server\n");
    goto exit;
  } else {
    nics = (struct ntc_nic_pci_ids_s*)data;
  }

  if (nics->num_nics <= 0) {
    fprintf(stderr, "No SmartNICs found\n");
    goto exit;
  }

  argparse_init(&argparse, options, usage, 0);
  argparse_describe(&argparse, "\nCommand description:", usageExamples);
  argparse_parse(&argparse, argc, argv);

  if (adapter != -1 && adapter >= nics->num_nics) {
    fprintf(stderr, "Adapter %i not found\n", adapter);
    goto exit;
  }

  len = ntconn_client_request(fd, nics->nic_pci_id[adapter], "adapter", "get,sensors", &data, NULL);
  if (len > 0) {
    sensors_info = (struct ntc_sensors_s *)data;
    DisplaySeparator1();
    printf("Adapter %d  %s\n", adapter, sensors_info->adapter_name);
    DisplaySeparator1();
    printf(SensorFmtStr, "", "Sensor", "Unit", "Value", "Value", "Value");
    printf(SensorFmtStr, "", "type"  , ""    , "curr.", "min"  , "max");
    DisplaySeparator1();

    // display board and AVR sensors
    for(uint16_t i = 0; i < sensors_info->adapter_sensors_cnt; i++) {
      sensor = ((struct sensor*)(data + sizeof(struct ntc_sensors_s)))[i];
      DisplaySensorsData(&sensor);
    }

    if (port != -1 && port >= sensors_info->ports_cnt) {
      fprintf(stderr, "Port %i not found\n", port);
      goto exit;
    }

    // display NIM sensors
    if (port == -1) {
      for(uint16_t p = 0; p < sensors_info->ports_cnt; p++) {
        printf("\nPort %u\n", p);
        for(uint16_t ns = 0; ns < sensors_info->nim_sensors_cnt[p]; ns++) {
          sensor = ((struct sensor*)(data + sizeof(struct ntc_sensors_s)))[ns + sensors_info->adapter_sensors_cnt];
          DisplaySensorsData(&sensor);
        }
      }
    } else {
      printf("\nPort %u\n", port);
      for(uint16_t ns = 0; ns < sensors_info->nim_sensors_cnt[port]; ns++) {
        sensor = ((struct sensor*)(data + sizeof(struct ntc_sensors_s)))[ns + sensors_info->adapter_sensors_cnt];
        DisplaySensorsData(&sensor);
      }
    }

    sensors_info = NULL;
    free(data);
  } else {
    fprintf(stderr, "No interfaces found\n");
    return 1;
  }

exit:
  close(fd);

  return 0;
}
