#include <libgen.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "argparse.h"
#include "inttypes.h"
#include "statistic.h"

static char queue_data[MAX_QUEUE_NUM][sizeof(QUEUE_HEADER)];
static char (*color_stat_data)[sizeof(COLOR_STAT_HEADER)];

static char *stats_opt = NULL;
static int port = 0;
static int adapter = 0;

struct argparse_option options[] = {
    OPT_HELP(),
    OPT_INTEGER('a', "adapter", &adapter, "Specifies the adapter that the tool is to operate on", NULL, 0, 0),
    OPT_INTEGER('p', "port", &port, "Specifies the port that the tool is to operate on", NULL, 0, 0),
    OPT_STRING('g', "get", &stats_opt, "Specifies the type of statistic to show", NULL, 0, 0),
    OPT_END(),
};

/*
 * Returns port stat index for the selected logical port index (dpdk_port_id)
 */
static int get_stat_port_index(int port_nr, struct stat_data_s *stats_info)
{
  return stats_info->ifs[stats_info->adap_sel].intf[port_nr].type != PORT_TYPE_VIRT
             ? port_nr
             : stats_info->ifs[stats_info->adap_sel].intf[port_nr].pci_id.devid * 8 +
                   stats_info->ifs[stats_info->adap_sel].intf[port_nr].pci_id.function;
}

static const char *find_pci_func(char *buf, int queue_idx, struct stat_data_s *stats_info)
{
  for (int i = 0; i < stats_info->ifs[stats_info->adap_sel].nb_ports; i++) {
    int ii = 0;
    while (ii < stats_info->ifs[stats_info->adap_sel].intf[i].num_queues) {
      if (stats_info->ifs[stats_info->adap_sel].intf[i].queue[ii].idx == queue_idx) {
        if (stats_info->ifs[stats_info->adap_sel].intf[i].type != PORT_TYPE_VIRT) {
          snprintf(buf, 31, "phy%i  ctl/in", i);
        } else {
          int num = stats_info->ifs[stats_info->adap_sel].intf[i].pci_id.devid * 8 + stats_info->ifs[stats_info->adap_sel].intf[i].pci_id.function;
          if (stats_info->ifs[stats_info->adap_sel].intf[i].queue[ii].dir == QUEUE_INPUT) {
            snprintf(buf, 31, "VF%i  ctl/in", num);
          } else {
            snprintf(buf, 31, "VF%i     out", num);
          }
        }
        return buf;
      }
      ii++;
    }
  }
  return "-unused-  ";
}

/*
 * Print statistic output functions
 */
static int print_que_stats(struct stat_data_s *stats_info)
{
  int queue_counters_number = stats_info->que->nb_counters;
  struct queue_type_fields_s *que = (struct queue_type_fields_s *)stats_info->que->data;
  char buf[32];

  if (queue_counters_number > MAX_QUEUE_NUM)
    queue_counters_number = MAX_QUEUE_NUM;

  fprintf(stdout, "Print queue stats:\n\n");
  fprintf(stdout, "%s\n", QUEUE_HEADER);
  for (int i = 0; i < queue_counters_number; ++i) {
    sprintf(queue_data[i], QUEUE_FORMAT, i, find_pci_func(buf, i, stats_info), que->fwd_pkts, que->fwd_octets, que->dbs_drop_pkts,
            que->dbs_drop_octets);
    fprintf(stdout, "%s\n", (queue_data[i]));
    que++;
  }
  fprintf(stdout, "------------------------------------------------------------\n\n");
  fprintf(stdout, "Timestamp: %lu\n", stats_info->que->timestamp);

  return 0;
}

static int print_flm_stats(struct stat_data_s *stats_info)
{
  int flm_counters_number = stats_info->flm->nb_counters;
  struct flowmatcher_type_fields_s *flm = (struct flowmatcher_type_fields_s *)stats_info->flm->data;

  fprintf(stdout, "Print flm stats:\n");
  for (int i = 0; i < flm_counters_number; ++i) {
    fprintf(stdout, "%-22s: %23" PRIu64 "\n", "current", flm->current);
    fprintf(stdout, "%-22s: %23" PRIu64 "\n", "learn_done", flm->learn_done);
    fprintf(stdout, "%-22s: %23" PRIu64 "\n", "learn_ignore", flm->learn_ignore);
    fprintf(stdout, "%-22s: %23" PRIu64 "\n", "learn_fail", flm->learn_fail);
    fprintf(stdout, "%-22s: %23" PRIu64 "\n", "unlearn_done", flm->unlearn_done);
    fprintf(stdout, "%-22s: %23" PRIu64 "\n", "unlearn_ignore", flm->unlearn_ignore);
    fprintf(stdout, "%-22s: %23" PRIu64 "\n", "auto_unlearn_done", flm->auto_unlearn_done);
    fprintf(stdout, "%-22s: %23" PRIu64 "\n", "auto_unlearn_ignore", flm->auto_unlearn_ignore);
    fprintf(stdout, "%-22s: %23" PRIu64 "\n", "auto_unlearn_fail", flm->auto_unlearn_fail);
    fprintf(stdout, "%-22s: %23" PRIu64 "\n", "timeout_unlearn_done", flm->timeout_unlearn_done);
    flm++;
  }
  fprintf(stdout, "------------------------------------------------------------\n\n");
  fprintf(stdout, "Timestamp: %lu\n", stats_info->flm->timestamp);

  return 0;
}

static int print_port_stats(int port_nr, struct stat_data_s *stats_info)
{
  int port_index = get_stat_port_index(port_nr, stats_info);
  int err = 0;

  if (stats_info->stat_layout_version == 6) {
    if (stats_info->rxc->is_virt) {
      // Virtualization counters
      struct rtx_type_fields_virt_s *rxc;
      struct rtx_type_fields_virt_s *txc;
      rxc = (struct rtx_type_fields_virt_s *)stats_info->rxc->data;
      if (port_index > (int)stats_info->rxc->nb_counters) {
        err = 1;
      }
      txc = (struct rtx_type_fields_virt_s *)stats_info->txc->data;
      if (port_index > (int)stats_info->txc->nb_counters) {
        err = 1;
      }

      if (err) {
        fprintf(stderr, "No statistic found\n");
        return 1;
      } else {
        txc += port_index;
        rxc += port_index;
      }

      fprintf(stdout, "Print port stats:\n\n");
      fprintf(stdout, "TX COUNTERS:\n");
      fprintf(stdout, "Packets:  %015lu\n", txc->pkts);
      fprintf(stdout, "Octets: %017lu\n", txc->octets);
      fprintf(stdout, "Qos drop packets: %015lu\n", txc->qos_drop_pkts);
      fprintf(stdout, "Qos drop octets: %017lu\n", txc->qos_drop_octets);
      fprintf(stdout, "Other drops: %015lu\n\n", txc->drop_events);

      fprintf(stdout, "RX COUNTERS:\n");
      fprintf(stdout, "Packets:  %015lu\n", rxc->pkts);
      fprintf(stdout, "Octets: %017lu\n", rxc->octets);
      fprintf(stdout, "Qos drop packets: %015lu\n", rxc->qos_drop_pkts);
      fprintf(stdout, "Qos drop octets: %017lu\n", rxc->qos_drop_octets);
      fprintf(stdout, "Other drops: %015lu\n", rxc->drop_events);

    } else {
      // Capture counters
      struct stat_rmon_s *tx_rmon;
      struct stat_rmon_s *rx_rmon;

      if (port_index > (int)stats_info->rxc->nb_counters) {
        return 1;
      }
      rx_rmon = &((struct rx_type_fields_cap_s *)stats_info->rxc->data + port_index)->rmon;
      if (port_index > (int)stats_info->txc->nb_counters) {
        return 1;
      }
      tx_rmon = &((struct tx_type_fields_cap_s *)stats_info->txc->data + port_index)->rmon;

      fprintf(stdout, "Print port stats:\n\n");
      fprintf(stdout, "RX COUNTERS:\n");
      fprintf(stdout, "Packets:  %017lu\n", rx_rmon->pkts);
      fprintf(stdout, "Octets:  %017lu\n", rx_rmon->octets);
      fprintf(stdout, "Broadcast:  %017lu\n", rx_rmon->broadcast_pkts);
      fprintf(stdout, "Multicast:  %017lu\n", rx_rmon->multicast_pkts);
      fprintf(stdout, "Unicast:  %017lu\n", rx_rmon->unicast_pkts);
      fprintf(stdout, "Crc errors:  %017lu\n", rx_rmon->pkts_crc);
      fprintf(stdout, "64 octets:  %017lu\n", rx_rmon->pkts_64_octets);
      fprintf(stdout, "65-127 octets:  %017lu\n", rx_rmon->pkts_65_to_127_octets);
      fprintf(stdout, "128-255 octets:  %017lu\n", rx_rmon->pkts_128_to_255_octets);
      fprintf(stdout, "256-511 octets:  %017lu\n", rx_rmon->pkts_256_to_511_octets);
      fprintf(stdout, "512-1023 octets:  %017lu\n", rx_rmon->pkts_512_to_1023_octets);
      fprintf(stdout, "1024-1518 octets:  %017lu\n", rx_rmon->pkts_1024_to_1518_octets);
      fprintf(stdout, "1519-2047 octets:  %017lu\n", rx_rmon->pkts_1519_to_2047_octets);
      fprintf(stdout, "2048-4095 octets:  %017lu\n", rx_rmon->pkts_2048_to_4095_octets);
      fprintf(stdout, "4096-8191 octets:  %017lu\n", rx_rmon->pkts_4096_to_8191_octets);
      fprintf(stdout, "8192-max octets:  %017lu\n", rx_rmon->pkts_8192_to_max_octets);
      fprintf(stdout, "Undersize:  %017lu\n", rx_rmon->undersize_pkts);
      fprintf(stdout, "Oversize:  %017lu\n", rx_rmon->oversize_pkts);
      fprintf(stdout, "Fragments:  %017lu\n", rx_rmon->fragments);
      fprintf(stdout, "Drop events:  %017lu\n", rx_rmon->drop_events);

      fprintf(stdout, "TX COUNTERS:\n");
      fprintf(stdout, "Packets:  %017lu\n", tx_rmon->pkts);
      fprintf(stdout, "Octets:  %017lu\n", tx_rmon->octets);
      fprintf(stdout, "Broadcast:  %017lu\n", tx_rmon->broadcast_pkts);
      fprintf(stdout, "Multicast:  %017lu\n", tx_rmon->multicast_pkts);
      fprintf(stdout, "Unicast:  %017lu\n", tx_rmon->unicast_pkts);
      fprintf(stdout, "Crc errors:  %017lu\n", tx_rmon->pkts_crc);
      fprintf(stdout, "64 octets:  %017lu\n", tx_rmon->pkts_64_octets);
      fprintf(stdout, "65-127 octets:  %017lu\n", tx_rmon->pkts_65_to_127_octets);
      fprintf(stdout, "128-255 octets:  %017lu\n", tx_rmon->pkts_128_to_255_octets);
      fprintf(stdout, "256-511 octets:  %017lu\n", tx_rmon->pkts_256_to_511_octets);
      fprintf(stdout, "512-1023 octets:  %017lu\n", tx_rmon->pkts_512_to_1023_octets);
      fprintf(stdout, "1024-1518 octets:  %017lu\n", tx_rmon->pkts_1024_to_1518_octets);
      fprintf(stdout, "1519-2047 octets:  %017lu\n", tx_rmon->pkts_1519_to_2047_octets);
      fprintf(stdout, "2048-4095 octets:  %017lu\n", tx_rmon->pkts_2048_to_4095_octets);
      fprintf(stdout, "4096-8191 octets:  %017lu\n", tx_rmon->pkts_4096_to_8191_octets);
      fprintf(stdout, "8192-max octets:  %017lu\n", tx_rmon->pkts_8192_to_max_octets);
      fprintf(stdout, "Undersize:  %017lu\n", tx_rmon->undersize_pkts);
      fprintf(stdout, "Oversize:  %017lu\n", tx_rmon->oversize_pkts);
      fprintf(stdout, "Fragments:  %017lu\n", tx_rmon->fragments);
      fprintf(stdout, "Drop events:  %017lu\n", tx_rmon->drop_events);
    }
    fprintf(stdout, "------------------------------------------------------------\n\n");
    fprintf(stdout, "Timestamp: %lu\n",
            stats_info->rxc->timestamp); // RX and TX timestamps are equal
  } else {
    fprintf(stderr, "Unknown statistics layout version. Cannot interpret "
                    "statistics information.");
    fprintf(stderr, "Version expected %i, version reported %i", 6, stats_info->stat_layout_version);
  }

  return 0;
}

static int print_color_stats(struct stat_data_s *stats_info)
{
  struct color_type_fields_s *clr = (struct color_type_fields_s *)stats_info->clr->data;
  int color_counters_number = stats_info->clr->nb_counters;

  color_stat_data = calloc(color_counters_number, sizeof(COLOR_STAT_HEADER));

  fprintf(stdout, "Print color stats:\n\n");
  fprintf(stdout, "%s\n", COLOR_STAT_HEADER);
  for (int i = 0; i < color_counters_number; ++i) {
    sprintf(color_stat_data[i], COLOR_STAT_FORMAT, i, clr->pkts, clr->octets, (uint8_t)clr->tcp_flgs);
    clr++;
    fprintf(stdout, "%s\n", color_stat_data[i]);
  }
  fprintf(stdout, "------------------------------------------------------------\n\n");
  fprintf(stdout, "Timestamp: %lu\n", stats_info->clr->timestamp);
  free(color_stat_data);

  return 0;
}

static int get_port_rxtx_stat(int fd, struct stat_data_s *stats_info, int port_nr)
{
  int len = 0;
  uint8_t *data = NULL;

  // Get all interfaces configured on this adapter
  len = ntconn_client_request(fd, stats_info->nics->nic_pci_id[stats_info->adap_sel], "adapter", "get,interfaces", &data, NULL);
  if (len > 0) {
    stats_info->ifs = (struct ntc_interfaces_s *)data;
  } else {
    fprintf(stderr, "No interfaces found\n");
    return 1;
  }

  len = ntconn_client_request(fd, stats_info->nics->nic_pci_id[stats_info->adap_sel], "stat", "get,tx_counters", &data, NULL);
  if (len > 0) {
    stats_info->txc = (struct ntc_stat_get_data_s *)data;
  } else {
    return 1;
  }

  len = ntconn_client_request(fd, stats_info->nics->nic_pci_id[stats_info->adap_sel], "stat", "get,rx_counters", &data, NULL);
  if (len > 0) {
    stats_info->rxc = (struct ntc_stat_get_data_s *)data;
  } else {
    return 1;
  }

  print_port_stats(port_nr, stats_info);

  return 0;
}

static int get_color_stats(int fd, struct stat_data_s *stats_info)
{
  int len = 0;
  uint8_t *data = NULL;

  len = ntconn_client_request(fd, stats_info->nics->nic_pci_id[stats_info->adap_sel], "stat", "get,colors", &data, NULL);
  if (len > 0) {
    stats_info->clr = (struct ntc_stat_get_data_s *)data;
  } else {
    return 1;
  }

  print_color_stats(stats_info);

  return 0;
}

static int get_que_stats(int fd, struct stat_data_s *stats_info)
{
  int len = 0;
  uint8_t *data = NULL;

  // Get all interfaces configured on this adapter
  len = ntconn_client_request(fd, stats_info->nics->nic_pci_id[stats_info->adap_sel], "adapter", "get,interfaces", &data, NULL);
  if (len > 0) {
    stats_info->ifs = (struct ntc_interfaces_s *)data;
  } else {
    fprintf(stderr, "No interfaces found\n");
    return 1;
  }

  if ((len = ntconn_client_request(fd, stats_info->nics->nic_pci_id[stats_info->adap_sel], "stat", "get,queues", &data, NULL)) < 0)
    return 1;
  if (len > 0) {
    stats_info->que = (struct ntc_stat_get_data_s *)data;
  }

  print_que_stats(stats_info);

  return 0;
}

static int get_flm_stats(int fd, struct stat_data_s *stats_info)
{
  int len = 0;
  uint8_t *data = NULL;

  // Get FLM stats
  len = ntconn_client_request(fd, stats_info->nics->nic_pci_id[stats_info->adap_sel], "stat", "get,flm", &data, NULL);

  if (len <= 0) {
    return 1;
  }

  stats_info->flm = (struct ntc_stat_get_data_s *)data;
  print_flm_stats(stats_info);

  return 0;
}

int main(int argc, const char **argv)
{
  uint64_t version;
  int len;
  uint8_t *data = NULL;
  struct stat_data_s stats_info = {0};
  int fd;
  struct argparse argparse;

  fd = ntconn_client_connect();

  if (fd < 0) {
    fprintf(stderr, "ERROR while connecting to %s\n", NTCONNECT_SOCKET);
    perror(NULL);
    return 0;
  }

  // Get version of NtConnect server
  version = ntconn_client_get_version(fd, "0000:00:00.0", "server");
  if (version != 1ULL) {
    fprintf(stderr, "Server version %u.%u not supported\n", VERSION_HI(version), VERSION_LO(version));
    goto exit;
  }

  // Get all physical adapters and their PCI IDs
  len = ntconn_client_request(fd, "0000:00:00.0", "server", "get,nic_pci_ids", &data, NULL);
  if (len > 0) {
    stats_info.nics = (struct ntc_nic_pci_ids_s *)data;
  } else {
    fprintf(stderr, "No modules registered on server\n");
    goto exit;
  }

  if (stats_info.nics->num_nics <= 0) {
    fprintf(stderr, "No SmartNICs found\n");
    goto exit;
  }

  argparse_init(&argparse, options, usage, 0);
  argparse_describe(&argparse, "\nCommand description:", usageExamples);
  argparse_parse(&argparse, argc, argv);

  stats_info.adap_sel = adapter;
  if (stats_info.adap_sel >= stats_info.nics->num_nics) {
    fprintf(stderr, "Adapter %i not found\n", stats_info.adap_sel);
    goto exit;
  }

  stats_info.stat_version = ntconn_client_get_version(fd, stats_info.nics->nic_pci_id[stats_info.adap_sel], "stat");
  if (stats_info.stat_version != 1ULL) {
    fprintf(stderr, "Server version %u.%u not supported\n", (unsigned)(stats_info.stat_version >> 32),
            (unsigned)(stats_info.stat_version & 0xffffffff));
    goto exit;
  }

  len = ntconn_client_request(fd, stats_info.nics->nic_pci_id[stats_info.adap_sel], "stat", "get,layout_version", &data, NULL);
  if (len) {
    if (len != 4) {
      free(data);
      goto exit;
    }
    stats_info.stat_layout_version = *(int *)data;
    free(data);
  } else
    goto exit;

  if (stats_opt) {
    if (memcmp(stats_opt, "flm_stats", 10) == 0) {
      get_flm_stats(fd, &stats_info);
    } else if (memcmp(stats_opt, "que_stats", 10) == 0) {
      get_que_stats(fd, &stats_info);
    } else if (memcmp(stats_opt, "color_stats", 12) == 0) {
      get_color_stats(fd, &stats_info);
    } else if (memcmp(stats_opt, "txrx_stats", 11) == 0) {
      get_port_rxtx_stat(fd, &stats_info, port);
    } else {
      fprintf(stderr, "Wrong params specified\n");
      argparse_usage(&argparse);
    }
  } else {
    fprintf(stderr, "No params specified\n");
    argparse_usage(&argparse);
  }

exit:
  close(fd);

  if (stats_info.nics) {
    free(stats_info.nics);
  }

  if (stats_info.ifs) {
    free(stats_info.ifs);
  }

  if (stats_info.que) {
    free(stats_info.que);
  }

  if (stats_info.clr) {
    free(stats_info.clr);
  }

  if (stats_info.rxc) {
    free(stats_info.rxc);
  }

  if (stats_info.txc) {
    free(stats_info.txc);
  }

  if (stats_info.flm) {
    free(stats_info.flm);
  }

  return 0;
}
